/****************************************************************************************

MMapInterruptDemo.c++ -- Copyright (c) by Jay Reynolds Freeman 2009.

Demonstration program for memory-mapping and interrupts in connection with Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

/****************************************************************************************

    USE:  This program, in connection with Wraith Scheme, demonstrates the use of
          memory-mapping and interrupts in the Wraith Scheme foreign-function interface.
          See the section in the Wraith Scheme help file titled

            "Demonstrating the Foreign-Function Interface with Other Programs"

          Copy the present file somewhere handy, and compile it with

            g++ -o /tmp/MMapInterruptDemo MMapInterruptDemo.c++

          which will put an executable copy of the program in your /tmp directory.
          Then follow the instructions in the help file.

****************************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <libkern/OSAtomic.h>
#include <unistd.h>

extern int errno;

#define MMAP_FLAGS (MAP_SHARED | MAP_FILE | MAP_HASSEMAPHORE)
#define MMAP_FILE_FLAGS (O_CREAT | O_RDWR)
#define ABSOLUTE_MAX_KITTENS 16
#define NUMBER_OF_INTERRUPTS 1024
#define SIGUSR1_NUMBER 30

typedef struct interruptInformationFileContent_struct {

    volatile int KittenPID[ ABSOLUTE_MAX_KITTENS ];             // Must be a 32-bit int.
    volatile char interruptRequested[ NUMBER_OF_INTERRUPTS ];
    volatile char interruptInstalled[ NUMBER_OF_INTERRUPTS ];

    } interruptInformationFileContent;

int main( int argc, char **argv )
{
    int mmappedFile = open("/tmp/foo", MMAP_FILE_FLAGS, 0666 );

    if( mmappedFile == -1 ) {
        printf("Open of mmapped file failed with error %d, \"%s\"\n", errno, strerror( errno ));
        exit( errno );
        }

    int interruptFile
        = open("/Users/JayFreeman/Library/Application Support/WraithScheme/Wraith Scheme Interrupt Information",
               O_RDWR, 0666 );

    if( interruptFile == -1 ) {
        printf("Open of interrupt file failed with error %d, \"%s\"\n", errno, strerror( errno ));
        exit( errno );
        }

    char *sharedSpace = (char *)mmap( 0, 1000, PROT_READ | PROT_WRITE, MMAP_FLAGS, mmappedFile, 0 );

    if( sharedSpace == MAP_FAILED ) {
        printf("mmap of shared area failed with error %d, \"%s\"\n", errno, strerror( errno ));
        exit( errno );
        }

    interruptInformationFileContent *info 
        = (interruptInformationFileContent *)mmap( 0, 1088, PROT_READ | PROT_WRITE, MMAP_FLAGS, interruptFile, 0 );

    // All we do is copy the first 16 bytes -- which the user has presumably put there
    // from Wraith Scheme -- into the second 16 bytes, so the user can see that they 
    // have been moved.

    for( int counter = 0; counter < 16; ++counter )
        sharedSpace[ counter +16 ] = sharedSpace[ counter ];

    // Now let's wait a while, to build suspense.

    sleep( 5 );

    // And now we are going to interrupt Wraith Scheme to tell it to get the copied 
    // data and print it out.  This next line *could* loop forever if there is a
    // mistake somewhere ...

    while( OSAtomicTestAndSet( 0, &((info->interruptRequested)[ 42 ]) ) )
        ;
    
    char systemCallBuffer[ 101 ];
    snprintf( systemCallBuffer, 100, "kill -%d %d", SIGUSR1_NUMBER, (info->KittenPID)[ 0 ]);

    printf("About to do a system call with \"%s\".\n", systemCallBuffer );
    system( systemCallBuffer );
    }
/*** END OF SECOND PROGRAM *****************************************************************/
