/****************************************************************************************

SocketServer.c++ -- Copyright (c) by Jay Reynolds Freeman 2009.

Demonstration program for memory-mapping and interrupts in connection with Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

/****************************************************************************************

    USE:  This is one of two C++ programs, which in connection with Wraith Scheme,
          demonstrates the use of Unix sockets,  memory-mapping and interrupts in
          the Wraith Scheme foreign-function interface.  The two programs' source
          files are

              SocketServer.c++
              SocketClient.c++

          See the section in the Wraith Scheme help file titled

            "Demonstrating the Foreign-Function Interface with Other Programs"

          Copy the two files somewhere handy, and compile them with

            g++ -o SocketServer SocketServer.c++
            g++ -o SocketClient SocketClient.c++

          Then follow the instructions in the help file.

****************************************************************************************/

/****************************************************************

;; For reference, here is the Scheme code required for the demonstration.

(define server-mmapped-block (e::make-memory-mapped-block 1000 "/tmp/serverMMap"))
(c::set-interrupt-handler! 42 0
  '(begin (display (e::peek-c-string server-mmapped-block 0)) (newline)))

(define client-mmapped-block (e::make-memory-mapped-block 1000 "/tmp/clientMMap"))

(e::memory-map-block-in-current-kitten server-mmapped-block)  ;; In the MomCat

(e::poke-c-string client-mmapped-block 0 "Your favorite string.")

(e::system "kill -2 <Unix process identification of the SocketClient program>")

****************************************************************/

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <libkern/OSAtomic.h>

#define SOCKET_ERROR          -1
#define BUFFER_SIZE           100
#define MESSAGE               "Test message."
#define SOCKET_NAME    "/tmp/TestSocket"
#define SOCKET_BACKLOG 100
#define MMAP_FLAGS (MAP_SHARED | MAP_FILE | MAP_HASSEMAPHORE)
#define MMAP_FILE_FLAGS (O_CREAT | O_RDWR)
#define SERVER_MMAP_AREA_NAME "/tmp/serverMMap"
#define ABSOLUTE_MAX_KITTENS 16
#define NUMBER_OF_INTERRUPTS 1024
#define SIGUSR1_NUMBER 30

typedef struct interruptInformationFileContent_struct {

    volatile int KittenPID[ ABSOLUTE_MAX_KITTENS ];             // Must be a 32-bit int.
    volatile char interruptRequested[ NUMBER_OF_INTERRUPTS ];
    volatile char interruptInstalled[ NUMBER_OF_INTERRUPTS ];

    } interruptInformationFileContent;

extern int errno;
typedef struct socketaddr_un *pSAun;

int main(int argc, char* argv[])
{
    int messageCounter = 0;
    int clientSocketID = 0;
    int serverSocketID = 0;
    int backLog = 0;
    char *socketName = 0;
    struct sockaddr_un serverSocketAddress;
    int serverSocketAddressSize = sizeof(struct sockaddr_un);
    char *messageBuffer = 0;

    if( argc != 1 ) {
        fprintf( stderr, "\nUsage: SocketServer\n");
        fflush( stderr );
	    exit( 1 );
        }

    int mmappedFile = open( SERVER_MMAP_AREA_NAME, MMAP_FILE_FLAGS, 0666 );

    if( mmappedFile == -1 ) {
        fprintf( stderr, "Server mmapped file open failed with error %d, \"%s\"\n", errno, strerror( errno ));
        fflush( stderr );
        exit( errno );
        }

    int interruptFile
        = open("/Users/JayFreeman/Library/Application Support/WraithScheme/Wraith Scheme Interrupt Information",
               O_RDWR, 0666 );

    if( interruptFile == -1 ) {
        fprintf( stderr, "Open of interrupt file failed with error %d, \"%s\"\n", errno, strerror( errno ));
        exit( errno );
        }

    char *sharedSpace = (char *)mmap( 0, 1000, PROT_READ | PROT_WRITE, MMAP_FLAGS, mmappedFile, 0 );
 
    if( sharedSpace == MAP_FAILED ) {
        fprintf( stderr, "Server mmap failed with error %d, \"%s\"\n", errno, strerror( errno ));
        fflush( stderr );
        exit( errno );
        }

    messageBuffer = sharedSpace;
    interruptInformationFileContent *info 
        = (interruptInformationFileContent *)mmap( 0, 1088, PROT_READ | PROT_WRITE, MMAP_FLAGS, interruptFile, 0 );

    fprintf( stderr, "Making server socket:\n");
    fflush( stderr );

    if( (serverSocketID = socket( AF_UNIX, SOCK_STREAM, 0)) == SOCKET_ERROR ) {
        fprintf( stderr, "\nCall to \"socket\" failed.\n");
        fflush( stderr );
        exit( 1 );
        }

    // Fill address struct ...

    serverSocketAddress.sun_family = AF_UNIX;
    strcpy( serverSocketAddress.sun_path, SOCKET_NAME );

    // Bind ...

    fprintf( stderr, "Binding...\n");
    fflush( stderr );

    unlink( SOCKET_NAME );
    if( bind( serverSocketID, (const sockaddr *)&serverSocketAddress, serverSocketAddressSize ) == SOCKET_ERROR ) {
        fprintf( stderr, "Socket bind failed with error %d, \"%s\".\n", errno, strerror( errno ) );
        fflush( stderr );
        exit( 1 );
        }

    // Listen ...

    fprintf( stderr, "Starting to listen with a queue of %d elements ...\n", SOCKET_BACKLOG );

    if( listen( serverSocketID, SOCKET_BACKLOG ) == SOCKET_ERROR ) {
        fprintf( stderr, "\nThe \"listen\" operation failed.\n");
        fflush( stderr );
        exit( 1 );
        }

    fprintf( stderr, "Server waiting for a connection ...\n");
    fflush( stderr );

    // Get the socket from the client ...

    clientSocketID
        = accept( serverSocketID,
                 (struct sockaddr*)&serverSocketAddress,
                 (socklen_t *)&serverSocketAddressSize
                 );
    fprintf( stderr, "Server has a connection.\n");
    fflush( stderr );

    for(;;) {
        // Write to the client.

        fprintf( stderr, "################################ Message %d ################################\n", messageCounter++ );
        fflush( stderr );

        read( clientSocketID, messageBuffer, BUFFER_SIZE );
        fprintf( stderr, "Server read \"%s\" from client.\n", messageBuffer );

        while( OSAtomicTestAndSet( 0, &((info->interruptRequested)[ 42 ]) ) )
            ;
    
        char systemCallBuffer[ 101 ];
        snprintf( systemCallBuffer, 100, "kill -%d %d", SIGUSR1_NUMBER, (info->KittenPID)[ 0 ]);

        printf("About to do a system call with \"%s\".\n", systemCallBuffer );
        system( systemCallBuffer );
        }
    }
