/****************************************************************************************

DataTypes.h -- Copyright (c) by Jay Reynolds Freeman 1987, 1988, 1989, 1990, 1991, 2006

Header file for the specific data types used in Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import <stdio.h>
#import "TaggedData.h"

#define ABSOLUTE_MAX_KITTENS  16

/**** Data Descriptors ****/

typedef struct {
    Text        *pathName;
    Text        *oldPathName;
    FILE *      fp;                 /* The Unix file pointer */
    Boolean     fdOpen;             /* Is this entry in the table of file descriptors now in use? */
    Boolean     fdInUse;            // Kind of a garbage-collection flag just for file descriptors; the idea is to
                                    // set this to TRUE if, and only if
                                    //     (1) at least one pointer in the heap, after GC, points to this file descriptor, OR
                                    //     (2) a register points to it, OR
                                    //     (3) it is one of the default ports.
    Text        *type;              /* Access rights for file; a parameter to C "fopen", etc */
    LongWord    placeInBigArray;    // At what index in "AllFileDescriptors" is this?
    LongWord    kittenNumber;       /* What kitten number is for this fileDescriptor? */
    } fileDescriptor;

typedef struct {
    UnsignedLongWord pNameNumber;                       /* Cursor into a resource list (of print names). */
    const char *pName;                                  /* Print name of the data type, e. g. "Symbol". */
    Tag tag;                                            /* Tag which characterizes this data type. */
    memoryBlockSize length;                             /* Datum length, in Bytes.  "0" means use the ... */    
    memoryBlockSize (*lengthFinder)( taggedAval *a );   /*   ... function (of pointer to datum) instead. */
    void (*identifier)();                               /* Predicate to test whether a tag is of this type. */
    Boolean (*display)( taggedAval *aPort);             /* How to display an instance of this dataDescriptor */
    Boolean (*write)( taggedAval *aPort);               /* How to write an instance of this dataDescriptor */
    Boolean (*inspect)( taggedAval *aPort);             /* How to inspect an instance of this dataDescriptor */

    /****************

    The next two items have to do with making saved worlds that can be read and written on either big-endian
    or little-endian machines, whether or not the machine doing the reading is the same endian as the machine
    that did the writing.  The file format will be such that straight Unix "read" and "write" will be correct
    on little-endian machines, which means that when operating on a big-endian machine, the stuff to be written
    will have to be switched to little-endian format before the write, and the stuff read will have to be switched
    to big-endian forma after the read.  The next two functions help do that.
    
    The software itself knows how to do the swapping of just a tagged aval.  When I wrote these words, it was
    two four-byte fields, and the blg-endian / little-endian transformation is easy and symmetrical.  The
    problem occurs when data items are not simple tagged avals; perhaps the pointer field of the tagged aval
    is a pointer to something else in main memory, and that must be transformed in a way that requires
    knowing what it is.  These functions are for transforming the *rest* of the data item, anything
    beyond just the tagged aval itself.  
    
    There is a noteworthy gotcha:  When a big-endian machine has read a world file, it cannot know anything
    about a data item until it has transformed the tagged aval that is part of it; those bits are meaningless
    for it, until they have been converted to big-endian format.  Thus these transformations must be done
    sequentially, in a pass through the memory image, knowing that the memory image starts with the tagged
    aval portion of some data item.  The tag gets transformed, then anything it may point to, then the
    transformed data item may be used to determine its own length and thus the address of the next
    untransformed data item to deal with.

    Since porting Wraith Scheme to a 64-bit application, I have not had a chance to test the "swapping" stuff,
    because the 64-bit port runs only on Snow Leopard, which runs only on Macintoshes with Intel processors,
    which are little-endian.  The relevant code, and these fields in this struct, are left in, in case they
    are ever needed.

    ****************/

    /* Flips a big-endian instance to little-endian. */
    memoryBlockSize (*flipToLittleEndian)( taggedAval *a, void *worldEnd, void *world, void *alternateWorld );

    /* Flips a little-endian instance to big-endian. */
    memoryBlockSize (*flipToBigEndian)( taggedAval *a, void *worldEnd, void *world, void *alternateWorld );
    } dataDescriptor;
        
typedef struct memorySpaceStruct {
    volatile Byte *end;                      /* Points one past the end of usable memory for this space. */
    volatile Byte *free;                     /* Pointer to next (high) location usable for this space. */
    volatile Byte *lowFree;                  /* Pointer to next (low) location usable for this space. */
    volatile Byte *space;                    /* Pointer to start of usable memory for this space. */
    volatile UnsignedLongWord size;          /* How many bytes in this space? */
    volatile Boolean inUse;                  /* Set when a space contains at least some non-garbage; used (e. g.) by the garbage
                                                collector, to detect recursive garbage-collection. */

    volatile struct memorySpaceStruct *allocateDomain;  // Where does allocation take place?
    volatile UnsignedLongWord allocateSize;  // How many bytes in the space, considered as an allocatable domain
    volatile Byte *allocateMax;              /* Pointer to the low end of allocatable memory for the allocate routines. */

    // Start of additions for generational garbage collector.

    volatile struct memorySpaceStruct *generationA;
    volatile struct memorySpaceStruct *generationB;
    volatile struct memorySpaceStruct *activeGeneration;
    volatile struct memorySpaceStruct *agingGeneration;

    volatile UnsignedLongWord *activeGenerationStackEnd;

    volatile UnsignedLongWord *agingGenerationStackEnd;

#ifdef PIXIE_SCHEME_III
    Word pad;                                // This is here so that the entire structure length is a multiple of 8 bytes.
#endif

    } memorySpace;

typedef struct {

    // The next two variables are on a per-kitten basis.

    Boolean mmapWasSuccessful[ ABSOLUTE_MAX_KITTENS ];      // Initialized false, set true if and only if mmap returned non-null.
    Boolean regionHasBeenUnmapped[ ABSOLUTE_MAX_KITTENS ];  // Initialized false, set true if and only if munmap is called hereon.

    Boolean fileHasBeenWritten;              // Initialized to false, set to true if and only if the file has been
                                             //   loaded full of initial data.

    // The next few arguments are a record of what mmap is called with.

    void *absoluteAddress;                   // Pointer to start of block returned by mmap.
    size_t size;                             // Size is in bytes.
    int prot;                                // Region accessibility.
    int flags;                               // Miscellaneous flags, or'd together.
    int fd;                                  // File descriptor.
    int offset;                              // Offset within file.
    
    } mmappedBlockDescriptor;

