/****************************************************************************************

MemoryManager.h -- Copyright (c) by Jay Reynolds Freeman 2007.

Memory management class, for Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#ifndef __MEMORYMANAGER_H__
#define __MEMORYMANAGER_H__

#import "WraithHeader.h"
#import "DataTypes.h"
#import "MMapStuff.e"

#define N_MMAP_TRIES 3
#define MMAP_FAILURE_SLEEP_DELAY 100000

#define inSpaceP( pointer, mSpace ) \
	( (((Byte *)((mSpace)->allocateMax)) <= (Byte *)(pointer)) && \
	  ( (Byte *)(pointer)         <  (Byte *)((mSpace)->end)) )

#define inWholeSpaceP( pointer, mSpace ) \
	( (((Byte *)((mSpace)->space)) <= (Byte *)(pointer)) && \
	  ( (Byte *)(pointer)         <  (Byte *)((mSpace)->end)) )

class GarbageCollector;
class WorldHandler;
class Clowder;

extern void printMemorySpace( volatile memorySpace *m, Boolean showSpaceP );
extern void clearMemorySpace( volatile memorySpace *m );
extern void fillMemorySpace( volatile memorySpace *m, Byte value );

// Instrumentation and display in human-readable form:

extern void describeMemorySpace( volatile memorySpace *m, Boolean showSpaceP );
extern void describeMemorySpaceNonGarbage( volatile memorySpace *m, Boolean showSpaceP );
extern void printfMemorySpace( volatile memorySpace *m, Boolean showSpaceP );
extern void describeRoom( volatile memorySpace *m );
extern UnsignedLongWord room( volatile memorySpace *m );
extern UnsignedLongWord storeSize( volatile memorySpace *m );

class MemoryManager {

  friend class GarbageCollector;
  friend class WorldHandler;
  friend class Clowder;

  public:

    // Initialization and allocation of heap:

    void initialize();
    void initializeMemoryManager( Boolean bEFlag,
                                  Boolean useGenerationalSpaces,
                                  Text *externalNameString,
                                  size_t externalNameStringSize,
                                  size_t spaceSize,
                                  size_t stackSize,
                                  size_t continuationSize,
                                  size_t generationSize,
                                  size_t *createdSpaceSize,
                                  Byte *spaceAGivenTargetAddress,
                                  Byte *spaceBGivenTargetAddress,
                                  Byte *scratchStackGivenTargetAddress,
                                  Byte *gGCScratchSpaceGivenTargetAddress
                                  );
    void initializeMemoryManagerByMalloc( Boolean bEFlag,
                                          Boolean useGenerationalSpaces,
                                          Text *externalNameString,
                                          size_t externalNameStringSize,
                                          size_t spaceSize,
                                          size_t stackSize,
                                          size_t continuationSize,
                                          size_t generationSize,
                                          size_t *createdSpaceSize
                                          );
    volatile memorySpace *initializeMemorySpace( size_t bytesWanted );
    volatile memorySpace *initializeGenerationalSpace( volatile memorySpace *m );
    void setClowder( Clowder *aClowder );

    // Release of memory-mapped heap space.

    void releaseMemoryFiles();

    // Instrumentation, and also display in human-readable form:

    UnsignedLongWord nOfTaggedAvalInHeap( taggedAval *tp );
    void describeMemorySpaces( Boolean showSpaceP );
    void describeCurrentMainSpace();
    void describeCurrentActiveGeneration();
    void describeCurrentAgingGeneration();
    void describeCurrentStackSpace();
    void describeCurrentContinuationSpace();
    void describeLocksInCurrentMainSpace();
    void releaseLocksInCurrentMainSpace();
    void describeCurrentRoom();
    UnsignedLongWord currentStoreSize();
    UnsignedLongWord currentActiveStoreSize();
    UnsignedLongWord currentInUse();
    UnsignedLongWord currentActiveInUse();
    UnsignedLongWord totalInUse();
    UnsignedLongWord currentRoom();
    UnsignedLongWord currentActiveRoom();
    UnsignedLongWord currentAgingRoom();
    UnsignedLongWord allocateDomainCurrentRoom();
    LongWord currentMainSpaceCurrentRoom();
    Byte *schemeStackAddress();
    Byte *schemeStackEnd();
    taggedAval *schemeStackSpace();
    UnsignedLongWord schemeStackSize();
    Byte *schemeContinuationAddress();
    Byte *schemeContinuationEnd();
    taggedAval *schemeContinuationSpace();
    UnsignedLongWord schemeContinuationSize();
    Boolean checkSpaceParameters( volatile memorySpace *m, Word position );
    volatile memorySpace *getStack();
    volatile memorySpace *getContinuation();
    volatile memorySpace *getScratchStack();
    volatile memorySpace *getGGCScratchSpace();
    Boolean get90PercentFlag();
    void padForFullGC( volatile memorySpace *m );
    
    // Allocation within heap:

    void allocate( memoryBlockSize bytesWanted );
    void allocateOne();
    void allocateTwo();
    void allocateWithoutInitialHeapLock( memoryBlockSize bytesWanted );
    void allocateR();
    void allocateVectorR();
    void allocateByteBlock( memoryBlockSize bytesWanted );
    void allocateStringBlock( const Text *str );
    void allocateStringNBlock( size_t n );
//    Boolean consLuckOut( Byte *p, UnsignedLongWord spaceWanted );
    Boolean consLuckOutOneTaggedAval( Byte *p );

    // Pass-throughs to other classes:

    void examineWorld( Boolean compareP, Text *fileName );
    UnsignedLongWord garbageCollectorGCCount();
    UnsignedLongWord garbageCollectorGenerationalGCCount();
    void loadWorld();
    void loadWorldGuts( fileDescriptor *wrldFile, Boolean describeP );
    Boolean worldIsUseable( Text *pathToFile );
    void saveWorld();
    Boolean triggerGC();
    Boolean triggerOnRebuildObList();
    WorldHandler *getWorldHandler();
    void cleanUpPorts();
    void setSpaceForFullGC( volatile memorySpace *m );
    void setSpaceForGenerationalGC( volatile memorySpace *m );
    void revertToFullGC();

    // Instrumentation:

    inline void addToByteAllocCounter( UnsignedLongWord n ) { byteAllocCounter += n; }
    inline UnsignedLongWord getByteAllocCounter() { return byteAllocCounter; }

    inline void addToSetBangCounter( UnsignedLongWord n ) { setBangCounter += n; }
    inline UnsignedLongWord getSetBangCounter() { return setBangCounter; }

 private:

    static const LongWord smallPressureBufferSize = MEGABYTE / 2;
    static const LongWord pressureBufferSize = MEGABYTE;

    volatile memorySpace *CurrentMainSpace;
    volatile memorySpace *AlternateMainSpace;
    volatile memorySpace *SchemeStack;
    volatile memorySpace *SchemeContinuation;
    volatile memorySpace *ScratchStack;
    volatile memorySpace *GGCScratchSpace;

    UnsignedLongWord byteAllocCounter;
    UnsignedLongWord setBangCounter;

    Text scratch[ MAX_PATHNAME_LENGTH + 1 ];
    Text spaceAFileName[ MAX_PATHNAME_LENGTH + 1 ];
    Text spaceBFileName[ MAX_PATHNAME_LENGTH + 1 ];
    Text scratchStackFileName[ MAX_PATHNAME_LENGTH + 1 ];
    Text gGCScratchSpaceFileName[ MAX_PATHNAME_LENGTH + 1 ];
    Boolean BigEndianFlag;

    UnsignedLongWord allocateMaxOffset;
    size_t generationSpaceSize;
    void swapGenerationSpaces( volatile memorySpace *consHere );
    void swapGenerationSpacesKitten();
    void clearGenerationSpace( volatile memorySpace *m );

    GarbageCollector *myGarbageCollector;
    WorldHandler *myWorldHandler;
    Clowder *myClowder;

    volatile memorySpace *initializeMemorySpaceGivenByteBlock( 
        Boolean useGenerationalSpaces,
        volatile memorySpace **m,
        Byte *block,
        UnsignedLongWord bytesGiven,
        UnsignedLongWord genSpaceSize
        );
    void copyMemorySpaceMiscellanea( volatile memorySpace *from, volatile memorySpace *to, UnsignedLongWord *toMinusFrom );
    void copyMemorySpacePartialContent( volatile memorySpace *from, volatile memorySpace *to,
                                        UnsignedLongWord *toMinusFrom, Byte *startHere );
    void allocateLow( volatile memorySpace *m, memoryBlockSize bytesWanted );
    void swapSpaces();
    UnsignedLongWord standardGenerationalReserve();
    UnsignedLongWord smallGenerationalReserve();
    UnsignedLongWord standardFullReserve();
    UnsignedLongWord standardLargeFullReserve();
    Boolean enoughMemoryToAllocate( memoryBlockSize bytesWanted );
    Boolean setReturn( Boolean returnValue, LongWord errorNumber, Boolean couldRebuildObList, Boolean neededToRebuildOblist,
                       memoryBlockSize bytesWanted, memoryBlockSize previousFree );
    Boolean gcSequence( memoryBlockSize bytesWanted );

// Inline methods:

/**** getCurrentMainSpace ****/

    inline volatile memorySpace *getCurrentMainSpace() { return CurrentMainSpace; }

    };

#endif
