/****************************************************************************************

Registers.h -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006, 2008

Header file for Register-manipulating macros used in Wraith Scheme.

****************************************************************************************/

#import "TaggedData.h"
#import "WraithConsts.h"
#import "Registers.e"
#import "IO.e"

const UnsignedLongWord ESR_applicativeBitMask             = 0x0000000000000001LL;
const UnsignedLongWord ESR_interruptInProgressMask        = 0x0000000000000002LL;  // FIXME -- don't forget.

/* Non-simple registers, declared elsewhere, these #defines are syntactic sugar. */

#define TopLevelE (*(theClowder->globalTopLevelE))
#define ObList (*(theClowder->globalObList))
#define TLC (*(theClowder->globalTLC))
#define KIQ (*(theClowder->globalKIQ))
#define IVR (*(theClowder->globalIVR))

#define SDrop()                 (++S)
#define SDropN( n )             (S += (n))
#define SPopA()                 (A = *S++)
#define SPopE()                 (E = *S++)
#define SPopR()                 (R = *S++)
#define SPushR()                ( *(--S) = R )
#define SPushA()                ( *(--S) = A )
#define SPushE()                ( *(--S) = E )
#define SDupR()                 (R = *S)

/* NB:  Offsets from the "frame pointer" entry of the frame". */

#define FrameSize               6

#define FRAME_E_OFFSET          (-1)
#define FRAME_FP_OFFSET         ( 0)
#define FRAME_CF_OFFSET         ( 1)
#define FRAME_NEW_MN_OFFSET     ( 2)
#define FRAME_NEW_L_OFFSET      ( 3)
#define FRAME_ESR_OFFSET        ( 4)

#define CSPush( x )             ( *(--CS) = (x) )
#define CSPopR()                (R = *CS++)
#define CSPopA()                (A = *CS++)
#define CSDup()                 (R = *CS)
#define CSDrop()                (++CS)
#define CSTopR()                (R = *CS)
#define CPushFrame( x )         (CF = CS, *(--CS) = (x))
#define CPushEmptyFrame()       (CF = CS)

#define emptyCSP()              (CS == CF)

#define LoadFrame( startOffset, stackStatus ) \
        (S[(startOffset)] = E, \
         S[(startOffset) + 1].u.a = F, \
         S[(startOffset) + 1].t = exactShortFixnumBitsMask, \
         S[(startOffset) + 2].u.a = CF, \
         S[(startOffset) + 2].t = exactShortFixnumBitsMask, \
         S[(startOffset) + 3].u.c = stackStatus, \
         S[(startOffset) + 3].t = exactShortFixnumBitsMask, \
         S[(startOffset) + 4].t = emptyListTag, \
         S[(startOffset) + 5].u.c = ESR, \
         S[(startOffset) + 5].t = exactShortFixnumBitsMask, \
         F = (S + (startOffset) + 1), \
         CF = CS \
         );

#define MakeFrame()             S -= FrameSize

//  "PushFrame" makes a frame right on top of the stack, as it is now.
//  NOTA BENE -- code to pop the stack in SECDEval.c++, search for "Must pop the stack".

#define PushFrame( stackStatus ) { \
     ((--S)->u).c = ESR; \
     S->t = exactShortFixnumBitsMask; \
     (--S)->t = emptyListTag; \
     ((--S)->u).c = (stackStatus); \
     S->t = exactShortFixnumBitsMask, \
     ((--S)->u).a = CF; \
     S->t = exactShortFixnumBitsMask, \
     ((--S)->u).a = F; \
     S->t = exactShortFixnumBitsMask, \
     F = S; \
     *(--S) = E; \
     CF = CS; \
     }

//  Some things for the sensory device registers:

#define N_SLIDERS 4                // Must match the definition in "WraithSchemeOuterController.mm".
#define N_LEVEL_INDICATORS 4       // Must match the definition in "WraithSchemeOuterController.mm".
#define N_PUSHBUTTONS 4            // Must match the definition in "WraithSchemeOuterController.mm".

#define SLIDER_INITIAL_VALUE 127
#define LEVEL_INDICATOR_MAX_VALUE 1024

/**** spliceOutNotCarForwardingPointers ****/

inline taggedAval spliceOutNotCarForwardingPointers( taggedAval x )
{
    // While x points to a forwarding pointer, make x point where that pointer pointed.

    while( not_car_forwarding_pointer_p( (x.u.a)->t ) )
        x.u.a = ((x.u.a)->u).a;
    return x;
    }

/**** spliceOutNotCarForwardingPointersIndirect ****/

inline taggedAval *spliceOutNotCarForwardingPointersIndirect( taggedAval *xp )
{
    // While xp points to a forwarding pointer, make xp point where that pointer pointed.

    while( not_car_forwarding_pointer_p( xp->t ) )
       xp = (xp->u).a;
    return xp;
    }

/**** exchangeFunction ****/

inline void exchangeFunction( taggedAval *foo, taggedAval *bar )
{
    taggedAval dummy = *foo;
    *foo = *bar;
    *bar = dummy;
    }

/**** Swap ****/

inline void Swap()
{
    taggedAval swapDummy = R;
    
    R = S[0];
    S[0] = swapDummy;
    }

/**** ESR Bit Field Macros ****/

inline Boolean applicativeMode() { return( (ESR & ESR_applicativeBitMask) != 0 ); }
inline Boolean interruptInProgress() { return( (ESR & ESR_interruptInProgressMask) != 0 ); }

/**** applicativeCheck ****/

#define ENABLE_APPLICATIVE_BIT_USE 1

inline void applicativeCheckDropNone( const Text *where )
{
#if ENABLE_APPLICATIVE_BIT_USE
    if( applicativeMode() ) {
        SDrop();
        tprintf3( &ErrorPort, MiscStrings[ 511 ], where );
        resetError( 328 );
        }
#endif
    }

inline void applicativeCheckDropOne( const Text *where )
{
#if ENABLE_APPLICATIVE_BIT_USE
    if( applicativeMode() ) {
        SDrop();
        tprintf3( &ErrorPort, MiscStrings[ 511 ], where );
        resetError( 328 );
        }
#endif
    }

inline void applicativeCheckDropTwo( const Text *where )
{
#if ENABLE_APPLICATIVE_BIT_USE
    if( applicativeMode() ) {
        SDrop();
        SDrop();
        tprintf3( &ErrorPort, MiscStrings[ 511 ], where );
        resetError( 328 );
        }
#endif
    }

inline void setApplicativeMode( Boolean value )
{
#if ENABLE_APPLICATIVE_BIT_USE
    if( value )
        ESR |= ESR_applicativeBitMask;
    else
        ESR &= ~(ESR_applicativeBitMask);
#endif
    }


