/****************************************************************************************

SenseLightCommands.h -- Copyright (c) by Jay Reynolds Freeman 2010

Header file for commands controlling the Wraith Scheme sense lights.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

/****************************************************************************************

    The idea of the sense light commands is to be able to pack commands for any or all
    of the eight sense lights into a single 64-bit long, for transmission from the
    Scheme Machine to the view/controller.  The commands for the individual sense lights
    correspond to the individual bytes of the long, with byte 0 -- the least-significant
    byte -- controlling sense light 0, and so on.

    Each sense light command has eight bits, though as of this writing we need only six.

    The MSB of the command, bit 7, tells whether the given sense light is affected by
    the command; that is, a zero here means "Don't change what that sense light is doing
    -- ignore the rest of the bits of this command."

    Bit 6 of the command tells whether the command is to change the color (if bit 6 is a
    1) or to change the visibility (if bit 6 is a zero); you cannot do both at once,
    although colors may be changed even when the lights are invisible: The lights will
    appear with their new, correct colors when they are made visible again.  Bit 6 is
    meaningful only if bit 7 is a 1.

    Bit 5 of the command tells whether the sense light is to be made visible or not; a
    zero means it is not visible and does not appear, not even in the "turned-off"
    state.  Bit 5 is meaningful only if bit 6 is a 0.
    
    Bits 0-2 of the command describe the illumination of the sense light, which will
    be one of the seven states described below, e.g., "SENSE_LIGHT_RED".  Bits 0-2 are
    meaningful only if bit 6 is a 1.

    Bits 3 and 4 of the command are not used at present.


****************************************************************************************/

#ifndef __SENSE_LIGHT_COMMANDS_H__
#define __SENSE_LIGHT_COMMANDS_H__

#define N_SENSE_LIGHTS 8

#define SENSE_LIGHT_OFF     0
#define SENSE_LIGHT_RED     1
#define SENSE_LIGHT_ORANGE  2
#define SENSE_LIGHT_YELLOW  3
#define SENSE_LIGHT_GREEN   4
#define SENSE_LIGHT_CYAN    5
#define SENSE_LIGHT_MAGENTA 6
#define SENSE_LIGHT_BLACK   7

#define N_SENSE_LIGHT_ILLUMINATIONS 8

#define COMMAND_MASK 0xff

#define COMMAND_BIT_POSITION 7
#define COMMAND_BIT_MASK 0x80

#define COMMAND_TYPE_BIT_POSITION 6
#define COMMAND_TYPE_BIT_MASK 0x40

#define VISIBLE_BIT_POSITION 5
#define VISIBLE_BIT_MASK 0x20

#define ILLUMINATION_BITS_POSITION 0
#define ILLUMINATION_BITS_MASK 0x7

#define N_SENSE_LIGHTS_MASK 0x7

#endif

