/****************************************************************************************

Clowder.c++ -- Copyright (c) by Jay Reynolds Freeman 1987, 1988, 1989, 1990, 1991, 2006, 2008

A clowder is a group of cats.

This file is the header for the object that coordinates a group of parallel
Wraith Scheme processes.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import <sys/types.h>
#import <sys/mman.h>
#import <fcntl.h>
#import <sys/ipc.h>
#import <sys/shm.h>
#import "LocalTypes.h"
#import "WraithHeader.h"
#import "Clowder.h"
#import "OpCodes.h"
#import "GarbageCollector.h"
#import "WorldHandler.h"
#import "DataTypes.e"
#import "Globals.e"
#import "IO.e"
#import "MMapStuff.e"

// #define SPIN_DELAY_IN_NANOSECONDS 50  // CHANGED FOR TEST
#define SPIN_DELAY_IN_NANOSECONDS 500

struct timespec SPIN_ON_GC_DELAY_struct          = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec TAG_BLOCK_DELAY_struct           = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec OBLIST_BLOCK_DELAY_struct        = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec OBLIST_COUNT_DELAY_struct        = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec ALLOCATE_BLOCK_DELAY_struct      = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec COLOR_BLOCK_DELAY_struct         = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec GENSYM_BLOCK_DELAY_struct        = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec TLE_BLOCK_DELAY_struct           = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec NFMMS_BLOCK_DELAY_struct         = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec INTERRUPT_BLOCK_DELAY_struct     = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec INTERRUPT_HANDLER_DELAY_struct   = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec GGC_BLOCK_DELAY_struct           = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec GC_BYTE_DELAY_struct             = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds

struct timespec SPIN_ON_GC_IN_PROGRESS_DELAY_struct          = { 0, 1000000 };           // nanoseconds

extern int errno;

/****  initialize  --  Canonical init method.  There is a lot to do.  ****/

void Clowder::initialize( Byte *clowderSpaceGivenTargetAddress, Byte *interruptInformationSpaceGivenTargetAddress )
{
    int clowder_fd = 0;
#ifndef PIXIE_SCHEMES
    int interruptCounter = 0;
    int interruptInformation_fd = 0;
#endif
    LongWord writeResult = 0;
    Word aKitten = 0;
    int mmapFlags = 0;
    int openFlags = 0;
    UnsignedLongWord counter = 0;
    Word aZero = 0;

    if( theNumberOfKittens > ABSOLUTE_MAX_KITTENS ) {
        fatalError( 122 );
        return;
        }
    if( theKittenNumber >= theNumberOfKittens  ) {
        useFatalErrorOutput();
        fprintf( stderr, "%d > %d", theKittenNumber, theNumberOfKittens );
        fatalError( 121 );
        return;
        }

    // Set flags for mmap according to whether or not we are the MomCat.

    if( isMomCatP ) {
        mmapFlags = MOMCAT_MMAP_FLAGS;
        openFlags = MOMCAT_MMAP_CREATE_FLAGS;
        }
    else {
        mmapFlags = KITTEN_MMAP_FLAGS;
        openFlags = KITTEN_MMAP_CREATE_FLAGS;
        }

    // Setting up the file for swapping.

    snprintf( clowderFileName, MAX_PATHNAME_LENGTH, "%s/", appSupDirName );

    strncat( clowderFileName,
#ifdef USE_COCOA_IO
#ifdef PIXIE_SCHEMES
             "Pixie Scheme II Clowder", MAX_PATHNAME_LENGTH - strlen( clowderFileName )
#else
             "Wraith Scheme Clowder", MAX_PATHNAME_LENGTH - strlen( clowderFileName )
#endif
#else
             "Wraith Scheme Unix Clowder", MAX_PATHNAME_LENGTH - strlen( clowderFileName )
#endif
             );

    if( isMomCatP )
        unlink( clowderFileName );

    clowder_fd = open( clowderFileName, openFlags, 0666 );
    if( clowder_fd == -1 ) {
        fatalError( 132 );
        return;
        }

    // If we are the MomCat we must actually write something to the file.

    if( isMomCatP ) {
        for( counter = 0; counter < sizeof( clowderFileContent ); ++counter ) {
            writeResult = write( clowder_fd, &aZero, 1 );
            if( writeResult == -1 ) {
                fprintf( stderr, "Writing to clowder file: %s\n", strerror( errno ) );
                fatalError( 130 );
                return;
                }
            }
        }

    if( (long)(sizeof( clowderFileContent )) > CLOWDER_RESERVE_SIZE ) {
        fatalError( 157 );
        return;
        }

    // MMap the clowder file content.  All this does is set a pointer to the block of mmapped
    // space, which is presently all filled with zeros.  Coincidentally, zeros are the correct
    // intialization values for most of the clowder file content.

    thisClowderFileContent = (clowderFileContent *)mmap( clowderSpaceGivenTargetAddress, sizeof( clowderFileContent ),
                                                         MMAP_PROT, mmapFlags, clowder_fd, 0 );
    if( thisClowderFileContent == (void *)(-1) ) {
        fprintf( stderr, "Mapping clowder space: %s.\n", strerror( errno ) );
        fatalError( 131);
        return;
        }
    else if( (Byte *)thisClowderFileContent != clowderSpaceGivenTargetAddress ) {
        fprintf( stderr, "mmap failed for clowder space address of 0x%016jx -- got 0x%016jx\n",
               (intmax_t)clowderSpaceGivenTargetAddress, (intmax_t)thisClowderFileContent );
        fatalError( 131);
        return;
        }
    close( clowder_fd );


#ifndef PIXIE_SCHEMES

    // Similarly to the clowder file, set up an mmapped file for information about the Wraith Scheme
    // interrupt system.

    snprintf( interruptInformationFileName, MAX_PATHNAME_LENGTH, "%s/", appSupDirName );

    strncat( interruptInformationFileName,
#ifdef USE_COCOA_IO
             "Wraith Scheme Interrupt Information", MAX_PATHNAME_LENGTH - strlen( interruptInformationFileName )
#else
             "Wraith Scheme Unix Interrupt Information", MAX_PATHNAME_LENGTH - strlen( interruptInformationFileName )
#endif
             );

    if( isMomCatP )
        unlink( interruptInformationFileName );

    interruptInformation_fd = open( interruptInformationFileName, openFlags, 0666 );
    if( interruptInformation_fd == -1 ) {
        fatalError( 145 );
        return;
        }
    if( isMomCatP ) {
        for( counter = 0; counter < sizeof( interruptInformationFileContent ); ++counter ) {
            writeResult = write( interruptInformation_fd, &aZero, 1 );
            if( writeResult == -1 ) {
                fprintf( stderr, "Writing to interruptInformation file: %s\n", strerror( errno ) );
                fatalError( 163 );
                return;
                }
            }
        }

    if( (long)(sizeof( interruptInformationFileContent )) > INTERRUPT_INFORMATION_RESERVE_SIZE ) {
        fatalError( 165 );
        return;
        }
    thisInterruptInformationFileContent
        = (interruptInformationFileContent *)mmap( interruptInformationSpaceGivenTargetAddress,
                                                   sizeof( interruptInformationFileContent ),
                                                   MMAP_PROT,
                                                   mmapFlags,
                                                   interruptInformation_fd,
                                                   0
                                                   );
    if( thisInterruptInformationFileContent == (void *)(-1) ) {
        fprintf( stderr, "Mapping interrupt information space: %s.\n", strerror( errno ) );
        fatalError( 168 );
        return;
        }
    else if( (Byte *)thisInterruptInformationFileContent != interruptInformationSpaceGivenTargetAddress ) {
        fprintf( stderr, "mmap failed for interrupt information space address of 0x%016lx -- got 0x%016lx\n",
               (UnsignedLongWord)interruptInformationSpaceGivenTargetAddress, (UnsignedLongWord)thisInterruptInformationFileContent );
        fatalError( 168 );
        return;
        }
    close( interruptInformation_fd );

#endif

    // It is the MomCat who is responsible for actually initializing the content of
    // those mmapped files.  Recall that much of the content is properly initialized to
    // zeros; here is where we put in the non-zero stuff.

    if( isMomCatP ) {

        memset( (void *)thisClowderFileContent, 0, sizeof( clowderFileContent ) );

        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) {

            ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedObListBin = -1;
            ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedTLEBin = -1;
            ((thisClowderFileContent->kittenData)[ aKitten ]).clowderKIQ = empty;
            ((thisClowderFileContent->kittenData)[ aKitten ]).genSymNumberForKitten = 23000 + aKitten;
      
            }
    
#ifndef PIXIE_SCHEMES
        for( interruptCounter = 0; interruptCounter < NUMBER_OF_INTERRUPTS; ++interruptCounter )
            (thisClowderFileContent->interruptBlock)[ interruptCounter ] = 0;
#endif

        thisClowderFileContent->scratch1 = empty;
        thisClowderFileContent->scratch2 = empty;

        // Here we initialize some GLOBAL Scheme machine registers.  The main memory is
        // not yet up and running yet, so what is going on here is that we are putting
        // in recognizable init values, so that if the code which later loads in the
        // actual values -- after main memory initialization -- is broken, we can
        // more easily recognize the problem at that time.

        (thisClowderFileContent->clowderTopLevelE).t    = shortFixnumBitsMask;
        (thisClowderFileContent->clowderTopLevelE).u.c  = 0xdeadbeef;
        (thisClowderFileContent->clowderIVR).t          = shortFixnumBitsMask;
        (thisClowderFileContent->clowderIVR).u.c        = 0xd0d0d0d0;
        (thisClowderFileContent->clowderObList).t       = shortFixnumBitsMask;
        (thisClowderFileContent->clowderObList).u.c     = 0xd0bed0be;
        (thisClowderFileContent->clowderTLC).t          = shortFixnumBitsMask;
        (thisClowderFileContent->clowderTLC).u.c        = 0xfeedbeef;
    
        thisClowderFileContent->GCRequired = FALSE;
        thisClowderFileContent->interruptsDisabledGlobally = FALSE;
        
#ifndef PIXIE_SCHEMES
        memset( (void *)thisInterruptInformationFileContent, 0, sizeof( interruptInformationFileContent ) );

        for( counter = 0; counter < NUMBER_OF_INTERRUPTS; ++counter )
            (thisInterruptInformationFileContent->interruptRequested)[ counter ] = FALSE;
#endif
        }

    globalIVR       = &(thisClowderFileContent->clowderIVR);
    globalTopLevelE = &(thisClowderFileContent->clowderTopLevelE);
    globalObList    = &(thisClowderFileContent->clowderObList);
    globalTLC       = &(thisClowderFileContent->clowderTLC);
    globalKIQ       = &(((thisClowderFileContent->kittenData)[ theKittenNumber ]).clowderKIQ);

    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenPID  = getpid();
#ifndef PIXIE_SCHEMES
    (thisInterruptInformationFileContent->KittenPID)[ theKittenNumber ] = getpid();
#endif

    thisClowderFileContent->nextFreeMMappableSpace = schemeMMapReserve;
    if( ((UnsignedLongWord)(thisClowderFileContent->nextFreeMMappableSpace)) % PAGE_SIZE != 0 ) {
        while( ((UnsignedLongWord)(thisClowderFileContent->nextFreeMMappableSpace)) % PAGE_SIZE != 0 ) {
            ++(thisClowderFileContent->nextFreeMMappableSpace);
            }
        }
        
    cachedLiveAddressStackIndex = 0;

    OSMemoryBarrier();

    // TESTS HERE
    }


/****  initializeByMalloc  --  Canonical init method.  There is a lot to do.  ****/

void Clowder::initializeByMalloc()
{
    int interruptCounter = 0;
    Word aKitten = 0;

    if( theNumberOfKittens > ABSOLUTE_MAX_KITTENS ) {
        fatalError( 122 );
        return;
        }
    if( theKittenNumber >= theNumberOfKittens  ) {
        useFatalErrorOutput();
        fprintf( stderr, "%d > %d", theKittenNumber, theNumberOfKittens );
        fatalError( 121 );
        return;
        }

    thisClowderFileContent = (clowderFileContent  *)calloc( sizeof( clowderFileContent ), 1 );
    if( thisClowderFileContent == 0 ) {
        fatalError( 2 );
        return;
        }

    // NOTE:  No interrupt system, mmapped file would be created here...


    // It is the MomCat who is responsible for actually initializing the content of
    // those mmapped files.  Recall that much of the content is properly initialized to
    // zeros; here is where we put in the non-zero stuff.

    memset( (void *)thisClowderFileContent, 0, sizeof( clowderFileContent ) );

    for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) {

        ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedObListBin = -1;
        ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedTLEBin = -1;
        ((thisClowderFileContent->kittenData)[ aKitten ]).clowderKIQ = empty;
        ((thisClowderFileContent->kittenData)[ aKitten ]).genSymNumberForKitten = 23000 + aKitten;
  
        }
    
    for( interruptCounter = 0; interruptCounter < NUMBER_OF_INTERRUPTS; ++interruptCounter )
        (thisClowderFileContent->interruptBlock)[ interruptCounter ] = 0;

    thisClowderFileContent->scratch1 = empty;
    thisClowderFileContent->scratch2 = empty;

    // Here we initialize some GLOBAL Scheme machine registers.  The main memory is
    // not yet up and running yet, so what is going on here is that we are putting
    // in recognizable init values, so that if the code which later loads in the
    // actual values -- after main memory initialization -- is broken, we can
    // more easily recognize the problem at that time.

    (thisClowderFileContent->clowderTopLevelE).t    = shortFixnumBitsMask;
    (thisClowderFileContent->clowderTopLevelE).u.c  = 0xdeadbeef;
    (thisClowderFileContent->clowderIVR).t          = shortFixnumBitsMask;
    (thisClowderFileContent->clowderIVR).u.c        = 0xd0d0d0d0;
    (thisClowderFileContent->clowderObList).t       = shortFixnumBitsMask;
    (thisClowderFileContent->clowderObList).u.c     = 0xd0bed0be;
    (thisClowderFileContent->clowderTLC).t          = shortFixnumBitsMask;
    (thisClowderFileContent->clowderTLC).u.c        = 0xfeedbeef;
    
    thisClowderFileContent->GCRequired = FALSE;
    thisClowderFileContent->interruptsDisabledGlobally = FALSE;
        
    globalIVR       = &(thisClowderFileContent->clowderIVR);
    globalTopLevelE = &(thisClowderFileContent->clowderTopLevelE);
    globalObList    = &(thisClowderFileContent->clowderObList);
    globalTLC       = &(thisClowderFileContent->clowderTLC);
    globalKIQ       = &(((thisClowderFileContent->kittenData)[ theKittenNumber ]).clowderKIQ);

    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenPID  = getpid();

    // NOTE:  No foreign function interface, reserve would be set here.

    cachedLiveAddressStackIndex = 0;

    OSMemoryBarrier();

    // TESTS HERE
    }


// The "coloring bits" are used for determining whether a list is proper.
// See their use in SECDMachine.c++

// In a 64-bit implementation, there are (heap size in bytes)/16 possible
// taggedAval locations, and a byte has enough bits to flag eight of them.
// So we need (memorySize / 128) + 1 bytes.

/**** allocateBitsForProper -- originally, the only use of these bits was for finding out whether a list was "proper".  ****/

void Clowder::allocateBitsForProper( UnsignedLongWord memorySize )
{
    nColoringBits = (size_t)(memorySize / 128) + 1;

    if( (coloringBits = (Byte *)calloc( (size_t)nColoringBits, 1 )) == 0 )
        fatalError( 160 );
    }

/**** setColoringBitN ****/

void Clowder::setColoringBitN( UnsignedLongWord n ) { coloringBits[ n >> 3 ] |= (1 << (n & 7)); }

/**** clearColoringBitN ****/

void Clowder::clearColoringBitN( UnsignedLongWord n ) { coloringBits[ n >> 3 ] &= ~(1 << (n & 7)); }

/**** getColoringBitN ****/

UnsignedLongWord Clowder::getColoringBitN( UnsignedLongWord n ) { return (coloringBits[ n >> 3 ] >> (n & 7) & 1); }

/**** clearColoringBits ****/

void Clowder::clearColoringBits() { bzero( coloringBits, nColoringBits ); }


// Stuff for the kitten input queue pointers.

/**** getKIQPointer ****/

taggedAval *Clowder::getKIQPointer( UnsignedLongWord n ) { return( &(((thisClowderFileContent->kittenData)[ n ]).clowderKIQ) ); }

/**** spliceOutKIQPointer ****/

void Clowder::spliceOutKIQPointer( UnsignedLongWord n )
{
    ((thisClowderFileContent->kittenData)[ n ]).clowderKIQ = spliceOutNotCarForwardingPointers( ((thisClowderFileContent->kittenData)[ n ]).clowderKIQ );
    }

/**** complainAbout  --  auxiliary routine for debugging blocking functions, and flags that enable/disable it.  ****/

//Boolean complain = TRUE;
Boolean complain = FALSE;

void complainAbout( UnsignedWord location, volatile taggedAval *tp )
{
    tprintf4( &CurrentOutputPort, MiscStrings[ 416 ], (Word)theKittenNumber, location );
    printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)tp );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)((tp->u).a) );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    theHeap->describeLocksInCurrentMainSpace();
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    }


// Routines for locking particular bins in the top-level environment hash table.

/**** lockTLEBin ****/

Boolean Clowder::lockTLEBin( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return TRUE;

    Word loopDelay = 1;
    Word theByte = SYMBOL_HASH_BYTE_FOR_BIT( n );
    Word theBit  = SYMBOL_HASH_BIT_IN_BYTE( n );

    while( OSAtomicTestAndSet( theBit, (void *)(thisClowderFileContent->TLEBinLockBits + theByte) ) ) {
        if( (++loopDelay & LOOP_MASK) == LOOP_MASK )
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
        nanosleep( &TLE_BLOCK_DELAY_struct, 0 );
        }
    setKittenHasLockedTLEBin( theKittenNumber, n );
    return TRUE;
    }

/**** unlockTLEBin ****/

void Clowder::unlockTLEBin( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return;

    Word theByte = SYMBOL_HASH_BYTE_FOR_BIT( n );
    Word theBit  = SYMBOL_HASH_BIT_IN_BYTE( n );

    OSAtomicTestAndClear( theBit, (void *)(thisClowderFileContent->TLEBinLockBits + theByte) );
    setKittenHasLockedTLEBin( theKittenNumber, -1 );
    }

/**** TLEBinLockBit ****/

Boolean Clowder::TLEBinLockBit( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return FALSE;

    Byte theBit =   ((thisClowderFileContent->TLEBinLockBits)[ SYMBOL_HASH_BYTE_FOR_BIT( n ) ]) 
                  & (0x80 >> SYMBOL_HASH_BIT_IN_BYTE( n ));

    return( theBit != 0 );
    }


// Routines for locking particular bins in the oblist, which is implemented as a hash table.

/**** lockObListBin ****/

Boolean Clowder::lockObListBin( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return TRUE;

    Word loopDelay = 1;
    Word theByte = SYMBOL_HASH_BYTE_FOR_BIT( n );
    Word theBit  = SYMBOL_HASH_BIT_IN_BYTE( n );

    while( OSAtomicTestAndSet( theBit, (void *)(thisClowderFileContent->ObListBinLockBits + theByte) ) ) {
        if( (++loopDelay & LOOP_MASK) == LOOP_MASK )
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
        nanosleep( &OBLIST_BLOCK_DELAY_struct, 0 );
        }
    setKittenHasLockedObListBin( theKittenNumber, n );
    return TRUE;
    }
 
/**** unlockObListBin ****/

void Clowder::unlockObListBin( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return;

    Word theByte = SYMBOL_HASH_BYTE_FOR_BIT( n );
    Word theBit  = SYMBOL_HASH_BIT_IN_BYTE( n );

    OSAtomicTestAndClear( theBit, (void *)(thisClowderFileContent->ObListBinLockBits + theByte) );
    setKittenHasLockedObListBin( theKittenNumber, -1 );
    }

/**** obListBinLockBit ****/

Boolean Clowder::obListBinLockBit( uint32_t n )
{
    if( theNumberOfKittens == 1 )
        return FALSE;

    Byte theBit =   ((thisClowderFileContent->ObListBinLockBits)[ SYMBOL_HASH_BYTE_FOR_BIT( n ) ]) 
                  & (0x80 >> SYMBOL_HASH_BIT_IN_BYTE( n ));

    return( theBit != 0 );
    }


/**** lockObListCount ****/

Boolean Clowder::lockObListCount()
{
    if( theNumberOfKittens == 1 )
        return TRUE;

    while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->oblistCountLock)) ) )
        nanosleep( &OBLIST_COUNT_DELAY_struct, 0 );
    return TRUE;
    }
 
/**** unlockObListCount ****/

void Clowder::unlockObListCount()
{
    if( theNumberOfKittens == 1 )
        return;

    OSAtomicTestAndClear( 0, (void *)(&(thisClowderFileContent->oblistCountLock)) );    
    }

/**** obListCountLock ****/

Boolean Clowder::obListCountLock()
{
    if( theNumberOfKittens == 1 )
        return FALSE;

    return( thisClowderFileContent->oblistCountLock != 0 );
    }

/**** getObListCount ****/

UnsignedLongWord Clowder::getObListCount() { return thisClowderFileContent->oblistCount; }

/**** getObListSize ****/

UnsignedLongWord Clowder::getObListSize() { return thisClowderFileContent->oblistSize; }

/**** lockGCBytes ****/

Boolean Clowder::lockGCBytes()
{
    if( theNumberOfKittens == 1 )
        return TRUE;

    while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->gCBytesLock)) ) )
        nanosleep( &GC_BYTE_DELAY_struct, 0 );
    return TRUE;
    }

/**** unlockGCBytes ****/

void Clowder::unlockGCBytes()
{
    if( theNumberOfKittens == 1 )
        return;

    OSAtomicTestAndClear( 0, (void *)(&(thisClowderFileContent->gCBytesLock)) );    
    }

/**** gCBytesLock ****/

Boolean Clowder::gCBytesLock()
{
    if( theNumberOfKittens == 1 )
        return FALSE;

    return( thisClowderFileContent->gCBytesLock );
    }

/**** setGCBytesValid ****/

void Clowder::setGCBytesValid() { thisClowderFileContent->gCBytesValid = TRUE; }

/**** clearGCBytesValid ****/

void Clowder::clearGCBytesValid() { thisClowderFileContent->gCBytesValid = FALSE; }

/**** gCBytesValid ****/

Boolean Clowder::gCBytesValid() { return( thisClowderFileContent->gCBytesValid ); }

/**** setGCBytes ****/

void Clowder::setGCBytes( UnsignedLongWord nBytes ) { thisClowderFileContent->gCBytes = nBytes; }

/**** getGCBytes ****/

UnsignedLongWord Clowder::getGCBytes() { return thisClowderFileContent->gCBytes; }

/**** incrementObListCount ****/

void Clowder::incrementObListCount()
{
    lockObListCount();
    (thisClowderFileContent->oblistCount) += 1;
    (thisClowderFileContent->oblistSize)  += 2 * sizeof( taggedAval );
    unlockObListCount();
    }

/**** setObListCount ****/

void Clowder::setObListCount( UnsignedLongWord n )
{
    lockObListCount();
    (thisClowderFileContent->oblistCount) = n;
    (thisClowderFileContent->oblistSize) = (1 + SYMBOL_HASH_CONSTANT + n) * sizeof( taggedAval );
    unlockObListCount();
    }


// Routines for locking particular interrupts.

/**** lockInterrupt ****/

Boolean Clowder::lockInterrupt( UnsignedLongWord n )
{
    Word loopDelay = 1;

    if( theNumberOfKittens == 1 )
        return TRUE;

    while( OSAtomicTestAndSet( 0, (void *)(&((thisClowderFileContent->interruptBlock)[ n ])) ) ) {
        if( (++loopDelay & LOOP_MASK) == LOOP_MASK )
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
        nanosleep( &INTERRUPT_HANDLER_DELAY_struct, 0 );
        }
    return TRUE;
    }
 
/**** unlockInterrupt ****/

void Clowder::unlockInterrupt( UnsignedLongWord n )
{
    if( theNumberOfKittens == 1 )
        return;

    OSAtomicTestAndClear( 0, (void *)(&((thisClowderFileContent->interruptBlock)[ n ])) );
    }

/**** interruptLockBit ****/

Boolean Clowder::interruptLockBit( UnsignedLongWord n ) { return( ((thisClowderFileContent->interruptBlock)[ n ]) != 0 ); }


// Routines for releasing or resetting kitten flags and locks, for error recovery or reset.

/**** releaseMyStuff ****/

void Clowder::releaseMyStuff()
{
    if( ! isMomCatP ) {
        OSMemoryBarrier();
        ((thisClowderFileContent->kittenData)[ theKittenNumber]).KittenEvaluatorActivity = KITTEN_ACTIVITY_NOT_STARTED;
        ((thisClowderFileContent->kittenData)[ theKittenNumber ]).ObListInRegister = FALSE;
        ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenSays = KITTEN_SAYS_NOTHING;
        ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenBlock = 0;
        OSMemoryBarrier();
        }
    }

/**** resetMyStuff ****/

void Clowder::resetMyStuff()
{
    OSMemoryBarrier();
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).ObListInRegister = FALSE;
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenSays = KITTEN_SAYS_NOTHING;
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenBlock = 0;
    OSMemoryBarrier();
    }


// Routines for transmitting messages between MomCat and kittens.  Used garbage collection.

/**** setMomSaysToKitten ****/

void Clowder::setMomSaysToKitten( messageFromMom aMessage, Word aKitten )
{
    // Note that Mom does not talk to herself.

    if( isMomCatP && (aKitten > 0) && (aKitten < ABSOLUTE_MAX_KITTENS) )
        ((thisClowderFileContent->kittenData)[ aKitten ]).momSays = aMessage;
    }

/**** setMomSaysToEverybody ****/

void Clowder::setMomSaysToEverybody( messageFromMom aMessage )
{
    UnsignedWord aKitten = 0;

    // Note that Mom does not talk to herself.

    if( theKittenNumber == 0 )
        for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten )
            ((thisClowderFileContent->kittenData)[ aKitten ]).momSays = aMessage;
    }

/**** getMomSays ****/

volatile messageFromMom Clowder::getMomSays() { return ((thisClowderFileContent->kittenData)[ theKittenNumber ]).momSays;  }

volatile messageFromMom Clowder::getMomSays( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).momSays; }

/**** setMomSaysToKittenController ****/

void Clowder::setMomSaysToKittenController( messageFromMomToController aMessage, Word aKitten )
{
    // Note that Mom does not talk to herself.

    if( isMomCatP && (aKitten > 0) && (aKitten < ABSOLUTE_MAX_KITTENS) )
        ((thisClowderFileContent->kittenData)[ aKitten ]).momSaysToController = aMessage;
    }

/**** setMomSaysToAllKittenControllers ****/

void Clowder::setMomSaysToAllKittenControllers( messageFromMomToController aMessage )
{
    UnsignedWord aKitten = 0;

    // Note that Mom does not talk to herself.

    if( theKittenNumber == 0 )
        for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten ) {
            ((thisClowderFileContent->kittenData)[ aKitten ]).momSaysToController = aMessage;
            }
    }

/**** getMomSaysToController ****/

volatile messageFromMomToController Clowder::getMomSaysToController()
{
    return ((thisClowderFileContent->kittenData)[ theKittenNumber ]).momSaysToController;
    }

volatile messageFromMomToController Clowder::getMomSaysToController( Word aKitten )
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).momSaysToController;
    }

/**** setKittenSays ****/

void Clowder::setKittenSays( messageFromKitten aMessage ) { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenSays = aMessage; }

/**** getKittenSays ****/

volatile messageFromKitten Clowder::getKittenSays( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).kittenSays; }


void Clowder::setKittenRequestingGC( Word aKitten ) { thisClowderFileContent->kittenRequestingGC = aKitten; }

Word Clowder::getKittenRequestingGC() { return thisClowderFileContent->kittenRequestingGC; }

void Clowder::setGCErrorNumber( UnsignedLongWord errorNumber ) { thisClowderFileContent->gCErrorNumber = errorNumber; }

UnsignedLongWord Clowder::getGCErrorNumber() { return thisClowderFileContent->gCErrorNumber; }

// Kitten blocks are used to keep garbage collection from happening at certain times.
// The first is when the "coloring bits" are in use, to see whether a list is proper or not.
// The second is to keep garbage collection from happening when a kitten is waiting to exit.
// These two uses are respectively in SECDMachine.c++ and in WraithSchemeController.mm.

/**** increaseKittenBlock ****/

void Clowder::increaseKittenBlock() { ++(((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenBlock); }

/**** reduceKittenBlock ****/

void Clowder::reduceKittenBlock()
{ 
    if( ((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenBlock  <= 0 ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 141 ], theKittenNumber );
        fatalError( 146 );
        }
    --(((thisClowderFileContent->kittenData)[ theKittenNumber ]).kittenBlock);
    }

/**** getKittenBlock ****/

volatile UnsignedLongWord Clowder::getKittenBlock( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).kittenBlock; }


// The kitten deadlock mechanism is used in the blocking functions herein, and in an enhancement in 
// SchemeFunctions2.c++, to allow deadlocks in blocking to time out and create an error message, for debugging.

/**** setBreakKittenDeadlockFlag ****/

void Clowder::setBreakKittenDeadlockFlag( Word aKitten) { ((thisClowderFileContent->kittenData)[ aKitten ]).breakKittenDeadlockFlag = TRUE; }

/**** clearBreakKittenDeadlockFlag ****/

void Clowder::clearBreakKittenDeadlockFlag( Word aKitten ) { ((thisClowderFileContent->kittenData)[ aKitten ]).breakKittenDeadlockFlag = FALSE; }

/**** clearBreakKittenDeadlockFlag ****/

void Clowder::clearBreakKittenDeadlockFlag() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).breakKittenDeadlockFlag = FALSE; }

/**** getBreakKittenDeadlockFlag ****/

volatile UnsignedLongWord Clowder::getBreakKittenDeadlockFlag( Word aKitten ) 
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).breakKittenDeadlockFlag;
    }


// Functions to deal with the return value from gcSequence:

/**** setGCSequenceReturnValue ****/

void Clowder::setGCSequenceReturnValue( Boolean toThis ) { thisClowderFileContent->gcSequenceReturnValue = toThis; }

/**** getGCSequenceReturnValue ****/

Boolean Clowder::getGCSequenceReturnValue() { return thisClowderFileContent->gcSequenceReturnValue; }


// Functions to record that something went wrong when a kitten was being initialized.

/**** setKittenFailedOnInit ****/

void Clowder::setKittenFailedOnInit() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenFailedOnInit = TRUE; }

/**** clearKittenFailedOnInit ****/

void Clowder::clearKittenFailedOnInit() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenFailedOnInit = FALSE; }

/**** getKittenFailedOnInit ****/

volatile Boolean Clowder::getKittenFailedOnInit( Word aKitten) { return ((thisClowderFileContent->kittenData)[ aKitten ]).KittenFailedOnInit; }


// Functions to record that a kitten is evaluating.

/**** setKittenEvaluating ****/

void Clowder::setKittenEvaluating() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenEvaluating = TRUE; }

/**** clearKittenEvaluating ****/

void Clowder::clearKittenEvaluating() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenEvaluating = FALSE; }

/**** getKittenEvaluating ****/

volatile Boolean Clowder::getKittenEvaluating( UnsignedLongWord aKitten)
{
    return ((thisClowderFileContent->kittenData)[aKitten ]).KittenEvaluating;
    }

volatile Boolean Clowder::anyKittenEvaluating()
{
    UnsignedWord aKitten = 0;    

    for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenEvaluating( aKitten ) )
            return TRUE;
    return FALSE;
    }


// Functions to record that a kitten has locked a particular oblist bin.

/**** setKittenHasLockedObListBin ****/

void Clowder::setKittenHasLockedObListBin( Word aKitten, Word aBin )
{ 
    ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedObListBin = aBin;
    }

/**** getKittenHasLockedObListBin ****/

volatile Word Clowder::getKittenHasLockedObListBin( Word aKitten )
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedObListBin;
    }

/**** anyKittenHasLockedObListBin ****/

volatile Boolean Clowder::anyKittenHasLockedObListBin()
{
    UnsignedWord aKitten = 0;    

    for( aKitten = 0; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenHasLockedObListBin( (Word)aKitten ) != -1 )
            return TRUE;
    return FALSE;
    }


// Functions to record that a kitten has locked a particular top-level environment bin.

/**** setKittenHasLockedTLEBin ****/

void Clowder::setKittenHasLockedTLEBin( Word aKitten, Word aBin ) { ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedTLEBin = aBin; }

/**** getKittenHasLockedTLEBin ****/

volatile Word Clowder::getKittenHasLockedTLEBin( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).kittenHasLockedTLEBin; }

/**** anyKittenHasLockedTLEBin ****/

volatile Boolean Clowder::anyKittenHasLockedTLEBin()
{
    UnsignedWord aKitten = 0;    

    for( aKitten = 0; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenHasLockedTLEBin( (Word)aKitten ) != -1 )
            return TRUE;
    return FALSE;
    }


// Functions to record that a kitten has set a block from a particular location.

/**** setCurrentActiveBlock ****/

void Clowder::setCurrentActiveBlock( Word aKitten, UnsignedWord location )
{
    ((thisClowderFileContent->kittenData)[ aKitten ]).currentActiveBlock = location;
    }

/**** setCurrentActiveBlock ****/

void Clowder::setCurrentActiveBlock( UnsignedWord location )
{
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).currentActiveBlock = location;
    }

/**** getCurrentActiveBlock ****/

volatile UnsignedWord Clowder::getCurrentActiveBlock( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).currentActiveBlock; }

/**** anyCurrentActiveBlock ****/

volatile Boolean Clowder::anyCurrentActiveBlock()
{
    UnsignedWord aKitten = 0;    

    for( aKitten = 0; aKitten < theNumberOfKittens; ++aKitten )
        if( getCurrentActiveBlock( (Word)aKitten ) != 0 )
            return TRUE;
    return FALSE;
    }


// Functions to record that a kitten has set a block of a particular tagged aval.

/**** setCurrentActiveTaggedAval ****/

void Clowder::setCurrentActiveTaggedAval( Word aKitten, volatile taggedAval *tp )
{
    ((thisClowderFileContent->kittenData)[ aKitten ]).currentActiveTaggedAval = tp;
    }

/**** setCurrentActiveTaggedAval ****/

void Clowder::setCurrentActiveTaggedAval( volatile taggedAval *tp )
{
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).currentActiveTaggedAval = tp;
    }

/**** getCurrentActiveTaggedAval ****/

volatile taggedAval *Clowder::getCurrentActiveTaggedAval( Word aKitten )
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).currentActiveTaggedAval; 
    }

/**** anyCurrentActiveTaggedAval ****/

volatile Boolean Clowder::anyCurrentActiveTaggedAval()
{
    UnsignedWord aKitten = 0;    

    for( aKitten = 0; aKitten < theNumberOfKittens; ++aKitten )
        if( getCurrentActiveTaggedAval( (Word)aKitten ) != 0 )
            return TRUE;
    return FALSE;
    }


// Functions to record a kitten's progress through initialization, toward "ready to go".

/**** setKittenEvaluatorActivity ****/

void Clowder::setKittenEvaluatorActivity( kittenEvaluatorActivityLevel activityLevel )
{
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenEvaluatorActivity = activityLevel;
    }

/**** getKittenEvaluatorActivity ****/

volatile kittenEvaluatorActivityLevel Clowder::getKittenEvaluatorActivity( Word aKitten)
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).KittenEvaluatorActivity;
    }

/**** setKittenGUIActivity ****/

void Clowder::setKittenGUIActivity( kittenGUIActivityLevel activityLevel )
{
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).KittenGUIActivity = activityLevel;
    }

/**** getKittenGUIActivity ****/

volatile kittenGUIActivityLevel Clowder::getKittenGUIActivity( Word aKitten)
{
    return ((thisClowderFileContent->kittenData)[ aKitten ]).KittenGUIActivity;
    }


// Functions to record that the oblist is in register.

/**** setObListInRegister ****/

void Clowder::setObListInRegister() { ((thisClowderFileContent->kittenData)[ theKittenNumber ]).ObListInRegister = TRUE; }

/**** clearObListInRegister ****/

void Clowder::clearObListInRegister()
{
    ((thisClowderFileContent->kittenData)[ theKittenNumber ]).ObListInRegister = FALSE;
    }

/**** getObListInRegister ****/

volatile Boolean Clowder::getObListInRegister( Word aKitten) { return ((thisClowderFileContent->kittenData)[ aKitten ]).ObListInRegister; }


// Functions to record that GC is required.  These functions do not care whether the GC required is full or generational.

/**** setGCIsRequired ****/

void Clowder::setGCIsRequired() { thisClowderFileContent->GCRequired = TRUE; }

/**** clearGCIsRequired ****/

void Clowder::clearGCIsRequired() { thisClowderFileContent->GCRequired = FALSE; }

/**** GCIsRequired ****/

volatile Boolean Clowder::GCIsRequired() { return thisClowderFileContent->GCRequired; }


// Functions to record that a full GC is required.

/**** setFullGCIsRequired ****/

void Clowder::setFullGCIsRequired() { thisClowderFileContent->FullGCRequired = TRUE; }

/**** clearFullGCIsRequired ****/

void Clowder::clearFullGCIsRequired() { thisClowderFileContent->FullGCRequired = FALSE; }

/**** FullGCIsRequired ****/

volatile Boolean Clowder::FullGCIsRequired() { return thisClowderFileContent->FullGCRequired; }


// Functions to block access to kitten input queues.

/**** getQueueBlock ****/

volatile UnsignedLongWord Clowder::getQueueBlock( Word aKitten )
{
    return( (volatile UnsignedLongWord)((thisClowderFileContent->kittenData)[ aKitten ]).QueueAccessFlag );
    }

/**** getQueueBlockPointer ****/

volatile UnsignedLongWord *Clowder::getQueueBlockPointer( Word aKitten )
{
    return (volatile UnsignedLongWord *)&(((thisClowderFileContent->kittenData)[ aKitten ]).QueueAccessFlag );
    }


// More functions for kitten/MomCat message interchange.

/**** kittenWaitForNewMessage ****/

volatile messageFromMom Clowder::kittenWaitForNewMessage( messageFromMom oldMessage )
{
    messageFromMom currentMessage;

    while( (currentMessage = getMomSays()) == oldMessage )
        usleep( SPIN_ON_MESSAGE_DELAY );
    return currentMessage;
    }

/**** kittenWaitForMessage ****/

volatile messageFromMom Clowder::kittenWaitForMessage( messageFromMom aMessage )
{
    messageFromMom currentMessage = MOM_SAYS_NOTHING;

    while( (currentMessage = getMomSays()) != aMessage )
        usleep( SPIN_ON_MESSAGE_DELAY );
    return currentMessage;
    }

/**** momWaitForMessage ****/

volatile Boolean Clowder::momWaitForMessage( messageFromKitten aMessage, UnsignedWord aKitten )
{
    messageFromKitten currentMessage = KITTEN_SAYS_NOTHING;

    while(    ((currentMessage = getKittenSays( aKitten )) != aMessage)
           && (getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL) ) {
        usleep( SPIN_ON_MESSAGE_DELAY );
        ;
        }
    return( currentMessage == aMessage );
    }


// Functions whereby a kitten writes its registers to scratch space where the garbage collector can use them as roots.

/**** writeRegistersToScratch ****/

void Clowder::writeRegistersToScratch()
{
    OSMemoryBarrier();
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localE = E;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localR = R;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localA = A;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCurrentInputPort = CurrentInputPort;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCurrentOutputPort = CurrentOutputPort;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localErrorPort = ErrorPort;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localSavedE = SavedE;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localL = L;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localDefaultInputPort = DefaultInputPort;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localMasterDefaultInputPort = MasterDefaultInputPort;
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localTranscriptPort = TranscriptPort;
    OSMemoryBarrier();
    }

/**** copyRegistersFromScratch ****/

void Clowder::copyRegistersFromScratch()
{
    OSMemoryBarrier();
    E = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localE;
    R = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localR;
    A = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localA;
    CurrentInputPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCurrentInputPort;
    CurrentOutputPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCurrentOutputPort;
    ErrorPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localErrorPort;
    SavedE = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localSavedE;
    L = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localL;
    DefaultInputPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localDefaultInputPort;
    MasterDefaultInputPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localMasterDefaultInputPort;
    TranscriptPort = (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localTranscriptPort;
    OSMemoryBarrier();
    }


// Functions whereby a kitten writes its stack to scratch space where the garbage collector can use it as roots.

/**** writeStackToScratch ****/

void Clowder::writeStackToScratch()
{
    UnsignedLongWord toMinusFrom = 0;

    OSMemoryBarrier();
    theHeap->copyMemorySpacePartialContent( theHeap->SchemeStack, theHeap->ScratchStack, &toMinusFrom, (Byte *)S );
    theHeap->copyMemorySpaceMiscellanea( theHeap->SchemeStack, theHeap->ScratchStack, &toMinusFrom );
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localS = (taggedAval *)((Byte *)S + toMinusFrom);
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localF = (taggedAval *)((Byte *)F + toMinusFrom);
    OSMemoryBarrier();
    }

/**** copyStackFromScratch ****/

void Clowder::copyStackFromScratch()
{
    UnsignedLongWord toMinusFrom = 0;

    OSMemoryBarrier();
    theHeap->copyMemorySpacePartialContent( theHeap->ScratchStack, theHeap->SchemeStack, &toMinusFrom,
                                            ((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localS) );
    S = (taggedAval *)((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localS + toMinusFrom);
    F = (taggedAval *)((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localF + toMinusFrom);
    OSMemoryBarrier();
    }


// Functions whereby a kitten writes its continuation to scratch space where the garbage collector can use it as roots.

/**** writeContinuationToScratch ****/

void Clowder::writeContinuationToScratch()
{
    UnsignedLongWord toMinusFrom = 0;

    OSMemoryBarrier();
    theHeap->copyMemorySpacePartialContent( theHeap->SchemeContinuation, theHeap->ScratchStack, &toMinusFrom, (Byte *)CS );
    theHeap->copyMemorySpaceMiscellanea( theHeap->SchemeContinuation, theHeap->ScratchStack, &toMinusFrom );
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCS = (taggedAval *)((Byte *)CS + toMinusFrom);
    (((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCF = (taggedAval *)((Byte *)CF + toMinusFrom);
    OSMemoryBarrier();
    }

/**** copyContinuationFromScratch ****/

void Clowder::copyContinuationFromScratch()
{
    UnsignedLongWord toMinusFrom = 0;

    OSMemoryBarrier();
    theHeap->copyMemorySpacePartialContent( theHeap->ScratchStack, theHeap->SchemeContinuation, &toMinusFrom,
                                            ((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCS) );
    CS = (taggedAval *)((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCS + toMinusFrom);
    CF = (taggedAval *)((Byte *)(((thisClowderFileContent->kittenData)[ theKittenNumber ]).localRegisters).localCF + toMinusFrom);
    OSMemoryBarrier();
    }


// More functions to use when the kittens are receiving messages from the MomCat and responding to them.

/**** kittenHandleGCSequenceMessages -- kitten handles gcSequence messages till the gcSequence is over ****/

Boolean Clowder::kittenHandleGCSequenceMessages()
{
    Boolean startMessageReceived = FALSE;
    Word unfortunateKitty = 0;

    messageFromMom currentMessage = getMomSays();

    while( TRUE ) {       // Act on current message, wait for a different one.
        switch( currentMessage ) {
            case MOM_SAYS_NOTHING:                    // No action.
            case MOM_SAYS_NO_MORE_MESSAGES:           // No action -- is intended for "kittenHandleMessages", rename it.  FIXME
                break;
            case MOM_SAYS_GC_SEQUENCE_STARTING:
                startMessageReceived = TRUE;
                setKittenSays( KITTEN_ACKNOWLEDGES_GC_SEQUENCE_STARTING );
                break;
            case MOM_SAYS_FULL_GC:
                setKittenSays( KITTEN_ACKNOWLEDGES_FULL_GC_STARTING );
                flushCachedAddresses();
                kittenGCSequence( FALSE );
                setKittenSays( KITTEN_ACKNOWLEDGES_FULL_GC_ENDING );
                break;     
            case MOM_SAYS_GENERATIONAL_GC:
                setKittenSays( KITTEN_ACKNOWLEDGES_GENERATIONAL_GC_STARTING );
                flushCachedAddresses();
                kittenGCSequence( TRUE );
                setKittenSays( KITTEN_ACKNOWLEDGES_GENERATIONAL_GC_ENDING );
                break;     
            case MOM_SAYS_GC_HAS_FAILED:
                setKittenSays( KITTEN_ACKNOWLEDGES_GC_HAS_FAILED );
                resetError( getGCErrorNumber() );
                break;       // Not reached.
            case MOM_SAYS_RESET_FOR_LOW_MEMORY:
//                setKittenSays( KITTEN_ACKNOWLEDGES_RESET_FOR_LOW_MEMORY );
                unfortunateKitty = getKittenRequestingGC();
                if( unfortunateKitty == 0 )
                    resetError( 326 );
                else {
                    useErrorOutput();
                    tprintf3( &CurrentOutputPort, MiscStrings[ 141 ], unfortunateKitty );
                    resetError( 327 );
                    }
                break;       // Not reached.
          
            case MOM_SAYS_GC_SEQUENCE_ENDING:
                if( startMessageReceived ) {
                    setKittenSays( KITTEN_ACKNOWLEDGES_GC_SEQUENCE_ENDING );

                    return getGCSequenceReturnValue();    // This is the only exit from kitten-message-handling.

                    }
                break;       // Not reached.

            default:
                fatalError( 187 );
                break;     
            }
        currentMessage = kittenWaitForNewMessage( currentMessage );
        }                
    return FALSE;  // Not reached.
    }

/**** kittenHandleMessages -- kittens handle messages till Mom tells them to stop. ****/

void Clowder::kittenHandleMessages()
{
    messageFromMom currentMessage = getMomSays();
    Word unfortunateKitty = 0;

    while( TRUE ) {       // Act on current message, wait for a different one.
        switch( currentMessage ) {
            case MOM_SAYS_NOTHING:                    // No action.
            case MOM_SAYS_FULL_GC:                    // No action.
            case MOM_SAYS_GENERATIONAL_GC:            // No action.
            case MOM_SAYS_GC_HAS_FAILED:              // No action.
                break;
            case MOM_SAYS_EXIT:
                exitTopLoop();
                break;
            case MOM_SAYS_WRITE_STACK_TO_SCRATCH:
                writeStackToScratch();
                setKittenSays( KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
                break;
            case MOM_SAYS_COPY_STACK_FROM_SCRATCH:
                copyStackFromScratch();
                setKittenSays( KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
                break;
            case MOM_SAYS_WRITE_REGISTERS_TO_SCRATCH:
                writeRegistersToScratch();
                setKittenSays( KITTEN_HAS_WRITTEN_REGISTERS_TO_SCRATCH );
                break;
            case MOM_SAYS_COPY_REGISTERS_FROM_SCRATCH:
                copyRegistersFromScratch();
                setKittenSays( KITTEN_HAS_COPIED_REGISTERS_FROM_SCRATCH );
                break;
            case MOM_SAYS_SWAP_SPACES:
                theHeap->swapSpaces();
                setKittenSays( KITTEN_HAS_SWAPPED_SPACES );
                break;
            case MOM_SAYS_SWAP_GENERATION_SPACES:
                theHeap->swapGenerationSpacesKitten( );
                setKittenSays( KITTEN_HAS_SWAPPED_GENERATION_SPACES );
                break;
            case MOM_SAYS_CLEAN_UP_PORTS:
                theHeap->cleanUpPorts();
                setKittenSays( KITTEN_HAS_CLEANED_UP_PORTS );
                break;
            case MOM_SAYS_CLEAR_COLOR_BITS:
                clearColoringBits();
                setKittenSays( KITTEN_HAS_CLEARED_COLOR_BITS );
                break;
            case MOM_SAYS_MOM_ACKNOWLEDGES:           // No action.
                break;
            case MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH:
                writeContinuationToScratch();
                setKittenSays( KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
                break;
            case MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH:
                copyContinuationFromScratch();
                setKittenSays( KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
                break;

            case MOM_SAYS_RESET_FOR_LOW_MEMORY:
//                setKittenSays( KITTEN_ACKNOWLEDGES_RESET_FOR_LOW_MEMORY );
                unfortunateKitty = getKittenRequestingGC();
                if( unfortunateKitty == 0 )
                    resetError( 326 );
                else {
                    useErrorOutput();
                    tprintf3( &CurrentOutputPort, MiscStrings[ 141 ], unfortunateKitty );
                    resetError( 327 );
                    }
                break;       // Not reached.

            case MOM_SAYS_NO_MORE_MESSAGES:
                setKittenSays( KITTEN_ACKNOWLEDGES_NO_MORE_MESSAGES );

                return;                               // This is the only exit from kitten-message-handling.

                break;
            default:
                fatalError( 144 );
                break;
            }
        currentMessage = kittenWaitForNewMessage( currentMessage );
        }
    }

/**** momDialogWithKitten ****/

void Clowder::momDialogWithKitten( UnsignedWord aKitten, messageFromMom momSays, messageFromKitten kittenReplies )
{
    if( getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
        setMomSaysToKitten( momSays, aKitten );
        momWaitForMessage( kittenReplies, aKitten );
        }
    }

/**** momDialogWithEverybody -- but Mom doesn't talk to herself. ****/

void Clowder::momDialogWithEverybody( messageFromMom momSays, messageFromKitten kittenReplies )
{
    UnsignedWord aKitten = 0;
    
    for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten ) {
        if( getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            setMomSaysToKitten( momSays, aKitten );
            momWaitForMessage( kittenReplies, aKitten );
            }
        }
    }

/**** momReceiveFromEverybody -- but Mom doesn't listen to herself. ****/

void Clowder::momReceiveFromEverybody( messageFromKitten kittenReplies )
{
    UnsignedWord aKitten = 0;
    
    for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            momWaitForMessage( kittenReplies, aKitten );
    }

// Check whether kittens and/or MomCat are in a particular state with respect to initialization completeness or GC.

/**** allKittensActiveAt ****/

volatile Boolean Clowder::allKittensActiveAt( kittenEvaluatorActivityLevel theLevel)
{
    UnsignedWord aKitten = 0;

    for( aKitten = 0; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenEvaluatorActivity( aKitten ) < theLevel ) {
            return FALSE;
            }
    return TRUE;
    }

/**** momActiveAt ****/

volatile Boolean Clowder::momActiveAt( kittenEvaluatorActivityLevel theLevel)
{
    return( getKittenEvaluatorActivity( 0 ) >= theLevel );
    }

/**** anyObListInRegister ****/

volatile Boolean Clowder::anyObListInRegister()
{
    UnsignedWord aKitten = 0;

    for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten )
        if( getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            if( getObListInRegister( aKitten ) ) {
                return TRUE;
                }
    return FALSE;
    }


// What a kitten must do for doing its part in GC.

/**** kittenGCSequence ****/

void Clowder::kittenGCSequence( Boolean doingGenerational )
{
    eventLoopOnce( 0 );
    reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
    reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
    reportEvent( GC_START_EVENT );
    eventLoopOnce( 0 );
    if( ! doingGenerational )
        reportEvent( GC_START_EVENT_1 );
    else
        reportEvent( GC_START_EVENT_2 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );

    kittenHandleMessages();

    if( ! doingGenerational )
        reportEvent( GC_END_EVENT );
    else
        reportEvent( GC_END_EVENT_4 );
    eventLoopOnce( 0 );
    reportEvent( GC_END_EVENT_1 );
    eventLoopOnce( 0 );
    if( ! doingGenerational )
        reportEvent( GC_END_EVENT_2 );
    else
        reportEvent( GC_END_EVENT_3 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
    reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
    }


void Clowder::setRebuildObList() { thisClowderFileContent->rebuildObList = TRUE; }

void Clowder::clearRebuildObList() { thisClowderFileContent->rebuildObList = FALSE; }

Boolean Clowder::getRebuildObList() { return thisClowderFileContent->rebuildObList; }

// Block the heap for allocation.

/**** blockOnAllocateLock ****/

void Clowder::blockOnAllocateLock()
{
    Word loopDelay = 0;

    if( theNumberOfKittens == 1 )
        return;

    while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->allocateLock)) ) ) {
        nanosleep( &ALLOCATE_BLOCK_DELAY_struct, 0 );
        if( loopDelay++ > LOOPS_BEFORE_GC ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            loopDelay = 0;
            }
        }
    }

/**** releaseAllocateLock ****/

void Clowder::releaseAllocateLock()
{ 
    thisClowderFileContent->allocateLock = 0;
    }


// Block the "gensym" mechanism, to avoid duplicate symbols.

/**** nextGenSymNumberForKitten ****/

UnsignedLongWord Clowder::nextGenSymNumberForKitten( UnsignedWord aKitten )
{
     UnsignedLongWord returnValue = ((thisClowderFileContent->kittenData)[ aKitten ]).genSymNumberForKitten;

     ((thisClowderFileContent->kittenData)[ aKitten ]).genSymNumberForKitten += ABSOLUTE_MAX_KITTENS;

     return returnValue;
     }


// Disable interrupts -- a global lock so the caller can do something with the interrupt table, et cetera.

/**** blockOnInterruptsDisabledGlobally ****/

void Clowder::blockOnInterruptsDisabledGlobally()
{
    Word loopDelay = 0;

    while( OSAtomicTestAndSetBarrier( 0, (void *)(&(thisClowderFileContent->interruptsDisabledGlobally)) ) ) {
        nanosleep( &INTERRUPT_BLOCK_DELAY_struct, 0 );
        if( loopDelay++ > LOOPS_BEFORE_GC ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            loopDelay = 0;
            }
        }
    }

/**** releaseInterruptsDisabledGlobally ****/

void Clowder::releaseInterruptsDisabledGlobally() { thisClowderFileContent->interruptsDisabledGlobally = FALSE; }

/**** getInterruptsDisabledGlobally ****/

Boolean Clowder::getInterruptsDisabledGlobally() { return thisClowderFileContent->interruptsDisabledGlobally; }

void Clowder::clearInterruptRequested( UnsignedLongWord n )
{
    (thisInterruptInformationFileContent->interruptRequested)[ n ] = FALSE;
 }

Boolean Clowder::getInterruptRequested( UnsignedLongWord n )
{
    return (thisInterruptInformationFileContent->interruptRequested)[ n ];
    }

void Clowder::setInterruptRequested( UnsignedLongWord n )
{
    (thisInterruptInformationFileContent->interruptRequested)[ n ] = TRUE;
    }


void Clowder::clearInterruptInstalled( UnsignedLongWord n )
{
    (thisInterruptInformationFileContent->interruptInstalled)[ n ] = FALSE;
 }

Boolean Clowder::getInterruptInstalled( UnsignedLongWord n )
{
    return (thisInterruptInformationFileContent->interruptInstalled)[ n ];
    }

void Clowder::setInterruptInstalled( UnsignedLongWord n )
{
    (thisInterruptInformationFileContent->interruptInstalled)[ n ] = TRUE;
    }


// See if a particular taggedAval is locked.

/**** lockedP ****/

Boolean lockedP( volatile Tag *tp ) { OSMemoryBarrier(); return (((*tp) & locationByteBitsMask) != 0); }


// Some assert functions related to locking, for debugging.

/**** assertLocked ****/

void assertLocked( volatile Tag *tp, UnsignedWord location, Text *message )
{
    if( ((*tp) & locationByteBitsMask) == 0 ) {
        useErrorOutput();
        tprintf6( &CurrentOutputPort, MiscStrings[ 411 ], location, message, *tp, GCInProgressFlag );
        }
    }

/**** assertNotLocked ****/

void assertNotLocked( volatile Tag *tp, UnsignedWord location, Text *message, UnsignedWord aKitten )
{
    UnsignedWord lockingKitten     = (Word)((((*tp) & kittenByteBitsMask) >> kittenByteOffset) & WORD_MASK);
    UnsignedWord lockingPlace      = (Word)((((*tp) & locationByteBitsMask) >> locationByteOffset) & WORD_MASK);
    UnsignedLongWord currentTag    = (UnsignedLongWord)*tp;

    if( lockingPlace != 0 ) {
        useErrorOutput();
        tprintf9( &CurrentOutputPort, MiscStrings[ 413 ], lockingPlace, lockingKitten,
                  location, aKitten, message, currentTag, GCInProgressFlag );
        }
    }

/**** assertCreatedByKitten ****/

void assertCreatedByKitten( volatile Tag *tp, UnsignedWord location, Text *message, UnsignedWord aKitten )
{
    UnsignedWord lockingKitten     = (Word)((((*tp) & kittenByteBitsMask) >> kittenByteOffset) & WORD_MASK);
    UnsignedWord lockingPlace      = (Word)((((*tp) & locationByteBitsMask) >> locationByteOffset) & WORD_MASK);
    UnsignedLongWord currentTag    = (UnsignedLongWord)*tp;

    if( ((*tp) & (locationByteBitsMask | kittenByteBitsMask)) == 0 ) {
        useErrorOutput();
        tprintf6( &CurrentOutputPort, MiscStrings[ 411 ], location, message, currentTag, GCInProgressFlag );
        }
    if( (lockingKitten != 0) && (lockingKitten != aKitten) ) {
        useErrorOutput();
        tprintf9( &CurrentOutputPort, MiscStrings[ 412 ], lockingKitten, lockingPlace, aKitten, location, message, currentTag, GCInProgressFlag ); 
        resetError( 296 );
        }
    }


// Force break of a deadlock in a blocking function.

/**** breakDeadlock ****/

void breakDeadlock( taggedAval *tp, Tag thePresentTag, UnsignedWord location )
{
    theClowder->clearBreakKittenDeadlockFlag();
    useErrorOutput();
    tprintf9( &CurrentOutputPort, MiscStrings[ 142 ],
              (UnsignedLongWord)((tp->u).a), (UnsignedLongWord)(thePresentTag), (UnsignedLongWord)((((tp->u).a)->u).a),
              theKittenNumber, location, getKittenByte( thePresentTag ), getLocationByte( thePresentTag ) );
    clearLocationByteNonAtomic( ((tp->u).a)->t );
    clearKittenByteNonAtomic( ((tp->u).a)->t );
    SPushR();
    R = *tp;
    DISPLAY
    SPopR();
    }

/**** lockedS0  -- Single use in SchemeFunctions4.c++.  ****/

Boolean lockedS0()
{
    OSMemoryBarrier();
    return( ((((S[0]).u.a)->t) & (locationByteBitsMask)) != 0);
    }

//  Fundamental routine for making the generational garbage collector work; use to record that we have a new
//  address in tenured space that points outside of it.
//
//  For this to work, we must guarantee in the calling code that nothing modified the taggedAval that is
//  pointed to by "addressOfInterest" while this function is working on it.

/**** addToLiveAddressStack -- addresses of interest are those in tenured space that point outside of it.  ****/

void Clowder::addToLiveAddressStack( taggedAval *addressOfInterest )
{
    memorySpace *tenured = 0;
    taggedAval *whereAddressOfInterestPoints = 0;

    if( ! usingGenerationalGC )
        return;

    // If the address of interest is a leaf( first clause), or is already stacked (second clause), or is not
    //   in tenured space (second if), forget it.

    if( leaf_or_points_to_active_p( addressOfInterest->t ) )
//    if( leaf_p( addressOfInterest->t ) || points_to_active_p( addressOfInterest->t ) )
        return;

    tenured = (memorySpace *)(theHeap->getCurrentMainSpace());
    if( ! inSpaceP( addressOfInterest, tenured ) )
        return;

    whereAddressOfInterestPoints = (addressOfInterest->u).a;

    // If the address of interest points into tenured space, forget it.
    //   (We are only interested in pointers that point outside of tenured space.)

    if( inSpaceP( whereAddressOfInterestPoints, tenured ) )
        return;
 
    if( (cachedLiveAddressStackIndex + 1) >= N_CACHED_ADDRESSES_PER_KITTEN )
        flushCachedAddresses();

    cachedLiveAddresses[ cachedLiveAddressStackIndex++ ] = addressOfInterest;
    }    

/**** flush the local cache of addresses of interest to the real stack.  ****/

void Clowder::flushCachedAddresses()
{
    taggedAval *currentAddress = 0;

    if( ! usingGenerationalGC )
        return;

    if( theNumberOfKittens != 1 ) {    // Lock required.
        if( thisClowderFileContent->liveStackIsA )
            while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->gGCAddressStackALock)) ) )
                nanosleep( &GGC_BLOCK_DELAY_struct, 0 );
        else
            while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->gGCAddressStackBLock)) ) )
                nanosleep( &GGC_BLOCK_DELAY_struct, 0 );
        }

    if( thisClowderFileContent->liveStackIsA ) {
        while( cachedLiveAddressStackIndex > 0 ) {
            currentAddress = cachedLiveAddresses[ --cachedLiveAddressStackIndex ];
            if( ! points_to_active_p( currentAddress->t ) ) {
                setPointsToActiveBitNonAtomic( currentAddress->t );
                *(--(thisClowderFileContent->gGCAddressStackAPointer)) = currentAddress;
                }
            }
        }
    else {
        while( cachedLiveAddressStackIndex > 0 ) {
            currentAddress = cachedLiveAddresses[ --cachedLiveAddressStackIndex ];
            if( ! points_to_active_p( currentAddress->t ) ) {
                setPointsToActiveBitNonAtomic( currentAddress->t );
                *(--(thisClowderFileContent->gGCAddressStackBPointer)) = currentAddress;
                }
            }
        }

    if( theNumberOfKittens != 1 ) {    // Lock required.
        if( thisClowderFileContent->liveStackIsA )
            thisClowderFileContent->gGCAddressStackALock = 0;
        else
            thisClowderFileContent->gGCAddressStackBLock = 0;
        }

    }

/****************

void Clowder::addToLiveAddressStack( taggedAval *addressOfInterest )
{
    memorySpace *tenured = 0;
    taggedAval *whereAddressOfInterestPoints = 0;

    if( ! usingGenerationalGC )
        return;

    // If the address of interest is a leaf( first clause), or is already stacked (second clause), or is not
    //   in tenured space (second if), forget it.

    if( leaf_p( addressOfInterest->t ) || points_to_active_p( addressOfInterest->t ) )
        return;

    tenured = (memorySpace *)(theHeap->getCurrentMainSpace());
    if( ! inSpaceP( addressOfInterest, tenured ) )
        return;

    whereAddressOfInterestPoints = (addressOfInterest->u).a;

    // If the address of interest points into tenured space, forget it.
    //   (We are only interested in pointers that point outside of tenured space.)

    if( inSpaceP( whereAddressOfInterestPoints, tenured ) )
        return;
 
    if( theNumberOfKittens == 1 ) {    // Lock not required.
        setPointsToActiveBitNonAtomic( addressOfInterest->t );
        if( thisClowderFileContent->liveStackIsA )
            *(--(thisClowderFileContent->gGCAddressStackAPointer)) = addressOfInterest;
        else  // Live stack is B.
            *(--(thisClowderFileContent->gGCAddressStackBPointer)) = addressOfInterest;
        return;
        }
    else {                             // Many kittens, lock required.
        if( thisClowderFileContent->liveStackIsA ) {
            while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->gGCAddressStackALock)) ) )
                nanosleep( &GGC_BLOCK_DELAY_struct, 0 );
                ;
            setPointsToActiveBitNonAtomic( addressOfInterest->t );
             *(--(thisClowderFileContent->gGCAddressStackAPointer)) = addressOfInterest;
            thisClowderFileContent->gGCAddressStackALock = 0;
            }
        else {  // Live stack is B.
            while( OSAtomicTestAndSet( 0, (void *)(&(thisClowderFileContent->gGCAddressStackBLock)) ) )
                nanosleep( &GGC_BLOCK_DELAY_struct, 0 );
                ;
            setPointsToActiveBitNonAtomic( addressOfInterest->t );
             *(--(thisClowderFileContent->gGCAddressStackBPointer)) = addressOfInterest;
            thisClowderFileContent->gGCAddressStackBLock = 0;
            }            
        }
    }

****************/

/**** clearDeadAddressStack  --  empty out the address stack not in use. ****/

void Clowder::clearDeadAddressStack()
{
    if( isMomCatP ) {
        if( thisClowderFileContent->liveStackIsA )
            thisClowderFileContent->gGCAddressStackBPointer = thisClowderFileContent->gGCAddressStackBEnd;
        else
            thisClowderFileContent->gGCAddressStackAPointer = thisClowderFileContent->gGCAddressStackAEnd;
        }
    }

/**** addToDeadAddressStack  --  Priming the address stack not in use when we switch generations.
                                 Simpler than "addToLiveAddressStack", because called in controlled circumstances. ****/

void Clowder::addToDeadAddressStack( taggedAval *addressOfInterest )
{
    if( thisClowderFileContent->liveStackIsA )
        *(--(thisClowderFileContent->gGCAddressStackBPointer)) = addressOfInterest;
    else  // Live stack is B.
        *(--(thisClowderFileContent->gGCAddressStackAPointer)) = addressOfInterest;

    setPointsToActiveBitNonAtomic( addressOfInterest->t );
    }


// The major locking functions, that lock objects in the heap to avoid heap corruption due to simultaneous low-level access. ****/

/**** blockOnTagLock1 ****/

Boolean blockOnTagLock1( taggedAval *tp, UnsignedLongWord location )
{
#ifndef PIXIE_SCHEMES
    UnsignedLongWord loopCount    = 0;
    Tag thePresentTag = 0;
    Boolean result = FALSE;
    Boolean notComplained = TRUE;
    
    // Differs from blockOnTagLock1Force only in the next two lines.

//    if( theNumberOfKittens == 1 )
    if( (theNumberOfKittens == 1) || immutable_bit( ((tp->u).a)->t ) )
        return TRUE;

    theClowder->setCurrentActiveBlock( (Word)location );
    theClowder->setCurrentActiveTaggedAval( (volatile taggedAval *)tp );
    while( TRUE ) {
        thePresentTag = (volatile Tag)(((tp->u).a)->t) & (~(locationByteBitsMask));
        result =
            OSAtomicCompareAndSwapLong( thePresentTag,
                                      (  (thePresentTag & (~(kittenByteBitsMask)))
                                       | (location << locationByteOffset)
                                       | (((UnsignedLongWord)theKittenNumber) << kittenByteOffset)),
                                      (LongWord *)(&(((tp->u).a)->t)) );
        if( result )
            break;
        nanosleep( &TAG_BLOCK_DELAY_struct, 0 );

        if( (++loopCount & LOOP_MASK) == 0 ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            if( complain && notComplained ) {
                notComplained = FALSE;
                complainAbout( (Word)location, tp );
                }
            if( (loopCount >= WORRY_VALUE) && theClowder->getBreakKittenDeadlockFlag( theKittenNumber ) ) {
                breakDeadlock( tp, thePresentTag, (Word)location );
                return FALSE;
                }
            }
        }

    theClowder->clearBreakKittenDeadlockFlag();
    theClowder->setCurrentActiveBlock( 0 );
#endif
    return TRUE;
    }

/**** blockOnTagLock1Force ****/

Boolean blockOnTagLock1Force( taggedAval *tp, UnsignedLongWord location )
{
#ifndef PIXIE_SCHEMES
    UnsignedLongWord loopCount    = 0;
    Tag thePresentTag = 0;
    Boolean result = FALSE;
    Boolean notComplained = TRUE;
    
    // Differs from blockOnTagLock1 only in the next two lines.

    if( theNumberOfKittens == 1 )
//    if( (theNumberOfKittens == 1) || immutable_bit( ((tp->u).a)->t ) )
        return TRUE;

    theClowder->setCurrentActiveBlock( (Word)location );
    theClowder->setCurrentActiveTaggedAval( (volatile taggedAval *)tp );
    while( TRUE ) {
        thePresentTag = (volatile Tag)(((tp->u).a)->t) & (~(locationByteBitsMask));
        result =
            OSAtomicCompareAndSwapLong( thePresentTag,
                                        (  (thePresentTag & (~(kittenByteBitsMask)))
                                         | (location << locationByteOffset)
                                         | (((UnsignedLongWord)theKittenNumber) << kittenByteOffset)),
                                        (LongWord *)(&(((tp->u).a)->t)) );
        if( result )
            break;
        nanosleep( &TAG_BLOCK_DELAY_struct, 0 );
        if( (++loopCount & LOOP_MASK) == 0 ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            if( complain && notComplained ) {
                notComplained = FALSE;
                complainAbout( (Word)location, tp );
                }
            if( (loopCount >= WORRY_VALUE) && theClowder->getBreakKittenDeadlockFlag( theKittenNumber ) ) {
                breakDeadlock( tp, thePresentTag, (Word)location );
                return FALSE;
                }
            }
        }

    theClowder->clearBreakKittenDeadlockFlag();
    theClowder->setCurrentActiveBlock( 0 );
#endif
    return TRUE;
    }

/**** blockQOnTagLock1 ****/

Boolean blockQOnTagLock1( UnsignedWord which, UnsignedLongWord location )
{
#ifndef PIXIE_SCHEMES
    UnsignedLongWord loopCount    = 0;
    Tag thePresentTag = 0;
    Boolean result = FALSE;
    
    if( theNumberOfKittens == 1 )
        return TRUE;

    theClowder->setCurrentActiveBlock( (Word)location );
    while( TRUE ) {
        thePresentTag = ((*(theClowder->getQueueBlockPointer( which ))) & (~(locationByteBitsMask)));
        result =
            OSAtomicCompareAndSwapLong( thePresentTag,
                                        (  (thePresentTag & (~(kittenByteBitsMask)))
                                          | (location << locationByteOffset)
                                          | (((UnsignedLongWord)theKittenNumber) << kittenByteOffset)),
                                        (volatile LongWord *)(theClowder->getQueueBlockPointer( which )) );
        if( result )
            break;
        nanosleep( &TAG_BLOCK_DELAY_struct, 0 );
        if( (++loopCount & LOOP_MASK) == 0 ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            if( (loopCount >= WORRY_VALUE) && theClowder->getBreakKittenDeadlockFlag( theKittenNumber ) ) {
                theClowder->clearBreakKittenDeadlockFlag();
                useErrorOutput();
                tprintf7( &CurrentOutputPort, MiscStrings[ 343 ], which,
                          theKittenNumber, location, getKittenByte( thePresentTag ), getLocationByte( thePresentTag ) );
                clearLocationByteNonAtomic( ((*(theClowder->getKIQPointer( which ))).u.a)->t );
                clearKittenByteNonAtomic( ((*(theClowder->getKIQPointer( which ))).u.a)->t );
                return FALSE;
                }
            }
        }

    theClowder->clearBreakKittenDeadlockFlag();
    theClowder->setCurrentActiveBlock( 0 );
#endif
    return TRUE;
    }

/**** releaseTagLock1 ****/

void releaseTagLock1( volatile taggedAval *tp, UnsignedLongWord location )
{ 
    Tag thePresentTag = 0;

    if( (theNumberOfKittens == 1) || immutable_bit( tp->t ) )
        return;

    if( ! hasLocationByte( thePresentTag = tp->t ) ) {
        useErrorOutput();
        tprintf7( &CurrentOutputPort, MiscStrings[ 145 ],
                  (UnsignedLongWord)tp, (UnsignedLongWord)(thePresentTag), (UnsignedLongWord)((tp->u).a),
                  theKittenNumber, location );
        printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)tp );
        resetError( 307 );
        }

    tp->t &= ~(locationByteBitsMask);
    }

/**** releaseTagLock1Barrier ****/

void releaseTagLock1Barrier( volatile taggedAval *tp, UnsignedLongWord location )
{ 
    Tag thePresentTag = 0;

    if( (theNumberOfKittens == 1) || immutable_bit( tp->t ) )
        return;

    if( ! hasLocationByte( thePresentTag = tp->t ) ) {
        useErrorOutput();
        tprintf7( &CurrentOutputPort, MiscStrings[ 145 ],
                  (UnsignedLongWord)tp, (UnsignedLongWord)(thePresentTag), (UnsignedLongWord)((tp->u).a),
                  theKittenNumber, location );
        printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)tp );
        resetError( 307 );
        }

    OSMemoryBarrier();
    tp->t &= ~(locationByteBitsMask);
    }

/**** releaseTagLock1Force ****/

void releaseTagLock1Force( volatile taggedAval *tp, UnsignedLongWord location )
{ 
    Tag thePresentTag = 0;

    if( theNumberOfKittens == 1 )                         
        return;

    if( ! hasLocationByte( thePresentTag = tp->t ) ) {
        useErrorOutput();
        tprintf7( &CurrentOutputPort, MiscStrings[ 145 ],
                  (UnsignedLongWord)tp, (UnsignedLongWord)(thePresentTag), (UnsignedLongWord)((tp->u).a),
                  theKittenNumber, location );
        printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)tp );
        resetError( 307 );
        }

    tp->t &= ~(locationByteBitsMask);
    }

/**** releaseQueue ****/

void releaseQueue( UnsignedWord which, UnsignedLongWord location )
{ 
    Tag thePresentTag = 0;

    if( theNumberOfKittens == 1 )
        return;

    if( ! hasLocationByte( thePresentTag = *(theClowder->getQueueBlockPointer( which )) ) ) {
        useErrorOutput();
        useErrorOutput();
        tprintf5( &CurrentOutputPort, MiscStrings[ 345 ], which, theKittenNumber, location );
        resetError( 307 );
        }

    (*(theClowder->getQueueBlockPointer( which ))) &= ~(locationByteBitsMask);
    }


// Functions for accessing parameters passed to kittens in connection with messages from MomCat during garbage collection. ****/

/**** setScratch1 ****/

void Clowder::setScratch1( taggedAval *aTaggedAval )
{
    OSMemoryBarrier();
    thisClowderFileContent->scratch1 = *aTaggedAval;
    OSMemoryBarrier();
    }

/**** setScratch2 ****/

void Clowder::setScratch2( taggedAval *aTaggedAval )
{
    OSMemoryBarrier();
    thisClowderFileContent->scratch2 = *aTaggedAval;
    OSMemoryBarrier();
    }

/**** getScratch1 ****/

taggedAval Clowder::getScratch1() { return thisClowderFileContent->scratch1; }

/**** getScratch2 ****/

taggedAval Clowder::getScratch2() { return thisClowderFileContent->scratch2; }

/**** getKittenPID  --  Return the UNIX process ID of the given kitten.  ****/

Word Clowder::getKittenPID( Word aKitten ) { return ((thisClowderFileContent->kittenData)[ aKitten ]).kittenPID; }

/**** releaseClowderFiles  --  Unlink the Clowder's mmapped files, as in prior to exiting. ****/

void Clowder::releaseClowderFiles()
{
    if( isMomCatP ) {
        unlink( clowderFileName );
        unlink( interruptInformationFileName );
        }
    }


// Functions associated with the main window of a kitten becoming the frontmost window of all Wraith Scheme processes.  ****/

/**** setKittenFrontRequest ****/

void Clowder::setKittenFrontRequest( Word aKitten, Word aValue ) { ((thisClowderFileContent->kittenData)[ aKitten ]).KittenFront = aValue; }

/**** getKittenFrontRequest ****/

volatile Word Clowder::getKittenFrontRequest( Word aKitten) { return ((thisClowderFileContent->kittenData)[ aKitten ]).KittenFront; }

/**** setFrontKitten ****/

void setFrontKitten( int aKitten, int aValue ) { theClowder->setKittenFrontRequest( aKitten, aValue ); }

/**** getFrontKitten ****/

int getFrontKitten( int aKitten ) { return( theClowder->getKittenFrontRequest( aKitten ) ); }

/**** allKittensSay  --  Are all kittens transmitting the same message?  ****/

Boolean Clowder::allKittensSay( messageFromKitten aMessage )
{
    UnsignedWord aKitten = 0;

    for( aKitten = 1; aKitten < theNumberOfKittens; ++aKitten ) {
        if( ! (getKittenEvaluatorActivity( aKitten ) >= KITTEN_ACTIVITY_READY_TO_EVAL) )
            continue;
        OSMemoryBarrier();
        if( ((thisClowderFileContent->kittenData)[ aKitten ]).kittenSays != aMessage )
            return FALSE;
        }
    return TRUE;
    }


// Functions associated with access to the foreign-function mmappable space area, and to the mmapped interrupt stuff.

/**** incrementNextFreeMMapableSpace -- returns the nextFreeMMappable space location BEFORE THE INCREMENT ****/

void *Clowder::incrementNextFreeMMappableSpace( UnsignedLongWord size )
{
    Word loopDelay = 0;
    void *answer   = 0;

    while( OSAtomicTestAndSetBarrier( 0, (void *)(&(thisClowderFileContent->nextFreeMMappableSpaceLock)) ) ) {
        nanosleep( &NFMMS_BLOCK_DELAY_struct, 0 );
        if( loopDelay++ > LOOPS_BEFORE_GC ) {
            theHeap->triggerGC();  // Reduce chances that other processes will spin waiting for GC.
            loopDelay = 0;
            }
        }

    if( (thisClowderFileContent->nextFreeMMappableSpace + size) > (schemeMMapReserve + schemeMMapReserveSize) ) {
        thisClowderFileContent->nextFreeMMappableSpaceLock = 0;
        useErrorOutput();
        tprintf4( &CurrentOutputPort, MiscStrings[ 437 ], size,
                  schemeMMapReserve + schemeMMapReserveSize - thisClowderFileContent->nextFreeMMappableSpace );
        resetError( 285 );
        }

    answer = (void *)thisClowderFileContent->nextFreeMMappableSpace;

    thisClowderFileContent->nextFreeMMappableSpace += size;
    if( ((UnsignedLongWord)(thisClowderFileContent->nextFreeMMappableSpace)) % PAGE_SIZE != 0 ) {
        while( ((UnsignedLongWord)(thisClowderFileContent->nextFreeMMappableSpace)) % PAGE_SIZE != 0 ) {
            ++(thisClowderFileContent->nextFreeMMappableSpace);
            }
        }
        
    thisClowderFileContent->nextFreeMMappableSpaceLock = 0;

    return answer;
    }

volatile UnsignedLongWord Clowder::getMMappableSpaceSize()
{
    return( (schemeMMapReserve + schemeMMapReserveSize) - thisClowderFileContent->nextFreeMMappableSpace );
    }

Text *Clowder::getInterruptInformationFileName(){ return interruptInformationFileName; }

UnsignedLongWord Clowder::getInterruptInformationFileSize(){ return sizeof( interruptInformationFileContent ); }

void Clowder::initializeGGCAddressStacks( volatile memorySpace *m )
{
    if( isMomCatP ) {
        Byte *halfWay = (Byte *)(m->space +((m->end - m->space)/2));

        thisClowderFileContent->gGCAddressStackAPointer = (taggedAval **)(m->end);
        thisClowderFileContent->gGCAddressStackAEnd = (taggedAval **)(m->end);

        thisClowderFileContent->gGCAddressStackBPointer = (taggedAval **)halfWay;
        thisClowderFileContent->gGCAddressStackBEnd = (taggedAval **)halfWay;
        }
    cachedLiveAddressStackIndex = 0; 
    }

// Functions for control of and access to, the generational garbage collection address stacks.

void Clowder::swapGGCAddressStacks()
{
    if( isMomCatP ) {
       if( thisClowderFileContent->liveStackIsA )
            thisClowderFileContent->liveStackIsA = FALSE;
        else
            thisClowderFileContent->liveStackIsA = TRUE;
        }
    }

taggedAval **Clowder::getGGCLiveAddressStackPointer()
{
    if( thisClowderFileContent->liveStackIsA )
        return thisClowderFileContent->gGCAddressStackAPointer;
    else
        return thisClowderFileContent->gGCAddressStackBPointer;
    }

taggedAval **Clowder::getGGCLiveAddressStackEnd()
{
    if( thisClowderFileContent->liveStackIsA )
        return thisClowderFileContent->gGCAddressStackAEnd;
    else
        return thisClowderFileContent->gGCAddressStackBEnd;
    }

taggedAval **Clowder::getGGCDeadAddressStackPointer()
{
    if( thisClowderFileContent->liveStackIsA )
        return thisClowderFileContent->gGCAddressStackBPointer;
    else
        return thisClowderFileContent->gGCAddressStackAPointer;
    }

taggedAval **Clowder::getGGCDeadAddressStackEnd()
{
    if( thisClowderFileContent->liveStackIsA )
        return thisClowderFileContent->gGCAddressStackBEnd;
    else
        return thisClowderFileContent->gGCAddressStackAEnd;
    }

void Clowder::setLiveStackIsA() { thisClowderFileContent->liveStackIsA = TRUE; }

UnsignedLongWord Clowder::activeStackLength() { return( getGGCLiveAddressStackEnd() - getGGCLiveAddressStackPointer() ); }

// Debugging functions involving the generational garbage collection stacks.

void Clowder::checkLiveStack( const Text *message)
{
    taggedAval **scratchPointer = 0;
    taggedAval **start = getGGCLiveAddressStackPointer();
    taggedAval **stop = getGGCLiveAddressStackEnd();
    taggedAval *tp = 0;
    Boolean anyBad = FALSE;
    volatile memorySpace *tenured = theHeap->getCurrentMainSpace();   // FIXME -- this should be an inline method.

    for( scratchPointer = start; scratchPointer < stop; ++scratchPointer ) {
        tp = *scratchPointer;        
        if( ! inSpaceP( tp, tenured ) ) {
            tprintf3( &CurrentOutputPort, MiscStrings[ 470 ], message);
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tp );
            printAnyTaggedAval( &CurrentOutputPort, tp );            
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            anyBad = TRUE;
            }
        if( ! points_to_active_p( tp->t ) ) {
            tprintf3( &CurrentOutputPort, MiscStrings[ 471 ], message);
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tp );
            printAnyTaggedAval( &CurrentOutputPort, tp );            
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            anyBad = TRUE;
            }
         }
     if( anyBad )
         tprintf3( &CurrentOutputPort, MiscStrings[ 472 ], message );
     }

void Clowder::seeIfStacked( taggedAval *anAddress, const Text *message )
{
    taggedAval **scratchPointer = 0;
    taggedAval **start = getGGCLiveAddressStackPointer();
    taggedAval **stop = getGGCLiveAddressStackEnd();

    for( scratchPointer = start; scratchPointer < stop; ++scratchPointer )
        if( *scratchPointer == anAddress ) {
            tprintf2( &CurrentOutputPort, MiscStrings[ 473 ]);
            if( ! points_to_active_p( (*scratchPointer)->t ) )
                tprintf2( &CurrentOutputPort, MiscStrings[ 474 ]);
            tprintf3( &CurrentOutputPort, "marked (%s).\n", message );
            return;
            }
    tprintf2( &CurrentOutputPort, MiscStrings[ 475 ]);
    if( ! points_to_active_p( anAddress->t ) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 474 ]);
    tprintf3( &CurrentOutputPort, MiscStrings[ 476 ], message );
    }

void Clowder::seeIfOn( taggedAval *anAddress, const Text *message )
{
    if( points_to_active_p( anAddress->t ) )
        tprintf3( &CurrentOutputPort, "On (%s).\n", message );
    else
        tprintf3( &CurrentOutputPort, "Not on (%s).\n", message );
    }

//  This next function is for things like the push-buttons, which are
//  intended to cause an interrupt on the local processor.

#define SIGUSR1_NUMBER 30
#define SYSTEM_COMMAND_BUFFER_SIZE 100

void Clowder::runInterruptLocally( UnsignedLongWord interruptNumber )
{
    char systemCallBuffer[ 101 ];

    if( interruptNumber > NUMBER_OF_INTERRUPTS ) {
        tprintf3( &ErrorPort, MiscStrings[ 514 ], interruptNumber );
        resetError( 342 );
        }

    if( ! (thisInterruptInformationFileContent->interruptInstalled)[ interruptNumber ] )
        return;

    while( OSAtomicTestAndSet( 0, &((thisInterruptInformationFileContent->interruptRequested)[ interruptNumber ]) ) )
        ;
    
    snprintf( systemCallBuffer, SYSTEM_COMMAND_BUFFER_SIZE, "kill -%d %d", SIGUSR1_NUMBER,
              (thisInterruptInformationFileContent->KittenPID)[ theKittenNumber ]);

    system( systemCallBuffer );
    
    }

