/****************************************************************************************

CocoaIO.c++ -- Copyright (c) by Jay Reynolds Freeman 2006.

Simple (*ahem*) I/O routines for Wraith Scheme to run as a Cocoa
application.  This module is intended to be self-contained in that it
hides all details of Cocoa from the rest of the program.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#define SHOW_THREAD_INTERACTIONS FALSE

#import <stdlib.h>
#import <stdio.h>
#import <string.h>
#import <setjmp.h>
#import <ctype.h>
#import <sys/types.h>
#import <unistd.h>

#import "LocalTypes.h"
#import "WraithHeader.h"
#import "MemoryManager.h"
#import "ASCIITypes.h"
#import "DataTypes.h"
#import "CocoaIO.h"
#import "CocoaInterface.h"

#import "OpCodes.h"
#import "Globals.e"
#import "DataTypes.e"
#import "Events.e"
#import "PutBack.e"
#import "NumberFormats.e"
#import "Reader.e"
#import "Registers.e"
#import "SchemeAux.e"
#import "StreamHandling.e"

extern Boolean consoleOutputEnabledFlag;
extern Text *FatalErrorStrings[];
extern Text *MiscStrings[];
extern void eventLoopOnce( LongWord sleepTimeInMicroseconds );
extern jmp_buf ExitBuffer;
extern void resetError( UnsignedLongWord n );
extern fileDescriptor *transcriptFD;
extern int waitingToExit;

/**** Forward declarations: ****/

void tellController( int mID, LongWord mLong, const char *mString );
void pollControllerOnce();
void dingDong();
void fatalError( UnsignedLongWord n);
Boolean emitStatusMessage( const Text *msg );
void processDialogResponse( Text *response );
Boolean Cocoa_printfAtMostOneLine( Text *msg );
Boolean Cocoa_printf( Text *msg );
void addToInput( Text *newInput );
void setNostalgiaMode();
void clearNostalgiaMode();
void doResize();

/**** Cocoa statics: ****/

Boolean cocoaIOInitialized = FALSE;
volatile Boolean nostalgiaMode = FALSE;
Boolean doingDialog = FALSE;
Boolean collectingFatalErrorMessage = FALSE;
Boolean fatalErrorOccurred = FALSE;

/**** Cocoa input buffer statics: ****/

#define CI_BUFSIZE 2048

Text CocoaMessageBuffer[ COCOA_MB_SIZE + 1 ];
static Text fatalErrorMessage[ CI_BUFSIZE + 1 ];

static Text *CIBuffer;
static Text *CIBp;
static LongWord CIn;
static Text *bufferA = 0;
static Text *bufferB = 0;
static Text **InputBuffer = &bufferA;
static Text **nextInputBuffer = &bufferB;

static Boolean enlargeBufferOnSwap = FALSE;  // Set when input buffer has been enlarged.

static Boolean waitingForDialogResult =  FALSE;

static LongWord bufferSize = CI_BUFSIZE;

static UnsignedLongWord fatalErrorNestingLevel = 0;

/**** Externals for inter-thread communication: ****/

extern threadMessage toController;
extern threadMessage toEvaluator;

//######## Stuff for new evaluator-controller interface.

volatile int evaluatorOutboundMailboxFlag  = MAILBOX_EMPTY;
volatile int controllerOutboundMailboxFlag = MAILBOX_EMPTY;

/**** True names ... ****/

const char *messagesFromEvaluatorNames[] = {

    // System administration.

    "E_NO_MESSAGE_YET",                 // Initialization state -- no messages sent yet.
    "E_COMPLETED_INITIALIZATION",       // And ready to handle messages.
    "E_HELLO_THERE",                    // Please acknowledge your existence.
    "E_FATAL_PROBLEM",                  // Display the string in a modal dialog, and quit.
    "E_PLEASE_TIDY_UP_FOR_EXIT",        // Please flush and close files, et cetera.
    "E_TIDIED_UP",                      // Evaluator has tidied up as necessary for exit -- close ports, perhaps.
    "E_COGNITO_ERGO_SUM",               // I think, therefore ... I AM!!
    "E_MAKE_FRONT",                     // Make the main window visible, deminiaturized, and at the front.
    "E_NEW_OBLIST_SYMBOL",              // Here is a new symbol to put in the spelling-completion table.
    "E_CLEAR_OBLIST_SYMBOLS",           // Clear the spelling-completion table.

    // Simple state transitions.

    "E_EVALUATING",                     // Evaluator is evaluating; some things are not allowed now.
    "E_NOT_EVALUATING",                 // Evaluator is not evaluating.
    "E_ACKNOWLEDGED_RESET",             // Evaluator acknowledges request by controller to reset.
    "E_GARBAGE_COLLECTING",             // Evaluator is garbage-collecting.
    "E_NOT_GARBAGE_COLLECTING",         // Evaluator is not garbage-collecting.
    "E_SLEEPING",                       // Evaluator is sleeping.
    "E_WAKED_UP",                       // Evaluator has waked up.
    "E_COMPILING_DEFINES",              // Evaluator will compile defines.
    "E_NOT_COMPILING_DEFINES",          // Evaluator will not compile defines.
    "E_USING_LOW_PRECISION",            // Evaluator will display numbers with low precision in top-level loop.
    "E_USING_HIGH_PRECISION",           // Evaluator will display numbers with high precision in top-level loop.

    // Routine I/O.

    "E_OUTPUT_READY",                   // Display the string in the main window.
    "E_MESSAGE_TO_DISPLAY",             // Message to display is in the string.
    "E_DIALOG_TO_START",                // Dialog starting, prompt is in the string.
    "E_FILE_LOAD_BEGINNING",            // Opened the in file in the string.
    "E_FILE_LOAD_SUCCEEDED",            // Read in file in the string.
    "E_FILE_LOAD_FAILED",               // Couldn't find, or open, or something, the file in the string.
    "E_WORLD_LOAD_SUCCEEDED",           // Loaded the world in the string.  (Failure of a world load might be fatal, and is so reported if so.)
    "E_WORLD_LOAD_FAILED",              // Couldn't find, or open, or something, the file in the string.  (Might also be fatal!)
    "E_WORLD_SAVE_SUCCEEDED",           // Evaluator has saved a world to the file in string.
    "E_WORLD_SAVE_FAILED",              // Attempt to save world to the file in string has failed.

    // Input parsing.

    "E_EXPECT_MORE_INPUT",              // Processing multi-line input, need more (expression is not finished).
    "E_DO_NOT_EXPECT_MORE_INPUT",       // Parsing finished; interpreter is doing something else.
    "E_EXPECT_PARENS",                  // Processing multi-line input, need N (in the long right parens to finish).
    "E_EXPECT_DOUBLE_QUOTE",            // Processing multi-line input, need a double-quote (e.g., in a string).
    "E_DISCARDED_INPUT",                // Evaluator has abandoned processing of the current S-expression.

    // Status information, possibly for display.

    "E_NEW_GC_COUNT",                   // New total of full GCs is in the long.
    "E_NEW_GENERATIONAL_GC_COUNT",      // New total of generational GCs is in the long.
    "E_MEMORY_SIZE",                    // The memory size is in the long.
    "E_MEMORY_IN_USE",                  // The memory in use is in the long.
    "E_LIST_PRINT_LENGTH",              // The list print length is in the long.
    "E_LIST_PRINT_DEPTH",               // The list print depth is in the long.
    "E_VECTOR_PRINT_LENGTH",            // The vector print length is in the long.
    "E_VECTOR_PRINT_DEPTH",             // The vector print depth is in the long.
    "E_GENERATIONAL_MEMORY_SIZE",       // The generational memory size is in the long.
    "E_GENERATIONAL_MEMORY_IN_USE",     // The generational memory in use is in the long.
    "E_SENSE_LIGHT_COMMAND",            // A command for zero or more sense lights is in the long.
    "E_SENSE_LIGHT_COMMAND",            // A command for zero or more sense lights is in the long.
    "E_SENSE_LIGHT_ROTATE",             // A command for how far to rotate the sense light array is in the long.
    "E_TOTAL_MEMORY_IN_USE",            // The total memory size in use in in the long.
    "E_LEVEL_INDICATOR_0_VALUE",        // New value for level indicator 0 is in the long.
    "E_LEVEL_INDICATOR_1_VALUE",        // New value for level indicator 1 is in the long.
    "E_LEVEL_INDICATOR_2_VALUE",        // New value for level indicator 2 is in the long.
    "E_LEVEL_INDICATOR_3_VALUE",        // New value for level indicator 3 is in the long.

    // Progress bar manipulation.

    "E_INIT_PROGRESS",                  // Make the progress bar visible with an initial setting of zero.
    "E_INCREMENT_PROGRESS",             // Set the progress to the number of tenths of percents that is the long.
    "E_WRAPUP_PROGRESS",                // Make the progress bar invisible.

    "E_NUMBER_OF_MESSAGES"              // Handy for counting.
    };

const char *messagesFromControllerNames[] = {

    // System administration.

    "C_NO_MESSAGE_YET",                 // Initialization state -- no messages sent yet.
    "C_PLEASE_CLEAR_STATUS",            // Clear the status-message window.
    "C_COMPLETED_INITIALIZATION",       // And ready to handle messages.
    "C_HELLO_THERE",                    // Please acknowledge your existence.
    "C_PLEASE_TIDY_UP_FOR_EXIT",        // Please flush and close files, et cetera.
    "C_TIDIED_UP",                      // Evaluator has tidied up as necessary for exit -- close ports, perhaps.
    "C_COGNITO_ERGO_SUM",               // I think, therefore ... I AM!!

    // Simple state transitions.

    "C_PLEASE_STOP_EVALUATING",         // Evaluator please stop evaluating and return control to top-level loop.
    "C_PLEASE_COMPILE_DEFINES",         // Evaluator please compile defines.
    "C_PLEASE_DO_NOT_COMPILE_DEFINES",  // Evaluator please do not compile defines.
    "C_PLEASE_USE_LOW_PRECISION",       // Evaluator please display numbers with low precision in top-level loop.
    "C_PLEASE_USE_HIGH_PRECISION",      // Evaluator please display numbers with high precision in top-level loop.

    // Routine I/O.

    "C_INPUT_READY",                    // This input is ready for parsing.
    "C_DIALOG_RESPONSE_READY",          // Dialog ending, response is in the string.
    "C_PLEASE_LOAD_FILE",               // Please load the file in the string.
    "C_PLEASE_LOAD_WORLD",              // Please load the world in the string.
    "C_PLEASE_SAVE_WORLD",              // Please save a world to the file in string.
    "C_PLEASE_SAVE_FINAL_WORLD",        // Please reset, and then save a world; see code for resetSchemeFinal, etc..

    // Input parsing.

    "C_PLEASE_DISCARD_INPUT",           // Please discard what has thus far been read of a multi-line S-expression.

    // Status information, possibly for display.

    "C_NUMBER_OF_MESSAGES"              // Handy for counting.
    };

/**** tellController -- the main routine for sending messages to the Cocoa side of the interface. ****/

void tellController( int mID, LongWord mLong, const char *mString )
{
    static Text stringToPass[ CI_BUFSIZE + 1 ];

    if( ! GUIInitialized ) {
        fputs( mString, stderr );
        return;
        }

#if SHOW_THREAD_INTERACTIONS
    printf("E: tellController called with message \"%s\".\n", messagesFromEvaluatorNames[ mID ] ); // DEBUG
#endif

    while( evaluatorOutboundMailboxFlag == MAILBOX_FULL )
        ;

#if SHOW_THREAD_INTERACTIONS
    printf("E: tellController has the controller's mailbox\n" ); // DEBUG
#endif
        
    if( mString != 0 )
        strncpy( stringToPass, mString, CI_BUFSIZE );
    else
        stringToPass[ 0 ] = 0;
    toController.messageID = mID;
    toController.theLong   = mLong;
    toController.theString = stringToPass;

    evaluatorOutboundMailboxFlag = MAILBOX_FULL;
    doPollEvaluatorOnce();
    }
        
/**** pollControllerOnce -- the main routine for receiving messages from the Cocoa side of the interface. ****/

//  To avoid deadlocks, all this routine is supposed to do is report events.

#define COUNTER_MAX  6000

void pollControllerOnce()
{
    static int counter = 0;

    static Word newMessage;
    static LongWord newLong = 0;

    if( counter++ == COUNTER_MAX ) {
        reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
        reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
        counter = 0;
        }

    if( controllerOutboundMailboxFlag == MAILBOX_FULL ) {

        newMessage = toEvaluator.messageID;
        newLong = toEvaluator.theLong;

        switch( newMessage ) {       
    
            // System administration.
        
            case C_NO_MESSAGE_YET: break;                   // Initialization state; evaluator has not sent any messages.
            case C_PLEASE_CLEAR_STATUS:
                emitStatusMessage( MiscStrings[ 337 ] );
                break;
            case C_COMPLETED_INITIALIZATION: break;
            case C_HELLO_THERE: break;                      // No action required.
            case C_TIDIED_UP: break;
            case C_PLEASE_TIDY_UP_FOR_EXIT: break;
            case C_COGNITO_ERGO_SUM: break;
        
            // Simple state transitions.
        
            case C_PLEASE_STOP_EVALUATING: 
                reportEvent( REQUEST_RESET_EVENT );
                break;
            case C_PLEASE_COMPILE_DEFINES:
                reportEvent( COMPILE_DEFINES_EVENT );
                break;
            case C_PLEASE_DO_NOT_COMPILE_DEFINES:
                reportEvent( DO_NOT_COMPILE_DEFINES_EVENT );
                break;
            case C_PLEASE_USE_LOW_PRECISION:
                reportEvent( USE_LOW_PRECISION_EVENT );
                break;
            case C_PLEASE_USE_HIGH_PRECISION:
                reportEvent( USE_HIGH_PRECISION_EVENT );
                break;
    
            // Routine I/O.
    
            case C_INPUT_READY:
                addToInput( toEvaluator.theString );
                break;
            case C_DIALOG_RESPONSE_READY:
                processDialogResponse( toEvaluator.theString );
                break;
            case C_PLEASE_LOAD_FILE:
                break;
            case C_PLEASE_LOAD_WORLD:
                break;
            case C_PLEASE_SAVE_WORLD:
                break;
            case C_PLEASE_SAVE_FINAL_WORLD:
                reportEvent( REQUEST_RESET_EVENT_2 );
                break;
    
            // Input parsing.
    
            case C_PLEASE_DISCARD_INPUT: break;
    
            default:
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, "Evaluator received message with ID %d from controller.\n", newMessage );
                fatalError( 88 );
                break;
            }
        controllerOutboundMailboxFlag = MAILBOX_EMPTY;
        }
    }
    
/**** dingDong ****/
    
void dingDong() { fputs( "\07", stderr ); }

/**** emitStatusMessage ****/

Boolean emitStatusMessage( const Text *msg ) { tellController( E_MESSAGE_TO_DISPLAY, 0, msg ); return TRUE; }

/**** performDialogAndReturnString ****/

static Text *placeForResult = 0;
static UnsignedWord lengthOfResult = 0;

Boolean performDialogAndReturnString( const Text *prompt, Text *result, UnsignedWord resultLength )
{

    waitingForDialogResult = TRUE;

    placeForResult = result;
    lengthOfResult = resultLength;
    tellController( E_DIALOG_TO_START, 0, prompt );
    if( transcriptFD->fdOpen ) {
        fprintf( transcriptFD->fp, MiscStrings[ 505 ], prompt );
        fflush( transcriptFD->fp );
        }
    
    while( waitingForDialogResult ) {
        usleep( NEXT_CHAR_DELAY );
        pollControllerOnce();
        eventLoopOnce( 0 );
        }

    return TRUE;
    }

/**** processDialogResponse ****/

void processDialogResponse( Text *response )
{
    strncpy( placeForResult, response, lengthOfResult - 1 );
    if( transcriptFD->fdOpen ) {
        fprintf( transcriptFD->fp, MiscStrings[ 507 ], response );
        fflush( transcriptFD->fp );
        }
    waitingForDialogResult = FALSE;
    }

/**** initializeLowLevelIO -- for cocoa ****/

void initializeLowLevelIO()
{
    CocoaMessageBuffer[ 0 ] = 0;

    initializePutBack();

    bufferA = (Text *)calloc( (size_t)bufferSize, sizeof( Byte ) );
    if( bufferA == 0 )
        fatalError( 51 );

    bufferB = (Text *)calloc( (size_t)bufferSize, sizeof( Byte ) );
    if( bufferB == 0 )
        fatalError( 51 );

    CIBuffer = *nextInputBuffer;
    CIBuffer[ 0 ] = 0;
    CIBp = CIBuffer;
    CIn = 0;

    (*InputBuffer)[ 0 ] = 0;

    waitingForDialogResult =  FALSE;
    cocoaIOInitialized = TRUE;
    waitingToExit = FALSE;

    }

/**** reinitializeLowLevelIO -- for cocoa ****/

void reinitializeLowLevelIO( Text *reinitializationAction )
{
    reinitializePutBack();

    CIBp = CIBuffer;
    CIn = 0;

    if( reinitializationAction && (strlen( reinitializationAction ) > 0) ) {
        ;  // Do something clever here.
        }

    // Caution:  The mailbox stuff is initialized on *both* sides of the interface.
    //           See "WraithSchemeController.mm".

    // But it is not reinitialized at all after a reset:
    // The idea here is that the evaluator side of things has to be out of
    // critical sections when it resets, so the variables it has set are in
    // good states, and we shouldn't reset anything that the controller may 
    // have set, incase it is spinning, waiting for the evaluator to respond
    // after resetting.

    tellController( E_WAKED_UP,                   0,                        "");

    waitingForDialogResult =  FALSE;
    cocoaIOInitialized = TRUE;
//    waitingToExit = FALSE;    // Removed for test, 2010.07.05
    }

/**** emitIOInitializationMessages -- just for Cocoa ****/

void emitIOInitializationMessages()
{
    emitStatusMessage( MiscStrings[ 322 ] );
    }

/**** nextLineOrFragment ****/

//  This routine is passed a buffer, which it does not modify.  If the buffer contains
//    any newlines, the result returns a copy of the substring of the buffer from its beginning
//    up to and including the first newline -- the substring is null-terminated after the
//    newline -- and sets *pastFragment to the location in the buffer just past the
//    newline found.  It is possible that *pastFragment points to NULL, of course.
//
//  If the buffer is non-null and contains no newlines, this function returns a copy of the
//    entire buffer, and sets *pastFragment to NULL.
//
//  If the buffer is empty or NULL, this routine sets *pastFragment to NULL and returns NULL.
//
//  The returned copy is statically allocated, so it won't go away; however, subsequent
//    calls to this function will overwrite it.

Text *nextLineOrFragment( Text *buffer, Text **pastFragment )
{
    static Text scratch[ COCOA_MB_SIZE + 1 ];
    Text *pNewline = 0;
 
    if( (buffer == NULL) || (*buffer == 0) ) {                 // Null or empty buffer has no fragments to return.
        *pastFragment = NULL;
        return NULL;
        }        
    else {
        strncpy( scratch, buffer, COCOA_MB_SIZE );                  // We will be returning this copy, possibly modified.

        if( (pNewline = strchr( scratch, '\n')) != NULL ) {    // Found a newline, must null-terminate the copy after it.
            *(pNewline + 1) = 0;
            *pastFragment = strchr( buffer, '\n') + 1;         // Report where the next fragment starts (might point to '0').
            }
        else                                                   // "scratch", as is, is the final fragment.
            *pastFragment = NULL;

        return scratch;
        }
    }       

/**** Cocoa_printfAtMostOneLine ****/

#define PRINTS_PER_MD 64
#define NOSTALGIA_NORMAL_DELAY     25000 // microseconds
#define NOSTALGIA_SWAPPING_DELAY  500000 // microseconds

Boolean Cocoa_printfAtMostOneLine( Text *msg )
{
    Word counter = 0;
    LongWord lnth = ( (msg != 0) ? strlen( msg ) : 0 );
    Text oneCharString[ 2 ];
    static Word PPMDCounter = PRINTS_PER_MD;

    if( cocoaIOInitialized )  {
        if( PPMDCounter-- < 0 ) {
            reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
            reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
            PPMDCounter = PRINTS_PER_MD;
            }
        eventLoopOnce( 0 );
        if( consoleOutputEnabledFlag ) {
            if( ! nostalgiaMode ) {
                if( ! useStdinP )
                    tellController( E_OUTPUT_READY, 0, msg );
                else
                    fprintf( stdout, MiscStrings[ 201 ], msg );
                return TRUE;
                }
            else {
                oneCharString[ 1 ] = 0;
                for( counter = 0; counter < lnth; ++counter ) {
                    oneCharString[ 0 ] = msg[ counter ];
                    oneCharString[ 1 ] = 0;
                    if( ! useStdinP )
                        tellController( E_OUTPUT_READY, 0, oneCharString );
                    else
                        fprintf( stdout, "%c", oneCharString[0] );
                    if( (random() % 512) == 0 )
                        usleep( NOSTALGIA_SWAPPING_DELAY );
                    else
                        usleep( NOSTALGIA_NORMAL_DELAY );
                    }
                return TRUE;
                }
            }
        else
            return TRUE;
        }
    else
        return fprintf( stderr, MiscStrings[ 201 ], msg );
    }

/**** Cocoa_printf ****/

Boolean Cocoa_printf( Text *msg )
{
    Text *copyFromHere = msg;
    
    if( ! collectingFatalErrorMessage ) {
        while( (copyFromHere != NULL) && (*copyFromHere != 0) ) {
            Cocoa_printfAtMostOneLine( nextLineOrFragment( copyFromHere, &copyFromHere ) );
//            dingDong(); sleep( 3 );
            }
        }
    else {  // Collecting text for a fatal error message, to be output from "fatalError" subsequently.
        snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), MiscStrings[ 201 ], msg );
        }
    return TRUE;
    }

/**** swapBuffers -- auxiliary for input routines. ****/

Text *swapBuffers( LongWord *bufferLength )
{
    Text *scratchBuffer = 0;

    reportEvent( PARSER_STATUS_EVENT );
    if( ReadingString )
        reportEvent( EXPECT_DOUBLE_QUOTE_EVENT );
    else
        reportEvent( EXPECT_RPARENS_EVENT );
    while( (*InputBuffer)[0] == 0 ) {   // Wait until there is new data
        eventLoopOnce( 0 );
        theHeap->triggerGC();
        usleep( NEXT_CHAR_DELAY );
        pollControllerOnce();
        eventLoopOnce( 0 );
        }

    *bufferLength = strlen( *InputBuffer );

    if( enlargeBufferOnSwap ) {
        free( *nextInputBuffer );
        *nextInputBuffer = (Text *)calloc( (size_t)bufferSize, sizeof( Byte ) );
        if( *nextInputBuffer == 0 )
            fatalError( 51 );
        enlargeBufferOnSwap = FALSE;      
        }

    scratchBuffer = *InputBuffer;
    *InputBuffer = *nextInputBuffer;
    *nextInputBuffer = scratchBuffer;
    (*InputBuffer)[ 0 ] = 0;
    return scratchBuffer;
    }        

/**** addToInput -- Put some new input in the current input buffer, enlarging it if necessary. ****/

void addToInput( Text *newInput )
{
    LongWord newInputLength = strlen( newInput );
    LongWord oldLength = strlen( *InputBuffer );
    LongWord lengthNeeded = newInputLength + oldLength + 1;
    Text *scratchPointer = 0;
    LongWord newBufferSize = 0;

    if( lengthNeeded > bufferSize ) {  //  Need a bigger hammer ... er, buffer.
        for( newBufferSize = bufferSize; newBufferSize < lengthNeeded; newBufferSize = newBufferSize * 2 )
             ;
        scratchPointer = (Text *)calloc( (Word)newBufferSize, sizeof( Byte ) );
        if( scratchPointer == 0 )
            fatalError( 51 );
        strcpy( scratchPointer, *InputBuffer );
        free( *InputBuffer );
        *InputBuffer = scratchPointer;
        bufferSize = newBufferSize;
        enlargeBufferOnSwap = TRUE;
        }
    strcat( *InputBuffer, newInput );
    }

/**** setNostalgiaMode ****/

void setNostalgiaMode()
{
    if( cocoaIOInitialized )
        nostalgiaMode = TRUE;
    }

/**** clearNostalgiaMode ****/

void clearNostalgiaMode()
{
    if( cocoaIOInitialized )
        nostalgiaMode = FALSE;
    }

/**** cleanUpLowLevelIO -- for cocoa ****/

void cleanUpLowLevelIO() { cocoaIOInitialized = FALSE; }

/**** dealWithResize ****/

void dealWithResize() {}

/**** tgetchar -- low-level console read routine ****/

Word tgetchar()
{
    eventLoopOnce( 0 );
    theHeap->triggerGC();
    if( hasPutBack() )
        return getPutBack();
    if( ! useStdinP ) {
        if( CIn == 0 ) {     // Empty buffer.
            CIBp = CIBuffer = swapBuffers( &CIn );
            if( (CIn > 0) && transcriptFD->fdOpen ) {
                fprintf( transcriptFD->fp, MiscStrings[ 201 ], CIBuffer );
                fflush( transcriptFD->fp );
                }
            }
        return( (--CIn >= 0) ? *CIBp++ : EOF );
        }
    else {
        while( ! evaluatorHasInitialized )
            usleep( NEXT_CHAR_DELAY );
        return fgetc( stdin );
        }
    }

/**** tcharreadyConsole -- Is there a character ready on the console? ****/

Boolean tcharreadyConsole()
{
    pollControllerOnce();
    eventLoopOnce( 0 );
    if( ! useStdinP ) {
        return( (CIn > 0) || ((*InputBuffer)[0] != 0) || hasPutBack() );
        }
    else {
        while( ! evaluatorHasInitialized )
            usleep( NEXT_CHAR_DELAY );
        return TRUE;
        }
    }

/**** tfflushConsoleInput ****/

void tfflushConsoleInput()
{ 
    CIn = 0;
    CIBp = CIBuffer;
    clearPutBack();
    }

/**** fatalErrorWithText ****/

#define FATAL_ERROR_WAIT_DELAY  1000 // microseconds

void fatalErrorWithText( Text *message )
{
    snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), "%s", MiscStrings[ 115 ] );
    snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), MiscStrings[ 234 ], message );
    snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), "%s", MiscStrings[ 115 ] );
    if( strstr( fatalErrorMessage, MiscStrings[ 339 ] ) )
        snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), "%s", MiscStrings[ 338 ] );
    if( strstr( fatalErrorMessage, MiscStrings[ 341 ] ) )
        snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ), "%s", MiscStrings[ 340 ] );
    tellController( E_FATAL_PROBLEM, 0, fatalErrorMessage );
    while( TRUE )
        usleep( FATAL_ERROR_WAIT_DELAY );
    }

/**** fatalError ****/

void fatalError( UnsignedLongWord n )
{
     fatalErrorOccurred = TRUE;

     if( ++fatalErrorNestingLevel > 1 )
        tellController( E_FATAL_PROBLEM, 0, FatalErrorStrings[ 152 ]);

    if( GUIInitialized && descriptorsInitializedP ) {
        fflush( getPortFilePointer( &CurrentOutputPort ) );
        fflush( getPortFilePointer( &ErrorPort ) );
        }
    if( (n < 1) || (n > NumberOfFatalErrorStrings) ) {
        snprintf( fatalErrorMessage + strlen( fatalErrorMessage ), CI_BUFSIZE - strlen( fatalErrorMessage ),
                  MiscStrings[ 93 ], n );
        tellController( E_FATAL_PROBLEM, 0, fatalErrorMessage );
        exit( 1 );
        }               
    fatalErrorWithText( FatalErrorStrings[ n ] );
    }

/**** useFatalErrorOutput ****/

void useFatalErrorOutput()
{
    cleanUpLowLevelIO();
//    defaultOutputFD = defaultErrorFD;
//    writeNewline( defaultOutputFD );

    fatalErrorOccurred = TRUE;
    collectingFatalErrorMessage = TRUE;
    fatalErrorMessage[ 0 ] = 0;
    strcpy( fatalErrorMessage, "\n");
    }

Boolean tprintfX( taggedAval *f )
{
    FILE *fp = 0;

    if( ! GUIInitialized ) {
        fputs( CocoaMessageBuffer, stderr );
        return TRUE;
        }

    fp = getPortFilePointer( f );
    if( (fp == stdout) || (fp == stderr) ) {
        if( transcriptFD->fdOpen ) {
            fputs( CocoaMessageBuffer, transcriptFD->fp );
            fflush( transcriptFD->fp );
            }
        return Cocoa_printf( CocoaMessageBuffer );
        }
    else
        return fputs( CocoaMessageBuffer, fp );
    return TRUE;
    }

