/****************************************************************************************

CursesIO.c++ -- Copyright (c) by Jay Reynolds Freeman 2006.

Simple (I hope!) Curses-based I/O routines for Wraith Scheme.  This
module is intended to be self-contained in that it hides all details
of Curses from the rest of the program.  In particular, this is the
only file that should need to include <curses.h>.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import <stdlib.h>
#import <stdio.h>
#import <string.h>
#import <setjmp.h>
#import <ctype.h>
#import <curses.h>
#import <sys/types.h>
#import <unistd.h>

#import "LocalTypes.h"
#import "ASCIITypes.h"
#import "DataTypes.h"
#import "CursesIO.h"

#import "OpCodes.h"
#import "DataTypes.e"
#import "Events.e"
#import "Globals.e"
#import "PutBack.e"
#import "SchemeAux.e"
#import "StreamHandling.e"

extern Boolean consoleOutputEnabledFlag;
extern Boolean resizeRequired;
extern Text *FatalErrorStrings[];
extern Text *MiscStrings[];
extern void fatalError( LongWord n);
extern void eventLoopOnce( LongWord sleepTimeInMicroseconds );
extern jmp_buf ExitBuffer;
extern void resetError( UnsignedLongWord n );
extern fileDescriptor *transcriptFD;

#define DEFAULT_COLOR_PAIR 1
#define NOSTALGIA_COLOR_PAIR 2
#define MINIMUM_WINDOW_X 72
#define MINIMUM_WINDOW_Y 20

Text CursorMessageBuffer[ CMB_SIZE + 1 ];

/**** Forward declarations: ****/

void addToCommandHistory( Text *buffer );
Boolean retrieveCommand( Word howFarBack, Text *buffer, Word sizeOfBuffer );
void dingDong();
Boolean emitStatusMessage( const Text *msg );
void addToScrollHistory( Text *buffer );
Boolean scrollWW( Word whatLineGoesToBottom );
Boolean cprintfAtMostOneLine( Text *msg, Boolean saveP );
Boolean cprintf( Text *msg, Boolean saveP );
void updateStatusLine();
void updateDialog( Boolean bold );
void manageUpdate( WINDOW *win );
void whatWas( WINDOW *win, Word c );
Word cursesRead( WINDOW *win, Word y, Word x, Text *buffer, Word sizeOfBuffer, Boolean saveP );
void setNostalgiaMode();
void clearNostalgiaMode();
void doResize();

/**** Curses statics: ****/

Boolean cursesIOInitialized = FALSE;
Boolean nostalgiaMode = FALSE;
Boolean useCurses = FALSE;
Boolean doingDialog = FALSE;

static WINDOW *MainWindow;
static Word maxWinX = 0;
static Word maxWinY = 0;

/**** Curses input buffer statics: ****/

#define CI_BUFSIZE 1024

static Text CIBuffer[ CI_BUFSIZE + 1 ];
static Text ScratchCIBuffer[ CI_BUFSIZE + 1 ];
static Text *CIBp = 0;
static Word CIn = 0;

/**** Curses status-line statics: ****/

#define STATUS_LINE_SIZE CMB_SIZE
#define STATUS_LINE_X 0
#define STATUS_LINE_Y 0

static Text StatusLine[ STATUS_LINE_SIZE + 1 ];

/**** Dialog statics ****/

static Text DialogPrompt[ STATUS_LINE_SIZE + 1 ];
Text dialogReturnString[ CMB_SIZE + 1 ];
Boolean dialogIsBold = false;

#define DIALOG_PROMPT_X 1
#define DIALOG_PROMPT_Y 1 
#define DIALOG_RESULT_X 1
#define DIALOG_RESULT_Y 2 
#define DIALOG_DIVIDER_X 1
#define DIALOG_DIVIDER_Y 3 

/**** Command History Mechanism statics: ****/

#define COMMAND_BUFFER_HISTORY_SIZE 100
// #define COMMAND_BUFFER_HISTORY_SIZE 10  // For debugging.
static Text PastCommandBuffers[ COMMAND_BUFFER_HISTORY_SIZE ][ CMB_SIZE + 1];
static Word firstFreeCommandBuffer = 0;
static Word mostRecentCommandBuffer = 0;
static Word nCommandsInBuffer = 0;

/**** addToCommandHistory ****/

void addToCommandHistory( Text *buffer )
{
//    Text foo[ 1000 ];  // debug

//    if( firstFreeCommandBuffer < 0 )  // debug
//        emitStatusMessage("firstFreeCommandBuffer is negative.");  // debug
//    if( firstFreeCommandBuffer >= COMMAND_BUFFER_HISTORY_SIZE )  // debug
//        emitStatusMessage("firstFreeCommandBuffer is too large.");  // debug
//    sprintf( foo, "Copying into buffer at %d: \"%s\".", firstFreeCommandBuffer, buffer );  // debug
//    emitStatusMessage( foo ); // debug
    strncpy( PastCommandBuffers[ firstFreeCommandBuffer ], buffer, CMB_SIZE );
    mostRecentCommandBuffer = firstFreeCommandBuffer;
    
    if( firstFreeCommandBuffer++ >= (COMMAND_BUFFER_HISTORY_SIZE - 1) )
        firstFreeCommandBuffer = 0;
    if( nCommandsInBuffer < COMMAND_BUFFER_HISTORY_SIZE )
        ++nCommandsInBuffer;
    }    

/**** retrieveCommand ****/

Boolean retrieveCommand( Word howFarBack, Text *buffer, Word sizeOfBuffer )
{
    // Caution fencepost error:  howFarBack == 1 <==> command at offset zero from mostRecentCommandBuffer.

    Word indexSought = 0;

    if( nCommandsInBuffer == 0 )
        return FALSE;
    if( (howFarBack - 1) < 0 )
        return FALSE;
    if( howFarBack > nCommandsInBuffer ) {
//        emitStatusMessage( MiscStrings[ 315 ] );
        return FALSE;
        }

    indexSought = mostRecentCommandBuffer - (howFarBack - 1);
    while( indexSought < 0 )
        indexSought += COMMAND_BUFFER_HISTORY_SIZE;
    indexSought = indexSought % COMMAND_BUFFER_HISTORY_SIZE;
//    if( indexSought < 0 )  // debug
//        emitStatusMessage("Command retrieval index is negative.");  // debug
//    if( indexSought >= COMMAND_BUFFER_HISTORY_SIZE )  // debug
//        emitStatusMessage("Command retrieval index is too large.");  // debug
    strncpy( buffer, PastCommandBuffers[ indexSought ], sizeOfBuffer );

    return TRUE;
    }    

/**** Scroll History Mechanism statics: ****/

// #define SCROLL_BUFFER_HISTORY_SIZE 1000
#define SCROLL_BUFFER_HISTORY_SIZE 100  // For debugging.
static Text PastLineBuffers[ SCROLL_BUFFER_HISTORY_SIZE ][ CMB_SIZE + 1];
static Word firstFreeLineBuffer = 0;
static Word mostRecentLineBuffer = 0;
static Word nLinesInBuffer = 0;

/**** addToScrollHistory ****/

void addToScrollHistory( Text *buffer )
{
    Word lnth = 0;
//    Text foo[ 1000 ];  // debug

//    sprintf( foo, "Adding a line to ScrollHistory"); //debug
//    emitStatusMessage( foo );  dingDong(); sleep( 2 ); // debug
    
    strncpy( PastLineBuffers[ firstFreeLineBuffer ], buffer, CMB_SIZE );
    lnth = strlen( PastLineBuffers[ firstFreeLineBuffer ] );
    PastLineBuffers[ firstFreeLineBuffer ][ lnth - 1 ] = 0;   // Strip newline.
    mostRecentLineBuffer = firstFreeLineBuffer;
    
    if( firstFreeLineBuffer++ >= (SCROLL_BUFFER_HISTORY_SIZE - 1) )
        firstFreeLineBuffer = 0;
    if( nLinesInBuffer < SCROLL_BUFFER_HISTORY_SIZE )
        ++nLinesInBuffer;
    }    

/**** scrollWW ****/

Boolean scrollWW( Word whatLineGoesToBottom )
{
    Word windowHeight = maxWinY - DIALOG_DIVIDER_Y - 1;
    Word indexSought = 0;
    Word firstVisibleLine = whatLineGoesToBottom + windowHeight - 1;
    Word windowLineCounter = 0;
    Word scrollLineCounter = 0;
//    Text foo[ 1000 ];  // debug

//    sprintf( foo, "%d -> bottom, ht. %d, fV %d, %d in buffer, %d most recent",              // debug
//             whatLineGoesToBottom, windowHeight, firstVisibleLine, nLinesInBuffer, mostRecentLineBuffer ); // debug
//    emitStatusMessage( foo );  // debug

    if( nLinesInBuffer < windowHeight )             // Nothing has scrolled out of sight.
        return FALSE;
    if( whatLineGoesToBottom < 0 )                  // Haven't seen that line yet.
        return FALSE;
    if( firstVisibleLine > (nLinesInBuffer - 1) ) { // Line sought is past the end of scrollback.
//        emitStatusMessage( MiscStrings[ 326 ] );
        return FALSE;
        }

    // Caution fencepost error:  whatLineGoesToBottom == 1 <==> line at offset zero from mostRecentLineBuffer,
    //   <==> first visible line at offset windowHeight - 1 from mostRecentLineBuffer

    scrollLineCounter = firstVisibleLine;
    for( windowLineCounter = DIALOG_DIVIDER_Y + 1; windowLineCounter < maxWinY; ++windowLineCounter ) {
        wmove( MainWindow, windowLineCounter, 0 );
        wclrtoeol( MainWindow );
        indexSought = mostRecentLineBuffer - (scrollLineCounter - 1);
        while( indexSought < 0 )
            indexSought += SCROLL_BUFFER_HISTORY_SIZE;
        indexSought = indexSought % SCROLL_BUFFER_HISTORY_SIZE;
        waddstr( MainWindow, PastLineBuffers[ indexSought ] );
        if( scrollLineCounter-- == 0 )
            break;
        }
    return TRUE;
    }

/**** dingDong ****/

void dingDong() { fprintf( stderr, "\07" ); }

/**** emitStatusMessage ****/

Boolean emitStatusMessage( const Text *msg )
{
    strncpy( StatusLine, " ", sizeof( StatusLine ) - strlen( StatusLine ) );
    strncat( StatusLine, MiscStrings[ 301 ], sizeof( StatusLine ) - strlen( StatusLine ) - 1 );
    strncat( StatusLine, msg, sizeof( StatusLine ) - strlen( StatusLine ) - 1 );
    manageUpdate( MainWindow );
    return TRUE;
    }

/**** performDialogAndReturnString ****/

Boolean performDialogAndReturnString( const Text *prompt, Text *result, UnsignedWord resultLength )
{
    Word x = 0;
    Word y = 0;
    Word charsToGet = 0;
    Word charsGotten = 0;
    getyx( MainWindow, y, x );

    result[ 0 ] = 0;
    doingDialog = TRUE;

    if( resultLength > sizeof( dialogReturnString ) )
        charsToGet = sizeof( dialogReturnString ) - 1;
    else
        charsToGet = resultLength - 1;
    if( charsToGet <= 0 )
        charsToGet = 0;

    strncpy( DialogPrompt, prompt, sizeof( DialogPrompt ) - strlen( DialogPrompt ) );
    dialogReturnString[ 0 ] = 0;
    manageUpdate( MainWindow );
    dingDong();

    charsGotten = cursesRead( MainWindow, DIALOG_RESULT_Y, DIALOG_RESULT_X,  dialogReturnString, charsToGet, FALSE );
    addToCommandHistory( dialogReturnString );

    // At this point, the result buffer ends in a newline.  We must strip it before returning.
   
    dialogReturnString[ charsGotten - 1 ] = 0;
    strncpy( result, dialogReturnString, resultLength );

    doingDialog = FALSE;
    wmove( MainWindow, y, x );
    manageUpdate( MainWindow );

    return TRUE;
    }

/**** initializeLowLevelIO -- for curses ****/

void initializeLowLevelIO()
{
    CIBuffer[ 0 ] = 0;
    ScratchCIBuffer[ 0 ] = 0;
    CIBp = &(CIBuffer[ 0 ]);
    CursorMessageBuffer[ 0 ] = 0;
    dialogReturnString[ 0 ] = 0;

    initializePutBack();

    MainWindow = initscr();                      // Create a curses window for subsequent use.
    getmaxyx( MainWindow, maxWinY, maxWinX );    // Make a note of how big it is.

    if(    (maxWinY < MINIMUM_WINDOW_Y)
        || (maxWinX < MINIMUM_WINDOW_X) ) {      // Window is too small
        endwin();

        //  This initialization routine is called very early,
        //    so we cannot count on "fatalError" working.

        fprintf( stderr, "%s.\n", FatalErrorStrings[ 86 ] );
        exit( 1 );
        }

    keypad( MainWindow, TRUE );                  // Let's see what this does -- 2006.11.29
    start_color();
    use_default_colors();
    init_pair(DEFAULT_COLOR_PAIR, -1, -1 );
    init_pair(NOSTALGIA_COLOR_PAIR, COLOR_GREEN,COLOR_BLACK);
    wbkgd( MainWindow,COLOR_PAIR(1));

    scrollok( MainWindow, TRUE );                // Allow scrolling ...
    idlok( MainWindow, TRUE );                   //   ... which requires idlok( ... , TRUE )
    cbreak();                                    // Arrange for character-at-a-time input, excepting certain signals.
    nodelay( MainWindow, TRUE );
//    halfdelay( 10 );                              // Arrange for char-at-a-time input, excepting certain signals, with timeout.
    noecho();                                    // Disable echoing of entered characters.
//    endwin();                                    // ... and put it away till it is needed.o

    StatusLine[ 0 ] = 0;
    DialogPrompt[ 0 ] = 0;

    cursesIOInitialized = TRUE;
    useCurses = TRUE;

    emitStatusMessage( MiscStrings[ 322 ] );
    }

/**** reinitializeLowLevelIO ****/

void reinitializeLowLevelIO( Text *reinitializationAction )
{
    reinitializePutBack();
    }

/**** nextLineOrFragment ****/

//  This routine is passed a buffer, which it does not modify.  If the buffer contains
//    any newlines, the result returns a copy of the substring of the buffer from its beginning
//    up to and including the first newline -- the substring is null-terminated after the
//    newline -- and sets *pastFragment to the location in the buffer just past the
//    newline found.  It is possible that *pastFragment points to NULL, of course.
//
//  If the buffer is non-null and contains no newlines, this function returns a copy of the
//    entire buffer, and sets *pastFragment to NULL.
//
//  If the buffer is empty or NULL, this routine sets *pastFragment to NULL and returns NULL.
//
//  The returned copy is statically allocated, so it won't go away; however, subsequent
//    calls to this function will overwrite it.

Text *nextLineOrFragment( Text *buffer, Text **pastFragment )
{
    static Text scratch[ CMB_SIZE + 1 ];
    Text *pNewline = 0;
 
    if( (buffer == NULL) || (*buffer == 0) ) {                 // Null or empty buffer has no fragments to return.
        *pastFragment = NULL;
        return NULL;
        }        
    else {
        strncpy( scratch, buffer, CMB_SIZE );                  // We will be returning this copy, possibly modified.

        if( (pNewline = strchr( scratch, '\n')) != NULL ) {    // Found a newline, must null-terminate the copy after it.
            *(pNewline + 1) = 0;
            *pastFragment = strchr( buffer, '\n') + 1;         // Report where the next fragment starts (might point to '0').
            }
        else                                                   // "scratch", as is, is the final fragment.
            *pastFragment = NULL;

        return scratch;
        }
    }       

/**** cprintfAtMostOneLine ****/

#define NOSTALGIA_NORMAL_DELAY     25000 // microseconds
#define NOSTALGIA_SWAPPING_DELAY  500000 // microseconds

Boolean cprintfAtMostOneLine( Text *msg, Boolean saveP )
{
    Word counter = 0;
    Word lnth = strlen( msg );

    if( cursesIOInitialized )  {
        if( consoleOutputEnabledFlag ) {
            if( saveP ) {
                strncat( ScratchCIBuffer, msg, CI_BUFSIZE - strlen( ScratchCIBuffer ) );
                if( (lnth > 0) && (msg[ lnth - 1 ] == '\n') ) {
                    addToScrollHistory( ScratchCIBuffer );
                    ScratchCIBuffer[ 0 ] = 0;
                    }
                }
            if( ! nostalgiaMode) {
                waddstr( MainWindow, msg );
                manageUpdate( MainWindow );
                return TRUE;
                }
            else {
                for( counter = 0; counter < lnth; ++counter ) {
                    waddch( MainWindow, msg[ counter ] );
                    manageUpdate( MainWindow );
                    if( (random() % 512) == 0 )
                        usleep( NOSTALGIA_SWAPPING_DELAY );
                    else
                        usleep( NOSTALGIA_NORMAL_DELAY );
                    }
                return TRUE;
                }
            }
        else
            return TRUE;
        }
    else
        return fprintf( stderr, MiscStrings[ 201 ], msg );
    }

/**** cprintf ****/

Boolean cprintf( Text *msg, Boolean saveP )
{
    Text *copyFromHere = msg;

    while( (copyFromHere != NULL) && (*copyFromHere != 0) ) {
        cprintfAtMostOneLine( nextLineOrFragment( copyFromHere, &copyFromHere ), saveP );
//        dingDong(); sleep( 3 );
        }

    return TRUE;
    }

/**** updateStatusLine ****/

void updateStatusLine()
{
    Word counter = 0;
    Word lnth = strlen( StatusLine );
    Word x = 0;
    Word y = 0;

    getyx( MainWindow, y, x );
    wmove( MainWindow, STATUS_LINE_Y, STATUS_LINE_X );
    for( counter = 0; counter < maxWinX; ++counter )
        if( counter < lnth )
            waddch( MainWindow, StatusLine[ counter ] | A_REVERSE );
        else
            waddch( MainWindow, ' ' | A_REVERSE );
    wmove( MainWindow, y, x );
    }

/**** updateDialog ****/

void updateDialog( Boolean bold )
{
    Word counter = 0;
    Word lnth = 0;
    Word x = 0;
    Word y = 0;

    dialogIsBold = bold;
    Word cAttrs = 0;

    if( bold )
        cAttrs = A_BOLD;

    getyx( MainWindow, y, x );
    wmove( MainWindow, DIALOG_PROMPT_Y, 0 );
    lnth = strlen( DialogPrompt );
    for( counter = 0; counter < maxWinX; ++counter )
        if( counter < DIALOG_PROMPT_X )
            waddch( MainWindow, ' ' );
        else if( counter < (lnth + DIALOG_PROMPT_X) )
            waddch( MainWindow, DialogPrompt[ counter -  DIALOG_PROMPT_X ] | cAttrs );
        else
            waddch( MainWindow, ' ' );

    wmove( MainWindow, DIALOG_RESULT_Y, 0 );
    lnth = strlen( dialogReturnString );
    for( counter = 0; counter < maxWinX; ++counter )
        if( counter < DIALOG_RESULT_X )
            waddch( MainWindow, ' ' );
        else if( counter < (lnth + DIALOG_RESULT_X) )
            waddch( MainWindow, dialogReturnString[ counter - DIALOG_RESULT_X ] );
        else
            waddch( MainWindow, ' ' );

    wmove( MainWindow, DIALOG_DIVIDER_Y, 0 );
    for( counter = 0; counter < maxWinX; ++counter )
        waddch( MainWindow, '-' );
    wmove( MainWindow, y, x );
    }

/**** manageUpdate ****/

void manageUpdate( WINDOW *win )
{
    Word x = 0;
    Word y = 0;

    eventLoopOnce( 0 );
    if( cursesIOInitialized && useCurses ) {
        getyx( MainWindow, y, x );
        if( resizeRequired ) {
            endwin();
            doResize();
            }
        updateStatusLine();
        updateDialog( doingDialog );
        wmove( MainWindow, y, x );
        wrefresh( MainWindow );
        }
    }

/**** whatWas -- debugging routine -- type out a character's decimal equivalent ****/

void whatWas( WINDOW *win, Word c )
{
    Text message[ 256 ];

    if( (c < 128) && isprint( c ) ) 
        snprintf( message, sizeof( message ), "Illegal character:  '%c' <ASCII %d>", c, c );
    else
        snprintf( message, sizeof( message ), "Illegal non-printing character:  <ASCII %d>", c );
    emitStatusMessage( message );
    }

/**** cursesRead -- Fill a buffer up till a newline, and return number of characters in it.
                    Don't read more characters than the window is wide.                      ****/

Word cursesRead( WINDOW *win, Word y, Word x, Text *buffer, Word sizeOfBuffer, Boolean saveP )
{
    Word cRaw = 0;
    Word insertPoint = 0;
    Word savedInsertPoint = 0;
    Word counter = 0;
    static Text KillBuffer[ STATUS_LINE_SIZE + 1 ];
    static Text SavedCurrentLine[ STATUS_LINE_SIZE + 1 ];
    static Text SavedCurrentLine2[ STATUS_LINE_SIZE + 1 ];
    Word positionX = 0;
    Word positionY = 0;
    Word nChars = 0;
    Word commandHowFarBack = 0;
    Word scrollHowFarBack = 0;
    Boolean scrollingP = false;

    if( sizeOfBuffer <= 0 )
        return 0;

    buffer[ 0 ] = 0;
    if( ! cursesIOInitialized )
        return 0;

    while( TRUE ) {
        cRaw = wgetch( win );
        if( scrollingP && (cRaw != ERR ) && (cRaw != CONTROL_CARET) && (cRaw != CONTROL_UNDERBAR)) {
            scrollingP = false;
            scrollWW( 0 );
            strncpy( buffer, SavedCurrentLine2, sizeOfBuffer );
            wmove( win, y, x );
            wclrtoeol( win );
            if( strlen( buffer ) > 0 )
                buffer[ strlen( buffer ) - 1 ] = 0;    // Strip final '\n'
            mvwaddstr( win, y, x, buffer );
            nChars = strlen( buffer );
            insertPoint = savedInsertPoint;
            wmove( win, y, x + insertPoint );
//            insertPoint = nChars = strlen( buffer );
            }
        switch( cRaw ) {
            case ERR:     // Means "no input ready" -- Wraith Scheme will hang on uncleverly redirected input.
                eventLoopOnce( 50000 );
                break;
            case CONTROL_A:
                insertPoint = 0;
                break;
            case CONTROL_B:
            case KEY_LEFT:
                if( insertPoint > 0 ) 
                    --insertPoint;
                else
                    dingDong();
                break;
            case CONTROL_D:
                if( nChars > 0 ) {
                    wdelch( win );
                    --nChars;
                    }
                else
                    dingDong();
                break;
            case CONTROL_E:
                insertPoint = nChars;
                break;
            case CONTROL_F:
            case KEY_RIGHT:
                if( insertPoint < nChars )
                    ++insertPoint;
                else
                    dingDong();
                break;
            case CONTROL_H:
            case KEYBOARD_DELETE:
            case KEY_BACKSPACE:
            case KEY_DC:
                if( insertPoint > 0 ) {
                   --insertPoint;
                   --nChars;
                    mvwdelch( win, y, x + insertPoint );
                    }
                else
                    dingDong();
                break;
            case CONTROL_K:
                if( insertPoint < nChars ) {
                    if( (nChars - insertPoint) < STATUS_LINE_SIZE ) {
                        innstr( KillBuffer, nChars - insertPoint );
                        KillBuffer[ nChars - insertPoint ] = 0;
                        }                     
                    else {
                        innstr( KillBuffer, STATUS_LINE_SIZE );
                        KillBuffer[ STATUS_LINE_SIZE ] = 0;
                        }                     
                    wclrtoeol( win );
                    nChars = insertPoint;
                    }                     
                break;
            case CONTROL_L:
                wmove( win, y, x + insertPoint );
                redrawwin( win );
                manageUpdate( win );
                break;
            case CONTROL_R:
                if( strlen( KillBuffer ) > 0 )
                    for( counter = strlen( KillBuffer ); counter > 0; )
                        ungetch( KillBuffer[ --counter ] );
                else
                    dingDong();
                break;
            case CONTROL_U:
                if( nChars > 0 ) {
                    wmove( win, y, x );
                    wclrtoeol( win );
                    nChars = insertPoint = 0;
                    }
                else
                    dingDong();
                break;
            case CONTROL_CARET:
                scrollingP = true;
                if( scrollHowFarBack == 0 ) {      // Save a copy of the current command line and insert point.
                    savedInsertPoint = insertPoint;
                    mvwinnstr( win, y, x, SavedCurrentLine2, nChars );
                    SavedCurrentLine2[ nChars ] = '\n';
                    SavedCurrentLine2[ nChars + 1 ] = 0;
                    insertPoint = nChars = 0;
                    }
                if( scrollWW( scrollHowFarBack + 1 ) )
                    ++scrollHowFarBack;
                else
                    dingDong();
                break;
            case CONTROL_UNDERBAR:
                scrollingP = true;
                if( scrollHowFarBack == 0 )
                    dingDong();
                else if( scrollWW( scrollHowFarBack - 1 ) ) {
                    if( scrollHowFarBack == 1 ) {   // We just scrolled back to zero.
                        strncpy( buffer, SavedCurrentLine2, sizeOfBuffer );
                        wmove( win, y, x );
                        wclrtoeol( win );
                        if( strlen( buffer ) > 0 )
                            buffer[ strlen( buffer ) - 1 ] = 0;    // Strip final '\n'
                        mvwaddstr( win, y, x, buffer );
//                        insertPoint = nChars = strlen( buffer );
                        nChars = strlen( buffer );
                        insertPoint = savedInsertPoint;
                        wmove( win, y, x + insertPoint );
                        }
                    --scrollHowFarBack;
                    }                   
                else
                    dingDong();
                break;
            case KEY_UP:
                if( commandHowFarBack == 0 ) {      // Save a copy of the current command line and insert point.
                    savedInsertPoint = insertPoint;
                    mvwinnstr( win, y, x, SavedCurrentLine, nChars );
                    SavedCurrentLine[ nChars ] = '\n';
                    SavedCurrentLine[ nChars + 1 ] = 0;
                    }
                if( retrieveCommand( commandHowFarBack + 1, buffer, sizeOfBuffer ) ) {
                    ++commandHowFarBack;                    
                    wmove( win, y, x );
                    wclrtoeol( win );
                    if( strlen( buffer ) >0 )
                        buffer[ strlen( buffer ) - 1 ] = 0;    // Strip final '\n'
                    mvwaddstr( win, y, x, buffer );
                    insertPoint = nChars = strlen( buffer );
                    nChars = strlen( buffer );
                    }
                else
                    dingDong();
                break;
            case KEY_DOWN:
                if( commandHowFarBack <= 0 )
                    dingDong();
                else if( commandHowFarBack == 1 ) {
                    --commandHowFarBack;
                    strncpy( buffer, SavedCurrentLine, sizeOfBuffer );
                    wmove( win, y, x );
                    wclrtoeol( win );
                    if( strlen( buffer ) > 0 )
                        buffer[ strlen( buffer ) - 1 ] = 0;    // Strip final '\n'
                    mvwaddstr( win, y, x, buffer );
                    nChars = strlen( buffer );
                    insertPoint = savedInsertPoint;
                    wmove( win, y, x + insertPoint );
//                    insertPoint = nChars = strlen( buffer );
                    }
                else if( retrieveCommand( commandHowFarBack - 1, buffer, sizeOfBuffer ) ) {
                    --commandHowFarBack;
                    wmove( win, y, x );
                    wclrtoeol( win );
                    if( strlen( buffer ) > 0 )
                        buffer[ strlen( buffer ) - 1 ] = 0;    // Strip final '\n'
                    mvwaddstr( win, y, x, buffer );
                    insertPoint = nChars = strlen( buffer );
                    }
                else
                    dingDong();
                break;
            case NEWLINE:
                insertPoint = nChars;
                if( insertPoint >= maxWinX )
                    dingDong();
                else {
                    mvwaddch( win, y, x + insertPoint, cRaw );
                    ++nChars;
                    getyx( win, positionY, positionX );
                    mvwinnstr( win, positionY - 1, x, buffer, nChars );
                    buffer[ nChars - 1 ] = '\n';    // Curses returns the last char of a line as a space, not a newline.  ???
                    buffer[ nChars ] = 0;
                    wmove( win, positionY, x );
                    ++y;
                    insertPoint = 0;
                    }
                break;
                    savedInsertPoint = insertPoint;
                    savedInsertPoint = insertPoint;
            default:
                if( (isprint( cRaw )) || (cRaw == '\t') ) {
                    if( insertPoint >= (maxWinX - 1) )
                        dingDong();
                    else {
                        winsch( win, cRaw );
                        ++insertPoint;
                        ++nChars;
                        }
                    }
                else {
                    whatWas( win, cRaw );  // Debug
                    dingDong();
                    }
                break;
            }
           wmove( win, y, x + insertPoint );
        if( cRaw == '\n' )
            break;
        }
    commandHowFarBack = 0;
    manageUpdate( win );
    if( saveP ) {
        addToCommandHistory( buffer );
        addToScrollHistory( buffer );
        }
    return strlen( buffer );
    }

/**** setNostalgiaMode ****/

void setNostalgiaMode()
{
    if( cursesIOInitialized ) {
        nostalgiaMode = TRUE;
        wbkgd( MainWindow, COLOR_PAIR( NOSTALGIA_COLOR_PAIR ) );
        manageUpdate( MainWindow );
        }
    }

/**** clearNostalgiaMode ****/

void clearNostalgiaMode()
{
    if( cursesIOInitialized ) {
        nostalgiaMode = FALSE;
        wbkgd( MainWindow,COLOR_PAIR( DEFAULT_COLOR_PAIR ) );
        manageUpdate( MainWindow );
        }
    }

/**** cleanUpLowLevelIO -- for curses ****/

void cleanUpLowLevelIO() 
{
    if( cursesIOInitialized )
        endwin();
    cursesIOInitialized = FALSE;
    }

/**** dealWithResize ****/

void dealWithResize() { resizeRequired = TRUE; ungetch( CONTROL_L ); }

/**** doResize ****/

void doResize()
{
    resizeRequired = FALSE;    // debug; I can't make "resize" work.
    return;                    // debug; I can't make "resize" work.
    }

/**** switchToSystemIO  ****/

void switchToSystemIO()   // FIXME -- clear conin. (??)
{
    useCurses = FALSE;
    if( cursesIOInitialized ) {
        nodelay( MainWindow, false );
        endwin();
        }
    }

/**** switchFromSystemIO ****/


void switchFromSystemIO()
{
    useCurses = TRUE;
    if( cursesIOInitialized ) {
        nodelay( MainWindow, true );
        manageUpdate( MainWindow );
        }
    }

/**** tgetchar -- low-level console read routine ****/

Word tgetchar()
{
    Word x = 0;
    Word y = 0;

    getyx( MainWindow, y, x );
    if( hasPutBack() )
        return getPutBack();
    if( CIn == 0 ) {     // Empty buffer.
        CIn = cursesRead( MainWindow, y, x, CIBuffer , CI_BUFSIZE, TRUE );
        CIBp = CIBuffer;
        if( (CIn > 0) && transcriptFD->fdOpen ) {
            fprintf( transcriptFD->fp, MiscStrings[ 201 ], CIBuffer );
            fflush(  transcriptFD->fp );
            }
        }
    return( (--CIn >= 0) ? *CIBp++ : EOF );
    }

/**** tcharreadyConsole -- Is there a character ready on the console? ****/

Boolean tcharreadyConsole()
{
    return ( (CIn >0) || hasPutBack() );   
    }

/**** tfflushConsoleInput ****/

void tfflushConsoleInput()
{ 
    CIn = 0;
    CIBp = CIBuffer;
    clearPutBack();
    }

/**** fatalErrorWithText ****/

static Boolean recursiveCall = FALSE;

void fatalErrorWithText( Text *fatalErrorMessage )
{
    useFatalErrorOutput();
    tprintf3( &ErrorPort, MiscStrings[ 234 ], fatalErrorMessage );
    tprintf2( &ErrorPort, MiscStrings[ 115 ] );
    if( !recursiveCall ) {                  /* try to provide helpful information */
        recursiveCall = TRUE;
        }
    if( strstr( (const char *)fatalErrorMessage, MiscStrings[ 339 ] ) )
        tprintf2( &ErrorPort, MiscStrings[ 338 ] );
    if( strstr( (const char *)fatalErrorMessage, MiscStrings[ 341 ] ) )
        tprintf2( &ErrorPort, MiscStrings[ 340 ] );
    resetCleanUp();
    longjmp( ExitBuffer, 2 );
    }

/**** fatalError ****/

void fatalError( UnsignedLongWord n )
{
    useFatalErrorOutput();
    if( descriptorsInitializedP ) {
        fflush( getPortFilePointer( &CurrentOutputPort ) );
        fflush( getPortFilePointer( &ErrorPort ) );
        }
    if( (n < 1) || (n > NumberOfFatalErrorStrings) ) {
        tprintf3( &ErrorPort, MiscStrings[ 93 ], n );
        exit( 1 );
        }               
    fatalErrorWithText( FatalErrorStrings[ n ] );
    }

/**** useFatalErrorOutput ****/

void useFatalErrorOutput()
{
    cleanUpLowLevelIO();
    CurrentOutputPort = ErrorPort;
    writeNewline( &CurrentOutputPort );
    }

Boolean tprintfX( taggedAval *f )
{
    FILE *fp = getPortFilePointer( f );

    if( (fp == stdout) || (fp == stderr) ) {
        if( transcriptFD->fdOpen ) {
            fputs( CursorMessageBuffer, transcriptFD->fp );
            fflush( transcriptFD->fp );
            }
        return cprintf( CursorMessageBuffer, TRUE );
        }
    else
        return fputs( CursorMessageBuffer, fp );
    return TRUE;
    }

