/****************************************************************************************

EventsForCocoa.c++ -- Copyright (c) by Jay Reynolds Freeman 2006

Eventloop and state-transition stuff for Wraith Scheme,
for the implementation involving Cocoa.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "MemoryManager.h"
#import "CocoaInterface.h"
#import "Registers.h"
#import "SECD.h"
#import "Globals.e"
#import "DataTypes.e"
#import "Intern.e"
#import "IO.e"
#import "MainAux.e"
#import "NumberFormats.e"
#import "Reader.e"
#import "Registers.e"
#import "SECDMachine.e"
#import "SchemeFunctionExterns.e"

Boolean         CompileDefinesFlag = false;
Boolean         GCInProgressFlag = false;
Boolean         EmptyListIsFalseFlag = false;
Boolean         ResetFlag = false;

static Text localLastFileLoadedName[ MAX_PATHNAME_LENGTH + 1 ];
static Text localLastWorldLoadedName[ MAX_PATHNAME_LENGTH + 1 ];

/**** Externs, for communication with Cocoa controller.  ****/

extern threadMessage toController;
extern void tellController( int mID, LongWord mLong, const Text *mString );
extern void pollControllerOnce();

/**** forward declarations: ****/

void reportEvent( int foo );

/**** reportControlC ****/

void reportControlC( int ignore ) { reportEvent( REQUEST_RESET_EVENT ); }

/**** reportSIGUSR1 ****/

void reportSIGUSR1( int ignore ) { reportEvent( SIGUSR1_EVENT ); }

/**** reportSIGALRM ****/

void reportSIGALRM( int ignore ) { reportEvent( SIGALRM_EVENT ); }

extern void tellKittenCode();

/**** doSIGQUIT ****/

extern void makeKittensExit();

void doSIGQUIT( int ignore )
{
#ifndef PIXIE_SCHEMES
    makeKittensExit();
#endif
    cleanUpForExit();
    exit( 1 );
    }        

/**** dealWithSIGUSR1 ****/

void dealWithSIGUSR1() {

    UnsignedLongWord interruptCounter = 0;

    SPushR();
    for( interruptCounter = 0; interruptCounter < NUMBER_OF_INTERRUPTS; ++interruptCounter ) {
        theClowder->lockInterrupt( interruptCounter );
        if( theClowder->getInterruptRequested( interruptCounter ) ) {
            if( theClowder->getInterruptInstalled( interruptCounter ) ) {
                R = IVR;
                vectorRefN( interruptCounter );
                if( ! empty_list_p( R.t ) ) {
                    SPushR();                     // R: kitten/handler; Stack: kitten/handler, saved R.
                    car();                        // R: kitten;         Stack: kitten/handler, saved R.
                    SPushR();                     // R: kitten;         Stack: kitten, kitten/handler, saved R.
                    R = S[1];                     // R: kitten/handler; Stack: kitten, kitten/handler, saved R.
                    cdr();                        // R: handler;        Stack: kitten, kitten/handler, saved R.
                    tellKittenCode();             // After call --      Stack: kitten/handler, saved R.
                    SDrop();                      //                    Stack: saved R.
                    }
                }
            theClowder->clearInterruptRequested( interruptCounter );
            }
        theClowder->unlockInterrupt( interruptCounter );
        }
    SPopR();
    }

/**** dealWithSIGALRM ****/

extern void tellKittenCode();

void dealWithSIGALRM() {

    SPushR();
    R.t = exactShortFixnumBitsMask;
    R.u.a = (taggedAval *)NameStrings[ 434 ];
    intern();
    SPushR();
    boundSymbolP();
    SPopR();
    if( B ) {                 // e::alarm is bound...
        lookUp();
        lambdaPairPMC();
        if( B ) {             // ... to a lambda
            nilCons();
            SPushR();
            R.t = exactShortFixnumBitsMask;
            R.u.c = theKittenNumber;
            Swap();
            tellKittenCode();
            }
        }
    SPopR();
    isAlarmSet = FALSE;
    }

/**** initializeSignalHandlers ****/

void initializeSignalHandlers() {}

/**** reinitializeSignalHandlers ****/

void reinitializeSignalHandlers() {}

static Boolean eventRegister[ NUMBER_OF_EVENTS ];

/**** initializeEvents ****/

void initializeEvents()
{
    int counter = 0;

    for( counter = 0; counter < NUMBER_OF_EVENTS; ++counter )
        eventRegister[ counter ] = FALSE;

#ifndef PIXIE_SCHEMES
    signal( SIGINT,  reportControlC );
    signal( SIGQUIT, doSIGQUIT );
    signal( SIGUSR1, reportSIGUSR1 );
    signal( SIGALRM, reportSIGALRM );
#endif
    }

/**** reinitializeEvents ****/

void reinitializeEvents()
{
    theClowder->clearKittenEvaluating();
    reportEvent( END_EVALUATION_EVENT );

#ifndef PIXIE_SCHEMES
    signal( SIGINT,  reportControlC );
    signal( SIGQUIT, doSIGQUIT );
    signal( SIGUSR1, reportSIGUSR1 );
    signal( SIGALRM, reportSIGALRM );
#endif
    }

/**** reportEvent ****/

static Boolean anyEvents = FALSE;

void reportEvent( int foo )
{
    switch( foo ) {
        case REINITIALIZATION_COMPLETE:
            GCInProgressFlag = FALSE;
            break;
        case GC_START_EVENT_3:
            GCInProgressFlag = TRUE;
            if( isMomCatP )
                theClowder->momDialogWithEverybody( MOM_SAYS_GC_SEQUENCE_STARTING, KITTEN_ACKNOWLEDGES_GC_SEQUENCE_STARTING );
            break;
        case GC_END_EVENT_5:
            if( isMomCatP ) {
                theClowder->clearGCBytesValid();
                theClowder->clearGCIsRequired();
                theClowder->momDialogWithEverybody( MOM_SAYS_GC_SEQUENCE_ENDING, KITTEN_ACKNOWLEDGES_GC_SEQUENCE_ENDING );
                }
            GCInProgressFlag = FALSE;
            break;
        case GC_END_EVENT_6:
            if( isMomCatP ) {
                theClowder->clearGCBytesValid();
                theClowder->clearGCIsRequired();
                }
            GCInProgressFlag = FALSE;
            break;
        case FILE_LOAD_BEGINNING_EVENT:
            strncpy( localLastFileLoadedName, lastFileLoadedName, MAX_PATHNAME_LENGTH - 1 );
            break;
        case WORLD_LOAD_SUCCEEDED_EVENT:
            strncpy( localLastWorldLoadedName, lastWorldLoadedName, MAX_PATHNAME_LENGTH - 1 );
            break;

        default: 
            break;
        }

    if( eventRegister[ foo ] )
        ;     // Worry, maybe, at some time in the future ...
    eventRegister[ foo ] = TRUE;
   
    anyEvents = TRUE;
    }

/**** dealWithEvent ****/

void dealWithEvent( int foo )
{
    LongWord memorySizeNow = 0;
    LongWord generationalMemorySizeNow = 0;

    switch( foo ) {

        case REINITIALIZATION_COMPLETE:
            tellController( E_ACKNOWLEDGE_RESET, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_START_EVENT:
            tellController( E_GARBAGE_COLLECTING, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;
 
        case GC_START_EVENT_1:
            emitStatusMessage( MiscStrings[ 299 ] );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_START_EVENT_2:
            emitStatusMessage( MiscStrings[ 452 ] );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_START_EVENT_3:                    // Reserved for possible use by generational GC.
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT:
            tellController( E_NEW_GC_COUNT, theHeap->garbageCollectorGCCount(), "" );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_1:
            tellController( E_NOT_GARBAGE_COLLECTING, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_2:
            emitStatusMessage( MiscStrings[ 300 ] );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_3:
            emitStatusMessage( MiscStrings[ 453 ] );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_4:
            tellController( E_NEW_GENERATIONAL_GC_COUNT, theHeap->garbageCollectorGenerationalGCCount(), "" );
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_5:
            eventRegister[ foo ] = FALSE;
            break;

        case GC_END_EVENT_6:
            eventRegister[ foo ] = FALSE;
            break;

         case SIGUSR1_EVENT:
            if( (! GCInProgressFlag) && (! theClowder->GCIsRequired()) ) {
                eventRegister[ foo ] = FALSE;
                dealWithSIGUSR1();
                }
            break; 

        case SIGALRM_EVENT:
            if( (! GCInProgressFlag) && (! theClowder->GCIsRequired()) ) {
                eventRegister[ foo ] = FALSE;
                dealWithSIGALRM();
                }
            break; 

        case INITIALIZATION_FINISHED_EVENT:
            tellController( E_COMPLETED_INITIALIZATION, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case COMPILE_DEFINES_EVENT:
            CompileDefinesFlag = TRUE;
            emitStatusMessage( MiscStrings[ 377 ] );
            tellController( E_COMPILING_DEFINES, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case DO_NOT_COMPILE_DEFINES_EVENT:
            CompileDefinesFlag = FALSE;
            emitStatusMessage( MiscStrings[ 378 ] );
            tellController( E_NOT_COMPILING_DEFINES, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case USE_LOW_PRECISION_EVENT:
            useFullNumericPrecision = FALSE;
            emitStatusMessage( MiscStrings[ 380 ] );
            tellController( E_USING_LOW_PRECISION, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case USE_HIGH_PRECISION_EVENT:
            useFullNumericPrecision = TRUE;
            emitStatusMessage( MiscStrings[ 379 ] );
            tellController( E_USING_HIGH_PRECISION, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case START_EVALUATION_EVENT:
            if( (theClowder->getKittenEvaluating( theKittenNumber )) == TRUE )
                tellController( E_EVALUATING, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case END_EVALUATION_EVENT:
            if( (theClowder->getKittenEvaluating( theKittenNumber )) == FALSE )
                tellController( E_NOT_EVALUATING, 0, "" );
            eventRegister[ foo ] = FALSE;
            break;

        case REQUEST_RESET_EVENT:
            if( (! GCInProgressFlag) && (! theClowder->GCIsRequired()) ) {
                if( ! isMomCatP )
                    theClowder->setKittenSays( KITTEN_ACKNOWLEDGES_GC_SEQUENCE_ENDING );
                evaluatorHasInitialized = FALSE;
                eventRegister[ foo ] = FALSE;
                resetScheme();
                }
            break; 

        case REQUEST_RESET_EVENT_2:
            if( (! GCInProgressFlag) && (! theClowder->GCIsRequired()) ) {
                if( ! isMomCatP )
                    theClowder->setKittenSays( KITTEN_ACKNOWLEDGES_GC_SEQUENCE_ENDING );
//                evaluatorHasInitialized = FALSE;   // Without this, tcharreadyConsole will not read.
                eventRegister[ foo ] = FALSE;
                resetSchemeFinal();
                }
            break; 

        case REQUEST_RESET_EVENT_3:
            eventRegister[ foo ] = FALSE;
            break; 

        case REQUEST_RESET_EVENT_4:
            eventRegister[ foo ] = FALSE;
            break; 

        case CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT:
            static LongWord savedCurrentGenerationalInUse = -1;
            LongWord currentGenerationalInUse;
            if( savedCurrentGenerationalInUse == (currentGenerationalInUse = theHeap->currentActiveInUse()) ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_GENERATIONAL_MEMORY_IN_USE, currentGenerationalInUse, "");
            savedCurrentGenerationalInUse = currentGenerationalInUse;
            eventRegister[ foo ] = FALSE;
            break; 

        case CURRENT_MEMORY_IN_USE_EVENT:
            static LongWord savedCurrentInUse = -1;
            LongWord currentInUse;
            if( savedCurrentInUse == (currentInUse = theHeap->currentInUse()) ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_MEMORY_IN_USE, currentInUse, "");
            savedCurrentInUse = currentInUse;
            eventRegister[ foo ] = FALSE;
            break; 

        case TOTAL_MEMORY_IN_USE_EVENT:
            static LongWord savedTotalInUse = -1;
            LongWord totalInUse;
            if( savedTotalInUse == (totalInUse = theHeap->totalInUse()) ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_TOTAL_MEMORY_IN_USE, totalInUse, "");
            savedTotalInUse = totalInUse;
            eventRegister[ foo ] = FALSE;
            break; 

        case FILE_LOAD_BEGINNING_EVENT:
            emitStatusMessage( MiscStrings[ 381 ] );
            tellController( E_FILE_LOAD_BEGINNING, 0, localLastFileLoadedName );
            eventRegister[ foo ] = FALSE;
            break; 

        case FILE_LOAD_SUCCEEDED_EVENT:
            emitStatusMessage( MiscStrings[ 382 ] );
            tellController( E_FILE_LOAD_SUCCEEDED, 0, localLastFileLoadedName );
            eventRegister[ foo ] = FALSE;
            break; 

        case FILE_LOAD_FAILED_EVENT:
            emitStatusMessage( MiscStrings[ 383 ] );
            tellController( E_FILE_LOAD_SUCCEEDED, 0, "");
            eventRegister[ foo ] = FALSE;
            break; 

        case WORLD_LOAD_SUCCEEDED_EVENT:
            tellController( E_WORLD_LOAD_SUCCEEDED, 0, localLastWorldLoadedName );
            eventRegister[ foo ] = FALSE;
            break; 

        case WORLD_LOAD_FAILED_EVENT:
            tellController( E_WORLD_LOAD_SUCCEEDED, 0, "");
            eventRegister[ foo ] = FALSE;
            break; 

        case WORLD_SAVE_SUCCEEDED_EVENT:
            tellController( E_WORLD_SAVE_SUCCEEDED, 0, "");
            eventRegister[ foo ] = FALSE;
            break; 

        case PARSER_STATUS_EVENT:
            if( ((theClowder->getKittenEvaluating( theKittenNumber)) == FALSE) && ParserActive )
                tellController( E_EXPECT_MORE_INPUT, 0, "");
            else
                tellController( E_DO_NOT_EXPECT_MORE_INPUT, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        case LIST_PRINT_DEPTH_EVENT:
            static UnsignedLongWord savedPrintDepth = -1;
            if( savedPrintDepth == printDepth ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_LIST_PRINT_DEPTH, printDepth, "");
            savedPrintDepth = printDepth;
            eventRegister[ foo ] = FALSE;
            break;

        case LIST_PRINT_LENGTH_EVENT:
            static UnsignedLongWord savedPrintLength = -1;
            if( savedPrintLength == printLength ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_LIST_PRINT_LENGTH, printLength, "");
            savedPrintLength = printLength;
            eventRegister[ foo ] = FALSE;
            break;

        case VECTOR_PRINT_DEPTH_EVENT:
            static UnsignedLongWord savedVPrintDepth = -1;
            if( savedVPrintDepth == vPrintDepth ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_VECTOR_PRINT_DEPTH, vPrintDepth, "");
            eventRegister[ foo ] = FALSE;
            savedVPrintDepth = vPrintDepth;
            break;

        case VECTOR_PRINT_LENGTH_EVENT:
            static UnsignedLongWord savedVPrintLength = -1;
            if( savedVPrintLength == vPrintLength ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_VECTOR_PRINT_LENGTH, vPrintLength, "");
            eventRegister[ foo ] = FALSE;
            savedVPrintLength = vPrintLength;
            break;

        case WAKED_UP_EVENT:
            tellController( E_WAKED_UP, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        case MEMORY_SIZE_EVENT:
            static LongWord savedMemorySize = -1;
            memorySizeNow = theHeap->currentStoreSize();
            if( savedMemorySize == memorySizeNow ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_MEMORY_SIZE, memorySizeNow, "");
            eventRegister[ foo ] = FALSE;
            savedMemorySize = memorySizeNow;
            break;

        case GENERATIONAL_MEMORY_SIZE_EVENT:
            static LongWord savedGenerationalMemorySize = -1;
            generationalMemorySizeNow = theHeap->currentActiveStoreSize();
            if( savedGenerationalMemorySize == generationalMemorySizeNow ) {
                eventRegister[ foo ] = FALSE;
                break;
                }
            tellController( E_GENERATIONAL_MEMORY_SIZE, generationalMemorySizeNow, "");
            eventRegister[ foo ] = FALSE;
            savedGenerationalMemorySize = generationalMemorySizeNow;
            break;

        case EXPECT_DOUBLE_QUOTE_EVENT:
            if( ReadingString )
                tellController( E_EXPECT_DOUBLE_QUOTE, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        case EXPECT_RPARENS_EVENT:
            if( RParensNeeded > 0 )
                tellController( E_EXPECT_PARENS, RParensNeeded, "");
            eventRegister[ foo ] = FALSE;
            break;

        case MAKE_FRONT_EVENT:
            tellController( E_MAKE_FRONT, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        case INIT_PROGRESS_EVENT:
            tellController( E_INIT_PROGRESS, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        case INCREMENT_PROGRESS_EVENT:
            tellController( E_INCREMENT_PROGRESS, progressFractionInThousandths, "");
            eventRegister[ foo ] = FALSE;
            break;

        case WRAPUP_PROGRESS_EVENT:
            tellController( E_WRAPUP_PROGRESS, 0, "");
            eventRegister[ foo ] = FALSE;
            break;

        default:
            useFatalErrorOutput();
            tprintf3( &ErrorPort, "%d\n", foo );
            fatalError( 50 );
            break;
        }
//    printf("E: Evaluator has dealt with event %d.\n", foo );
    }

/**** pollMom ****/

void pollMom()
{
    if( momSaysQuit() )
        exitTopLoop();

    if( momSaysWorldOp() ) {
        theClowder->setKittenSays( KITTEN_ACKNOWLEDGES_WORLD_OPERATION );
        while( momSaysWorldOp() )
            usleep( WORLD_OPERATION_DELAY );
        }

    //  The ifs must be in this order so that after a world has been loaded,
    //  all kittens will immediately reset.

    if( momSaysReset() ) {
        reportEvent( REQUEST_RESET_EVENT   );
        reportEvent( REQUEST_RESET_EVENT_2 );
        reportEvent( REQUEST_RESET_EVENT_3 );
        reportEvent( REQUEST_RESET_EVENT_4 );
        theClowder->setKittenSays( KITTEN_ACKNOWLEDGES_RESET );
        }

    }

/**** eventLoopOnce ****/

void eventLoopOnce( Word sleepTimeInMicroseconds )
{
    int counter = 0;
 
    if( ! isMomCatP )
        pollMom();

    pollControllerOnce();

    if( ! anyEvents )
        return;

    anyEvents = FALSE;

    if( eventRegister[ MAKE_FRONT_EVENT ] )
         dealWithEvent( MAKE_FRONT_EVENT );
    for( counter = 0; counter < NUMBER_OF_EVENTS; ++counter )
        if( eventRegister[ counter ] )
             dealWithEvent( counter );

    if( sleepTimeInMicroseconds > 0 )
        usleep( sleepTimeInMicroseconds );
    }

/**** reportInitialEvents ****/

void reportInitialEvents()
{
    reportEvent( LIST_PRINT_LENGTH_EVENT );
    reportEvent( LIST_PRINT_DEPTH_EVENT );
    reportEvent( VECTOR_PRINT_LENGTH_EVENT );
    reportEvent( VECTOR_PRINT_DEPTH_EVENT );
    reportEvent( MEMORY_SIZE_EVENT );
    reportEvent( GENERATIONAL_MEMORY_SIZE_EVENT );
    reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
    reportEvent( TOTAL_MEMORY_IN_USE_EVENT );
    reportEvent( INITIALIZATION_FINISHED_EVENT );
    }
