/****************************************************************************************

GarbageCollector.c++ -- Copyright (c) by Jay Reynolds Freeman 2007.

Garbage collection class for Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import <sys/mman.h>
#import "LocalTypes.h"
#import "WraithHeader.h"
#import "MemoryManager.h"
#import "OpCodes.h"
#import "GarbageCollector.h"
#import "WorldHandler.h"
#import "Registers.h"
#import "DataTypes.e"
#import "Events.e"
#import "Globals.e"
#import "Intern.e"
#import "IO.e"
#import "StreamHandling.e"

#define RUNNING_AS_BIG_ENDIAN (BigEndianFlag == TRUE)
#define WORLDS_ARE_LITTLE_ENDIAN TRUE

Boolean spliceGarbageOutOfList();

void GarbageCollector::initializeGC( MemoryManager *aMemoryManager )
{
    myMemoryManager = aMemoryManager;
    garbageCollectorRunning = FALSE;
    garbageCollectorEntryRoutineRunning = FALSE;
    gcRandomNumber = 0;
    scratchRandomNumber = 0;
    isMomCatLocalP = FALSE;
    myClowder = NULL;
    }

/**** setClowder ****/

void GarbageCollector::setClowder( Clowder *aClowder ) { myClowder = aClowder; }

/**** postGCAuxiliary -- portmanteau for common operations  ****/

void GarbageCollector::postGCAuxiliary( volatile memorySpace *currentSpace, volatile memorySpace *alternateSpace )
{
    theHeap->setSpaceForGenerationalGC( currentSpace );
    theHeap->setSpaceForGenerationalGC( alternateSpace );
    myClowder->setLiveStackIsA();
    theHeap->clearGenerationSpace( currentSpace->activeGeneration );
    theHeap->clearGenerationSpace( currentSpace->agingGeneration );
    theHeap->clearGenerationSpace( alternateSpace->activeGeneration );
    theHeap->clearGenerationSpace( alternateSpace->agingGeneration );
    }

/**** reportCheckOneRegisterError -- Auxiliary for "checkOneRegister", which follows. ****/

void GarbageCollector::reportCheckOneRegisterError( taggedAval *aReg, const Text* what, Word messageIndex, Boolean quitOnErrorP )
{
    if( quitOnErrorP )
        useFatalErrorOutput();
    tprintf3( &CurrentOutputPort, MiscStrings[ messageIndex], what );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    printAnyTaggedAval( &CurrentOutputPort, aReg );
    if( ! leaf_p( aReg->t ) ) {
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        printAnyTaggedAval( &CurrentOutputPort, (aReg->u).a );
        }
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    if( quitOnErrorP )
        fatalError( 66 );
    }

/****  checkOneRegister -- For debugging and paranoia ... ****/

Boolean GarbageCollector::checkOneRegister( taggedAval *aReg, volatile memorySpace *tenured,
                                            volatile memorySpace *active, volatile memorySpace *aging, const Text *what,
                                            Boolean quitOnErrorP )
{
    taggedAval *tap = 0;
    Boolean badP = FALSE;

    if( leaf_p( aReg->t ) )
        return badP;
    tap = (aReg->u).a;
    if( inSpaceP( tap, tenured ) ) {
        if( any_gc_bit_on( tap->t ) || any_recursion_bit_on( tap->t ) ) {
            reportCheckOneRegisterError( aReg, what, 454, quitOnErrorP );                             
            badP |= TRUE;
            }
        }
   if( (aging != 0) && inSpaceP( tap, aging ) ) {
        if( any_gc_bit_on( tap->t ) || any_recursion_bit_on( tap->t ) ) {
            reportCheckOneRegisterError( aReg, what, 455, quitOnErrorP );
            badP |= TRUE;
            }
        }
    if( (active != 0) && inSpaceP( tap, active ) ) {
        reportCheckOneRegisterError( aReg, what, 456, quitOnErrorP );
        badP |= TRUE;
        }
    if( (! inSpaceP( tap, tenured )) && (! inSpaceP( tap, active)) && (! inSpaceP( tap, aging )) ) {
        reportCheckOneRegisterError( aReg, what, 457, quitOnErrorP );
        badP |= TRUE;
        }
    if( ggc_forwarding_pointer_p( tap->t ) ) {
        reportCheckOneRegisterError( aReg, what, 462, quitOnErrorP );
        badP |= TRUE;
        }
    if( ggc_forwarding_pointer_p( aReg->t ) ) {
        reportCheckOneRegisterError( aReg, what, 461, quitOnErrorP );
        badP |= TRUE;
        }
    return badP;
    }

/****  checkStackRegisters -- For debugging and paranoia ... ****/

Boolean GarbageCollector::checkStackRegisters( volatile memorySpace *aStack, volatile memorySpace *tenured,
                                            volatile memorySpace *active, volatile memorySpace *aging, const Text *what,
                                            taggedAval *aStackPointer, Boolean quitOnErrorP )
{
    taggedAval *scratch = 0;
    Boolean badP = FALSE;

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        badP |= checkOneRegister( scratch, tenured, active, aging, what, quitOnErrorP );
    return badP;
    }

/**** checkRegistersForKitten -- For debugging and paranoia ... ****/

Boolean GarbageCollector::checkRegistersForKitten( UnsignedWord aKittenNumber, volatile memorySpace *tenured,
                                                   volatile memorySpace *active, volatile memorySpace *aging, const Text *message,
                                                   Boolean quitOnErrorP )
{
    Text kittenMessageBuffer[ 200 ];
    Boolean badP = FALSE;

    if( ! (myClowder->getKittenEvaluatorActivity( aKittenNumber ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
        return badP;

    sprintf( kittenMessageBuffer, "%s kitten %d localE", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localE),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localSavedE", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localSavedE),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localR", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localR),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localA", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localA),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localL", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localL),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localCurrentInputPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentInputPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localCurrentOutputPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentOutputPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localErrorPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localErrorPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localDefaultInputPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localDefaultInputPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localMasterDefaultInputPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localMasterDefaultInputPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    sprintf( kittenMessageBuffer, "%s kitten %d localTranscriptPort", message, aKittenNumber );
    badP |= checkOneRegister( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localTranscriptPort),
                              tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    return badP;
    }

/**** checkRegisters -- For debugging and paranoia... ****/

void GarbageCollector::checkRegisters( volatile memorySpace *tenured,
                                       volatile memorySpace *active, volatile memorySpace *aging, const Text *message,
                                       Boolean quitOnErrorP )
{
    Text kittenMessageBuffer[ 200 ];
    UnsignedWord counter = 0;
    Boolean badP = FALSE;

    sprintf( kittenMessageBuffer, "%s: %s", message, "E" );
    badP |= checkOneRegister( &(E), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "TopLevelE" );
    badP |= checkOneRegister( &(TopLevelE), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "SavedE" );
    badP |= checkOneRegister( &(SavedE), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "IVR" );
    badP |= checkOneRegister( &(IVR), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "R" );
    badP |= checkOneRegister( &(R), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "A" );
    badP |= checkOneRegister( &(A), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "L" );
    badP |= checkOneRegister( &(L), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "TLC" );
    badP |= checkOneRegister( &(TLC), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "CurrentInputPort" );
    badP |= checkOneRegister( &(CurrentInputPort), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "CurrentOutputPort" );
    badP |= checkOneRegister( &(CurrentOutputPort), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "ErrorPort" );
    badP |= checkOneRegister( &(ErrorPort), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "DefaultInputPort" );
    badP |= checkOneRegister( &(DefaultInputPort), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "MasterDefaultInputPort" );
    badP |= checkOneRegister( &(MasterDefaultInputPort), tenured, active, aging, kittenMessageBuffer,
                              quitOnErrorP );

    sprintf( kittenMessageBuffer, kittenMessageBuffer, message, "TranscriptPort" );
    badP |= checkOneRegister( &(TranscriptPort), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );

    for( counter = 0; counter < theNumberOfKittens; ++counter ) {
         sprintf( kittenMessageBuffer, "%s: kitten %d input queue", message, counter );
         badP |= checkOneRegister( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ,
                                   tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
         }
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_REGISTERS_TO_SCRATCH, KITTEN_HAS_WRITTEN_REGISTERS_TO_SCRATCH );

    OSMemoryBarrier();

    sprintf( kittenMessageBuffer, "%s:", message );
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        badP |= checkRegistersForKitten( counter, tenured, active, aging, "kittenMessageBuffer", quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "main stack" );
    badP |= checkStackRegisters( (myMemoryManager->SchemeStack), tenured, active, aging, kittenMessageBuffer, S, quitOnErrorP );

    sprintf( kittenMessageBuffer, "%s: %s", message, "main continuation" );
    badP |= checkStackRegisters( (myMemoryManager->SchemeContinuation), tenured, active, aging, kittenMessageBuffer, CS,
                                 quitOnErrorP );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            sprintf( kittenMessageBuffer, "%s: kitten %d stack", message, counter );
            badP |= checkStackRegisters( (myMemoryManager->ScratchStack), tenured, active, aging, kittenMessageBuffer,
                                         (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS,
                                         quitOnErrorP );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH,
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();

            sprintf( kittenMessageBuffer, "%s: kitten %d continuation", message, counter );
            badP |= checkStackRegisters( (myMemoryManager->ScratchStack), tenured, active, aging, kittenMessageBuffer,
                                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS,
                                        quitOnErrorP );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    sprintf( kittenMessageBuffer, "%s: %s", message, "ObList" );
    badP |= checkOneRegister( &(ObList), tenured, active, aging, kittenMessageBuffer, quitOnErrorP );
    }

/**** reportCheckSpaceError -- auxiliary for "checkSpace", which follows. ****/

void GarbageCollector::reportCheckSpaceError( volatile memorySpace *m, volatile Byte *bp, volatile Byte *lastbp,
                                              taggedAval *tap, taggedAval lastTag,
                                              Boolean lastValid, const Text *message, Word messageIndex, Boolean quitOnError )
{
    if( quitOnError )
        useFatalErrorOutput();
    printMemorySpace( m, FALSE );
    tprintf4( &CurrentOutputPort, MiscStrings[ messageIndex], message, (void *)bp );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    printAnyTaggedAval( &CurrentOutputPort, tap );
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    if( ! leaf_p( tap->t ) ) {
        printAnyTaggedAval( &CurrentOutputPort, (tap->u).a );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        }
    if( lastValid) {
        tprintf3( &CurrentOutputPort, MiscStrings[ 363 ], (void *)lastbp );
        if( byte_block_p( lastTag.t ) )
            displayByteBlockPointer( &CurrentOutputPort, (taggedAval *)lastbp );
        else
            printAnyTaggedAval( &CurrentOutputPort, &lastTag );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        }
    if( quitOnError )
        fatalError( 112 );
    }

/**** checkForPointsToActiveBits -- Report error if any pointsToActiveBits in the given space are on.  ****/

void GarbageCollector::checkForPointsToActiveBits( volatile memorySpace *m, const Text *message, Boolean quitOnError )
{
    Byte *tap;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        if( points_to_active_bit( ((taggedAval *)tap)->t ) ) {
            tprintf3( &CurrentOutputPort, "%s\n", message );
            printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)tap );
            tprintf2( &CurrentOutputPort, MiscStrings[ 465 ] );            
            if( quitOnError )
                fatalError( 112 );
            }
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    }

/**** checkSpace -- Integrity checks of the given space, used for tests before, during and after garbage collection. ****/

void GarbageCollector::checkSpace( volatile memorySpace *check, volatile memorySpace *pointInto, taggedAval* stackPointer,
                                   const Text *message, Boolean quitOnError )
{
    volatile Byte *bp = 0;
    volatile Byte *lastbp = 0;
    taggedAval *tap = 0;
    taggedAval lastTag = { (Tag)emptyListTag, {0} };
    Boolean lastValid = FALSE;
    taggedAval *whereItPoints = 0;
    UnsignedLongWord objectCount = 0;

    if( stackPointer == 0 )
        bp = check->free;
    else
        bp = (Byte *)stackPointer;

    for( ; bp < check->end; bp += lengthTaggedAvalPointer( (taggedAval *)bp ) ) {
        ++objectCount;
        tap = (taggedAval *)bp;

        if( (UnsignedLongWord)tap & (UnsignedLongWord)0x7 )
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 424, quitOnError );
        if( any_gc_bit_on( tap->t ) || any_recursion_bit_on( tap->t ) )
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 372, quitOnError );
        if( ggc_forwarding_pointer_p( tap->t ) )
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 461, quitOnError );
        if( byte_block_p( tap->t ) && car_p( tap->t ) )
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 369, quitOnError );
        if( forwarding_pointer_p( tap->t ) && car_p( tap->t ) )
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 365, quitOnError );
        if( ! leaf_p( tap->t ) ) {
            whereItPoints = (tap->u).a;
            if( whereItPoints == 0 )
                reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 417, quitOnError );
            if( ! inWholeSpaceP( whereItPoints, pointInto ) )
                reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 368, quitOnError );
            if( gc_bit( tap->t ) && (! gc_bit( whereItPoints->t )) )
                reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 371, quitOnError );
            }
        if( ! known_tag_p( tap->t ) ) {
            reportCheckSpaceError( check, bp, lastbp, tap, lastTag, lastValid, message, 364, quitOnError) ;
            break;
            }

        lastTag = *tap;
        lastbp = bp;
        lastValid = TRUE;
        }
    }

// Stuff for rebuilding the oblist follows...

/**** addIfPointsToSymbol  --  If the passed pointer points to a symbol, add the symbol to the growing new ObList. ****/

void GarbageCollector::addIfPointsToSymbol( taggedAval *tp )
{
    taggedAval *symbolPair = 0;

    if( symbol_p( tp->t ) ) {          // We are at a symbol -- which points to a list -- considering going to that list.
        symbolPair = spliceOutNotCarForwardingPointersIndirect( (tp->u).a );
        if( ! gc_bit( symbolPair->t ) ) {
            installSymbolInObList( tp );
            setGCBitNonAtomic( symbolPair->t );
            }
        }
    }

/**** addIfPointsToSymbolStack  --  Map "addIfPointsToSymbol" onto the passed stack.  ****/

void GarbageCollector::addIfPointsToSymbolStack( volatile memorySpace *aStack, taggedAval *aStackPointer )
{
    taggedAval *scratch = 0;

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        addIfPointsToSymbol( scratch );
    }

/**** anyBlockedSymbols -- Make a pass through memory, checking for symbols which are blocked;
                           that is, that have had blockOnTagLock1 or one of its friends called
                           on them, but have not (yet) been released from the lock so established.
                           The intent here is to assist "rebuildObList", which calls procedures
                           that in turn call "installSymbolInObList", which latter will hang
                           if it is passed a blocked symbol...    ****/
 

Boolean GarbageCollector::anyBlockedSymbols( volatile memorySpace *fromSpace )
{
    Byte *bp = 0;
    taggedAval *tp = 0;
    taggedAval *symbolPair = 0;        
    
    for( bp = (Byte *)(fromSpace->free); bp < (Byte *)(fromSpace->end); ) {
        tp = (taggedAval *)bp;
        if( symbol_p( tp->t ) ) {          // We are at a symbol -- which points to a list -- considering going to that list.
            symbolPair = spliceOutNotCarForwardingPointersIndirect( (tp->u).a );
            if( lockedP( &(tp->t) ) )
                return TRUE;
            }
        bp += lengthTaggedAvalPointer( tp );
        }

    return FALSE;
    }

/**** rebuildObList -- This function creates a new oblist, then populates it with non-garbage symbols.
                       It is called *after* a full garbage collection has been made, in which the oblist
                       was *not* used as a root of collection; thus the only symbols present in the new
                       memory space will be those which were not garbage because of some other
                       connection to a root than via the oblist.  It is thus only necessary to gather
                       them all into a new oblist.  To work properly, this function must be called
                       only when there is no pointer to the oblist stored in any register or in
                       the heap.  Code in "SECDEval.c++" reports that condition in a global.  ****/

void GarbageCollector::rebuildObList( volatile memorySpace *fromSpace )
{
    Byte *bp = 0;
    taggedAval *tp = 0;
    taggedAval *symbolPair = 0;        
    UnsignedWord counter = 0;

    PUSH
    MOVE( A, R )
    PUSH

    initializeObList( TRUE );              // This procedure should be run only by MomCat.

    for( bp = (Byte *)(fromSpace->free); bp < (Byte *)(fromSpace->end); ) {
        tp = (taggedAval *)bp;
        if( symbol_p( tp->t ) ) {          // We are at a symbol -- which points to a list -- considering going to that list.
            symbolPair = spliceOutNotCarForwardingPointersIndirect( (tp->u).a );
            if( ! gc_bit( symbolPair->t ) ) {         //  Probably unnecessary test, per above comment in description of method.
                installSymbolInObList( tp );          //    Leave in as defensive programming.
                setGCBitNonAtomic( symbolPair->t );
                }
            }
        bp += lengthTaggedAvalPointer( tp );
        }
    addIfPointsToSymbol( &E);
    addIfPointsToSymbol( &TopLevelE );
    addIfPointsToSymbol( &SavedE );
    addIfPointsToSymbol( &IVR );
    addIfPointsToSymbol( &R );
    addIfPointsToSymbol( &A );
    addIfPointsToSymbol( &L );
    addIfPointsToSymbol( &TLC );
    addIfPointsToSymbol( &CurrentInputPort );
    addIfPointsToSymbol( &CurrentOutputPort );
    addIfPointsToSymbol( &ErrorPort );
    addIfPointsToSymbol( &DefaultInputPort );
    addIfPointsToSymbol( &MasterDefaultInputPort );
    addIfPointsToSymbol( &TranscriptPort );

    for( counter = 0; counter < theNumberOfKittens; ++counter )
        addIfPointsToSymbol( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        if( ! (myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
            continue;
        myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_REGISTERS_TO_SCRATCH, KITTEN_HAS_WRITTEN_REGISTERS_TO_SCRATCH );
        }

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        if( ! (myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
            continue;
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localE) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localSavedE) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localR) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localA) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localL) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCurrentInputPort) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCurrentOutputPort) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localErrorPort) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localDefaultInputPort) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localMasterDefaultInputPort) );
        addIfPointsToSymbol( &((((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localTranscriptPort) );
        }

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        if( ! (myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
            continue;
        myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_REGISTERS_FROM_SCRATCH, KITTEN_HAS_COPIED_REGISTERS_FROM_SCRATCH );
        }

    addIfPointsToSymbolStack( myMemoryManager->SchemeStack, S );
    addIfPointsToSymbolStack( myMemoryManager->SchemeContinuation, CS );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        if( ! (myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
            continue;
        myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
        addIfPointsToSymbolStack( (myMemoryManager->ScratchStack),
                                  (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS );
        myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
        }

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        if( ! (myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
            continue;
        myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH, KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
        addIfPointsToSymbolStack( (myMemoryManager->ScratchStack),
                                  (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS );
        myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH, KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
        }

    clearAllGCAndRecursionBits( fromSpace );

    POP
    MOVE( R, A )
    POP
//    theClowder->setObListCount( countObList() );
    }
 
/**** clearAllGCBits -- Turn off all GC bits of both kinds in the given space.  ****/

void GarbageCollector::clearAllGCBits( volatile memorySpace *m )
{
    Byte *tap;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        clearBothGCBitsNonAtomic( ((taggedAval *)tap)->t );           
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    }

/**** clearAllPointsToActiveBits -- Turn off all pointsToActiveBits in the given space.  ****/

void GarbageCollector::clearAllPointsToActiveBits( volatile memorySpace *m )
{
    Byte *tap;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        clearPointsToActiveBitNonAtomic( ((taggedAval *)tap)->t );           
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    }

/**** clearAllGCAndRecursionBits -- Turn off all GC bits of both kinds, and all recursion
                                    bits of both kinds, in the given space.  ****/

void GarbageCollector::clearAllGCAndRecursionBits( volatile memorySpace *m )
{
    Byte *tap;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        clearBothGCBitsNonAtomic( ((taggedAval *)tap)->t );           
        clearBothRecursionBitsNonAtomic( ((taggedAval *)tap)->t );           
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    }

/**** clearAllGCAndRecursionBits -- Turn off all GC bits of both kinds, and all
                                    recursion bits of both kinds, in the given space.  ****/

void clearAllGCAndRecursionBits( volatile memorySpace *m )
{
    Byte *tap;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        clearBothGCBitsNonAtomic( ((taggedAval *)tap)->t );           
        clearBothRecursionBitsNonAtomic( ((taggedAval *)tap)->t );           
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    }

/**** markGarbageFromRegForKitten -- Mark all garbage from the roots of the given kitten. ****/

void GarbageCollector::markGarbageFromRegForKitten( UnsignedWord aKittenNumber,
                                                    volatile memorySpace *fromSpace, volatile memorySpace *toSpace,
                                                    UnsignedLongWord *nonGarbageCount )
{
    if( ! (myClowder->getKittenEvaluatorActivity( aKittenNumber ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
        return;
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localE),
                                            fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localSavedE),
                                            fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localR),
                                            fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localA),
                                            fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localL),
                                           fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg(
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentInputPort),
                           fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg(
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentOutputPort),
                           fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localErrorPort),
                                            fromSpace, toSpace );
    *nonGarbageCount
        += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localDefaultInputPort),
                               fromSpace, toSpace );
    *nonGarbageCount
        += markGarbageFromReg( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localMasterDefaultInputPort),
                               fromSpace, toSpace );
    *nonGarbageCount += markGarbageFromReg(
                            &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localTranscriptPort),
                            fromSpace, toSpace );
    }

/**** markGarbageFromRegForKittenGGC -- Mark all garbage from the roots of the given kitten. ****/

void GarbageCollector::markGarbageFromRegForKittenGGC( UnsignedWord aKittenNumber,
                                                       volatile memorySpace *tenured,
                                                       volatile memorySpace *aging,
                                                       volatile memorySpace *active,
                                                       volatile memorySpace *scratchSpace,
                                                       UnsignedLongWord *nonGarbageCount )
{
    if( ! (myClowder->getKittenEvaluatorActivity( aKittenNumber ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
        return;
    *nonGarbageCount += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localE),
                                               tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localSavedE),
                                               tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localR),
                                               tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localA),
                                               tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localL),
                                               tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC(
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentInputPort),
                           tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC(
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentOutputPort),
                           tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC( 
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localErrorPort),
                           tenured, aging, active, scratchSpace );
    *nonGarbageCount
        += markGarbageFromRegGGC( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localDefaultInputPort),
                                  tenured, aging, active, scratchSpace );
    *nonGarbageCount
        += markGarbageFromRegGGC(
                           &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localMasterDefaultInputPort),
                           tenured, aging, active, scratchSpace );
    *nonGarbageCount += markGarbageFromRegGGC(
                            &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localTranscriptPort),
                            tenured, aging, active, scratchSpace );
    }

/**** replaceRegisterPointersForKitten -- Replace all register pointers for the given kitten. ****/

void GarbageCollector::replaceRegisterPointersForKitten( UnsignedWord aKittenNumber,
                                                        volatile memorySpace *fromSpace,
                                                        volatile memorySpace *toSpace,
                                                        LongWord bumpDelta )
{
    if( ! (myClowder->getKittenEvaluatorActivity( aKittenNumber ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
        return;
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localE),
                            bumpDelta, "localE", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localSavedE),
                            bumpDelta, "localSavedE", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localR),
                            bumpDelta, "localR", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localA),
                            bumpDelta, "localA", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localL),
                            bumpDelta, "localL", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentInputPort),
                            bumpDelta, "localCurrentInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentOutputPort),
                            bumpDelta, "localCurrentOutputPort", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localErrorPort),
                            bumpDelta, "localErrorPort", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localDefaultInputPort),
                            bumpDelta, "localDefaultInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localMasterDefaultInputPort),
                            bumpDelta, "localMasterDefaultInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localTranscriptPort),
                            bumpDelta, "localTranscriptPort", fromSpace, toSpace );
    }

/**** fullGCMarkNonGarbage -- mark all non-garbage in Scheme main memory -- for fullGCType2.  ****/

UnsignedLongWord GarbageCollector::fullGCMarkNonGarbage( volatile memorySpace *fromSpace,
                                                         volatile memorySpace *toSpace, Boolean dealWithObList )
{
    UnsignedLongWord nonGarbageCount = 0;
    UnsignedWord counter = 0;

    nonGarbageCount += markGarbageFromReg( &(E), fromSpace, toSpace );
    eventLoopOnce( 0 );
    nonGarbageCount += markGarbageFromReg( &(TopLevelE), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(SavedE), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(IVR), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(R), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(A), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(L), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(TLC), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(CurrentInputPort), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(CurrentOutputPort), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(ErrorPort), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(DefaultInputPort), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(MasterDefaultInputPort), fromSpace, toSpace );
    nonGarbageCount += markGarbageFromReg( &(TranscriptPort), fromSpace, toSpace );
    eventLoopOnce( 0 );

    for( counter = 0; counter < theNumberOfKittens; ++counter )
        nonGarbageCount += markGarbageFromReg( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ, fromSpace, toSpace );
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_REGISTERS_TO_SCRATCH, KITTEN_HAS_WRITTEN_REGISTERS_TO_SCRATCH );

    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        markGarbageFromRegForKitten( counter, fromSpace, toSpace, &nonGarbageCount );
        eventLoopOnce( 0 );
        }

    nonGarbageCount += markGarbageFromStack( (myMemoryManager->SchemeStack),fromSpace, toSpace, S );
    nonGarbageCount += markGarbageFromStack( (myMemoryManager->SchemeContinuation),fromSpace, toSpace, CS );
    eventLoopOnce( 0 );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            nonGarbageCount += markGarbageFromStack( (myMemoryManager->ScratchStack), fromSpace, toSpace,
                                                     (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH,
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();
            nonGarbageCount += markGarbageFromStack( (myMemoryManager->ScratchStack), fromSpace, toSpace,
                                                     (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    if( dealWithObList )
        nonGarbageCount += markGarbageFromReg( &(ObList), fromSpace, toSpace );

    nonGarbageCount += expireForgettables( fromSpace, toSpace );

    return nonGarbageCount;

    }


#define KITTEN_NAME_LENGTH 100

/**** fullGCReplaceRegisterPointers -- replace all necessary register pointers, for fullGCType2.  ****/

void GarbageCollector::fullGCReplaceRegisterPointers( volatile memorySpace *fromSpace, volatile memorySpace *toSpace,
                                                      Boolean dealWithObList, LongWord bumpDelta )

{
    UnsignedWord counter = 0;
    Text kittenNameBuffer[ KITTEN_NAME_LENGTH ];

    replaceRegisterPointer( &E, bumpDelta, "E", fromSpace, toSpace );
    replaceRegisterPointer( &SavedE, bumpDelta, "SavedE",  fromSpace, toSpace );
    replaceRegisterPointer( &R, bumpDelta, "R", fromSpace, toSpace );
    replaceRegisterPointer( &A, bumpDelta, "A", fromSpace, toSpace );
    replaceRegisterPointer( &L, bumpDelta, "L", fromSpace, toSpace );
    replaceRegisterPointer( &TLC, bumpDelta, "TLC", fromSpace, toSpace );
    replaceRegisterPointer( &CurrentInputPort, bumpDelta, "CurrentInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &CurrentOutputPort, bumpDelta, "CurrentOutputPort", fromSpace, toSpace );
    replaceRegisterPointer( &ErrorPort, bumpDelta, "ErrorPort", fromSpace, toSpace );
    replaceRegisterPointer( &DefaultInputPort, bumpDelta, "DefaltInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &MasterDefaultInputPort, bumpDelta, "MasterDefaultInputPort", fromSpace, toSpace );
    replaceRegisterPointer( &TranscriptPort, bumpDelta, "TranscriptPort", fromSpace, toSpace );

    for( counter = 0; counter < theNumberOfKittens; ++counter )
        replaceRegisterPointer( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ, 
                                bumpDelta, "a KIQ", fromSpace, toSpace );

    for( counter = 1; counter < theNumberOfKittens; ++counter )
        replaceRegisterPointersForKitten( counter, fromSpace, toSpace, bumpDelta );

    replaceRegisterPointer( &TopLevelE, bumpDelta, "TopLevelE", fromSpace, toSpace );
    replaceRegisterPointer( &IVR, bumpDelta, "IVR", fromSpace, toSpace );

    if( dealWithObList )
        replaceRegisterPointer( &ObList, bumpDelta, "ObList", fromSpace, toSpace );

    replaceStackPointers( (myMemoryManager->SchemeStack), S, bumpDelta, 0, fromSpace, toSpace );
    eventLoopOnce( 0 );

    eventLoopOnce( 0 );

    replaceContinuationPointers( (myMemoryManager->SchemeContinuation), CS, bumpDelta, 0, fromSpace, toSpace );
    eventLoopOnce( 0 );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            replaceStackPointers( (myMemoryManager->ScratchStack), 
                                  (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS,
                                  bumpDelta, counter, fromSpace, toSpace );
            sprintf( kittenNameBuffer, "Stack for kitten %d", counter );
            eventLoopOnce( 0 );

            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH,
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();
            replaceContinuationPointers( (myMemoryManager->ScratchStack), 
                                         (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS,
                                         bumpDelta, counter, fromSpace, toSpace );
            sprintf( kittenNameBuffer, "Continuation for kitten %d", counter );
            eventLoopOnce( 0 );

            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();
    }


/**** fullGCType2 -- Perform a complete type 2 GC of everything in "fromSpace", into "toSpace". ****/

void GarbageCollector::fullGCType2( volatile memorySpace *fromSpace, volatile memorySpace *toSpace, Boolean dealWithObList )
{
    LongWord bumpDelta = 0;
    UnsignedLongWord nonGarbageCount = 0;
    UnsignedLongWord copiedAcrossCount = 0;
    UnsignedWord counter = 0;
    UnsignedLongWord oldLockNumber = 0;
    UnsignedLongWord newLockNumber = 0;

#if CHECK_SPACE_AFTER
    Text kittenNameBuffer[ KITTEN_NAME_LENGTH ];  
#endif

    taggedAval scratch1;
    taggedAval scratch2;

    if( garbageCollectorRunning )
        fatalError( 185 );
    garbageCollectorRunning = TRUE;

    eventLoopOnce( 0 );

    OSMemoryBarrier();

    if( toSpace->inUse )
        resetError( 7 );

    nonGarbageCount = fullGCMarkNonGarbage( fromSpace, toSpace, dealWithObList );

    if( theNumberOfKittens == 1 ) {                                            // No locks used for single Wraith Scheme process
        markNonGarbageCCNBlocksInSpaceAndMergeGarbage( fromSpace );            //  Mark car-cdr-next blocks in old space.
        }
    else {
        oldLockNumber = countNonGarbageLocksAndMergeGarbage( fromSpace );      //  Count the number of locks in old space.
        markNonGarbageCCNBlocksInSpace( fromSpace );                           //  Mark car-cdr-next blocks in old space.
        }
    eventLoopOnce( 0 );

    clearMemorySpace( toSpace );                                               //  Set up toSpace for the copy. 
    copiedAcrossCount = copyAcrossLow( fromSpace, toSpace, nonGarbageCount );  //  Copy non-garbage into low end of toSpace.
    eventLoopOnce( 0 );

    spliceOutAllForwardingPointers( toSpace, 0 );
    eventLoopOnce( 0 );

    bumpDelta = slideLowToHigh( toSpace, copiedAcrossCount );                  //  Slide toSpace content from low end to high.
    eventLoopOnce( 0 );

    //****************

    // At this point, no register has been touched.  Some of course pointed into fromSpace.  What is at
    // the places where they point, are forwarding pointers to toSpace locations which are no longer valid,
    // because of the "slideLowToHigh" of toSpace just done.  We must replace the pointer portion of any
    // such register with ((the "u.a" field of the forwarding pointer it pointed to) + bumpDelta).

    fullGCReplaceRegisterPointers( fromSpace, toSpace, dealWithObList, bumpDelta );

    //****************  Miscellaneous clean-up.

    // Clean up ports.

    scratch1.u.a = (taggedAval *)toSpace;
    theClowder->setScratch1( &scratch1 );
    scratch2.u.a = 0;
    theClowder->setScratch2( &scratch2 );

    cleanUpPorts( toSpace, 0 );
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) 
            myClowder->momDialogWithKitten( counter, MOM_SAYS_CLEAN_UP_PORTS, KITTEN_HAS_CLEANED_UP_PORTS );
    eventLoopOnce( 0 );

    OSMemoryBarrier();

    // Clean up some car-cdr-nil stuff.

    replaceCarCdrNilPointersToNilWithNil( toSpace );
    replaceCarCdrNextFollowedByNilWithCarCdrNil( toSpace );
    eventLoopOnce( 0 );

    // Clear GC and recursion bits.

    clearAllGCAndRecursionBits( toSpace );

    // Copy back the modified registers.

    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_REGISTERS_FROM_SCRATCH, KITTEN_HAS_COPIED_REGISTERS_FROM_SCRATCH );

    OSMemoryBarrier();

    // Check that lock numbers are consistent before and after GC.

    if( theNumberOfKittens > 1 ) {
        newLockNumber = countAllLocks( toSpace );
        if( newLockNumber != oldLockNumber ) {
            useFatalErrorOutput();
            tprintf4( &CurrentOutputPort, MiscStrings[ 406 ], oldLockNumber, newLockNumber );
            fatalError( 155 );
            }        
        }        

    OSMemoryBarrier();

    // Do some possibly-ifdefed-out checks.

#if CHECK_SPACE_AFTER
    checkSpace( toSpace, toSpace, 0, "Checking new space", TRUE );
    OSMemoryBarrier();
    checkSpace( myMemoryManager->SchemeContinuation, toSpace, CS, "Continuation for kitten 0, at end of GC", TRUE );
     for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH, 
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();
            sprintf( kittenNameBuffer, "Continuation for kitten %d, at end of GC", counter );
            checkSpace( myMemoryManager->ScratchStack, toSpace,
                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS,
                        kittenNameBuffer, TRUE );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();
    checkSpace( myMemoryManager->SchemeStack, toSpace, S,"Stack for kitten 0, at end of GC", TRUE );
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            sprintf( kittenNameBuffer, "Stack for kitten %d, at end of GC", counter );
            checkSpace( myMemoryManager->ScratchStack, toSpace, 0,
                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS,
                        kittenNameBuffer, TRUE );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
#endif

    eventLoopOnce( 0 );

    // Finally, swap the roles of main and alternate spaces.

    myMemoryManager->swapSpaces();
    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_SWAP_SPACES, KITTEN_HAS_SWAPPED_SPACES );

    OSMemoryBarrier();

    garbageCollectorRunning = FALSE;
    eventLoopOnce( 0 );
    }

/**** markFDIfPointedTo ****/

void markFDIfPointedTo( taggedAval *tp )
{
    UnsignedLongWord n = 0;

    if( port_p( tp->t ) ) {
        n = getPortBigArrayLocation( tp );              
        if( getPortKitten( tp ) == theKittenNumber ) {
            if( n < nInputFileDescriptors + nOutputFileDescriptors )
                (AllFileDescriptors[ n ]).fdInUse = TRUE;
            }
        }
    }

// Stuff for dealing with ports during garbage collection, follows.

/**** markFDIfPointedToFromStack -- wrapper to call "markFDIPointedTo" with each stack location. ****/

void markFDIfPointedToFromStack( volatile memorySpace *aStack, taggedAval *aStackPointer )
{
    taggedAval *scratch = 0;

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        markFDIfPointedTo( scratch );
    }

/**** cleanUpPorts -- Close every port that is garbage (that is, not in use). ****/

void GarbageCollector::cleanUpPorts( volatile memorySpace *m, volatile memorySpace *also )
{
    volatile Byte *bp = m->free;
    taggedAval *tp = 0;
    volatile Byte *stop = m->end;
    UnsignedLongWord i = 0;
    fileDescriptor *fd;
    UnsignedWord counter = 0;

    // Declare that all file descriptors are not in use.
 
    for( i = 0; i < nInputFileDescriptors + nOutputFileDescriptors; ++i )
        (AllFileDescriptors[i]).fdInUse = FALSE;

    // Find the ones that are in the heap; those are in use.

    for( bp = m->free;  bp < stop; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        markFDIfPointedTo( tp );
        }

    if( also ) {
        stop = also->end;
        for( bp = also->free;  bp < stop; bp += lengthTaggedAvalPointer( tp ) ) {
            tp = (taggedAval *)bp;
            markFDIfPointedTo( tp );
            }
        }

    markFDIfPointedTo( &E );
    markFDIfPointedTo( &TopLevelE );
    markFDIfPointedTo( &SavedE );
    markFDIfPointedTo( &IVR );
    markFDIfPointedTo( &R );
    markFDIfPointedTo( &A );
    markFDIfPointedTo( &L );
    markFDIfPointedTo( &TLC );
    markFDIfPointedTo( &CurrentInputPort );
    markFDIfPointedTo( &CurrentOutputPort );
    markFDIfPointedTo( &ErrorPort );
    markFDIfPointedTo( &DefaultInputPort );
    markFDIfPointedTo( &MasterDefaultInputPort );
    markFDIfPointedTo( &TranscriptPort );
    for( counter = 0; counter < theNumberOfKittens; ++counter )
        markFDIfPointedTo( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ );

    markFDIfPointedToFromStack( (myMemoryManager->SchemeStack), S );
    markFDIfPointedToFromStack( (myMemoryManager->SchemeContinuation), CS );

    conIn->fdInUse = TRUE;
    conOut->fdInUse = TRUE;
    conErr->fdInUse = TRUE;

    // If a file descriptor is not now marked in use, close it.

    for( i = 0; i < nInputFileDescriptors + nOutputFileDescriptors; ++i ) {
        fd = &(AllFileDescriptors[i]);
        if( ! fd->fdInUse )
            closePortGuts( fd );
        }

    // For cleanliness, assert that all file descriptors are not now in use.

    for( i = 0; i < nInputFileDescriptors + nOutputFileDescriptors; ++i )
        (AllFileDescriptors[i]).fdInUse = FALSE;

    // For paranoia ...

    OSMemoryBarrier();
    }

/**** markNonGarbageCCNBlocksInSpace  --  mark cdr-coded list blocks, for possible compacting during copy-across. ****/

void GarbageCollector::markNonGarbageCCNBlocksInSpace( volatile memorySpace *m )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;
    Boolean inBlockP = FALSE;

    for( bp = m->free; bp < m->end; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        if( inBlockP ) {
            if( ! gc_bit( tp->t ) ) {
                fatalError( 117 );
                }
            else if( ! car_cdr_next_p( tp->t ) )
                inBlockP = FALSE;
            // else nothing
            }
        else {
            if( ! gc_bit( tp->t ) )
                ;
            else if( car_cdr_next_p( tp->t ) ) {
                inBlockP = TRUE;
                setStartCCNBlockBitNonAtomic( tp->t );
                }
            // else nothing
            }
        }
    }

// Stuff for handling forgettable memory at garbage-collection time, follows:

/**** checkForgettables -- for debugging  ****/

void GarbageCollector::checkForgettables( volatile memorySpace *m, const Text *message )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;
    UnsignedLongWord nSeen = 0;

    tprintf3( &ErrorPort, "Start: %s", message );
    for( bp = m->free; bp < m->end; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        if( forgettable_unexpired_p( tp->t ) || forgettable_remembered_p( tp->t ) || forgettable_forgotten_p( tp->t ) ) {
            ++nSeen;
            printAnyTaggedAval( &ErrorPort, tp );
            tprintf2( &ErrorPort, MiscStrings[ 115 ] );
            if( ! inSpaceP( (tp->u).a, m ) )
                tprintf2( &ErrorPort, "POINTS OUT OF SPACE.\n");
            }
        }
    tprintf3( &ErrorPort, "Done: %s", message );
    }

/**** expireForgettables -- Changes expired forgettables to type forgettable-forgotten.
                            Must be used before GC bits of a GC are cleared.              ****/

UnsignedLongWord GarbageCollector::expireForgettables( volatile memorySpace *m, volatile memorySpace *scratch )
{
    taggedAval *tp = 0;
    taggedAval forgettableAsFP;
    taggedAval *tpNumber = 0;
    taggedAval *tpObject = 0;
    taggedAval *tpTime = 0;
    taggedAval *tpProbability = 0;
    volatile Byte *bp = 0;
    UnsignedLongWord timeNow = time( 0 );
    UnsignedLongWord expirationTime = 0;
    ShortFloat theProbability = 0.;
    UnsignedLongWord localNonGarbageCount = 0;
    unsigned int theRandom = 0;

    for( bp = m->free; bp < m->end; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        if( forgettable_unexpired_p( tp->t ) && gc_bit( tp->t) ) {     //  Found an unexpired, non-garbage forgettable.
            forgettableAsFP.u.a = (tp->u).a;
            forgettableAsFP.t = forwardingPointerTag;
            tpNumber = (tp->u).a;
            while( not_car_forwarding_pointer_p( tpNumber->t ) )
               tpNumber = (tpNumber->u).a;
            if( ! leaf_p( (tpNumber + 1)->t ) ) {
                tpObject = ((tpNumber + 1)->u).a;
                while( not_car_forwarding_pointer_p( tpObject->t ) )
                   tpObject = (tpObject->u).a;
                }
            else
                tpObject = tpNumber + 1;
            if( ! gc_bit( tpObject->t ) ) {                        //  Whose object is otherwise garbage.
                tpTime = tpNumber + 2;
                if( float_p( tpTime->t ) )
                    expirationTime = (UnsignedLongWord)((tpTime->u).f);
                else
                    expirationTime = (UnsignedLongWord)(tpTime->u).c;
                if( expirationTime < timeNow ) {                               //  Whose time is up.
                    tpProbability = tpNumber + 3;
                    if( fixnum_p( tpProbability->t ) ) {
                        tp->t &= ~pureTypeCodeBitsMask;
                        if( (tpProbability->u).c == 1 ) {          //  Must remember.
                            tp->t |= forgettableRememberedTag;
                            localNonGarbageCount += markGarbageFromReg( &forgettableAsFP, m, scratch );
                            }
                        else {
                            tp->t |= forgettableForgottenTag;          //  Must forget.
                            (tp->u).c = expirationTime;
                            }
                        }
                    else {                                             //  Must randomize its fate as directed.
                        theProbability = (tpProbability->u).f;
                        theRandom = (unsigned int)random();
                        if( (ShortFloat)theRandom < (theProbability * (ShortFloat)2147483648.0) ) {
                            tp->t |= forgettableRememberedTag;
                            localNonGarbageCount += markGarbageFromReg( &forgettableAsFP, m, scratch );
                            }
                        else {
                            tp->t |= forgettableForgottenTag;
                            (tp->u).c = expirationTime;
                            }
                        }
                    }
                else {                                                 // Forgettable has not expired, must de-garbage whole vector.
                    localNonGarbageCount += markGarbageFromReg( &forgettableAsFP, m, scratch );
                    }
                }
            else {                                                     // Object is not garbage, must de-garbage whole vector.
                tpObject->t &= ~gcBitMask;
                --localNonGarbageCount;
                if( gc_bit( tpNumber->t ) ) {
                    tpNumber->t &= ~gcBitMask;
                    --localNonGarbageCount;
                    }
                if( gc_bit( (tpNumber + 1)->t ) ) {
                    (tpNumber + 1)->t &= ~gcBitMask;
                    --localNonGarbageCount;
                    }
                localNonGarbageCount += markGarbageFromReg( &forgettableAsFP, m, scratch );
                }
            }
        }
    return localNonGarbageCount;
    }

/**** replaceCarCdrNilPointersToNilWithNil ****/

void GarbageCollector::replaceCarCdrNilPointersToNilWithNil( volatile memorySpace *m )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;

    for( bp = m->free; bp < m->end; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        if(    car_cdr_nil_p( tp->t )
            && forwarding_pointer_p( tp->t )
            && not_car_p( ((tp->u).a)->t )
            && empty_list_p( ((tp->u).a)->t )
            ) {
            clearCdrCodingBitsNonAtomic( tp->t );
            tp->t |= carCdrNil;
            }
        }
    }

/**** replaceCarCdrNextFollowedByNilWithCarCdrNil ****/

void GarbageCollector::replaceCarCdrNextFollowedByNilWithCarCdrNil( volatile memorySpace *m )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;

    for( bp = m->free; bp < m->end; bp += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)bp;
        if(    (! car_cdr_next_p( tp->t ))
            && not_car_p( (tp + 1)->t )
            && empty_list_p( (tp + 1)->t )
            ) {
            clearCdrCodingBitsNonAtomic( tp->t );
            tp->t |= carCdrNil;
            }
        }
    }

/**** slideLowToHigh -- Move the contents of the given space from low-justified to
                        high-justified.  Don't forget to bump the pointers in the process.   ****/

UnsignedLongWord GarbageCollector::slideLowToHigh( volatile memorySpace *m, UnsignedLongWord expectedObjectCount )
{
    volatile Byte *bp = m->space;
    taggedAval *tp = 0;
    UnsignedLongWord slide = m->end - m->lowFree;
    volatile Byte *stop = m->lowFree;
    size_t inUse = (size_t)(m->lowFree - m->space);
    UnsignedLongWord nObjectsCopied = 0;

    while( bp < stop ) {
        tp = (taggedAval *)bp;
        if( ! leaf_p( tp->t ) )
            (tp->u).a = (taggedAval *)(((Byte *)((tp->u).a)) + slide);
        ++nObjectsCopied;
        bp += lengthTaggedAvalPointer( tp );
        }

    memmove( (void *)(m->end - inUse), (void *)(m->space), inUse ); 
    m->free = m->end - inUse;

    if( nObjectsCopied != expectedObjectCount ) {
        useFatalErrorOutput();
        tprintf5( &CurrentOutputPort, MiscStrings[ 370 ], MiscStrings[ 482 ], nObjectsCopied, expectedObjectCount );
        fatalError( 104 );
        }
    return slide;   
    }
        
/**** spliceOutAllForwardingPointers -- Splice out all chains of forwarding pointers that originate in the given space.  ****/
                                                                           
void GarbageCollector::spliceOutAllForwardingPointers( volatile memorySpace *m, taggedAval *stackPointer )
{
    taggedAval *tap, *stop;
    Byte *nextByteAddress = 0;             // Messy; need this address both as pointer to taggedAval and as pointer to Byte.


    if( stackPointer == 0 ) {
        tap = (taggedAval *)(m->space);
        stop = (taggedAval *)(m->lowFree);
        }
    else {
        tap = stackPointer;
        stop = (taggedAval *)(m->end);
        }

    nextByteAddress = (Byte *)tap;

    while( tap < stop )
        if( leaf_p( tap->t ) ) {
            nextByteAddress += lengthTaggedAvalPointer( tap );
            tap = (taggedAval *)nextByteAddress;
            }
        else {
            (tap->u).a = spliceOutNotCarForwardingPointersIndirect( (tap->u).a );
            ++tap;
            nextByteAddress = (Byte *)tap;
            }
   }
                                
// Next batch of stuff is for doling out psuedorandom bits one at a time:  When "setGCRandomNumbers" is
// called, it fetches a new random number and initializes "scratchRandomNumber" to zero.  When "getRandomBit"
// is called, it in effect pulls the low bit off "gcRandomNumber" and right-rotates it (using "scratchRandomNumber"
// as an intermediary).  There need be only one random number per GC, so this stuff saves lots of calls to "random".

void GarbageCollector::setGCRandomNumbers() {
    gcRandomNumber = random();
    scratchRandomNumber = 0;
    }

UnsignedLongWord GarbageCollector::getRandomBit()
{
    UnsignedLongWord randomBit = 0;

    if( scratchRandomNumber == 0 )
        scratchRandomNumber = gcRandomNumber;

    randomBit = scratchRandomNumber & 1;
    scratchRandomNumber = scratchRandomNumber >> 1;
    return randomBit;
    }

/**** copyBlockLow -- Copy a block of given length into the given space.  Leave forwarding pointers
                      in the old space.  Return a pointer to the next position past the block
                      location in old space.  Set gcBit2 and clear gcBit1 on the objects
                      copied (that is, on the new forwarding pointers).                             */

Byte *GarbageCollector::copyBlockLow( taggedAval *blockPointer, memoryBlockSize blockLength, UnsignedLongWord nObjects,
                                      volatile memorySpace *to, calledFromWhere whence )
{
    Byte *transportedBlockPointer = 0;
    Byte *oldDummy = 0;
    Byte *transportedDummy = 0;
    UnsignedLongWord bump = 0;
    UnsignedLongWord counter = 0;

    myMemoryManager->allocateLow( to, blockLength );

    transportedBlockPointer = (Byte *)(to->lowFree) - blockLength;
    memcpy( transportedBlockPointer, (Byte *)blockPointer, (size_t)blockLength );

    oldDummy = (Byte *)blockPointer;
    transportedDummy = transportedBlockPointer;
    for( counter = 0; counter < nObjects; ++counter ) {
        if( gc_bit_2( ((taggedAval *)oldDummy)->t) ) {            // Object has already been copied.
            fatalError( 115 );
            }
        bump = lengthTaggedAvalPointer( (taggedAval *)transportedDummy );
        (((taggedAval *)oldDummy)->u).a = (taggedAval *)transportedDummy;
        ((taggedAval *)oldDummy)->t = (forwardingPointerTag | gcBit2Mask);
        oldDummy += bump;
        transportedDummy += bump;
        }
    
    return( (Byte *)blockPointer + blockLength );
    }           

// The next four functions are helpers for "copyAcrossLowAux".  The first three copy non-car objects, car-cdr-nil objects,
// and car-cdr-next blocks of objects, into memory space "to".  The fourth skips past a car-cdr-next block of objects.

Byte *GarbageCollector::copyNonCar( taggedAval *start, volatile memorySpace *m, volatile memorySpace *to )
{
    if( gc_bit_2( start->t) )             // Object has already been copied.
        fatalError( 113 );
    return copyBlockLow( start, lengthTaggedAvalPointer( start ), 1, to, COPY_NON_CAR);
    }

Byte *GarbageCollector::copyCCNil( taggedAval *start, volatile memorySpace *m, volatile memorySpace *to )
{
    if( gc_bit_2( start->t) )             // Object has already been copied.
        fatalError( 114 );
    return copyBlockLow( start, lengthTaggedAvalPointer( start ), 1, to, COPY_CC_NIL );
    }

/**** notCarFPDereferencesInSpace ****/

Boolean GarbageCollector::notCarFPDereferencesInSpace( taggedAval *xp, volatile memorySpace *m, taggedAval **whither,
                                                       UnsignedLongWord *nInSpaceFPsSeen )
{
   // Follow the non-car forwarding pointer pointed to till it either points to
   // something that is not a non-car forwarding pointer, or points outside
   // of space m.  Return true iff, it ends pointing within m,
   // and return what it points to in the passed parameter.

    UnsignedLongWord n = 0;
    taggedAval *last = 0;

    *whither = xp;

    while( inSpaceP( *whither, m ) && not_car_forwarding_pointer_p( (*whither)->t ) ) {
        if( (n++ > 0) && (! gc_bit_2( last->t )) ) {
            setGCBitNonAtomic( last->t );
            ++*nInSpaceFPsSeen;
            }
        last = *whither;
        *whither = ((*whither)->u).a;
        }
    if( (n++ > 0) && (! gc_bit_2( last->t )) ) {
        setGCBitNonAtomic( last->t );
        ++*nInSpaceFPsSeen;
        }
    
    return( inSpaceP( *whither, m ) );
    }

Byte *GarbageCollector::copyCCNextSection( taggedAval *start, volatile memorySpace *m, volatile memorySpace *to,
                                           UnsignedLongWord *nObjectsCopied, UnsignedLongWord *nNonGarbageObjects,
                                           UnsignedLongWord *nInSpaceFPsSeen )
{
    memoryBlockSize blockLength = 0;
    UnsignedLongWord nBlocksCopied = 0;
    UnsignedLongWord nObjectsInBlock = 0;

    taggedAval *scan = start;
    taggedAval *scanPlusOne = start + 1;
    taggedAval *scratchPtr = 0;

    Byte *returnValue = 0;                   // Should end up just past first block.

    while( TRUE ) {
    
        while( car_cdr_next_p( scan->t ) ) {
            if( ((Byte *)scan) >= m->end )       // This doesn't look like Kansas anymore ...
                fatalError( 101 );
            if( gc_bit_2( scan->t) )             // Object has already been copied.
                fatalError( 105 );
            if( ! gc_bit( scanPlusOne->t ) ) {   // If next object is garbage it had better be a non-car forwarding pointer.
                if( not_car_p( scanPlusOne->t ) && forwarding_pointer_p( scanPlusOne->t ) )
                    ;                            // Fine.
                else
                    fatalError( 99 );
                }
            ++scan;
            ++scanPlusOne;            
            ++nObjectsInBlock;
            }
    
        // With respect to the following test, note that when copying a CCN-coded circular list, the copy mechanism
        //   will first note that the end of the list points to an uncopied block (the list itself), then copy the
        //   list, then "continue", and find itself about to copy a block that is already copied.  Hence the 
        //   "nBlocksCopied == 0" prelude to the error 'cause that last is not an error.
        //
        // But if nBlocksCopied is zero, and we are somehow trying to copy a block with nothing in it, it
        //   means that the original, non-recursive call to this routine was in error.

        if( nObjectsInBlock == 0 )
            if( nBlocksCopied == 0 )
                fatalError( 118 );
            else {
                if( forwarding_pointer_p( scan->t ) && (! inSpaceP( (scan->u).a, m) ) )  {  //  Fine -- copy the FP.
                    blockLength += lengthTaggedAvalPointer( scan );
                    if( nBlocksCopied == 0 )                                                // We just checked and it wasn't ...
                        returnValue = (Byte *)start + blockLength;
                    clearGCBit2NonAtomic( scan->t );             // Clear the "already copied" bit, so "copyBlockLow" won't complain.
                    copyBlockLow( start, blockLength, nObjectsInBlock + 1, to, COPY_CC_NEXT_SECTION_NOT_CAR_POSITION_FIVE );
                    ++nBlocksCopied;
                    *nObjectsCopied += nObjectsInBlock + 1;
                    *nNonGarbageObjects += nObjectsInBlock;      // Not "+ 1" -- the forwarding pointer is already counted.
                    return returnValue;
                    }
                else
                    fatalError( 119 );
                }

        // End of block.  Tentatively set blockLength.
    
        blockLength = ((Byte *)scan) - ((Byte *)start);  // So far.
    
        switch( cdr_code( scan->t ) ) {
            case carCdrNil:
                blockLength += lengthTaggedAvalPointer( scan );
                if( nBlocksCopied == 0 )
                    returnValue = (Byte *)start + blockLength;
                if( blockLength > 0 )
                    copyBlockLow( start, blockLength, nObjectsInBlock + 1, to, COPY_CC_NEXT_SECTION_CAR_CDR_NIL );
                *nObjectsCopied += nObjectsInBlock + 1;
                *nNonGarbageObjects += nObjectsInBlock + 1;
                return returnValue;
                break;
            case notCar:
                if( ! (forwarding_pointer_p( scan->t )) ) {                              // Append the last cell to the block.
                    blockLength += lengthTaggedAvalPointer( scan );
                    if( nBlocksCopied == 0 )
                        returnValue = (Byte *)start + blockLength;
                    copyBlockLow( start, blockLength, nObjectsInBlock + 1, to, COPY_CC_NEXT_SECTION_NOT_CAR_POSITION_ONE);
                    ++nBlocksCopied;
                    *nObjectsCopied += nObjectsInBlock + 1;
                    *nNonGarbageObjects += nObjectsInBlock + 1;
                    return returnValue;
                    }
                else {                                                     // Not-car forwarding pointer.
                    if( ! inSpaceP( (scan->u).a, m) )                      // Error -- should not have been copied.
                        fatalError( 116 );
                    if( notCarFPDereferencesInSpace( scan, m, &scratchPtr, nInSpaceFPsSeen ) ) { // There is something to copy:
                        if( start_ccn_block_bit( scratchPtr->t ) ) {            // An un-copied CCN block:  Copy current block and ...
                            if( nBlocksCopied == 0 )
                                returnValue = (Byte *)start + blockLength;
                            copyBlockLow( start, blockLength, nObjectsInBlock, to, COPY_CC_NEXT_SECTION_NOT_CAR_POSITION_TWO );
                            ++nBlocksCopied;
                            *nObjectsCopied += nObjectsInBlock;
                            *nNonGarbageObjects += nObjectsInBlock;
                            start = scan = scratchPtr;                     //   ... set up to append the new block.
                            scanPlusOne = scan + 1;                    
                            blockLength = 0;
                            nObjectsInBlock = 0;
                            continue;
                            }
                        else {                                             // Put the dereferenced fp in place, and copy the whole.
                            (scan->u).a = scratchPtr;
                            blockLength += lengthTaggedAvalPointer( scan );
                            if( nBlocksCopied == 0 )
                                returnValue = (Byte *)start + blockLength;
                            copyBlockLow( start, blockLength, nObjectsInBlock + 1, to, COPY_CC_NEXT_SECTION_NOT_CAR_POSITION_THREE);
                            ++nBlocksCopied;
                            *nObjectsCopied += nObjectsInBlock + 1;
                            *nNonGarbageObjects += nObjectsInBlock + 1;
                            return returnValue;
                            }
                        }
                    else {                                                 // Forwarding pointer to an object already copied; 
                        (scan->u).a = scratchPtr;                          // Put dereferenced fp in place, and copy the whole.
                        blockLength += lengthTaggedAvalPointer( scan );
                        if( nBlocksCopied == 0 )
                            returnValue = (Byte *)start + blockLength;
                        copyBlockLow( start, blockLength, nObjectsInBlock + 1, to, COPY_CC_NEXT_SECTION_NOT_CAR_POSITION_FOUR);
                        ++nBlocksCopied;
                        *nObjectsCopied += nObjectsInBlock + 1;
                        *nNonGarbageObjects += nObjectsInBlock + 1;
                        return returnValue;
                        }
                    }
                break;
            default:
                fatalError( 94 );
                break;
            }
        }    
    return 0;  // Not reached
    }    

Byte *GarbageCollector::skipCCNextSection( taggedAval *start, volatile memorySpace *m )
{
    memoryBlockSize blockLength = 0;
    UnsignedLongWord nObjectsInBlock = 0;

    taggedAval *scan = start;
    taggedAval *scanPlusOne = start + 1;

    Byte *returnValue = 0;                   // Should end up just past first block.

    while( TRUE ) {
    
        while( car_cdr_next_p( scan->t ) ) {
            if( ((Byte *)scan) >= m->end )       // This doesn't look like Kansas anymore ...
                fatalError( 106 );
            if( gc_bit_2( scan->t) )             // Object has already been copied.
                fatalError( 107 );
            if( ! gc_bit( scanPlusOne->t ) ) {   // If next object is garbage it had better be a non-car forwarding pointer.
                if( not_car_p( scanPlusOne->t ) && forwarding_pointer_p( scanPlusOne->t ) )
                    ;                            // Fine.
                else {
                    fatalError( 108 );
                    }
                }
            ++scan;
            ++scanPlusOne;            
            ++nObjectsInBlock;
            }
    
        // End of block.  Note blocklength, so far.

        blockLength = ((Byte *)scan) - ((Byte *)start);  
    
        switch( cdr_code( scan->t ) ) {
            case carCdrNil:
            case notCar:
                blockLength += lengthTaggedAvalPointer( scan );
                returnValue = (Byte *)start + blockLength;
                return returnValue;
                break;
            default:
                fatalError( 109 );
                break;
            }
        }    
    return 0;  // Not reached
    }    

/**** copyAcrossLowAUX2 -- Auxiliary for "copyAcrossLow", that does most of the work. ****/

UnsignedLongWord GarbageCollector::copyAcrossLowAUX2( volatile memorySpace *m, volatile memorySpace *to, 
                                                      UnsignedLongWord *nNonGarbageObjects, Boolean useRandomizerP,
                                                      UnsignedLongWord *nInSpaceFPsSeen,
                                                      UnsignedWord whichKitten )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;
    volatile Byte *nextBP = 0;
    UnsignedLongWord nObjectsCopied = 0;
    UnsignedLongWord kittenShifted = (UnsignedLongWord)whichKitten << kittenByteOffset;
//    UnsignedWord nKittensShifted = theNumberOfKittens << kittenByteOffset;

    setGCRandomNumbers();

    for( bp = m->free; bp < m->end; bp = nextBP ) {
        tp = (taggedAval *)bp;
        if( gc_bit_2( tp->t ) ) {
            nextBP = bp + lengthTaggedAvalPointer( (tp->u).a );
            }
        else if( ! gc_bit( tp->t ) ) {
            nextBP = bp + lengthTaggedAvalPointer( tp );
            }
        else if( ((tp->t) & kittenByteBitsMask) == kittenShifted ) {
            if( car_cdr_next_p( tp->t ) ) {
    
                //  The randomization deals with the case where there are two car-cdr-next objects in memory:
                //
                //         ... AAAAAA ... BBBBBBBBBB ...  ( memory addresses increase this way --> )
                //
                //  and the last cdr of B points to the start of A.  These two blocks will not get conjoined
                //  if block A is always copied before block B.  Randomization ensures that eventually, B will
                //  get copied first, and the blocks will become ... BBBBBBBBBBAAAAAA ... , as they should.
                //
                //  This function is called twice by "copyAcrossLow", first with "useRandomizerP" TRUE, and then
                //  with it FALSE, so that every garbage collection does copy all blocks -- conjoined or not.
     
                if( (! useRandomizerP) || getRandomBit() )
                    nextBP = copyCCNextSection( tp, m, to, &nObjectsCopied, nNonGarbageObjects, nInSpaceFPsSeen );
                else
                    nextBP = skipCCNextSection( tp, m );
                }
            else if( car_cdr_nil_p( tp->t ) ) {
                nextBP = copyCCNil( tp, m, to );
                ++nObjectsCopied;
                ++(* nNonGarbageObjects);
                }
            else if( forwarding_pointer_p( tp->t ) ) {
                nextBP = copyNonCar( tp, m, to );
                ++(* nNonGarbageObjects);
                ++nObjectsCopied;
                }
            else  { // Not car, not forwarding-pointer
                nextBP = copyNonCar( tp, m, to );
                ++nObjectsCopied;
                ++(* nNonGarbageObjects);
                }
            }
        else {

            if( car_cdr_next_p( tp->t ) )
                nextBP = skipCCNextSection( tp, m );
            else
                nextBP = bp + lengthTaggedAvalPointer( tp );
            }
        }
    return nObjectsCopied;
    }

/**** copyAcrossLowAUX3 -- Auxiliary for "copyAcrossLow", that deals with leftover forwarding pointers:

    The problem here is that the "copyAcrossLowAUX2" algorithm does not copy all non-garbage
    forwarding pointers that are at the ends of cdr-coded blocks.  If such a pointer is 
    referenced by a register, then a subsequent part of the entire garbage-collection algorithm,
    that dereferences pointer chains that pass through old space, will fail to do its job, because
    the aforementioned forwarding pointers will not be properly redirected to new space.

    This procedure deals with those forwarding pointers.

****/

UnsignedLongWord GarbageCollector::copyAcrossLowAUX3( volatile memorySpace *m, volatile memorySpace *to, 
                                                      UnsignedLongWord *nNonGarbageObjects,
                                                      UnsignedLongWord *nInSpaceFPsSeen,
                                                      UnsignedWord whichKitten )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;
    volatile Byte *nextBP = 0;
    UnsignedLongWord nObjectsCopied = 0;
    UnsignedLongWord kittenShifted = (LongWord)whichKitten << kittenByteOffset;

    for( bp = m->free; bp < m->end; bp = nextBP ) {
        tp = (taggedAval *)bp;
        
        if( gc_bit_2( tp->t ) )               // Already copied -- bump pointer by length of copied object.
            nextBP = bp + lengthTaggedAvalPointer( (tp->u).a );
        else if( ! gc_bit( tp->t ) )          // Garbage, so has not been copied.
            nextBP = bp + lengthTaggedAvalPointer( tp );
        else if( (((tp->t) & kittenByteBitsMask) == kittenShifted)  && not_car_forwarding_pointer_p( tp->t ) ) {

            // The case of interest -- an uncopied non-garbage forwarding pointer.  Copy it.

            nextBP = copyNonCar( tp, m, to );
            ++nObjectsCopied;
            ++(* nNonGarbageObjects);
            }
        else if( car_cdr_next_p( tp->t ) )    // Not garbage, so can skip using length in the old space.
            nextBP = skipCCNextSection( tp, m );
        else                                  // Not garbage, so can skip using length in the old space.
            nextBP = bp + lengthTaggedAvalPointer( tp );

        }
    return nObjectsCopied;
    }

/**** checkSpaceForUncopiedNonGarbage -- Was used for debugging. ****/

void checkSpaceForUncopiedNonGarbage( volatile memorySpace *m )
{
    taggedAval *tp = 0;
    volatile Byte *bp = 0;
    volatile Byte *nextBP = 0;

    for( bp = m->free; bp < m->end; bp = nextBP ) {
        tp = (taggedAval *)bp;
        if( gc_bit_2( tp->t ) ) {       // Copied.
            nextBP = bp + lengthTaggedAvalPointer( (tp->u).a );
            }
        else if( ! gc_bit( tp->t ) ) {  // Garbage.
            nextBP = bp + lengthTaggedAvalPointer( tp );
            }
        else {                          // Not garbage, but not copied.
            printAnyTaggedAval( &CurrentOutputPort, tp );
            tprintf2( &CurrentOutputPort, "\n");
            nextBP = bp + lengthTaggedAvalPointer( tp );
            }
        }
    }

/**** copyAcrossLow -- copy non-garbage from "m" to "to", and check that the number of objects copied was correct. ****/

UnsignedLongWord GarbageCollector::copyAcrossLow( volatile memorySpace *m, volatile memorySpace *to,
                                                  UnsignedLongWord expectedObjectCount )
{
    UnsignedLongWord nObjectsCopied = 0;
    UnsignedLongWord nNonGarbageObjects = 0;
    UnsignedLongWord nInSpaceFPsSeen = 0;
    UnsignedWord kitten = 0;
    UnsignedWord maxKittensAndMomCat = 1 + ABSOLUTE_MAX_KITTENS;

    //  In the following for-loops, we iterate over the maximum possible number of kittens (counting the
    //  MomCat) in order to allow for the possibility that we have just loaded a world that was created
    //  in a Wraith Scheme instance that had more kittens that the one we are now running.

//    for( kitten = 0; kitten < theNumberOfKittens; ++kitten ) {
    for( kitten = 0; kitten < maxKittensAndMomCat; ++kitten ) {
        nObjectsCopied += copyAcrossLowAUX2( m, to, &nNonGarbageObjects, TRUE,  &nInSpaceFPsSeen, kitten );
        nObjectsCopied += copyAcrossLowAUX2( m, to, &nNonGarbageObjects, FALSE, &nInSpaceFPsSeen, kitten );
        }

//    for( kitten = 0; kitten < theNumberOfKittens; ++kitten )
    for( kitten = 0; kitten < maxKittensAndMomCat; ++kitten )
        nObjectsCopied += copyAcrossLowAUX3( m, to, &nNonGarbageObjects,        &nInSpaceFPsSeen, kitten );

    if( nNonGarbageObjects != expectedObjectCount ) {
        useFatalErrorOutput();
        tprintf7( &CurrentOutputPort, MiscStrings[ 370 ], MiscStrings[ 483 ], 
                  nNonGarbageObjects, expectedObjectCount, nObjectsCopied, nInSpaceFPsSeen );
        fatalError( 104 );
        }
    return nObjectsCopied;
    }

/**** markGarbageFromReg -- Perform a no-nonsense iterative marking of non-garbage, starting
                            from where the given register points, using the new space as
                            scratch storage for back-pointers.
   
  The function uses GC_BIT to mark non-garbage, and uses the recursion bits to record whether its
  traverse of the heap has left a given location in the car-wise / cdr-wise direction.           ****/

UnsignedLongWord GarbageCollector::markGarbageFromReg( taggedAval *aReg, volatile memorySpace *m, volatile memorySpace *scratchSpace )
{
    LongWord delta = scratchSpace->space - m->space;
    UnsignedLongWord count = 0;
    taggedAval *tp = 0;
    taggedAval *tpInScratch = 0;
    taggedAval *tpNext = 0;
    taggedAval *lastTP = 0;

    UnsignedLongWord recursionBitCdrsSet = 0;
    UnsignedLongWord recursionBitCdrsCleared = 0;

    if( leaf_p( aReg->t ) )
        return 0;

    tp = (aReg->u).a;
    tpInScratch = (taggedAval *)((Byte *)tp + delta);

    if( gc_bit( tp->t ) )
        return 0;

    while( tp != 0 ) {
        if( recursion_bit_car( tp->t ) ) {                        // Recursion bit for car is set
            if( recursion_bit_cdr( tp->t ) ) {                    // We have recursed -- or tried to -- in both directions from here.
                clearBothRecursionBitsNonAtomic( tp->t );
                ++recursionBitCdrsCleared;
                tp = (tpInScratch->u).a;                          // Recover stored back pointer.
                tpInScratch = (taggedAval *)((Byte *)tp + delta);
                }
            else {                                                // Recursion bit for cdr is not set, must consider "recursing"
                setRecursionBitCdrNonAtomic( tp->t );             //   in the cdr direction.
                ++recursionBitCdrsSet;
                tpNext = tp + 1;
                if(    car_cdr_next_p( tp->t )                    // There is some place to go, and the GC bit is not set.
                    && gc_bit_off( tpNext->t )
                    ) {
                    lastTP = tp;                                  // Set up for next save.
                    tp = tpNext;
                    }
                }
            }
        else {                                                   // Recursion bit for car is not set
            if( recursion_bit_cdr( tp->t ) ) {                   // Recursion bit for car is not set, but for cdr is -- error.
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ]);
                printAnyTaggedAval( &CurrentOutputPort, tp );
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ]);
                fatalError( 59 );
                }
            else {                                                // Neither recursion bit is set.
                setGCBitNonAtomic( tp->t );                       // Never been here, so set the gc bit to show "not garbage".
                ++count;
                tpInScratch = (taggedAval *)((Byte *)tp + delta);
                (tpInScratch->u).a = lastTP;                      // Consider recursing carwise, and since we just got here for the
                tpNext = (tp->u).a;                               //   first time, we save the pointer for returning from this cell.
                setRecursionBitCarNonAtomic( tp->t );
                if(    (! leaf_p( tp->t))                         // There is some place to go, and the GC bit is not set.
                    && gc_bit_off( tpNext->t )
                    ) {
                    if( ! forgettable_unexpired_p( tp->t ) ) {
                        lastTP = tp;                              // Set up for next save.
                        tp = tpNext;
                        }
                    }
                }
            }
        }
    if( recursionBitCdrsSet != recursionBitCdrsCleared )
        fatalError( 110 );
    return count;
    }

/**** markGarbageFromStack -- wrapper to call "markGarbageFromReg" with each stack location. ****/

UnsignedLongWord GarbageCollector::markGarbageFromStack( volatile memorySpace *aStack, volatile memorySpace *m,
                                                         volatile memorySpace *scratchSpace,
                                                         taggedAval *aStackPointer )
{
    UnsignedLongWord count = 0;
    taggedAval *scratch = 0;

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        count += markGarbageFromReg( scratch, m, scratchSpace );
    return count;
    }

/**** markGarbageFromRegGGC -- Perform a no-nonsense iterative marking of non-garbage, starting
                               from where the given register points, using the given scratch space
                               as scratch storage for back-pointers, and continuing only if the
                               place pointed to is *not* in space "tenured" -- which is intended to be
                               the tenured space.
   
  The function uses GC_BIT to mark non-garbage, and uses the recursion bits to record whether its
  traverse of the heap has left a given location in the car-wise / cdr-wise direction.           ****/

UnsignedLongWord GarbageCollector::markGarbageFromRegGGC( taggedAval *aReg,     // a tagged aval that may point into the heap
                                                          volatile memorySpace *tenured,
                                                          volatile memorySpace *aging,
                                                          volatile memorySpace *active,
                                                          volatile memorySpace *scratchSpace )
{
    LongWord delta = scratchSpace->space - tenured->space;  // Assumes that the space where garbage is is high in tenured space.
    UnsignedLongWord count = 0;
    taggedAval *tp = 0;
    taggedAval *tpInScratch = 0;
    taggedAval *tpNext = 0;
    taggedAval *lastTP = 0;

    UnsignedLongWord recursionBitCdrsSet = 0;
    UnsignedLongWord recursionBitCdrsCleared = 0;

    if( leaf_p( aReg->t ) )
        return 0;

    tp = (aReg->u).a;

    if( inSpaceP( tp, tenured ) )
        return 0;

    if( gc_bit( tp->t ) )
        return 0;

    while( tp != 0 ) {
        if( recursion_bit_car( tp->t ) ) {                        // Recursion bit for car is set
            if( recursion_bit_cdr( tp->t ) ) {                    // We have recursed -- or tried to -- in both directions from here.
                clearBothRecursionBitsNonAtomic( tp->t );
                ++recursionBitCdrsCleared;
                tp = (tpInScratch->u).a;                          // Recover stored back pointer.
                tpInScratch = (taggedAval *)((Byte *)tp + delta);
                }
            else {                                                // Recursion bit for cdr is not set, must consider "recursing"
                setRecursionBitCdrNonAtomic( tp->t );             //   in the cdr direction.
                ++recursionBitCdrsSet;
                tpNext = tp + 1;
                if(    car_cdr_next_p( tp->t )                    // There is some place to go, and the GC bit is not set.
                    && gc_bit_off( tpNext->t )
                    && (   inSpaceP( tpNext, active )
                        || inSpaceP( tpNext, aging ) )            // And we are in the spaces where we need to collect garbage
                    ) {
                    lastTP = tp;                                  // Set up for next save.
                    tp = tpNext;
                    }
                }
            }
        else {                                                   // Recursion bit for car is not set
            if( recursion_bit_cdr( tp->t ) ) {                   // Recursion bit for car is not set, but for cdr is -- error.
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ]);
                printAnyTaggedAval( &CurrentOutputPort, tp );
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ]);
                fatalError( 70 );
                }
            else {                                                // Neither recursion bit is set.
                setGCBitNonAtomic( tp->t );                       // Never been here, so set the gc bit to show "not garbage".
                ++count;
                tpInScratch = (taggedAval *)((Byte *)tp + delta);
                (tpInScratch->u).a = lastTP;                      // Consider recursing carwise, and since we just got here for the
                tpNext = (tp->u).a;                               //   first time, we save the pointer for returning from this cell.
                setRecursionBitCarNonAtomic( tp->t );
                if(    (! leaf_p( tp->t))                         // There is some place to go, and the GC bit is not set.
                    && gc_bit_off( tpNext->t )
                    && (   inSpaceP( tpNext, active )
                        || inSpaceP( tpNext, aging ) )            // And we are in the spaces where we need to collect garbage
                    ) {
                    if( ! forgettable_unexpired_p( tp->t ) ) {
                        lastTP = tp;                              // Set up for next save.
                        tp = tpNext;
                        }
                    }
                }
            }
        }
    if( recursionBitCdrsSet != recursionBitCdrsCleared )
        fatalError( 71 );
    return count;
    }

/**** markGarbageFromStackGGC -- wrapper to call "markGarbageFromRegGGC" with each stack location.
                                 "GGC" <==> generational garbage collection ****/

UnsignedLongWord GarbageCollector::markGarbageFromStackGGC( volatile memorySpace *aStack,
                                                            volatile memorySpace *tenured,
                                                            volatile memorySpace *aging,
                                                            volatile memorySpace *active,
                                                            volatile memorySpace *scratchSpace,
                                                            taggedAval *aStackPointer )
{
    UnsignedLongWord count = 0;
    taggedAval *scratch = 0;

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        count += markGarbageFromRegGGC( scratch, tenured, aging, active, scratchSpace );
    return count;
    }

/**** markGarbageFromAddressStackReferences -- wrapper to call "markGarbageFromRegGGC" with the location pointed to
                                               by non-leaf tagged avals at each stacked address.
                                               "GGC" <==> generational garbage collection  ****/

UnsignedLongWord GarbageCollector::markGarbageFromAddressStackReferences( taggedAval **start, taggedAval **end,
                                                                          volatile memorySpace *tenured,
                                                                          volatile memorySpace *aging,
                                                                          volatile memorySpace *active,
                                                                          volatile memorySpace *scratchSpace,
                                                                          const Text *message
                                                                          )
{
    UnsignedLongWord count = 0;
    taggedAval **scratch = 0;

    for( scratch = start; scratch < end; ++scratch ) {
        count += markGarbageFromRegGGC( *scratch, tenured, aging, active, scratchSpace );
        }
    return count;
    }

/**** replaceRegisterPointer -- Bump register pointer from oldspace to new.  ****/

void GarbageCollector::replaceRegisterPointer( taggedAval *aReg, LongWord bumpDelta, const Text *name,
                                                     volatile memorySpace *from, volatile memorySpace *to )
{
    taggedAval *locationOfForwardingPointer = 0;
    Byte *whereForwardingPointerPoints = 0;

    if( leaf_p( aReg->t ) )
        return;

    locationOfForwardingPointer = (aReg->u).a;

    if( ! forwarding_pointer_p( locationOfForwardingPointer->t ) ) {
        useFatalErrorOutput();
        printAnyTaggedAval( &CurrentOutputPort, aReg );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        printAnyTaggedAval( &CurrentOutputPort, locationOfForwardingPointer );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        tprintf3( &CurrentOutputPort, MiscStrings[ 458 ], name );
        fatalError( 103 );
        }

    if( ! inSpaceP( locationOfForwardingPointer, from ) ) {
        useFatalErrorOutput();
        printAnyTaggedAval( &CurrentOutputPort, aReg );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        printAnyTaggedAval( &CurrentOutputPort, locationOfForwardingPointer );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        tprintf3( &CurrentOutputPort, MiscStrings[ 458 ], name );
        fatalError( 166 );
        }

    whereForwardingPointerPoints = (Byte *)((locationOfForwardingPointer->u).a);

    if( ! inSpaceP( whereForwardingPointerPoints, to ) ) {
        useFatalErrorOutput();
        printAnyTaggedAval( &CurrentOutputPort, aReg );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        printAnyTaggedAval( &CurrentOutputPort, locationOfForwardingPointer );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        printAnyTaggedAval( &CurrentOutputPort, (taggedAval *)whereForwardingPointerPoints );
        tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
        tprintf3( &CurrentOutputPort, MiscStrings[ 458 ], name );
        fatalError( 167 );
        }

    (aReg->u).a = (taggedAval *)(whereForwardingPointerPoints + bumpDelta );
    }

/**** replaceStackPointers -- wrapper to call "replaceRegisterPointer" with each stack location. ****/

void GarbageCollector::replaceStackPointers( volatile memorySpace *aStack, taggedAval *aStackPointer,
                                             LongWord bumpDelta, UnsignedWord kitten,
                                             volatile memorySpace *from, volatile memorySpace *to )
{
    taggedAval *scratch = 0;
    Text message[ 32 ];

    sprintf( message, "kitten %d stack", kitten );

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        replaceRegisterPointer( scratch, bumpDelta, message, from, to );
    }

/**** replaceContinuationPointers -- wrapper to call "replaceRegisterPointer" with each stack location. ****/

void GarbageCollector::replaceContinuationPointers( volatile memorySpace *aContinuation, taggedAval *aContinuationPointer,
                                                    LongWord bumpDelta, UnsignedWord kitten,
                                                    volatile memorySpace *from, volatile memorySpace *to )
{
    taggedAval *scratch = 0;
    Text message[ 32 ];

    sprintf( message, "kitten %d continuation", kitten );

    for( scratch = aContinuationPointer; scratch < (taggedAval *)(aContinuation->end); ++scratch )
        replaceRegisterPointer( scratch, bumpDelta, message, from, to );
    }

/**** garbageCollectorGCCount  --  Instrumentation.  ****/

UnsignedLongWord GarbageCollector::garbageCollectorGCCount()
{
    return myClowder->thisClowderFileContent->GCCount;
    }

/**** garbageCollectorGenerationalGCCount  --  Instrumentation.  ****/

UnsignedLongWord GarbageCollector::garbageCollectorGenerationalGCCount()
{
    return myClowder->thisClowderFileContent->GenerationalGCCount;
    }

/**** setIsMomCat ****/

void GarbageCollector::setIsMomCat( Boolean trueOrFalse )
{
    isMomCatLocalP = trueOrFalse;
    }

/**** countNonGarbageLocks  --  How many non-garbage locks in the given space.  ****/

UnsignedLongWord GarbageCollector::countNonGarbageLocks( volatile memorySpace *m )
{
    Byte *tap = 0;
    UnsignedLongWord answer = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        if( gc_bit( ((taggedAval *)tap)->t ) && (getLocationByte( ((taggedAval *)tap)->t ) || getKittenByte( ((taggedAval *)tap)->t )) )
            ++answer;
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    return answer;
    }

/**** countAllLocks  --  How many locks, garbage or otherwise, in the given space.  ****/

UnsignedLongWord GarbageCollector::countAllLocks( volatile memorySpace *m )
{
    Byte *tap = 0;
    UnsignedLongWord answer = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        if( getLocationByte( ((taggedAval *)tap)->t ) || getKittenByte( ((taggedAval *)tap)->t ) )
            ++answer;
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    return answer;
    }

/**** mergeGarbage -- The idea here is to merge adjacent garbage blocks into large byte blocks which
                      can quickly be bypassed because their length is embedded in them.  ****/

void GarbageCollector::mergeGarbage( volatile memorySpace *m )
{
    Byte *tap = 0;
    Boolean inGarbageBlock = FALSE;
    Byte *startGarbageBlock = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        if( inGarbageBlock ) {
            if( ! garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = FALSE;
                ((taggedAval *)startGarbageBlock)->t = byteBlockTag;          // GC bit is off.
                (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
                }
            }
        else {   // Not in garbage block
            if( garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = TRUE;
                startGarbageBlock = tap;
                }
            }
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        } 
    if( inGarbageBlock ) {  // Space ends in garbage block.
        ((taggedAval *)startGarbageBlock)->t = byteBlockTag;                  // GC bit is off.
        (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
        }
    }

/**** markNonGarbageCCNBlocksInSpaceAndMergeGarbage  -- A do-something-and-merge-garbage-while-you-are-at-it function. ****/

void GarbageCollector::markNonGarbageCCNBlocksInSpaceAndMergeGarbage( volatile memorySpace *m )
{
    taggedAval *tp = 0;
    volatile Byte *tap = 0;
    Boolean inBlockP = FALSE;
    Boolean inGarbageBlock = FALSE;
    volatile Byte *startGarbageBlock = 0;

    for( tap = m->free; tap < m->end; tap += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)tap;
        if( inGarbageBlock ) {
            if( ! garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = FALSE;
                ((taggedAval *)startGarbageBlock)->t = byteBlockTag;          // GC bit is off.
                (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
                }
            }
        else {   // Not in garbage block
            if( garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = TRUE;
                startGarbageBlock = tap;
                }
            }
        if( inBlockP ) {
            if( ! gc_bit( tp->t ) ) {
                fatalError( 117 );
                }
            else if( ! car_cdr_next_p( tp->t ) )
                inBlockP = FALSE;

           // else nothing

            }
        else {
            if( ! gc_bit( tp->t ) )
                ;
            else if( car_cdr_next_p( tp->t ) ) {
                inBlockP = TRUE;
                setStartCCNBlockBitNonAtomic( tp->t );
                }

           // else nothing

            }
        }
    if( inGarbageBlock ) {  // Space ends in garbage block.
        ((taggedAval *)startGarbageBlock)->t = byteBlockTag;                  // GC bit is off.
        (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
        }
    }

/**** countNonGarbageLocksAndMergeGarbage  --  Another do-something-and-merge-garbage-while-you-are-at-it function.  ****/

UnsignedLongWord GarbageCollector::countNonGarbageLocksAndMergeGarbage( volatile memorySpace *m )
{
    Byte *tap = 0;
    UnsignedLongWord answer = 0;
    Boolean inGarbageBlock = FALSE;
    Byte *startGarbageBlock = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        if(    gc_bit( ((taggedAval *)tap)->t )
            && (getLocationByte( ((taggedAval *)tap)->t ) || getKittenByte( ((taggedAval *)tap)->t )) ) {
            ++answer;
            }
        if( inGarbageBlock ) {
            if( ! garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = FALSE;
                ((taggedAval *)startGarbageBlock)->t = byteBlockTag;          // GC bit is off.
                (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
                }
            }
        else {   // Not in garbage block
            if( garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = TRUE;
                startGarbageBlock = tap;
                }
            }
        tap += lengthTaggedAvalPointer( (taggedAval *)tap );
        }
    if( inGarbageBlock ) {  // Space ends in garbage block.
        ((taggedAval *)startGarbageBlock)->t = byteBlockTag;                  // GC bit is off.
        (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
        }
    return answer;
    }

// A large block of stuff for generational garbage-collection starts here.  Note that references to
// the "B" algorithm should tell you something about the success and/or merits of the "A" algorithm ...

/**** replaceRegisterPointersForKittenGenGCB  -- Replace all register pointers for the given kitten. ****/

void GarbageCollector::replaceRegisterPointersForKittenGenGCB( UnsignedWord aKittenNumber, volatile memorySpace *aging )
{
    if( ! (myClowder->getKittenEvaluatorActivity( aKittenNumber ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) )
        return;
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localE), aging );
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localSavedE), aging );
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localR), aging );
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localA), aging );
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localL), aging );
    replaceRegisterPointerGenGCB(
        &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentInputPort), aging );
    replaceRegisterPointerGenGCB(
        &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localCurrentOutputPort), aging );
    replaceRegisterPointerGenGCB( &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localErrorPort), aging );
    replaceRegisterPointerGenGCB(
        &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localDefaultInputPort), aging );
    replaceRegisterPointerGenGCB(
        &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localMasterDefaultInputPort), aging );
    replaceRegisterPointerGenGCB(
        &((((myClowder->thisClowderFileContent->kittenData)[ aKittenNumber ]).localRegisters).localTranscriptPort), aging );
    }

/**** replacePointersInAddressStackB -- wrapper to call "replaceRegisterPointerGenGCB" with each location in the
                                        stack of interesting addresses. ****/

void GarbageCollector::replacePointersInAddressStackB( taggedAval **start, taggedAval **stop, volatile memorySpace *aging )
{
    taggedAval **scratch = 0;

    for( scratch = start; scratch < stop; ++scratch ) {
        replaceRegisterPointerGenGCB( *scratch, aging );
        }
    }

/**** replaceRegisterPointerGenGCB -- Replace the given pointer, if it points into aging space, by following
                                      the ggc forwarding pointer that should be found there.  ****/

void GarbageCollector::replaceRegisterPointerGenGCB( taggedAval *aReg, volatile memorySpace *aging )
{
    taggedAval *tp = 0;

    if( leaf_p( aReg->t ) )
        return;

    tp = (aReg->u).a;

    if( inSpaceP( tp, aging ) ) {
        if( ggc_forwarding_pointer_p( tp->t ) )
            (aReg->u).a = (tp->u).a;
         else {
            useFatalErrorOutput();
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], tp );
            fatalError( 178 );
            }
        return;
        }
    return;
    }

/**** replaceStackPointersGenGCB -- wrapper to call "replaceRegisterPointerGenGCB" with each stack location. ****/

void GarbageCollector::replaceStackPointersGenGCB( volatile memorySpace *aStack, taggedAval *aStackPointer,
                                                   UnsignedWord kitten, volatile memorySpace *aging )
{
    taggedAval *scratch = 0;
    Text message[ 32 ];

    sprintf( message, "kitten %d stack", kitten );

    for( scratch = aStackPointer; scratch < (taggedAval *)(aStack->end); ++scratch )
        replaceRegisterPointerGenGCB( scratch, aging );
    }

/**** replaceContinuationPointersGenGCB -- wrapper to call "replaceRegisterPointerGenGCB" with each stack location. ****/

void GarbageCollector::replaceContinuationPointersGenGCB( volatile memorySpace *aContinuation,
                                                         taggedAval *aContinuationPointer,
                                                         UnsignedWord kitten,
                                                         volatile memorySpace *aging )
{
    taggedAval *scratch = 0;
    Text message[ 32 ];

    sprintf( message, "kitten %d continuation", kitten );

    for( scratch = aContinuationPointer; scratch < (taggedAval *)(aContinuation->end); ++scratch )
         replaceRegisterPointerGenGCB( scratch, aging );
    }

/**** countNonGarbageBytesAndMergeGarbage  --  hack to save one pass through the old space  ****/

UnsignedLongWord GarbageCollector::countNonGarbageBytesAndMergeGarbage( volatile memorySpace *m )
{
    Byte *tap = 0;
    UnsignedLongWord currentLength = 0;
    UnsignedLongWord answer = 0;
    Boolean inGarbageBlock = FALSE;
    Byte *startGarbageBlock = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        currentLength = lengthTaggedAvalPointer( (taggedAval *)tap );
        if( gc_bit( ((taggedAval *)tap)->t ) )
            answer += currentLength;
        if( inGarbageBlock ) {
            if( ! garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = FALSE;
                ((taggedAval *)startGarbageBlock)->t = byteBlockTag;          // GC bit is off.
                (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
                }
            }
        else {   // Not in garbage block
            if( garbage_p( ((taggedAval *)tap)->t ) ) {
                inGarbageBlock = TRUE;
                startGarbageBlock = tap;
                }
            }
        tap += currentLength;
        }
    if( inGarbageBlock ) {  // Space ends in garbage block.
        ((taggedAval *)startGarbageBlock)->t = byteBlockTag;                  // GC bit is off.
        (((taggedAval *)startGarbageBlock)->u).c = (tap - startGarbageBlock - sizeof( taggedAval ) );
        }
    return answer;
    }

/**** countNonGarbageBytesInSpace  --  How many garbage bytes in this space? ****/

UnsignedLongWord GarbageCollector::countNonGarbageBytesInSpace( volatile memorySpace *m )
{
    Byte *tap = 0;
    UnsignedLongWord currentLength = 0;
    UnsignedLongWord answer = 0;
        
    for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
        currentLength = lengthTaggedAvalPointer( (taggedAval *)tap );
        if( ! garbage_p( ((taggedAval *)tap)->t ) )
            answer += currentLength;
        tap += currentLength;
        }
    return answer;
    }

/**** copyDownB -- copy non-garbage from "m" to "to", and check that the number of objects copied was correct. ****/

UnsignedLongWord GarbageCollector::copyDownB( volatile memorySpace *m, volatile memorySpace *to )
{
    taggedAval *tp = 0;
    taggedAval *newTp = 0;
    volatile Byte *bp = 0;
    UnsignedLongWord nBytesCopied = 0;
    UnsignedLongWord currentLength = 0;

    for( bp = m->free; bp < m->end; bp += currentLength ) {
        tp = (taggedAval *)bp;
        newTp = (taggedAval *)(to->free + nBytesCopied);
        currentLength = lengthTaggedAvalPointer( tp );
        if( gc_bit( tp->t ) ) {
            memcpy( (Byte *)newTp, (Byte *)bp, (size_t)currentLength );
            (tp->u).a = (taggedAval *)(to->free + nBytesCopied);
            tp->t = ggcForwardingPointerTag;
            newTp->t &= ~(pointsToActiveBitMask | recursionBitsMask | gcBitsMask);
            nBytesCopied += currentLength;
            }
        }
    return nBytesCopied;
    }

/**** updatePointersInCopiedDownArea -- If any pointer in the area from start to stop still points to "lookForPointersToHere",
                                        update it so that it points to where the forwarding pointer at that location points.  ****/

void GarbageCollector::updatePointersInCopiedDownArea( Byte *start, Byte *stop, volatile memorySpace *lookForPointersToHere )
{
    Byte *bp = 0;
    taggedAval *tap = 0;

    for( bp = start; bp < stop; bp += lengthTaggedAvalPointer( (taggedAval *)bp ) ) {
        tap = (taggedAval *)bp;
        if( (! leaf_p( tap->t )) && inSpaceP( (tap->u).a, lookForPointersToHere ) ) {
            if( ggc_forwarding_pointer_p( ((tap->u).a)->t ) ) {     // Where we expect a forwarding pointer, which we use to update.
                (tap->u).a = (((tap->u).a)->u).a;
                }
            else {                                              // If there is no forwarding pointer there,
                useFatalErrorOutput();                          //   then something is very wrong.
                printAnyTaggedAval( &CurrentOutputPort, tap );
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
                printAnyTaggedAval( &CurrentOutputPort, (tap->u).a );
                fatalError( 74 );
                }
            }
        }
    }

/**** updatePointersFromActiveToAging -- Find pointers in active that point to aging, and if there is a ggc forwarding pointer there,
                                         update the pointer to point to where it points.  ****/

void GarbageCollector::updatePointersFromActiveToAging( volatile memorySpace *active, volatile memorySpace *aging )
{
    volatile Byte *bp = 0;
    taggedAval *tap = 0;
    volatile Byte *stop = active->end;


    for( bp = active->free; bp < stop; bp += lengthTaggedAvalPointer( (taggedAval *)bp ) ) {
        tap = (taggedAval *)bp;
        if( (! leaf_p( tap->t )) && inSpaceP( (tap->u).a, aging ) ) {
            if( ggc_forwarding_pointer_p( ((tap->u).a)->t ) ) {     // Where we expect a forwarding pointer, which we use to update.
                (tap->u).a = (((tap->u).a)->u).a;
                }
            }
        }
    }

/**** stackPointersToActiveInCopiedDownAreaB -- called after the copyDown; the issue is to find pointers in the copied-down
                                                area that point into active space, and stack them.  ****/

void GarbageCollector::stackPointersToActiveInCopiedDownAreaB( Byte *start, Byte *stop, volatile memorySpace *active )
{
    Byte *bp = 0;
    taggedAval *tap = 0;

    for( bp = start; bp < stop; bp += lengthTaggedAvalPointer( (taggedAval *)bp ) ) {
        tap = (taggedAval *)bp;
        if( (! leaf_p( tap->t )) && inSpaceP( (tap->u).a, active ) ) {
            theClowder->addToLiveAddressStack( tap );
            }
        }
    }

/**** primeDeadStackB -- After swapping generations, prime the stack of pointers into aging space with pointers
                         that were into active space until it just got swapped.  ****/

void GarbageCollector::primeDeadStackB( taggedAval **start, taggedAval **stop, 
                                       volatile memorySpace *tenured, volatile memorySpace *active, volatile memorySpace *aging )
{
    taggedAval **scratch = 0;
    taggedAval *whereStarScratchPoints = 0;

    // This version of "primeDeadStackB" is a very conservative routine, with a lengthy check in the middle.
    // The code to run the check is commented out, at the start of the routine.

    // The idea of the test is this:  The intent is that a taggedAval's pointsToActiveBit will be set
    // if and only if the taggedAval in question has been put into the live address stack.  The way 
    // we test for that is to turn *off* all such bits associated with stacked addresses, and then run
    // a test of all spaces to see if there are any that we have somehow missed.

    // This test is done in this routine mostly for convenience; it is as good a place to do it as anywhere.

//    for( scratch = start; scratch < stop; ++scratch )
//         clearPointsToActiveBitNonAtomic( (*scratch)->t );
//
//    checkForPointsToActiveBits( tenured, "Checking tenured space while priming the stack", FALSE );
//    checkForPointsToActiveBits( active, "Checking active space while priming the stack", FALSE );
   
    // We don't check the aging space because it is empty when "primeDeadStackB" is called.

    theClowder->clearDeadAddressStack();

    for( scratch = start; scratch < stop; ++scratch ) {
        if( !leaf_p( (*scratch)->t ) ) {
            whereStarScratchPoints = ((*scratch)->u).a;
            if( inSpaceP( whereStarScratchPoints, active ) ) {
                theClowder->addToDeadAddressStack( *scratch );
                continue;
                }
             }
         clearPointsToActiveBitNonAtomic( (*scratch)->t );   // Make sure to clear the bits in case a pointer is put there later.
         }
    }

/**** checkActiveForPointersToAging -- when this test function is called, all non-garbage pointers from active space that used to
                                       point into aging space should should have been bumped to point into tenured space instead.
                                       Complain if they do not.  ****/

void GarbageCollector::checkActiveForPointersToAging( volatile memorySpace *active, volatile memorySpace *aging )
{
    Byte *scratch = 0;
    Byte *stop = (Byte *)(active->end);
    taggedAval *tp = 0;

    for( scratch = (Byte *)(active->free); scratch < stop; scratch += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)scratch;
        if(    (! leaf_p( tp->t ))
            && (! garbage_p( tp->t ))
            && inSpaceP( (tp->u).a, aging )
            ) {
            tprintf2( &CurrentOutputPort, MiscStrings[ 484 ] );
            if( ggc_forwarding_pointer_p( ((tp->u).a)->t ) )
                (tp->u).a = (((tp->u).a)->u).a;
             else {
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], tp );
                fatalError( 179 );
                }
            }
        }

    }

/**** checkAgingCdrCodedBlocks -- verify that everything after a non-garbage item coded ca-cdr-next, till the end of the
                                  cdr-coded block, has also been marked non-garbage.  ****/

void GarbageCollector::checkAgingCdrCodedBlocks( volatile memorySpace *aging )
{
    Byte *scratch = 0;
    Byte *stop = (Byte *)(aging->end);
    taggedAval *tp = 0;

    for( scratch = (Byte *)(aging->free); scratch < stop; scratch += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)scratch;
        if( car_cdr_next_p( tp->t ) && (! garbage_p( tp->t )) && garbage_p( (tp + 1)->t ) )
            tprintf2( &CurrentOutputPort, "Bad cdr-coded block in aging space.\n");        
        }
    }

/**** checkTenuredForMissedNonGarbage -- for debugging, make sure we have actually marked as non-garbage,
                                         what every pointer out of tenured space points to.  ****/

void GarbageCollector::checkTenuredForMissedNonGarbage( volatile memorySpace *tenured, 
                                                        volatile memorySpace *active,
                                                        volatile memorySpace *aging )
{
    Byte *scratch = 0;
    Byte *stop = (Byte *)(tenured->end);
    taggedAval *tp = 0;
    taggedAval *tpNext = 0;

    for( scratch = (Byte *)(tenured->free); scratch < stop; scratch += lengthTaggedAvalPointer( tp ) ) {
        tp = (taggedAval *)scratch;
        if( ! validDataType( tp->t ) ) {
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tp );
            printAnyTaggedAval( &CurrentOutputPort, tp );
            resetError( 80 );
            }
        if(    (! leaf_p( tp->t ))
            && (inSpaceP( (tp->u).a, active ) || inSpaceP( (tp->u).a, aging ))
            && garbage_p( ((tp->u).a)->t )
            ) {
            if( inSpaceP( (tp->u).a, active ) )
                tprintf2( &CurrentOutputPort, MiscStrings[ 485 ] );
            if( inSpaceP( (tp->u).a, aging ) )
                tprintf2( &CurrentOutputPort, MiscStrings[ 486 ]);
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tp );
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            printAnyTaggedAval( &CurrentOutputPort, tp );
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            printAnyTaggedAval( &CurrentOutputPort, (tp->u).a );
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            }
        tpNext = (taggedAval *)(scratch + lengthTaggedAvalPointer( tp ));
        if( ! validDataType( tpNext->t ) ) {
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tp );
            printAnyTaggedAval( &CurrentOutputPort, tp );
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );            
            if( ! leaf_p( tp->t ) ) {
                printAnyTaggedAval( &CurrentOutputPort, (tp->u).a );
                tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );            
                }
            tprintf3( &CurrentOutputPort, MiscStrings[ 260 ], (UnsignedLongWord)tpNext );
            printAnyTaggedAval( &CurrentOutputPort, tpNext );
            resetError( 80 );
            }
        }
    }

/**** gGCMarkNonGarbage -- portmanteau for marking garbage in the generational collector. ****/

UnsignedLongWord GarbageCollector::gGCMarkNonGarbage( volatile memorySpace *currentSpace )
{
    UnsignedLongWord nonGarbageCount = 0;
    UnsignedWord counter = 0;

    volatile memorySpace *agingSpace = currentSpace->agingGeneration;
    volatile memorySpace *activeSpace = currentSpace->activeGeneration;
    volatile memorySpace *tenuredSpace = currentSpace;
    volatile memorySpace *scratchSpace = theHeap->getGGCScratchSpace();

    nonGarbageCount += markGarbageFromRegGGC( &(E), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    eventLoopOnce( 0 );
    nonGarbageCount += markGarbageFromRegGGC( &(TopLevelE), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(SavedE), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(IVR), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(R), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(A), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(L), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(TLC), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(CurrentInputPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(CurrentOutputPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(ErrorPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(DefaultInputPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(MasterDefaultInputPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    nonGarbageCount += markGarbageFromRegGGC( &(TranscriptPort), tenuredSpace, agingSpace, activeSpace, scratchSpace );
    eventLoopOnce( 0 );

    for( counter = 0; counter < theNumberOfKittens; ++counter ) {
        nonGarbageCount += markGarbageFromRegGGC( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ,
                                                  tenuredSpace, agingSpace, activeSpace, scratchSpace );
        }

    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_REGISTERS_TO_SCRATCH, KITTEN_HAS_WRITTEN_REGISTERS_TO_SCRATCH );
    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) {
        markGarbageFromRegForKittenGGC( counter, tenuredSpace, agingSpace, activeSpace, scratchSpace, &nonGarbageCount );
        eventLoopOnce( 0 );
        }

    nonGarbageCount += markGarbageFromStackGGC( (myMemoryManager->SchemeStack),tenuredSpace, agingSpace, activeSpace, scratchSpace, S );
    nonGarbageCount += markGarbageFromStackGGC( (myMemoryManager->SchemeContinuation),tenuredSpace, agingSpace, activeSpace, scratchSpace, CS );
    eventLoopOnce( 0 );

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            nonGarbageCount += markGarbageFromStackGGC( (myMemoryManager->ScratchStack), tenuredSpace, agingSpace, activeSpace, scratchSpace,
                                                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH,
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();
            nonGarbageCount += markGarbageFromStackGGC( (myMemoryManager->ScratchStack), tenuredSpace, agingSpace, activeSpace, scratchSpace,
                                                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS );
            eventLoopOnce( 0 );
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();

    nonGarbageCount += markGarbageFromRegGGC( &(ObList), tenuredSpace, agingSpace, activeSpace, scratchSpace );

    nonGarbageCount += markGarbageFromAddressStackReferences( theClowder->getGGCLiveAddressStackPointer(),
                                                              theClowder->getGGCLiveAddressStackEnd(),
                                                              tenuredSpace, agingSpace, activeSpace, scratchSpace,
                                                              "address stack references" );

//    nonGarbageCount += expireForgettables( agingSpace, scratchSpace );

      return nonGarbageCount;

      }


/**** gGCReplaceRegisterPointers -- portmanteau for replacing pointers in the generational collector. ****/

void GarbageCollector::gGCReplaceRegisterPointers( volatile memorySpace *agingSpace )
{
    Text kittenNameBuffer[ KITTEN_NAME_LENGTH ];
    UnsignedWord counter = 0;

    replaceRegisterPointerGenGCB( &E, agingSpace );
    replaceRegisterPointerGenGCB( &SavedE, agingSpace );
    replaceRegisterPointerGenGCB( &R, agingSpace );
    replaceRegisterPointerGenGCB( &A, agingSpace );
    replaceRegisterPointerGenGCB( &L, agingSpace );
    replaceRegisterPointerGenGCB( &TLC, agingSpace );
    replaceRegisterPointerGenGCB( &CurrentInputPort, agingSpace );
    replaceRegisterPointerGenGCB( &CurrentOutputPort, agingSpace );
    replaceRegisterPointerGenGCB( &ErrorPort, agingSpace );
    replaceRegisterPointerGenGCB( &DefaultInputPort, agingSpace );
    replaceRegisterPointerGenGCB( &MasterDefaultInputPort, agingSpace );
    replaceRegisterPointerGenGCB( &TranscriptPort, agingSpace );

    for( counter = 0; counter < theNumberOfKittens; ++counter )
         replaceRegisterPointerGenGCB( &((myClowder->thisClowderFileContent->kittenData)[ counter ]).clowderKIQ, agingSpace );

    for( counter = 1; counter < theNumberOfKittens; ++counter )
        replaceRegisterPointersForKittenGenGCB( counter, agingSpace );

    replaceRegisterPointerGenGCB( &TopLevelE, agingSpace );
    replaceRegisterPointerGenGCB( &IVR, agingSpace );

    replaceRegisterPointerGenGCB( &ObList, agingSpace );

    replaceStackPointersGenGCB( (myMemoryManager->SchemeStack), S, 0, agingSpace );
    eventLoopOnce( 0 );

    replaceContinuationPointersGenGCB( (myMemoryManager->SchemeContinuation), CS, 0, agingSpace );
    eventLoopOnce( 0 );
    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_STACK_TO_SCRATCH, KITTEN_HAS_WRITTEN_STACK_TO_SCRATCH );
            OSMemoryBarrier();
            replaceStackPointersGenGCB( (myMemoryManager->ScratchStack), 
                                        (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localS,
                                        counter, agingSpace );
            sprintf( kittenNameBuffer, "Stack for kitten %d", counter );
            eventLoopOnce( 0 );

            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_STACK_FROM_SCRATCH, KITTEN_HAS_COPIED_STACK_FROM_SCRATCH );
            }
    OSMemoryBarrier();
    for( counter = 1; counter < theNumberOfKittens; ++counter ) 
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL ) {
            myClowder->momDialogWithKitten( counter, MOM_SAYS_WRITE_CONTINUATION_TO_SCRATCH,
                                                     KITTEN_HAS_WRITTEN_CONTINUATION_TO_SCRATCH );
            OSMemoryBarrier();
            replaceContinuationPointersGenGCB( (myMemoryManager->ScratchStack), 
                                               (((myClowder->thisClowderFileContent->kittenData)[ counter ]).localRegisters).localCS,
                                               counter, agingSpace );
            sprintf( kittenNameBuffer, "Continuation for kitten %d", counter );
            eventLoopOnce( 0 );

            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_CONTINUATION_FROM_SCRATCH,
                                                     KITTEN_HAS_COPIED_CONTINUATION_FROM_SCRATCH );
            }
    OSMemoryBarrier();
    replacePointersInAddressStackB( theClowder->getGGCLiveAddressStackPointer(),theClowder->getGGCLiveAddressStackEnd(), agingSpace );

    }


/**** generationalGCB -- Portmanteau for generational garbage collection.  ****/

void GarbageCollector::generationalGCB( volatile memorySpace *currentSpace, volatile memorySpace *alternateSpace )
{
    UnsignedLongWord newLockNumber = 0;
    UnsignedLongWord nonGarbageCount = 0;
    UnsignedLongWord nonGarbageBytes = 0;
    UnsignedLongWord nonGarbageBytesInAgingSpace = 0;
    UnsignedLongWord bytesCopied = 0;
    UnsignedWord counter = 0;
    UnsignedLongWord oldLockNumber = 0;

    volatile memorySpace *agingSpace = currentSpace->agingGeneration;
    volatile memorySpace *activeSpace = currentSpace->activeGeneration;
    volatile memorySpace *tenuredSpace = currentSpace;

#if FALSE    // DEBUG
//    checkSpace( currentSpace, currentSpace, 0, "Checking current space before a generational GC", TRUE );
//    checkSpace( agingSpace, currentSpace, 0, "Checking aging space before a generational GC", TRUE );
    checkSpace( activeSpace, currentSpace, 0, "Checking active space before a generational GC", TRUE );
#endif

    if( garbageCollectorRunning )
        fatalError( 98 );
    garbageCollectorRunning = TRUE;

    if( alternateSpace->inUse )
        resetError( 7 );

    // Mark non-garbage.

    nonGarbageCount = gGCMarkNonGarbage( currentSpace );

   // Check locks, and do some bagging of garbage while you are at it.

    nonGarbageBytes = countNonGarbageBytesAndMergeGarbage( agingSpace );
    nonGarbageBytesInAgingSpace = countNonGarbageBytesInSpace( agingSpace );
    if( theNumberOfKittens > 1 )
        oldLockNumber = countAllLocks( currentSpace ) + countAllLocks( agingSpace );

    // Set up tenured space for the copy-down.  

    tenuredSpace->free -= nonGarbageBytesInAgingSpace;

//  Test code commented out, left visible for possible future use.
//
//    checkAgingCdrCodedBlocks( agingSpace );
//
//    tprintf2( &CurrentOutputPort, " checked aging ..." );  // DEBUG
//
//    checkTenuredForMissedNonGarbage( tenuredSpace, activeSpace, agingSpace );
//
//    tprintf2( &CurrentOutputPort, " checked tenured ..." );  // DEBUG

    // Compact the non-garbage from the aging space onto the newly-adjusted end of the tenured space.

    bytesCopied = copyDownB( agingSpace, tenuredSpace );

    if( bytesCopied != nonGarbageBytesInAgingSpace ) {
        useFatalErrorOutput();
        tprintf4( &CurrentOutputPort, MiscStrings[ 362 ], nonGarbageBytesInAgingSpace, bytesCopied );
        fatalError( 89 );
        }

    // Fuss with pointers related to the copy-down.

    updatePointersInCopiedDownArea( (Byte *)(tenuredSpace->free),
                                    (Byte *)(tenuredSpace->free) + nonGarbageBytesInAgingSpace,
                                    agingSpace );

    updatePointersFromActiveToAging( activeSpace, agingSpace );

    stackPointersToActiveInCopiedDownAreaB( (Byte *)(tenuredSpace->free),
                                            (Byte *)(tenuredSpace->free) + nonGarbageBytesInAgingSpace,
                                            activeSpace );

    eventLoopOnce( 0 );

    clearAllGCAndRecursionBits( activeSpace );

    // Replace the register pointers that have been moved.

    gGCReplaceRegisterPointers( agingSpace );

    checkActiveForPointersToAging( activeSpace, agingSpace );

    //****************  Miscellaneous clean-up.

    eventLoopOnce( 0 );

    OSMemoryBarrier();

    // Transfer un-copied-down addresses of interest to the new stack.

    primeDeadStackB( theClowder->getGGCLiveAddressStackPointer(), theClowder->getGGCLiveAddressStackEnd(),
                     tenuredSpace, activeSpace, agingSpace );

    // Copy back the registers.

    for( counter = 1; counter < theNumberOfKittens; ++counter )
        if( myClowder->getKittenEvaluatorActivity( counter ) >= KITTEN_ACTIVITY_READY_TO_EVAL )
            myClowder->momDialogWithKitten( counter, MOM_SAYS_COPY_REGISTERS_FROM_SCRATCH, KITTEN_HAS_COPIED_REGISTERS_FROM_SCRATCH );

    OSMemoryBarrier();

    // Verify locks not messed up.

    if( theNumberOfKittens > 1 ) {
        newLockNumber = countAllLocks( currentSpace );
        if( newLockNumber != oldLockNumber ) {
            useFatalErrorOutput();
            tprintf4( &CurrentOutputPort, MiscStrings[ 406 ], oldLockNumber, newLockNumber );
            fatalError( 67 );
            }        
        }        

    OSMemoryBarrier();

    eventLoopOnce( 0 );

    // Swap active and aging spaces.

    myMemoryManager->swapGenerationSpaces( currentSpace );

    OSMemoryBarrier();

    garbageCollectorRunning = FALSE;
    eventLoopOnce( 0 );

    // And now some checking, possibly commented out.

#if TRUE    // DEBUG
//    checkForPointsToActiveBits( tenuredSpace, "Checking tenured space after a generational GC", FALSE );
//    checkSpace( tenuredSpace, currentSpace, 0, "Checking tenured space after a generational GC", TRUE );
    checkSpace( agingSpace, currentSpace, 0,  "Checking aging space after a generational GC", TRUE );
    checkSpace( activeSpace, currentSpace, 0, "Checking active space after a generational GC", TRUE );
#endif

    }

/**** startEvents -- Wrapper for events to start GC. ****/

void GarbageCollector::startEvents( Boolean isFullGC )
{
    eventLoopOnce( 0 );
    reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
    eventLoopOnce( 0 );
    reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
    eventLoopOnce( 0 );
    reportEvent( TOTAL_MEMORY_IN_USE_EVENT );
    eventLoopOnce( 0 );
    
    reportEvent( GC_START_EVENT);
    eventLoopOnce( 0 );
    if( isFullGC )
        reportEvent( GC_START_EVENT_1 );
    else
        reportEvent( GC_START_EVENT_2 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    }

/**** endEvents -- Wrapper for events to end GC. ****/

void GarbageCollector::endEvents( Boolean isFullGC )
{
    eventLoopOnce( 0 );
    if( isFullGC )
        reportEvent( GC_END_EVENT );
    else
        reportEvent( GC_END_EVENT_4 );
    eventLoopOnce( 0 );
    reportEvent( GC_END_EVENT_1 );
    eventLoopOnce( 0 );
    if( isFullGC )
        reportEvent( GC_END_EVENT_2 );
    else
        reportEvent( GC_END_EVENT_3 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    eventLoopOnce( 0 );
    reportEvent( CURRENT_MEMORY_IN_USE_EVENT );
    reportEvent( CURRENT_GENERATIONAL_MEMORY_IN_USE_EVENT );
    reportEvent( TOTAL_MEMORY_IN_USE_EVENT );
    }

/**** standardFullGC  --  Standard entry point for full GC. ****/

void GarbageCollector::standardFullGC( volatile memorySpace *currentSpace, volatile memorySpace *alternateSpace,
                                       Boolean *didRebuildObList )
{       
    if( garbageCollectorEntryRoutineRunning )
        fatalError( 184 );

    *didRebuildObList = FALSE;

    garbageCollectorEntryRoutineRunning = TRUE;

//    myClowder->initialMomAndKittenGCSynchronization();

    if( isMomCatLocalP ) {    //  We are the MomCat -- really do the GC.

        startEvents( TRUE);
        myClowder->momDialogWithEverybody( MOM_SAYS_FULL_GC, KITTEN_ACKNOWLEDGES_FULL_GC_STARTING );

        theHeap->padForFullGC( currentSpace );
        theHeap->setSpaceForFullGC( currentSpace );
        theHeap->setSpaceForFullGC( alternateSpace );

        clearAllPointsToActiveBits( currentSpace );

        if( myClowder->anyObListInRegister() || anyBlockedSymbols( currentSpace ) ) {  
            fullGCType2( currentSpace, alternateSpace, TRUE );
            postGCAuxiliary( currentSpace, alternateSpace );
            }
        else {        // Must fuss with oblist.

            *didRebuildObList = TRUE;

            fullGCType2( currentSpace, alternateSpace, FALSE );
            postGCAuxiliary( currentSpace, alternateSpace );

            theClowder->initializeGGCAddressStacks( currentSpace );

            if( usingGenerationalGC ) {  // Make new oblist in main space, not in generational space.
                alternateSpace->allocateDomain = alternateSpace;
                rebuildObList( alternateSpace );
                alternateSpace->allocateDomain = alternateSpace->activeGeneration;
                }
            else
                rebuildObList( alternateSpace );
            }
			
        myClowder->thisClowderFileContent->GCCount = myClowder->thisClowderFileContent->GCCount + 1;
        theClowder->clearFullGCIsRequired();

        myClowder->setMomSaysToEverybody( MOM_SAYS_NO_MORE_MESSAGES );
        myClowder->momReceiveFromEverybody( KITTEN_ACKNOWLEDGES_FULL_GC_ENDING );
        myClowder->setMomSaysToEverybody( MOM_SAYS_NOTHING );
        endEvents( TRUE );
        }
    else  // Kitten should not call this function!
        fatalError( 188 );

    eventLoopOnce( 0 );

    garbageCollectorEntryRoutineRunning = FALSE;
    }


/**** standardGenerationalGC  --  Standard entry point for generational GC. ****/

void GarbageCollector::standardGenerationalGC( volatile memorySpace *currentSpace, volatile memorySpace *alternateSpace )
{       
    if( garbageCollectorEntryRoutineRunning )
        fatalError( 184 );
    garbageCollectorEntryRoutineRunning = TRUE;

//    myClowder->initialMomAndKittenGCSynchronization();

    if( isMomCatLocalP ) {    //  We are the MomCat -- really do the GC.
        startEvents( FALSE );
        myClowder->momDialogWithEverybody( MOM_SAYS_GENERATIONAL_GC, KITTEN_ACKNOWLEDGES_GENERATIONAL_GC_STARTING );

        theClowder->flushCachedAddresses();
        generationalGCB( currentSpace, alternateSpace );
        checkRegisters( currentSpace, currentSpace->activeGeneration, currentSpace->agingGeneration, "Generational GC", TRUE );
        myClowder->thisClowderFileContent->GenerationalGCCount = myClowder->thisClowderFileContent->GenerationalGCCount + 1;

        myClowder->setMomSaysToEverybody( MOM_SAYS_NO_MORE_MESSAGES );
        myClowder->momReceiveFromEverybody( KITTEN_ACKNOWLEDGES_GENERATIONAL_GC_ENDING );
        myClowder->setMomSaysToEverybody( MOM_SAYS_NOTHING );
        endEvents( FALSE );
        }
    else  // Kitten should not call this function!
        fatalError( 189 );

    eventLoopOnce( 0 );

    garbageCollectorEntryRoutineRunning = FALSE;
    }
