/****************************************************************************************

IO.c++ -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006

Low-level input/output handlers for Wraith Scheme.  Includes things to print error messages.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "OpCodes.h"
#import "DataTypes.e"
#import "Events.e"
#import "Globals.e"
#import "SchemeAux.e"
#import "StreamHandling.e"

#define TGETC_DELAY 1000        // microseconds

Boolean consoleOutputEnabledFlag = TRUE;

/**** forward declarations ****/

void tfflushConsoleOutput();

/**** Functions: ****/

/**** resetError ****/

static LongWord resetRecursiveLevel = 0;

void resetError( UnsignedLongWord n )
{
    Boolean traceHeaderPrintedP = false;
    taggedAval *frameBottom = (taggedAval *)theHeap->schemeStackEnd();
       
    fflush( getPortFilePointer( &CurrentOutputPort ) );
    if( ++resetRecursiveLevel >= 3 ) {
        resetCleanUp();
        }
    else {
        fflush( getPortFilePointer( &CurrentOutputPort ) );
        if( n > NumberOfErrorStrings ) {    /* Range-check error number.  Zero is special-cased in the switch. */
            useFatalErrorOutput();
            tprintf3( &ErrorPort, MiscStrings[ 104 ], n );
            fflush( getPortFilePointer( &ErrorPort ) );
            fatalError( 9 );
            }
        if( ! evaluatorHasInitialized ) {   // Hack for a fatal error occurring, eg, during load of a moused-on world.
            fatalErrorWithText( ErrorStrings[ n ] );
            }
        switch( n ) {
            case 0:                         /* Called from "e::reset".  No message.  Just clean up and reset. */
                resetCleanUp();
                break;
            default:                        /* Print message and do something useful, maybe. */
                tprintf3( &ErrorPort, MiscStrings[ 123 ], ErrorStrings[ n ] );
                if( (n != 1) && (n != 54) && (n != 326) && (n != 327) ) {  /* Insufficient heap space even after a GC, in which
                                                                              case we are wary of using the debugger, because there
                                                                              may not be enough heap space to run it. */
                    fflush( getPortFilePointer( &ErrorPort ) );
                    resetCleanUp();
                    if( debugBufferInitialized )
                        CALL( debugCode )       /* Returns if there is no debugger. */
                    }
                if( lambda_pair_p( L.t ) ) {
                    tprintf2( &ErrorPort, MiscStrings[ 115 ] );
                    R = L;
                    R.t = pairTag;
                    cdr();
                    cdr();
                    cdr();
                    if( empty_list_p( R.t ) ) {
                        tprintf2( &ErrorPort, MiscStrings[ 279 ] );
                        }
                    else {
                        tprintf2( &ErrorPort, MiscStrings[ 4 ] );
                        R = L;
                        displayLambdaNameOnDefaultPort();
                        tprintf2( &ErrorPort, MiscStrings[ 115 ] );
                       }
                    }
            while( (F != 0) && (F < frameBottom) ) {
                if( lambda_pair_p( (F + FRAME_NEW_L_OFFSET)->t ) ){
                    if( ! traceHeaderPrintedP ) {
                        R = *(F + FRAME_NEW_L_OFFSET);
                        R.t = pairTag;
                        cdr();
                        cdr();
                        cdr();
                        if( ! empty_list_p( R.t ) ) {
                            tprintf2( &ErrorPort, MiscStrings[ 280 ] );
                            traceHeaderPrintedP = true;
                            }
                        }
                    if( traceHeaderPrintedP ) {
                        tprintf2( &ErrorPort, MiscStrings[ 110 ] );
                        tprintf2( &ErrorPort, MiscStrings[ 110 ] );
                        R = *(F + FRAME_NEW_L_OFFSET);
                        displayLambdaNameOnDefaultPort();
                        tprintf2( &ErrorPort, MiscStrings[ 115 ] );
                        }
                    }
                F = ((F + FRAME_FP_OFFSET)->u).a;
                }
            tprintf2( &ErrorPort, MiscStrings[ 14 ] );
            break;
            }
        fflush( getPortFilePointer( &ErrorPort ) );
        }
    resetRecursiveLevel = 0;
    resetScheme();
    }

/**** tgetc -- get one character from the input, sleep for a while if there are none. ****/

LongWord tgetc( FILE *descriptor )
{
    Word c;

    if( descriptor == stdin ) {
        eventLoopOnce( 0 );
#ifdef PIXIE_SCHEMES
        if( ! tcharreadyConsole() ) {
            return EOF;
            }
#endif
        while( (c = tgetchar()) == EOF ) {
#ifdef USE_COCOA_IO
            eventLoopOnce( 0 );
            usleep( TGETC_DELAY );
#else
            if( ! feof( descriptor ) ) {
                clearerr( descriptor );
                usleep( TGETC_DELAY );
                }
            else
                break;
#endif
            }
        }
    else {
      while( (c = fgetc( descriptor )) == EOF ) {
            if( ! feof( descriptor ) ) {
                clearerr( descriptor );
                eventLoopOnce( 0 );
                usleep( TGETC_DELAY );
                }
            else
                break;
            }
        }
    return c;
    }       

/**** tgetcNoHangOnEOF -- get one character from the input, sleep for a while if there are none. ****/

LongWord tgetcNoHangOnEOF( FILE *descriptor )
{
    Word c;

    if( descriptor == stdin ) {
#ifdef USE_COCOA_IO
        eventLoopOnce( 0 );
        if( ! tcharreadyConsole() ) {
            return EOF;
            }
#endif
        while( (c = tgetchar()) == EOF ) {
#ifdef USE_COCOA_IO
            eventLoopOnce( 0 );
            break;
#else
            if( ! feof( descriptor ) ) {
                clearerr( descriptor );
                usleep( TGETC_DELAY );
                }
            else
                break;
#endif
            }
        }
    else
        while( (c = fgetc( descriptor )) == EOF )
            if( ! feof( descriptor ) ) {
                clearerr( descriptor );
                eventLoopOnce( 0 );
                usleep( TGETC_DELAY );
                }
            else
                break;
    return c;
    }       

/**** untgetc( FILE *descriptor ) ****/

void untgetc( Text c, FILE *descriptor )
{
    if( descriptor == stdin )
        untgetchar( c );
    else        
        ungetc( c, descriptor );
    }       

/**** tpeek ****/

LongWord tpeek( FILE *descriptor )
{
    LongWord c;
 
    c = tgetc( descriptor );
    if( c != EOF )
        untgetc( c, descriptor );   
    return c;
    }       

/**** tpeekNoHangOnEOF ****/

LongWord tpeekNoHangOnEOF( FILE *descriptor )
{
    LongWord c;
 
    c = tgetcNoHangOnEOF( descriptor );
    if( c != EOF )
        untgetc( c, descriptor );   
    return c;
    }       

/**** tcharready -- Is there a character ready? ****/

Boolean tcharready( FILE *descriptor )
{
    if( descriptor == stdin )
        return tcharreadyConsole();
    else                     // Files always have a character ready, but maybe it's EOF.
        return TRUE;
    }

/**** tputc -- Variation on "putc".      ****/

LongWord tputc( FILE *descriptor, char c ) { return fputc( c, descriptor ); }
                                
/**** tfflushConsoleOutput ****/

void tfflushConsoleOutput() { fflush( stdout ); fflush( stderr ); }

/**** tfflushOutput ****/

void tfflushOutput( FILE *descriptor ) { fflush( descriptor ); }


