/****************************************************************************************

MemoryManager.c++ -- Copyright (c) by Jay Reynolds Freeman 1987, 1988, 1989, 1990, 1991, 2006, 2009

Storage-allocation system for Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import <sys/types.h>
#import <sys/mman.h>
#import <fcntl.h>
#import <sys/ipc.h>
#import <sys/shm.h>
#import "LocalTypes.h"
#import "WraithHeader.h"
#import "MemoryManager.h"
#import "OpCodes.h"
#import "GarbageCollector.h"
#import "WorldHandler.h"
#import "DataTypes.e"
#import "Globals.e"
#import "IO.e"
#import "MainAux.e"
#import "MMapStuff.e"

static const UnsignedLongWord localOne = 1;

// NOTE that these next two macros are important; they are the guts of the storage allocator.
// The first allocates in the current allocation domain, whatever it may be.
// The second allocates in tenured space, even if the current allocation domain is not tenured space.

#define ALLOCATE_LOOP_GUTS( nBytes ) \
    CurrentMainSpace->allocateDomain->free = CurrentMainSpace->allocateDomain->free - (nBytes); \
    A.t = pointerToByteBlockTag; \
    A.u.a = (taggedAval *)(CurrentMainSpace->allocateDomain->free); \
    (A.u.a)->t = (byteBlockTag | (localOne << locationByteOffset) | ((LongWord)theKittenNumber << kittenByteOffset)); \
    ((A.u.a)->u).a = (taggedAval *)(nBytes - sizeof( taggedAval ))


#define TENURED_ALLOCATE_LOOP_GUTS( nBytes ) \
    CurrentMainSpace->free = CurrentMainSpace->free - (nBytes); \
    A.t = pointerToByteBlockTag; \
    A.u.a = (taggedAval *)(CurrentMainSpace->free); \
    (A.u.a)->t = (byteBlockTag | (localOne << locationByteOffset) | ((LongWord)theKittenNumber << kittenByteOffset)); \
    ((A.u.a)->u).a = (taggedAval *)((nBytes) - sizeof( taggedAval ))

#define SPIN_DELAY_IN_NANOSECONDS 500
#define LONG_SPIN_DELAY_IN_NANOSECONDS 10000000
struct timespec GC_BYTES_VALID_DELAY_struct             = { 0, SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds
struct timespec KITTEN_HALT_DELAY_struct                = { 0, LONG_SPIN_DELAY_IN_NANOSECONDS };     // nanoseconds

extern int errno;

/**** statics ****/

/****  initialize -- very simple, but there is much work to be done later, after various things have
                     been loaded in and calculated.  ****/

void MemoryManager::initialize()
{

    CurrentMainSpace = NULL;
    AlternateMainSpace = NULL;
    SchemeStack = NULL;
    ScratchStack = NULL;
    GGCScratchSpace = NULL;

    }

/****  initializeMemoryManager -- does the real work of initialization, and there is a lot of it.  ****/

void MemoryManager::initializeMemoryManager( Boolean bEFlag,
                                             Boolean useGenerationalSpaces,
                                             Text *externalNameString,
                                             size_t externalNameStringSize,
                                             size_t spaceSize,
                                             size_t stackSize,
                                             size_t continuationSize,
                                             size_t generationSize,
                                             size_t *createdSpaceSize,
                                             Byte *spaceAGivenTargetAddress,
                                             Byte *spaceBGivenTargetAddress,
                                             Byte *scratchStackGivenTargetAddress,
                                             Byte *gGCScratchSpaceGivenTargetAddress
                                             )
{
    volatile memorySpace *SpaceA = 0;
    volatile memorySpace *SpaceB = 0;
    Byte *spaceABytes = 0;
    Byte *spaceBBytes = 0;
    Byte *stackBytes = 0;
    Byte *continuationBytes = 0;
    Byte *scratchStackBytes = 0;
    Byte *gGCScratchSpaceBytes = 0;
    size_t nSpaceBytes = 0;
    size_t nStackBytes = 0;
    size_t nContinuationBytes = 0;
    size_t nGGCScratchSpaceBytes = 0;
    int mmapFlags = 0;
    int openFlags = 0;
    byteAllocCounter = 0;
    setBangCounter = 0;
    size_t generationScratchSpaceSize;

    int spaceA_fd = 0;
    int spaceB_fd = 0;
    int scratchStack_fd = 0;
    int gGCScratchSpace_fd = 0;
    Byte *scratchWriteBuffer = 0;
    size_t scratchWriteBufferSize = 0;
    size_t virtualBytesWritten = 0;
    LongWord writeResult = 0;
    LongWord progressDenominator = 0;

    BigEndianFlag = bEFlag;    

    if( externalNameString == 0 ) {
        externalNameString = &(scratch[ 0 ]);
        externalNameStringSize = MAX_PATHNAME_LENGTH;
        }

    if( spaceSize > MAXIMUM_MEMORY_SIZE ) {
        spaceSize = MAXIMUM_MEMORY_SIZE;
        }

    // Set up some sizes for the various spaces, taking sure make sure they are appropriate for correct alignment.

    stackSize = stackSize - (stackSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nStackBytes = sizeof( memorySpace ) + stackSize + PAGE_SIZE;
    nStackBytes = nStackBytes - (nStackBytes % PAGE_SIZE);

    generationSpaceSize = generationSize;
    if( generationSpaceSize > (spaceSize / 10) )
        generationSpaceSize = spaceSize / 10;
    if( (generationSpaceSize % PAGE_SIZE) != 0 ) {
        generationSpaceSize = generationSpaceSize + PAGE_SIZE;
        generationSpaceSize = generationSpaceSize - (generationSpaceSize % PAGE_SIZE);
        }
    if( generationSpaceSize > MAX_GENERATION_SPACE_SIZE )
        fatalError( 175 );
//    allocateMaxOffset = 4 * generationSpaceSize + PAGE_SIZE;
    allocateMaxOffset = 2 * generationSpaceSize + PAGE_SIZE;

    continuationSize = continuationSize - (continuationSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nContinuationBytes = sizeof( memorySpace ) + continuationSize + PAGE_SIZE;
    nContinuationBytes = nContinuationBytes - (nContinuationBytes % PAGE_SIZE);

    spaceSize = spaceSize - (spaceSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nSpaceBytes = 3 * sizeof( memorySpace ) + spaceSize + PAGE_SIZE;
    nSpaceBytes = nSpaceBytes - (nSpaceBytes % PAGE_SIZE);

    generationScratchSpaceSize = 2 * generationSpaceSize + 2 * PAGE_SIZE;
    nGGCScratchSpaceBytes = sizeof( memorySpace ) + generationScratchSpaceSize + PAGE_SIZE;
    nGGCScratchSpaceBytes = nGGCScratchSpaceBytes - (nGGCScratchSpaceBytes % PAGE_SIZE);

    // Deal with file names for the mmapped files.

    snprintf( spaceAFileName,          MAX_PATHNAME_LENGTH, "%s/", appSupDirName );
    snprintf( spaceBFileName,          MAX_PATHNAME_LENGTH, "%s/", appSupDirName );
    snprintf( scratchStackFileName,    MAX_PATHNAME_LENGTH, "%s/", appSupDirName );
    snprintf( gGCScratchSpaceFileName, MAX_PATHNAME_LENGTH, "%s/", appSupDirName );

#ifdef USE_COCOA_IO

#ifdef PIXIE_SCHEMES

    strncat( spaceAFileName,
             "Pixie Scheme II Space A", MAX_PATHNAME_LENGTH - strlen( spaceAFileName ) );
    strncat( spaceBFileName,
             "Pixie Scheme II Space B", MAX_PATHNAME_LENGTH - strlen( spaceBFileName ) );
    strncat( scratchStackFileName,
             "Pixie Scheme II Scratch Stack", MAX_PATHNAME_LENGTH - strlen( scratchStackFileName ) );
    strncat( gGCScratchSpaceFileName,
             "Pixie Scheme II GGC Scratch Space", MAX_PATHNAME_LENGTH - strlen( gGCScratchSpaceFileName ) );

#else

    strncat( spaceAFileName,
             "Wraith Scheme Space A", MAX_PATHNAME_LENGTH - strlen( spaceAFileName ) );
    strncat( spaceBFileName,
             "Wraith Scheme Space B", MAX_PATHNAME_LENGTH - strlen( spaceBFileName ) );
    strncat( scratchStackFileName,
             "Wraith Scheme Scratch Stack", MAX_PATHNAME_LENGTH - strlen( scratchStackFileName ) );
    strncat( gGCScratchSpaceFileName,
             "Wraith Scheme GGC Scratch Space", MAX_PATHNAME_LENGTH - strlen( gGCScratchSpaceFileName ) );

#endif

#else

    strncat( spaceAFileName,
             "Wraith Scheme Unix Space A", MAX_PATHNAME_LENGTH - strlen( spaceAFileName ) );
    strncat( spaceBFileName,
             "Wraith Scheme Unix Space B", MAX_PATHNAME_LENGTH - strlen( spaceBFileName ) );
    strncat( scratchStackFileName,
             "Wraith Scheme Unix Scratch Stack", MAX_PATHNAME_LENGTH - strlen( scratchStackFileName ) );
    strncat( gGCScratchSpaceFileName,
             "Wraith Scheme Unix GGC Scratch Space", MAX_PATHNAME_LENGTH - strlen( gGCScratchSpaceFileName ) );

#endif

    // Set up the mmap flags, which are different for the MomCat and the other kittens.

    if( isMomCatP ) {
        mmapFlags = MOMCAT_MMAP_FLAGS;
        openFlags = MOMCAT_MMAP_CREATE_FLAGS;
        }
    else {
        mmapFlags = KITTEN_MMAP_FLAGS;
        openFlags = KITTEN_MMAP_CREATE_FLAGS;
        }

    // Remove any preexisting files.

    if( isMomCatP )
        releaseMemoryFiles();

    // Open the various backing store files for mmap.

    spaceA_fd = open( spaceAFileName, openFlags, 0666 );
    if( spaceA_fd == -1 ) {
        fatalError( 134 );
        return;
        }
    spaceB_fd = open( spaceBFileName, openFlags, 0666 );
    if( spaceB_fd == -1 ) {
        fatalError( 134 );
        return;
        }
    scratchStack_fd = open( scratchStackFileName, openFlags, 0666 );
    if( scratchStack_fd == -1 )  {
        fatalError( 139 );
        return;
        }
    gGCScratchSpace_fd = open( gGCScratchSpaceFileName, openFlags, 0666 );
    if( gGCScratchSpace_fd == -1 )  {
        fatalError( 173 );
        return;
        }
   
    // If we are the MomCat, we must actually write something to all those file bytes.

    if( isMomCatP ) {

        // Make some guesses about an appropriate buffer size for writing to the files.

        if( nSpaceBytes      < 11000000 )
            scratchWriteBufferSize =    128 * PAGE_SIZE;
        else if( nSpaceBytes < 110000000 )
            scratchWriteBufferSize =   1280 * PAGE_SIZE;
        else if( nSpaceBytes < 1100000000 )
            scratchWriteBufferSize =  12800 * PAGE_SIZE;
        else
            scratchWriteBufferSize = 128000 * PAGE_SIZE;

        scratchWriteBuffer = (Byte *)calloc( scratchWriteBufferSize, sizeof( Byte ) );

        if( scratchWriteBuffer == 0 ) {
            fatalError( 123 );
            return;
            }

        progressDenominator = nSpaceBytes + nSpaceBytes + nStackBytes + nGGCScratchSpaceBytes;
        progressFractionInThousandths = 20;

        virtualBytesWritten = 0;
        while( virtualBytesWritten < nSpaceBytes ) {        
            if( (nSpaceBytes - virtualBytesWritten) >= (UnsignedLongWord)scratchWriteBufferSize )
                writeResult = write( spaceA_fd, scratchWriteBuffer, scratchWriteBufferSize );
            else
                writeResult = write( spaceA_fd, scratchWriteBuffer, nSpaceBytes - virtualBytesWritten );
            if( writeResult == -1 ) {
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, MiscStrings[ 487 ], strerror( errno ) );
                fatalError( 124 );
                return;
                }
            virtualBytesWritten += scratchWriteBufferSize;
            progressFractionInThousandths = (virtualBytesWritten * 980) / progressDenominator;
            }

        virtualBytesWritten = 0;
        while( virtualBytesWritten < nSpaceBytes ) {        
            if( (nSpaceBytes - virtualBytesWritten) >= (UnsignedLongWord)scratchWriteBufferSize )
                writeResult = write( spaceB_fd, scratchWriteBuffer, scratchWriteBufferSize );
            else
                writeResult = write( spaceB_fd, scratchWriteBuffer, nSpaceBytes - virtualBytesWritten );
            if( writeResult == -1 ) {
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, MiscStrings[ 488 ], strerror( errno ) );
                fatalError( 124 );
                return;
                }
            virtualBytesWritten += scratchWriteBufferSize;
            progressFractionInThousandths = ((nSpaceBytes + virtualBytesWritten) * 980) / progressDenominator;
            }

        virtualBytesWritten = 0;
        while( virtualBytesWritten < nStackBytes ) {        
            if( (nStackBytes - virtualBytesWritten) >= (UnsignedLongWord)scratchWriteBufferSize )
                writeResult = write( scratchStack_fd, scratchWriteBuffer, scratchWriteBufferSize );
            else
                writeResult = write( scratchStack_fd, scratchWriteBuffer, nStackBytes - virtualBytesWritten );
            if( writeResult == -1 ) {
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, MiscStrings[ 489 ], strerror( errno ) );
                fatalError( 124 );
                return;
                }
            virtualBytesWritten += scratchWriteBufferSize;
            progressFractionInThousandths = ((nSpaceBytes + nSpaceBytes + virtualBytesWritten) * 980) / progressDenominator;
            }

        virtualBytesWritten = 0;
        while( virtualBytesWritten < nGGCScratchSpaceBytes ) {        
            if( (nGGCScratchSpaceBytes - virtualBytesWritten) >= (UnsignedLongWord)scratchWriteBufferSize )
                writeResult = write( gGCScratchSpace_fd, scratchWriteBuffer, scratchWriteBufferSize );
            else
                writeResult = write( gGCScratchSpace_fd, scratchWriteBuffer, nGGCScratchSpaceBytes - virtualBytesWritten );
            if( writeResult == -1 ) {
                useFatalErrorOutput();
                tprintf3( &CurrentOutputPort, MiscStrings[ 490 ], strerror( errno ) );
                fatalError( 124 );
                return;
                }
            virtualBytesWritten += scratchWriteBufferSize;
            progressFractionInThousandths = ((nSpaceBytes + nSpaceBytes + virtualBytesWritten) * 980) / progressDenominator;
            }
        }

    if( nSpaceBytes > SPACE_RESERVE_SIZE ) {
        useFatalErrorOutput();
        tprintf4( &ErrorPort, MiscStrings[ 491 ], nSpaceBytes, SPACE_RESERVE_SIZE );
        fatalError( 154 );
        return;
        }
    if( nStackBytes > STACK_RESERVE_SIZE ) {
        useFatalErrorOutput();
        tprintf4( &ErrorPort, MiscStrings[ 492 ], nStackBytes, STACK_RESERVE_SIZE );
        fatalError( 156 );
        return;
        }
    if( nGGCScratchSpaceBytes > GGC_SCRATCH_RESERVE_SIZE ) {
        useFatalErrorOutput();
        tprintf4( &ErrorPort, MiscStrings[ 493 ],
                  nGGCScratchSpaceBytes, GGC_SCRATCH_RESERVE_SIZE );
        fatalError( 177 );
        return;
        }

    // Memory-map the files.

    spaceABytes = (Byte *)mmap( spaceAGivenTargetAddress, nSpaceBytes, MMAP_PROT, mmapFlags, spaceA_fd, 0 );
    spaceBBytes = (Byte *)mmap( spaceBGivenTargetAddress, nSpaceBytes, MMAP_PROT, mmapFlags, spaceB_fd, 0 );
    scratchStackBytes = (Byte *)mmap( scratchStackGivenTargetAddress, nStackBytes, MMAP_PROT, mmapFlags, scratchStack_fd, 0 );
    gGCScratchSpaceBytes = (Byte *)mmap( gGCScratchSpaceGivenTargetAddress, nGGCScratchSpaceBytes,
                                         MMAP_PROT, mmapFlags, gGCScratchSpace_fd, 0 );
    if( spaceABytes == (void *)(-1) ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 494 ], strerror( errno) );
        fatalError( 125);
        return;
        }
    else if( spaceABytes != spaceAGivenTargetAddress ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 495 ], (UnsignedLongWord)spaceAGivenTargetAddress );
        fatalError( 125);
        return;
        }
    if( spaceBBytes == (void *)(-1) ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 496 ], strerror( errno) );
        fatalError( 125);
        return;
        }
    else  if( spaceBBytes != spaceBGivenTargetAddress ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 497 ], (UnsignedLongWord)spaceBGivenTargetAddress );
        fatalError( 125);
        return;
        }
    if( scratchStackBytes == (void *)(-1) ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 498 ], strerror( errno) );
        fatalError( 125);
        return;
        }
    else if( scratchStackBytes != scratchStackGivenTargetAddress ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 499 ], (UnsignedLongWord)scratchStackGivenTargetAddress );
        fatalError( 125);
        return;
        }
    if( gGCScratchSpaceBytes == (void *)(-1) ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 500 ], strerror( errno) );
        fatalError( 125);
        return;
        }
    else if( gGCScratchSpaceBytes != gGCScratchSpaceGivenTargetAddress ) {
        useFatalErrorOutput();
        tprintf3( &CurrentOutputPort, MiscStrings[ 501 ],
                  (UnsignedLongWord)gGCScratchSpaceGivenTargetAddress );
        fatalError( 125);
        return;
        }
    close( spaceA_fd );
    close( spaceB_fd );
    close( scratchStack_fd );

    // If we are the MomCat, correctly load the initial values into the mmapped files.

    if( isMomCatP ) {
        initializeMemorySpaceGivenByteBlock( useGenerationalSpaces, &SpaceA, spaceABytes, spaceSize, generationSpaceSize );
        initializeMemorySpaceGivenByteBlock( useGenerationalSpaces, &SpaceB, spaceBBytes, spaceSize, generationSpaceSize );
        initializeMemorySpaceGivenByteBlock( FALSE, &ScratchStack, scratchStackBytes, stackSize, 0 );
        initializeMemorySpaceGivenByteBlock( FALSE, &GGCScratchSpace, gGCScratchSpaceBytes, generationScratchSpaceSize, 0 );
        }
    else {
        SpaceA = (volatile memorySpace *)spaceABytes;
        SpaceB = (volatile memorySpace *)spaceBBytes;
        ScratchStack = (volatile memorySpace *)scratchStackBytes;
        GGCScratchSpace = (volatile memorySpace *)gGCScratchSpaceBytes;
        }     

    if( (stackBytes = (Byte *)malloc( nStackBytes )) == 0 ) {
        fatalError( 126);
        return;
        }

    initializeMemorySpaceGivenByteBlock( FALSE, &SchemeStack, stackBytes, stackSize, 0 );

    if( (continuationBytes = (Byte *)malloc( nContinuationBytes )) == 0 ) {
        fatalError( 162);
        return;
        }

    initializeMemorySpaceGivenByteBlock( FALSE, &SchemeContinuation, continuationBytes, continuationSize, 0 );

    // Hook up with a world handler and garbage collector.

    myGarbageCollector = new GarbageCollector;
    myGarbageCollector->initializeGC( this );
    myGarbageCollector->setIsMomCat( isMomCatP );
    myWorldHandler = new WorldHandler;
    myWorldHandler->initializeWH( bEFlag, externalNameString, externalNameStringSize );
    myClowder = NULL;

    CurrentMainSpace = SpaceA;
    AlternateMainSpace = SpaceB;

    // Set up the locations of the active and aging stacks in the current and alternate main spaces.    

    CurrentMainSpace->activeGenerationStackEnd = (volatile UnsignedLongWord *)(AlternateMainSpace->end);
    CurrentMainSpace->agingGenerationStackEnd
        = (volatile UnsignedLongWord *)(AlternateMainSpace->end) - ((AlternateMainSpace->size) / (2 * sizeof( UnsignedLongWord *)));

    AlternateMainSpace->activeGenerationStackEnd = (volatile UnsignedLongWord *)(CurrentMainSpace->end);
    AlternateMainSpace->agingGenerationStackEnd
        = (volatile UnsignedLongWord *)(CurrentMainSpace->end) - ((CurrentMainSpace->size) / (2 * sizeof( UnsignedLongWord *)));

    theClowder->initializeGGCAddressStacks( AlternateMainSpace );

    CurrentMainSpace->inUse = TRUE;
    *createdSpaceSize = spaceSize;
    OSMemoryBarrier();

    // The default is to set up for generational garbage collection, here is where we change to full GC if only that is in use.

    if( ! usingGenerationalGC )
        revertToFullGC();
    }

/****  initializeMemoryManagerByMalloc -- does the real work of initialization, and there is a lot of it.  ****/

void MemoryManager::initializeMemoryManagerByMalloc( Boolean bEFlag,
                                                     Boolean useGenerationalSpaces,
                                                     Text *externalNameString,
                                                     size_t externalNameStringSize,
                                                     size_t spaceSize,
                                                     size_t stackSize,
                                                     size_t continuationSize,
                                                     size_t generationSize,
                                                     size_t *createdSpaceSize
                                                     )
{
    volatile memorySpace *SpaceA = 0;
    volatile memorySpace *SpaceB = 0;
    Byte *spaceABytes = 0;
    Byte *spaceBBytes = 0;
    Byte *stackBytes = 0;
    Byte *continuationBytes = 0;
    Byte *scratchStackBytes = 0;
    Byte *gGCScratchSpaceBytes = 0;
    size_t nSpaceBytes = 0;
    size_t nStackBytes = 0;
    size_t nContinuationBytes = 0;
    size_t nGGCScratchSpaceBytes = 0;
    byteAllocCounter = 0;
    setBangCounter = 0;
    size_t generationScratchSpaceSize;

    BigEndianFlag = bEFlag;    

    if( externalNameString == 0 ) {
        externalNameString = &(scratch[ 0 ]);
        externalNameStringSize = MAX_PATHNAME_LENGTH;
        }

    if( spaceSize > MAXIMUM_MEMORY_SIZE ) {
        spaceSize = MAXIMUM_MEMORY_SIZE;
        }

    // Set up some sizes for the various spaces, taking sure make sure they are appropriate for correct alignment.

    stackSize = stackSize - (stackSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nStackBytes = sizeof( memorySpace ) + stackSize + PAGE_SIZE;
    nStackBytes = nStackBytes - (nStackBytes % PAGE_SIZE);

    generationSpaceSize = generationSize;
    if( generationSpaceSize > (spaceSize / 10) )
        generationSpaceSize = spaceSize / 10;
    if( (generationSpaceSize % PAGE_SIZE) != 0 ) {
        generationSpaceSize = generationSpaceSize + PAGE_SIZE;
        generationSpaceSize = generationSpaceSize - (generationSpaceSize % PAGE_SIZE);
        }
    if( generationSpaceSize > MAX_GENERATION_SPACE_SIZE )
        fatalError( 175 );
//    allocateMaxOffset = 4 * generationSpaceSize + PAGE_SIZE;
    allocateMaxOffset = 2 * generationSpaceSize + PAGE_SIZE;

    continuationSize = continuationSize - (continuationSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nContinuationBytes = sizeof( memorySpace ) + continuationSize + PAGE_SIZE;
    nContinuationBytes = nContinuationBytes - (nContinuationBytes % PAGE_SIZE);

    spaceSize = spaceSize - (spaceSize % sizeof( taggedAval ));     // Tagged-aval aligned.
    nSpaceBytes = 3 * sizeof( memorySpace ) + spaceSize + PAGE_SIZE;
    nSpaceBytes = nSpaceBytes - (nSpaceBytes % PAGE_SIZE);

    generationScratchSpaceSize = 2 * generationSpaceSize + 2 * PAGE_SIZE;
    nGGCScratchSpaceBytes = sizeof( memorySpace ) + generationScratchSpaceSize + PAGE_SIZE;
    nGGCScratchSpaceBytes = nGGCScratchSpaceBytes - (nGGCScratchSpaceBytes % PAGE_SIZE);

    // If we are the MomCat, we must actually write something to all those file bytes.

    // Malloc the various chunks of memory:

    spaceABytes = (Byte *)calloc( nSpaceBytes, 1 );
    if( spaceABytes == 0 ) {
        fatalError( 2 );
        return;
        }
    spaceBBytes = (Byte *)calloc( nSpaceBytes, 1 );
    if( spaceBBytes == 0 ) {
        fatalError( 2 );
        return;
        }
    scratchStackBytes = (Byte *)calloc( nStackBytes, 1 );
    if( scratchStackBytes == 0 ) {
        fatalError( 2 );
        return;
        }
    gGCScratchSpaceBytes = (Byte *)calloc( nGGCScratchSpaceBytes, 1 );
    if( gGCScratchSpaceBytes == 0 ) {
        fatalError( 2 );
        return;
        }
   
    initializeMemorySpaceGivenByteBlock( useGenerationalSpaces, &SpaceA, spaceABytes, spaceSize, generationSpaceSize );
    initializeMemorySpaceGivenByteBlock( useGenerationalSpaces, &SpaceB, spaceBBytes, spaceSize, generationSpaceSize );
    initializeMemorySpaceGivenByteBlock( FALSE, &ScratchStack, scratchStackBytes, stackSize, 0 );
    initializeMemorySpaceGivenByteBlock( FALSE, &GGCScratchSpace, gGCScratchSpaceBytes, generationScratchSpaceSize, 0 );

    if( (stackBytes = (Byte *)malloc( nStackBytes )) == 0 ) {
        fatalError( 126);
        return;
        }

    initializeMemorySpaceGivenByteBlock( FALSE, &SchemeStack, stackBytes, stackSize, 0 );

    if( (continuationBytes = (Byte *)malloc( nContinuationBytes )) == 0 ) {
        fatalError( 162);
        return;
        }

    initializeMemorySpaceGivenByteBlock( FALSE, &SchemeContinuation, continuationBytes, continuationSize, 0 );

    // Hook up with a world handler and garbage collector.

    myGarbageCollector = new GarbageCollector;
    myGarbageCollector->initializeGC( this );
    myGarbageCollector->setIsMomCat( TRUE );
    myWorldHandler = new WorldHandler;
    myWorldHandler->initializeWH( bEFlag, externalNameString, externalNameStringSize );
    myClowder = NULL;

    CurrentMainSpace = SpaceA;
    AlternateMainSpace = SpaceB;

    // Set up the locations of the active and aging stacks in the current and alternate main spaces.    

    CurrentMainSpace->activeGenerationStackEnd = (volatile UnsignedLongWord *)(AlternateMainSpace->end);
    CurrentMainSpace->agingGenerationStackEnd
        = (volatile UnsignedLongWord *)(AlternateMainSpace->end) - ((AlternateMainSpace->size) / (2 * sizeof( UnsignedLongWord *)));

    AlternateMainSpace->activeGenerationStackEnd = (volatile UnsignedLongWord *)(CurrentMainSpace->end);
    AlternateMainSpace->agingGenerationStackEnd
        = (volatile UnsignedLongWord *)(CurrentMainSpace->end) - ((CurrentMainSpace->size) / (2 * sizeof( UnsignedLongWord *)));

    theClowder->initializeGGCAddressStacks( AlternateMainSpace );

    CurrentMainSpace->inUse = TRUE;
    *createdSpaceSize = spaceSize;
    OSMemoryBarrier();

    // The default is to set up for generational garbage collection, here is where we change to full GC if only that is in use.

    if( ! usingGenerationalGC )
        revertToFullGC();
    }

/**** initializeMemorySpaceGivenByteBlock -- given a block of space to use -- presumably an mmapped one -- establish
                                             a memory space within it.  There is a flag to tell whether to set up
                                             generational spaces within the given block.  ****/

volatile memorySpace *MemoryManager::initializeMemorySpaceGivenByteBlock(
     Boolean useGenerationalSpaces,
     volatile memorySpace **m,
     Byte *block,
     UnsignedLongWord bytesGiven,
     UnsignedLongWord genSpaceSize
     )
{
#ifdef PIXIE_SCHEME_III

#endif

    (*m) = (volatile memorySpace *)block;
    (*m)->space = (*m)->lowFree = block + sizeof( memorySpace );
    if( ((UnsignedLongWord)((*m)->space) & 0x7) != 0)
        fatalError( 161 );
    (*m)->end = (*m)->space + bytesGiven;
    (*m)->free = (*m)->end;
    (*m)->size = bytesGiven;
    (*m)->inUse = FALSE;
    (*m)->allocateDomain = *m;  // For not using generational
    (*m)->allocateMax = (*m)->space;
    (*m)->allocateSize = (*m)->size;

    // There are two generational spaces set a little apart at the top of the given block of space.
    // There needs to be enough space around the generational space that there is room to fill up
    // the space between them with a byte block marked as garbage, when it comes time to do a full
    // garbage collection of the entire space.

    if( useGenerationalSpaces ) {
        (*m)->generationA = initializeGenerationalSpace( (volatile memorySpace *)((*m)->end) );

        ((*m)->generationA)->space = (*m)->space;
        if( ((UnsignedLongWord)(((*m)->generationA)->space) & 0x7) != 0)
            fatalError( 62 );
        ((*m)->generationA)->allocateMax = ((*m)->generationA)->space;
        ((*m)->generationA)->end = ((*m)->generationA)->space + genSpaceSize;
        ((*m)->generationA)->size = genSpaceSize;
        ((*m)->generationA)->allocateSize = genSpaceSize;
        ((*m)->generationA)->inUse = true;
		clearGenerationSpace( ((*m)->generationA) );

        (*m)->generationB = initializeGenerationalSpace( (volatile memorySpace *)((*m)->end + sizeof( memorySpace)) );

        ((*m)->generationB)->space = ((*m)->generationA)->end + PAGE_SIZE;
        if( ((UnsignedLongWord)(((*m)->generationB)->space) & 0x7) != 0)
            fatalError( 62 );
        ((*m)->generationB)->allocateMax = ((*m)->generationB)->space;
        ((*m)->generationB)->end = ((*m)->generationB)->space + genSpaceSize;
        ((*m)->generationB)->size = genSpaceSize;
        ((*m)->generationB)->allocateSize = genSpaceSize;
        ((*m)->generationB)->inUse = true;
		clearGenerationSpace( ((*m)->generationB) );

        (*m)->activeGeneration = (*m)->generationA;
        (*m)->agingGeneration = (*m)->generationB;

        (*m)->allocateDomain = ((*m)->activeGeneration);
        (*m)->allocateMax = ((*m)->generationB)->end + allocateMaxOffset;
        (*m)->allocateSize = (*m)->end - (*m)->allocateMax;
        if( ((UnsignedLongWord)((*m)->allocateMax) & 0x7) != 0)
            fatalError( 63 );
        }

    return (*m);
    }

/**** copyMemorySpaceMiscelleanea -- Portmanteau for moving memory space stuff around during garbage collection. ****/ 

void MemoryManager::copyMemorySpaceMiscellanea( volatile memorySpace *from, volatile memorySpace *to, UnsignedLongWord *toMinusFrom )
{
    *toMinusFrom = (Byte *)to - (Byte *)from;

    if( to->size != from->size )
        fatalError( 141 );

    to->space = from->space + *toMinusFrom;
    to->lowFree = from->lowFree + *toMinusFrom;
    to->free = from->free + *toMinusFrom;
    to->size = from->size;
    to->inUse = from->inUse;
    to->free = from->free + *toMinusFrom;
    }

/**** copyMemorySpacePartialContent -- Portmanteau for moving memory space stuff around during garbage collection. ****/ 

void MemoryManager::copyMemorySpacePartialContent( volatile memorySpace *from, volatile memorySpace *to,
                                                   UnsignedLongWord *toMinusFrom, Byte *startHere )
{
    LongWord offsetInSpace = startHere - from->space;
    size_t copyCount = from->end - startHere;

    *toMinusFrom = (Byte *)to - (Byte *)from;

    if( to->size < copyCount )
        fatalError( 164 );

    memcpy( (void *)(to->space + offsetInSpace), (void *)startHere, copyCount );
    to->free = from->free + *toMinusFrom;          // Nota bene -- there were a couple of options here ...
    }

/**** setClowder -- identify the clowder associated with this memory manager.  FIXME -- may not be needed.  ****/

void MemoryManager::setClowder( Clowder *aClowder )
{
    myClowder = aClowder;
    myGarbageCollector->setClowder( aClowder );
    }

/**** releaseMemoryFiles  --  remove unnecessary files.  ****/

void MemoryManager::releaseMemoryFiles()
{
    if( isMomCatP ) {
        unlink( spaceAFileName );
        unlink( spaceBFileName );
        unlink( scratchStackFileName );
        unlink( gGCScratchSpaceFileName );
        }
    }

/**** nOfTaggedAvalInHeap  -- what is the greatest number of tagged avals that this heap can contain? ****/

UnsignedLongWord MemoryManager::nOfTaggedAvalInHeap( taggedAval *tp )
{
    return tp - ((taggedAval *)(CurrentMainSpace->space));
    }

// Various functions for printing the content of memory spaces.

/**** describeMemorySpace -- Display what's in a memory space, possibly verbosely.
                              When this primitive prints out the contents of a memory
                              cell, it uses Scheme print routines.  ****/

void describeMemorySpace( volatile memorySpace *m, Boolean showSpaceP )
{
    Byte *tap;
    UnsignedLongWord length;
        
    tprintf3( &CurrentOutputPort, MiscStrings[ 74 ], m );
    tprintf5( &CurrentOutputPort, MiscStrings[ 26 ], m->end - m->free, m->size, m->size );
    tprintf7( &CurrentOutputPort, MiscStrings[ 9 ], m->space, m->free, m->lowFree, m->end, m->inUse );
    if( (m->space > m->free) || (m->free > m-> end) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 35 ] );
    else if( showSpaceP )
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            length = describeTaggedAvalPointer( (taggedAval *)tap );                        
            tap += length;
            }
    // else nothing
    }

/**** describeMemorySpaceNonGarbage -- Display what's in a memory space, possibly verbosely.
                                       When this primitive prints out the contents of a memory
                                       cell, it uses Scheme print routines.  ****/

void describeMemorySpaceNonGarbage( volatile memorySpace *m, Boolean showSpaceP )
{
    Byte *tap;
    UnsignedLongWord length;
        
    tprintf3( &CurrentOutputPort, MiscStrings[ 74 ], m );
    tprintf5( &CurrentOutputPort, MiscStrings[ 26 ], m->end - m->free, m->size, m->size );
    tprintf7( &CurrentOutputPort, MiscStrings[ 9 ], m->space, m->free, m->lowFree, m->end, m->inUse );
    if( (m->space > m->free) || (m->free > m-> end) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 35 ] );
    else if( showSpaceP )
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            if( gc_bit( ((taggedAval *)tap)->t ) )
                length = describeTaggedAvalPointer( (taggedAval *)tap );                        
            else
                length = lengthTaggedAvalPointer( (taggedAval *)tap );
            tap += length;
            }
    // else nothing
    }

/**** describeLocksInMemorySpace -- Display what's locked in a memory space, possibly verbosely.
                                    When this primitive prints out the contents of a memory
                                    cell, it uses Scheme print routines.  ****/

void describeLocksInMemorySpace( volatile memorySpace *m )
{
    Byte *tap;
    UnsignedLongWord length;
    Boolean foundAnyLocksP = false;
    UnsignedWord counter = 0;
    UnsignedWord aKitten = 0;
    Word theBin = 0;
    UnsignedWord theLocation = 0;
    taggedAval *tp = 0;
        
    tprintf3( &CurrentOutputPort, MiscStrings[ 74 ], m );
    tprintf5( &CurrentOutputPort, MiscStrings[ 26 ], m->end - m->free, m->size, m->size );
    tprintf7( &CurrentOutputPort, MiscStrings[ 9 ], m->space, m->free, m->lowFree, m->end, m->inUse );
    if( (m->space > m->free) || (m->free > m-> end) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 35 ] );
    else {
        foundAnyLocksP = false;
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            if( getLocationByte( ((taggedAval *)tap)->t) ) {
                foundAnyLocksP = TRUE;
                length = describeTaggedAvalPointer( (taggedAval *)tap );                        
                }
            else
                length = lengthTaggedAvalPointer( (taggedAval *)tap );
            tap += length;
            }
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 209 ] );
        foundAnyLocksP = false;
        for( counter = 0; counter < SYMBOL_HASH_CONSTANT; ++counter ) {
            if( theClowder->obListBinLockBit( counter ) ) {
                foundAnyLocksP = true;
                tprintf3( &CurrentOutputPort, MiscStrings[ 239 ], counter );
                }
            }
        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) 
            if( (theBin = theClowder->getKittenHasLockedObListBin( aKitten )) != -1 ) {
                foundAnyLocksP = true;
                tprintf4( &CurrentOutputPort, MiscStrings[ 249 ], (Word)counter, theBin );
                }
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 248 ] );
        foundAnyLocksP = false;
        for( counter = 0; counter < SYMBOL_HASH_CONSTANT; ++counter ) {
            if( theClowder->TLEBinLockBit( counter ) ) {
                foundAnyLocksP = true;
                tprintf3( &CurrentOutputPort, MiscStrings[ 265 ], counter );
                }
            }
        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) 
            if( (theBin = theClowder->getKittenHasLockedTLEBin( aKitten )) != -1 ) {
                foundAnyLocksP = true;
                tprintf4( &CurrentOutputPort, MiscStrings[ 398 ], (Word)aKitten, theBin );
                }
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 397 ] );
        foundAnyLocksP = false;
        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) 
            if( (theLocation = theClowder->getCurrentActiveBlock( aKitten )) != 0 ) {
                tp = (taggedAval *)(theClowder->getCurrentActiveTaggedAval( aKitten ));
                foundAnyLocksP = true;
                fprintf( stdout, MiscStrings[ 409 ], (Word)aKitten, theLocation );
                printfAnyTaggedAval( stdout, tp );
                fputs( MiscStrings[ 115 ], stdout );
                printfAnyTaggedAval( stdout, (tp->u).a );
                fputs( MiscStrings[ 115 ], stdout );
                fputs(  MiscStrings[ 115 ], stdout );
                }
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 410 ] );
        }
    }

/**** releaseLocksInMemorySpace -- Unlocks any locks it finds in the given memory space.
                                   When this primitive prints out the contents of a memory
                                   cell, it uses Scheme print routines.  ****/

void releaseLocksInMemorySpace( volatile memorySpace *m )
{
    Byte *tap;
    UnsignedLongWord length;
    Boolean foundAnyLocksP = false;
    UnsignedWord counter = 0;
    UnsignedWord aKitten = 0;
    Word theBin = 0;
        
    tprintf2( &CurrentOutputPort, MiscStrings[ 207 ] );
    if( (m->space > m->free) || (m->free > m-> end) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 35 ] );
    else  {
        foundAnyLocksP = false;
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            length = lengthTaggedAvalPointer( (taggedAval *)tap );
            if( getLocationByte( ((taggedAval *)tap)->t) ) {
                foundAnyLocksP = TRUE;
                describeTaggedAvalPointer( (taggedAval *)tap );
                tprintf2( &CurrentOutputPort, MiscStrings[ 208 ] );
                releaseTagLock1Barrier( (volatile taggedAval *)tap, 7 );
                describeTaggedAvalPointer( (taggedAval *)tap );
                }         
            tap += length;
            }
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 209 ] );
        foundAnyLocksP = false;
        for( counter = 0; counter < SYMBOL_HASH_CONSTANT; ++counter )
            if( theClowder->obListBinLockBit( counter ) )
                theClowder->unlockObListBin( counter );
        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) 
            if( (theBin = theClowder->getKittenHasLockedObListBin( aKitten )) != -1 )
                theClowder->setKittenHasLockedObListBin( aKitten, -1 );
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 248 ] );
        foundAnyLocksP = false;
        for( counter = 0; counter < SYMBOL_HASH_CONSTANT; ++counter )
            if( theClowder->TLEBinLockBit( counter ) )
                theClowder->unlockTLEBin( counter );
        for( aKitten = 0; aKitten < ABSOLUTE_MAX_KITTENS; ++aKitten ) 
            if( (theBin = theClowder->getKittenHasLockedTLEBin( aKitten )) != -1 )
                theClowder->setKittenHasLockedTLEBin( aKitten, -1 );
        if( ! foundAnyLocksP )
            tprintf2( &CurrentOutputPort, MiscStrings[ 397 ] );
        }
    }

/**** printfMemorySpace -- Display what's in a memory space, possibly verbosely.
                           When this primitive prints out the contents of a memory
                           cell, it does not use Scheme print routines.  ****/

void printfMemorySpace( volatile memorySpace *m, Boolean showSpaceP )
{
    Byte *tap;
    UnsignedLongWord length;

    fprintf( stdout, MiscStrings[ 74 ], m );
    fprintf( stdout, MiscStrings[ 26 ], m->end - m->free, m->size, m->size );
    fprintf( stdout, MiscStrings[ 9 ], m->space, m->free, m->lowFree, m->end, m->inUse );
    if( (m->space > m->free) || (m->free > m-> end) )
        fputs( MiscStrings[ 35 ], stdout );
    else if( showSpaceP )
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            length = printfTaggedAvalPointer( (taggedAval *)tap );                        
            tap += length;
            }
    // else nothing
    }

/**** describeMemorySpaces -- Display what's in all memory spaces possibly verbosely.  ****/

void MemoryManager::describeMemorySpaces( Boolean showSpaceP )
{
    tprintf2( &CurrentOutputPort, MiscStrings[ 118 ] );
    describeMemorySpace( CurrentMainSpace, showSpaceP );
    if( showSpaceP ) {
        printMemorySpace( CurrentMainSpace, showSpaceP );
        tprintf2( &CurrentOutputPort, MiscStrings[ 117 ] );
        }
    tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
    }

/**** describeRoom -- Display how many bytes remain allocatable in the given space.  ****/

void describeRoom( volatile memorySpace *m )
{
//    tprintf5( &CurrentOutputPort, MiscStrings[ 58 ], m->size - (m->end - m->free), m->size, m->end - m->free );
    tprintf5( &CurrentOutputPort, MiscStrings[ 58 ], m->free - m->allocateMax, m->end - m->allocateMax, m->end - m->free );
    tprintf5( &CurrentOutputPort, MiscStrings[ 464 ],
              (m->agingGeneration)->free - (m->agingGeneration)->allocateMax,
              (m->agingGeneration)->end  - (m->agingGeneration)->allocateMax,
              (m->agingGeneration)->end  - (m->agingGeneration)->free );
    tprintf5( &CurrentOutputPort, MiscStrings[ 463 ],
              (m->activeGeneration)->free - (m->activeGeneration)->allocateMax,
              (m->activeGeneration)->end  - (m->activeGeneration)->allocateMax,
              (m->activeGeneration)->end  - (m->activeGeneration)->free );
    }

/**** allocateByteBlock -- Allocate a byte block containing the given number of bytes, with
                           that number installed in the header of the block.
                           Return an appropriately tagged pointer to the block, in A.

                           The "allocate" routine itself actually does all that, but it didn't use to.  ****/

void MemoryManager::allocateByteBlock( memoryBlockSize bytesWanted )
{
    allocate( sizeof(taggedAval) + bytesWanted );  
    }

/**** allocateStringBlock -- Allocate a string block containing the given string.
                             Return an appropriately tagged pointer to the block, in A.  ****/

void MemoryManager::allocateStringBlock( const Text *str )
{
    size_t sizeWanted = sizeof( Text ) * (strlen( str ) + 1);

    if( sizeWanted & (sizeof( taggedAval ) - 1) )
        sizeWanted = (sizeWanted & (~(sizeof( taggedAval ) - 1))) + sizeof( taggedAval );

    allocateByteBlock( sizeWanted );
    if( str != 0 )
        strcpy( (Text *)((Byte *)(A.u.c) + sizeof(taggedAval)), str );
    }

/**** allocateStringNBlock -- Allocate a string block with space for n characters.
                              Take care to null-terminate the space wherein the characters will go.
                              Return an appropriately tagged pointer to the block, in A.                ****/

void MemoryManager::allocateStringNBlock( size_t n )
{
    size_t sizeWanted = sizeof( Text ) * (n + 1);

    if( sizeWanted & (sizeof( taggedAval ) - 1) )
        sizeWanted = (sizeWanted & (~(sizeof( taggedAval ) - 1))) + sizeof( taggedAval );

    allocateByteBlock( sizeWanted );
    memset( (Byte *)(A.u.a) + sizeof(taggedAval), ' ', sizeWanted );
    *(Text *)((Byte *)(A.u.c) + sizeof(taggedAval) + n * sizeof( Text ) ) = '\0';
    }

/**** room -- Return how many bytes remain allocatable in the given space.  ****/

UnsignedLongWord room( volatile memorySpace *m ) { return m->free - m->allocateMax; }

/**** currentRoom -- Return how many bytes remain allocatable in the current mainspace.  ****/

UnsignedLongWord MemoryManager::currentRoom() { return( CurrentMainSpace->free - CurrentMainSpace->allocateMax ); }

/**** currentActiveRoom -- Return how many bytes remain allocatable in the current active generation.  ****/

UnsignedLongWord MemoryManager::currentActiveRoom()
{
    return( CurrentMainSpace->activeGeneration->free - CurrentMainSpace->activeGeneration->allocateMax );
    }

/**** currentRoom -- Return how many bytes remain allocatable in the current aging generation.  ****/

UnsignedLongWord MemoryManager::currentAgingRoom()
{
    return( CurrentMainSpace->agingGeneration->free - CurrentMainSpace->agingGeneration->allocateMax );
    }

/**** allocateDomainCurrentRoom -- Return how many bytes remain allocatable in the current allocate domain.  ****/

UnsignedLongWord MemoryManager::allocateDomainCurrentRoom()
{
    return(   CurrentMainSpace->allocateDomain->allocateSize
            - (CurrentMainSpace->allocateDomain->end - CurrentMainSpace->allocateDomain->free)
          );
    }

/**** currentMainSpaceCurrentRoom -- Return how many bytes remain allocatable in the current main space's allocate domain.  ****/

LongWord MemoryManager::currentMainSpaceCurrentRoom()
{
    return( CurrentMainSpace->free - CurrentMainSpace->allocateMax );
    }

/**** returnStoreSize -- Return how many bytes total in the given space.  ****/

UnsignedLongWord returnStoreSize( volatile memorySpace *m ) { return m->size; }

/**** currentStoreSize -- Return how many bytes total in the given space.  ****/

UnsignedLongWord MemoryManager::currentStoreSize() { return (CurrentMainSpace->end - CurrentMainSpace->allocateMax); }

/**** currentActiveStoreSize -- Return how many bytes total in the active space.  ****/

UnsignedLongWord MemoryManager::currentActiveStoreSize()
{
    return (CurrentMainSpace->activeGeneration->end - CurrentMainSpace->activeGeneration->allocateMax);
    }

/**** currentInUse -- Return how how many bytes are in use in the current main space. ****/

UnsignedLongWord MemoryManager::currentInUse() { return( CurrentMainSpace->end - CurrentMainSpace->free ); }

/**** currentActiveInUse -- Return how how many bytes are in use in the current active generational space. ****/

UnsignedLongWord MemoryManager::currentActiveInUse()
{ 
    return( CurrentMainSpace->activeGeneration->end - CurrentMainSpace->activeGeneration->free );
    }

/**** totalInUse -- Return how how many bytes are in use in tenured, active, and aging spaces.. ****/

UnsignedLongWord MemoryManager::totalInUse()
{
    return(    currentInUse() + currentActiveInUse()
             + CurrentMainSpace->agingGeneration->end - CurrentMainSpace->agingGeneration->free );
    }

/**** printMemorySpace  -- Display what's in a memory space, possibly verbosely.
                           When this primitive prints out the contents of a memory
                           cell, it uses non-Scheme print routines, which ought not to
                           fail even if the contents of the memory space are messed up. ****/

void printMemorySpace( volatile memorySpace *m, Boolean showSpaceP )
{
    Byte *tap;
    UnsignedLongWord length;
    taggedAval ta;
        
    tprintf3( &CurrentOutputPort, MiscStrings[ 75 ], m );
    tprintf6( &CurrentOutputPort, MiscStrings[ 10 ], m->space, m->size, m->free, m->lowFree );
    tprintf4( &CurrentOutputPort, MiscStrings[ 8 ], m->end, m->space + m->size );
    tprintf3( &CurrentOutputPort, MiscStrings[ 7 ], m->end - m->free );
    if( (m->space > m->free) || (m->free > m-> end) )
        tprintf2( &CurrentOutputPort, MiscStrings[ 35 ] );
    else if( showSpaceP )
        for( tap = (Byte *)m->free; tap < (Byte *)m->end; ) {
            ta = *((taggedAval *)tap);
            length = lengthTaggedAvalPointer( (taggedAval *)tap );
            tprintf3( &CurrentOutputPort, MiscStrings[ 55 ], tap );
            writeTaggedAvalSafelyNoScheme( &ta );
            tprintf2( &CurrentOutputPort, MiscStrings[ 115 ] );
            tap += length;
            }
    // else nothing
    }

/**** clearMemorySpace -- Set the "free" pointer to the start-allocation end of the given space.  ****/

void clearMemorySpace( volatile memorySpace *m )
{
    m->free = m->end;
    m->lowFree = m->space;
    m->inUse = FALSE;
    }
        
/**** swapSpaces -- Swap current and alternate main spaces. ****/

void MemoryManager::swapSpaces()
{
    volatile memorySpace *scratchSpace = AlternateMainSpace;

    AlternateMainSpace = CurrentMainSpace;
    CurrentMainSpace = scratchSpace;
    CurrentMainSpace->inUse = TRUE;
    AlternateMainSpace->inUse = FALSE;
    theClowder->initializeGGCAddressStacks( AlternateMainSpace );

    if( isMomCatP ) {
        clearGenerationSpace( CurrentMainSpace->activeGeneration );
        clearGenerationSpace( CurrentMainSpace->agingGeneration );
        clearGenerationSpace( AlternateMainSpace->activeGeneration );
        clearGenerationSpace( AlternateMainSpace->agingGeneration );
        }
    }

/**** swapGenerationSpaces -- Swap active and aging generation spaces in the given memory space. ****/

void MemoryManager::swapGenerationSpaces( volatile memorySpace *consHere )
{
    volatile memorySpace *scratchSpace = consHere->activeGeneration;
 
    consHere->activeGeneration = consHere->agingGeneration;
    consHere->agingGeneration = scratchSpace;

    consHere->allocateDomain = consHere->activeGeneration;
    clearGenerationSpace( consHere->activeGeneration );
    theClowder->swapGGCAddressStacks();
    }

/**** swapGenerationSpacesKitten -- Swap active and aging generation spaces in the given memory space. ****/

void MemoryManager::swapGenerationSpacesKitten()
{
    taggedAval scratch1 = theClowder->getScratch1();

    swapGenerationSpaces( (volatile memorySpace *)( scratch1.u.a ));
    }

/**** clearGenerationSpace -- Set the pointers and stuff of this space so it is empty.  ****/

void MemoryManager::clearGenerationSpace( volatile memorySpace *m )
{
    m->lowFree = m->space;
    m->free = m->end;
    }

/**** allocate -- The basic allocation primitive.  Allocate space for "bytesWanted" 
                  number of bytes in "space", and return a pointer thereto, in A
                  ****/

void MemoryManager::allocate( memoryBlockSize bytesWanted )
{
    while( TRUE ) {

        theClowder->blockOnAllocateLock();
        if( enoughMemoryToAllocate( bytesWanted ) ) {
            ALLOCATE_LOOP_GUTS( bytesWanted );
            theClowder->releaseAllocateLock();
            break;
            }
        theClowder->releaseAllocateLock();

        myClowder->setKittenRequestingGC( theKittenNumber );
        if( gcSequence( bytesWanted ) ) {
            theClowder->blockOnAllocateLock();
            if( (CurrentMainSpace->free - bytesWanted) >= CurrentMainSpace->allocateMax ) {
                TENURED_ALLOCATE_LOOP_GUTS( bytesWanted );
                theClowder->releaseAllocateLock();
                break;
                }
            theClowder->releaseAllocateLock();
            }
        }
    }

/**** allocateOne -- Allocate space for one taggedAval in "space", and return a pointer thereto, in A    ****/

void MemoryManager::allocateOne()
{
    while( TRUE ) {
        theClowder->blockOnAllocateLock();
        if( enoughMemoryToAllocate( sizeof( taggedAval ) ) ) {
            ALLOCATE_LOOP_GUTS( sizeof( taggedAval ) );
            theClowder->releaseAllocateLock();
            break;
            }
        theClowder->releaseAllocateLock();

        myClowder->setKittenRequestingGC( theKittenNumber );
        if( gcSequence( sizeof( taggedAval ) ) ) {
            theClowder->blockOnAllocateLock();
            if( (CurrentMainSpace->free - sizeof( taggedAval )) >= CurrentMainSpace->allocateMax ) {
                TENURED_ALLOCATE_LOOP_GUTS( sizeof( taggedAval ) );
                theClowder->releaseAllocateLock();
                break;
                }
            theClowder->releaseAllocateLock();
            }
        }
    }

/**** allocateTwo -- Allocate space for two taggedAvals in "space", and return a pointer thereto, in A    ****/

void MemoryManager::allocateTwo()
{
    while( TRUE ) {
        theClowder->blockOnAllocateLock();
        if( enoughMemoryToAllocate( 2 * sizeof( taggedAval ) ) ) {
            ALLOCATE_LOOP_GUTS( (2 * sizeof( taggedAval )) );
            theClowder->releaseAllocateLock();
            break;
            }
        theClowder->releaseAllocateLock();

        myClowder->setKittenRequestingGC( theKittenNumber );
        if( gcSequence( 2 * sizeof( taggedAval ) ) ) {
            theClowder->blockOnAllocateLock();
            if( (CurrentMainSpace->free - 2 * sizeof( taggedAval )) >= CurrentMainSpace->allocateMax ) {
                TENURED_ALLOCATE_LOOP_GUTS( 2 *sizeof( taggedAval ) );
                theClowder->releaseAllocateLock();
                break;
                }
            theClowder->releaseAllocateLock();
            }
        }
    }

/**** allocateWithoutInitialHeapLock -- A basic allocation primitive.  Allocate space for "bytesWanted" 
                                        number of bytes in "space", and return a pointer thereto, in A

                  DOES NOT INITIALLY LOCK THE HEAP.  Or unlock it.

                  ****/

void MemoryManager::allocateWithoutInitialHeapLock( memoryBlockSize bytesWanted )
{
    if( bytesWanted & 0x7 )
        bytesWanted = (bytesWanted & (~(0x7))) + 8;
    if( bytesWanted < sizeof( taggedAval ) )
        bytesWanted = sizeof( taggedAval );

    if( (CurrentMainSpace->allocateDomain->free - bytesWanted) >= CurrentMainSpace->allocateDomain->allocateMax ) {
        ALLOCATE_LOOP_GUTS( bytesWanted );
        }
    }

/**** allocateR -- CALLER MUST GUARANTEE bytesWanted is a multiple of taggedAval size.  ****/

void MemoryManager::allocateR()
{
    memoryBlockSize bytesWanted = (memoryBlockSize)(R.u.c);

    allocate( bytesWanted );

    R.u.a = A.u.a;
    R.t = forwardingPointerTag;
    }

/**** allocateVectorR ****/

void MemoryManager::allocateVectorR()
{
    memoryBlockSize bytesWanted = (memoryBlockSize)((R.u.c + 1) * sizeof( taggedAval ));

    UnsignedLongWord vLength = R.u.c;
    UnsignedLongWord counter = 0;

    allocate( bytesWanted );

    ((A.u.a)->u).c = vLength;
    if( vLength == 0 )
        (A.u.a)->t = exactShortFixnumBitsMask | carCdrNil;
    else {
        (A.u.a)->t = exactShortFixnumBitsMask | carCdrNext;

        for( counter = 1; counter < vLength ; ++counter )
            (A.u.a + counter)->t = emptyListTag | carCdrNext;

        (A.u.a + vLength)->t = emptyListTag | carCdrNil;
        }

    R.u.a = A.u.a;
    R.t = vectorTag;
    }

/**** fillMemorySpace -- fill the entire space with the (Byte) value given ****/

void fillMemorySpace( volatile memorySpace *m, Byte value )
{
    (void)memset( (void *)(m->space), (char)value, (long)(m->end - m->space) );
    }
        
/**** allocateLow -- Allocation primitive that hides knowledge of current space.  Uses no registers.  ****/ 

void MemoryManager::allocateLow( volatile memorySpace *m, memoryBlockSize bytesWanted )
{ 
    taggedAval *tp = 0;

    if( bytesWanted & 0x7 )
        bytesWanted = (bytesWanted & (~(0x7))) + 8;
    if( bytesWanted < sizeof( taggedAval ) )
        bytesWanted = sizeof( taggedAval );

    if( bytesWanted > (UnsignedLongWord)(m->end - m->lowFree) ) {
        useFatalErrorOutput();
#ifdef PIXIE_SCHEME_III

#endif
        tprintf3( &ErrorPort, MiscStrings[ 116 ], bytesWanted);
        fatalError( 1 );
        }

    tp = (taggedAval *)(m->lowFree);
    tp->t = byteBlockTag;
    (tp->u).a = (taggedAval *)(bytesWanted - sizeof( taggedAval ));
    m->lowFree += bytesWanted;
    }

/**** loadWorldGuts -- a pass-through to the world handler.  ****/

void MemoryManager::loadWorldGuts( fileDescriptor *wrldFile, Boolean describeP )
{
    myWorldHandler->loadWorldGuts( wrldFile, CurrentMainSpace, AlternateMainSpace, describeP );
    }

/**** worldIsUseable -- a pass-through to the world handler.  ****/

Boolean MemoryManager::worldIsUseable( Text *pathToFile ) { return myWorldHandler->worldIsUseable( pathToFile, CurrentMainSpace ); }

/**** getWorldHandler -- pass-through from the world handler. ****/

WorldHandler *MemoryManager::getWorldHandler() { return myWorldHandler; }

/**** initializeMemorySpace -- routine to initialize a (small) memory space, with no pre-allocated
                               block of memory to put it in.  This routine is used by the world-handler, in debugging.
                               It is probably obsolete.  (FIXME)  ****/

volatile memorySpace *MemoryManager::initializeMemorySpace( size_t bytesWanted )
{
    Byte *block;
    volatile memorySpace *m;
        
    if( bytesWanted & 0x7 )
        bytesWanted = (bytesWanted & (~(0x7))) + 8;

    if( bytesWanted > MAXIMUM_MEMORY_SIZE ) {
        useFatalErrorOutput();
        tprintf4( &ErrorPort, MiscStrings[ 312 ], bytesWanted, MAXIMUM_MEMORY_SIZE );
        fatalError( 52 );
        }
    block = (Byte *)calloc( sizeof(volatile memorySpace) + bytesWanted, sizeof(Byte) );
    if( ! block ) {
        useFatalErrorOutput();
        tprintf3( &ErrorPort, MiscStrings[ 116 ], bytesWanted);
        fatalError( 2 );
        }
    return initializeMemorySpaceGivenByteBlock( FALSE, &m, block, bytesWanted, 0 );
    }

// Not much to do in this function, we will have to set up the generational spaces
// repeatedly as they move around.

volatile memorySpace *MemoryManager::initializeGenerationalSpace( volatile memorySpace *m )
{
    memset( (void *)m, 0, sizeof( memorySpace ) );
    m->inUse = FALSE;
    return m;
    }

/**** consLuckOutOneTaggedAval -- can we cdr-code this item as part of a growing list? ****/

Boolean MemoryManager::consLuckOutOneTaggedAval( Byte *a )
{
    theClowder->blockOnAllocateLock();
    if(   (a == CurrentMainSpace->allocateDomain->free) 
       && (a < CurrentMainSpace->allocateDomain->end) 
       && (allocateDomainCurrentRoom() >= sizeof( taggedAval ) )
        ) {
        ALLOCATE_LOOP_GUTS( sizeof( taggedAval ) );
        theClowder->releaseAllocateLock();
        return TRUE;
        }
    else {
        theClowder->releaseAllocateLock();
        return FALSE;
        }
    }

// Wrappers for space-describing functions.

/**** describeCurrentMainSpace ****/

void MemoryManager::describeCurrentMainSpace() { describeMemorySpace( CurrentMainSpace, TRUE ); }

/**** describeCurrentActiveGeneration ****/

void MemoryManager::describeCurrentActiveGeneration() { describeMemorySpace( CurrentMainSpace->activeGeneration, TRUE ); }

/**** describeCurrentAgingGeneration ****/

void MemoryManager::describeCurrentAgingGeneration() { describeMemorySpace( CurrentMainSpace->agingGeneration, TRUE ); }

/**** describeCurrentStackSpace ****/

void MemoryManager::describeCurrentStackSpace() { describeMemorySpace( SchemeStack, TRUE ); }

/**** describeCurrentContinuationSpace ****/

void MemoryManager::describeCurrentContinuationSpace() { describeMemorySpace( SchemeContinuation, TRUE ); }

/**** describeLocksInCurrentMainSpace ****/

void MemoryManager::describeLocksInCurrentMainSpace() { describeLocksInMemorySpace( CurrentMainSpace ); }

/**** releaseLocksInCurrentMainSpace ****/

void MemoryManager::releaseLocksInCurrentMainSpace() { releaseLocksInMemorySpace( CurrentMainSpace ); }

/**** describeCurrentRoom ****/

void MemoryManager::describeCurrentRoom() { describeRoom( CurrentMainSpace ); }

/**** schemeStackAddress ****/

Byte* MemoryManager::schemeStackAddress() { return (Byte *)(SchemeStack->end - sizeof( taggedAval)); }

/**** schemeStackEnd ****/

Byte *MemoryManager::schemeStackEnd() { return (Byte *)(SchemeStack->end); }

/**** schemeStackSpace ****/

taggedAval *MemoryManager::schemeStackSpace() { return (taggedAval *)(SchemeStack->space); }

/**** schemeStackSize ****/

UnsignedLongWord MemoryManager::schemeStackSize() { return SchemeStack->size; }

/**** schemeContinuationAddress ****/

Byte* MemoryManager::schemeContinuationAddress() { return (Byte *)(SchemeContinuation->end - sizeof( taggedAval)); }

/**** schemeContinuationEnd ****/

Byte *MemoryManager::schemeContinuationEnd() { return (Byte *)(SchemeContinuation->end); }

/**** schemeContinuationSpace ****/

taggedAval *MemoryManager::schemeContinuationSpace() { return (taggedAval *)(SchemeContinuation->space); }

/**** schemeContinuationSize ****/

UnsignedLongWord MemoryManager::schemeContinuationSize() { return SchemeContinuation->size; }

/**** loadWorld ****/

void MemoryManager::loadWorld()
{
    myWorldHandler->loadWorld( CurrentMainSpace, AlternateMainSpace );
    }

/**** examineWorld ****/

 void MemoryManager::examineWorld( Boolean compareP, Text *fileName )
{
    myWorldHandler->examineWorld( compareP, fileName, CurrentMainSpace, AlternateMainSpace );
    }

/**** saveWorld ****/

void MemoryManager::saveWorld()
{
    myClowder->setFullGCIsRequired();
    myClowder->setGCIsRequired();
    triggerGC();
    myWorldHandler->saveWorld( CurrentMainSpace, AlternateMainSpace );
    }

/**** garbageCollectorGCCount ****/

UnsignedLongWord MemoryManager::garbageCollectorGCCount()
{
    return myGarbageCollector->garbageCollectorGCCount();
    }

/**** garbageCollectorGenerationalGCCount ****/

UnsignedLongWord MemoryManager::garbageCollectorGenerationalGCCount()
{
    return myGarbageCollector->garbageCollectorGenerationalGCCount();
    }

/**** checkSpaceParameters ****/

Boolean MemoryManager::checkSpaceParameters( volatile memorySpace *m, Word position )
{
    Boolean spaceOK = TRUE;

    if( m->free < m->space )
        spaceOK = FALSE;
    if( m->free > m->end )
        spaceOK = FALSE;
    if( m->size != (UnsignedLongWord)(m->end - m->space) )
        spaceOK = FALSE;
    if( (m->inUse != 0) && (m->inUse != 1) )
        spaceOK = FALSE;

    if( spaceOK )
        return TRUE;

    printf("Memory space problem at position %d:\n", position );
    printfMemorySpace( m, FALSE );

    return FALSE;
    }

/**** getStack ****/

volatile memorySpace *MemoryManager::getStack() { return SchemeStack; }

/**** getContinuation ****/

volatile memorySpace *MemoryManager::getContinuation() { return SchemeContinuation; }

/**** getScratchStack ****/

volatile memorySpace *MemoryManager::getScratchStack() { return ScratchStack; }

/**** getGGCScratchSpace ****/

volatile memorySpace *MemoryManager::getGGCScratchSpace() { return GGCScratchSpace; }

/**** get90PercentFlag -- is the memory space in use 90 percent or more full? ****/

Boolean MemoryManager::get90PercentFlag() { return( currentMainSpaceCurrentRoom() < (0.1 * currentStoreSize()) ); }

/**** waitForBytes ****/

memoryBlockSize waitForBytes()
{
    while( ! theClowder->gCBytesValid() )
        nanosleep( &GC_BYTES_VALID_DELAY_struct, 0 );
    return( (memoryBlockSize)(theClowder->getGCBytes()) );
    }

/**** triggerGC ****/

Boolean MemoryManager::triggerGC()
{
    memoryBlockSize bytesWanted = 0;
    Boolean didRebuildObList = FALSE;

    if( (theClowder->GCIsRequired()) && ((theClowder->getKittenBlock( theKittenNumber )) == 0) ) {

        if( theClowder->FullGCIsRequired() ) {
            myClowder->setKittenRequestingGC( theKittenNumber );
            if( isMomCatP ) {
                reportEvent( GC_START_EVENT_3 );
                eventLoopOnce( 0 );
                myGarbageCollector->standardFullGC( CurrentMainSpace, AlternateMainSpace, &didRebuildObList );
                reportEvent( GC_END_EVENT_5 );
                eventLoopOnce( 0 );
                }
            else {
                theClowder->kittenHandleGCSequenceMessages();
                }
            }
        else {
            if( isMomCatP ) {
                bytesWanted = waitForBytes();
                gcSequence( bytesWanted );
                }
            else {
                theClowder->kittenHandleGCSequenceMessages();
                }
            }
        return TRUE;
        }    
    else
        return FALSE;
    }

/**** triggerOnRebuildObList ****/

Boolean MemoryManager::triggerOnRebuildObList()
{
    if( theClowder->getRebuildObList() ) {
        theClowder->clearRebuildObList();
        theClowder->setFullGCIsRequired();
        theClowder->setGCIsRequired();
        theHeap->triggerGC();
        return TRUE;
        }        
    return FALSE;
    }

/**** cleanUpPorts ****/

void MemoryManager::cleanUpPorts()
{
    taggedAval scratch1 = theClowder->getScratch1();
    taggedAval scratch2 = theClowder->getScratch2();

    myGarbageCollector->cleanUpPorts( (memorySpace *)(scratch1.u.a), (memorySpace *)(scratch2.u.a) );
    }

/**** padForFullGC -- insert garbage byte blocks so full GC can start at generationA->free. ****/

void MemoryManager::padForFullGC( volatile memorySpace *m )
{
    taggedAval *scratchTP = 0;

    if( m->allocateDomain != m ) {

        scratchTP = (taggedAval *)(m->generationA->end);
        scratchTP->t = byteBlockTag;
        (scratchTP->u).c = m->generationB->free - m->generationA->end - sizeof( taggedAval );

        scratchTP = (taggedAval *)(m->generationB->end);
        scratchTP->t = byteBlockTag;
        (scratchTP->u).c = m->free - m->generationB->end - sizeof( taggedAval );

        }

    }

void MemoryManager::setSpaceForFullGC( volatile memorySpace *m )
{
    if( m->allocateDomain != m ) {
        m->allocateMax = m->space;
        m->free = m->generationA->free;
        }
    }

void MemoryManager::setSpaceForGenerationalGC( volatile memorySpace *m )
{
    if( m->allocateDomain != m )
        m->allocateMax = m->generationB->end + allocateMaxOffset;
    }

void MemoryManager::revertToFullGC()
{
    (CurrentMainSpace)->allocateDomain = CurrentMainSpace;
    (CurrentMainSpace)->allocateMax = (CurrentMainSpace)->space;
    (CurrentMainSpace)->allocateSize = (CurrentMainSpace)->size;

    (AlternateMainSpace)->allocateDomain = AlternateMainSpace;
    (AlternateMainSpace)->allocateMax = (AlternateMainSpace)->space;
    (AlternateMainSpace)->allocateSize = (AlternateMainSpace)->size;
    }

UnsignedLongWord MemoryManager::standardGenerationalReserve() 
{
   return( 2 * theClowder->getObListSize() + 2 * generationSpaceSize + pressureBufferSize );
   }

UnsignedLongWord MemoryManager::smallGenerationalReserve()
{
   return( 2 * theClowder->getObListSize() + 2 * generationSpaceSize + smallPressureBufferSize );
   }

UnsignedLongWord MemoryManager::standardFullReserve()
{
   return( 2 * theClowder->getObListSize()                           + pressureBufferSize );
   }

UnsignedLongWord MemoryManager::standardLargeFullReserve()
{
   return( 2 * theClowder->getObListSize()                           + pressureBufferSize + smallPressureBufferSize );
   }

Boolean MemoryManager::enoughMemoryToAllocate( memoryBlockSize bytesWanted )
{
#ifdef PIXIE_SCHEME_III

#endif
    if( usingGenerationalGC )
        return( (CurrentMainSpace->allocateDomain->free - bytesWanted) >= CurrentMainSpace->allocateDomain->allocateMax );
    else
        return( (CurrentMainSpace->free - standardFullReserve() - bytesWanted) >=  CurrentMainSpace->allocateMax );
    }

Boolean MemoryManager::setReturn( Boolean returnValue, LongWord errorNumber,
                                  Boolean didRebuildObList, Boolean neededToRebuildObList, memoryBlockSize bytesWanted,
                                  memoryBlockSize previousFree )
{
    UnsignedWord failingGCKitty = 0;
    UnsignedWord aKitty = 0;

    if( errorNumber >= 0 ) {
        theClowder->setGCErrorNumber( errorNumber );

        // Remember that the kitten running this procedure is necessarily the MomCat, and must reset itself somehow.

        if( (failingGCKitty = theClowder->getKittenRequestingGC()) != 0 ) {
            theClowder->momDialogWithKitten( failingGCKitty, MOM_SAYS_GC_HAS_FAILED, KITTEN_ACKNOWLEDGES_GC_HAS_FAILED );
            theClowder->setMomSaysToKitten( MOM_SAYS_NOTHING, failingGCKitty );
            for( aKitty = 1; aKitty < theNumberOfKittens; ++aKitty )
                if( aKitty != failingGCKitty )
                    theClowder->setMomSaysToKitten( MOM_SAYS_RESET_FOR_LOW_MEMORY, aKitty );
//                    theClowder->momDialogWithKitten( failingGCKitty,
//                                                     MOM_SAYS_RESET_FOR_LOW_MEMORY, KITTEN_ACKNOWLEDGES_RESET_FOR_LOW_MEMORY );
            reportEvent( GC_END_EVENT_6 );
            eventLoopOnce( 0 );
            useErrorOutput();
            tprintf3( &CurrentOutputPort, MiscStrings[ 141 ], failingGCKitty );
            resetError( 327 );           // MomCat resets, reporting that another kitten has found low memory.
            }
        else {
            for( aKitty = 1; aKitty < theNumberOfKittens; ++aKitty )
                theClowder->setMomSaysToKitten( MOM_SAYS_RESET_FOR_LOW_MEMORY, aKitty );
//                theClowder->momDialogWithKitten( failingGCKitty,
//                                                 MOM_SAYS_RESET_FOR_LOW_MEMORY, KITTEN_ACKNOWLEDGES_RESET_FOR_LOW_MEMORY );
            reportEvent( GC_END_EVENT_6 );
            eventLoopOnce( 0 );
            useErrorOutput();
            resetError( errorNumber );   // MomCat has found low memory, and resets with an appropriate report.
            }        
        }

    if( neededToRebuildObList & (! didRebuildObList ) )
        theClowder->setRebuildObList();

    reportEvent( GC_END_EVENT_5 );
    eventLoopOnce( 0 );
    return returnValue;
    }

Boolean MemoryManager::gcSequence( memoryBlockSize bytesWanted )
{
    Boolean needToRebuildObList = FALSE;
    Boolean didRebuildObList = FALSE;

    memoryBlockSize previousFree = CurrentMainSpace->free - CurrentMainSpace->allocateMax;

    myClowder->setGCIsRequired();

    theClowder->setGCBytes( bytesWanted );
    theClowder->setGCBytesValid();          // Remember to clear this after kittens are all done.  FIXME

    if( ! isMomCatP )
        return( theClowder->kittenHandleGCSequenceMessages() );

    reportEvent( GC_START_EVENT_3 );
    eventLoopOnce( 0 );

    if( usingGenerationalGC ) {

        //  So, do a generational, if possible:

        if( (CurrentMainSpace->free - generationSpaceSize) > CurrentMainSpace->allocateMax ) {
            myGarbageCollector->standardGenerationalGC( CurrentMainSpace, AlternateMainSpace );
            }
        else   // Catastrophe, the generational garbage-collector has failed 
            fatalError( 186 );
       
        //  Now see if there is pressure on tenured space, and if so, do a full garbage collection.

        if(    (CurrentMainSpace->free - standardGenerationalReserve() -  2 * generationSpaceSize - smallPressureBufferSize)
            <= CurrentMainSpace->allocateMax ) {

            needToRebuildObList = TRUE;

            didRebuildObList = FALSE;
            myGarbageCollector->standardFullGC( CurrentMainSpace, AlternateMainSpace, &didRebuildObList );

            //  If there is still pressure, try another garbage collection and hope for the best ...

            if( (CurrentMainSpace->free - standardGenerationalReserve()) <= CurrentMainSpace->allocateMax ) {
                didRebuildObList = FALSE;
                myGarbageCollector->standardFullGC( CurrentMainSpace, AlternateMainSpace, &didRebuildObList );
                }

            //  If there is still pressure, report out of memory,

            if( (CurrentMainSpace->free - standardGenerationalReserve()) <= CurrentMainSpace->allocateMax )
                return setReturn( FALSE, 54, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

            //  If there is enough generational space for the task (and generational space is now empty), return FALSE.

            if( (CurrentMainSpace->allocateDomain->free - bytesWanted) >= CurrentMainSpace->allocateDomain->allocateMax )
                return setReturn(FALSE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

            //  Or, if there is enough tenured space to allocate there, return TRUE.

            else if( (CurrentMainSpace->free - standardGenerationalReserve() - bytesWanted) >= CurrentMainSpace->allocateMax )
                return setReturn( TRUE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

            //  Else fail.

            else
                return setReturn( FALSE, 329, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );
            }

        //  There was no pressure on tenured space, so we have not (yet) done a full GC.
        //  Did the previous generational GC free up enough space?

        if( (CurrentMainSpace->allocateDomain->free - bytesWanted) >= CurrentMainSpace->allocateDomain->allocateMax )
            return setReturn( FALSE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

        //  If not, try another generational (which will leave both active and aging spaces filled, which
        //  we will need if we have to do a full GC in a moment).

        myGarbageCollector->standardGenerationalGC( CurrentMainSpace, AlternateMainSpace );

        if( (CurrentMainSpace->allocateDomain->free - bytesWanted) >= CurrentMainSpace->allocateDomain->allocateMax )
            return setReturn( FALSE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

        //  If that did not work, then it may be possible to allocate at the top of tenured space:  For that
        //  to work, there must be sufficient space there for the proposed allocation, PLUS space for one
        //  more generational GC, plus enough extra space that the system gets many chances to GC before it
        //  gives up.  (This last is because the GCs that free the most space will be those in which the
        //  oblist is not "in register", and that won't happen all the time, only stochastically.)  The
        //  "enough extra" is set at one generationSpaceSize.  NOTE that we have just done two generational GCs,
        //  So that generational space should be empty, so that anything that gets loaded into the newly allocated
        //  block cannnot contain any pointers to active or aging space.

        //  So do a full GC, if possible, and look again.

        if( (CurrentMainSpace->free - standardGenerationalReserve()) > CurrentMainSpace->allocateMax ) {
            needToRebuildObList = TRUE;
            didRebuildObList = FALSE;
            myGarbageCollector->standardFullGC( CurrentMainSpace, AlternateMainSpace, &didRebuildObList );
            }
        else    // Report out of memory.
            return setReturn( FALSE, 330, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

        //  If there is pressure on tenured space after the full GC, report out of memory.

        if( (CurrentMainSpace->free - standardGenerationalReserve()) <= CurrentMainSpace->allocateMax )
            return setReturn( FALSE, 331, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

        //  If there is enough tenured space now, return TRUE.

        if( (CurrentMainSpace->free - standardGenerationalReserve() - bytesWanted) >= CurrentMainSpace->allocateMax )
            return setReturn( TRUE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );

        //  All possibilities are exhausted:  Report out of memory.

        return setReturn( FALSE, 332, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );
        }

    else {  //  Using full GC.  Note that the test whereby this routine is called will have guaranteed
            //  that sufficient space remains to rebuild the Oblist.

        didRebuildObList = FALSE;
        myGarbageCollector->standardFullGC( CurrentMainSpace, AlternateMainSpace, &didRebuildObList );

        if( (CurrentMainSpace->free - standardLargeFullReserve() - bytesWanted) < CurrentMainSpace->allocateMax )
            return setReturn( FALSE, 1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );
        else if( (CurrentMainSpace->free - standardFullReserve() - bytesWanted) >= CurrentMainSpace->allocateMax ) {
            needToRebuildObList = TRUE;
            return setReturn( TRUE, -1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );
            }
        else
            return setReturn( FALSE, 1, didRebuildObList, needToRebuildObList, bytesWanted, previousFree );
        }
    }
