/****************************************************************************************

NumberFormats.c -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006

Input reader numeric parsing functions for Pixie.

        "A computer scientist is a person who refuses to believe that 1.0 is an integer."

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "Numbers.h"
#import "OpCodes.h"
#import "SECD.h"

#import "DataTypes.e"
#import "Globals.e"
#import "IO.e"

// LongWord maxFixDigits[17] = { 0, 0, 32, 0, 0, 0, 0, 0, 11, 0, 10, 0, 0, 0, 0, 0, 8 };
LongWord maxLongDoubleDigits[17] = {  0, 0, 65, 0, 0, 0, 0, 0, 22, 0, 19, 0, 0, 0, 0, 0, 17 };  // Maybe
LongWord maxFixDigits[17] = { 0, 0, 64, 0, 0, 0, 0, 0, 21, 0, 20, 0, 0, 0, 0, 0, 16 };

#ifdef FLOATS_ARE_64_BIT

UnsignedLongWord maxShortFloatDigits[17] = {  0, 0, 53, 0, 0, 0, 0, 0, 18, 0, 16, 0, 0, 0, 0, 0, 14 };  // ShortFloat <--> double

#else  // *not* FLOATS_ARE_64_BIT

LongWord maxShortFloatDigits[17] = {  0, 0, 65, 0, 0, 0, 0, 0, 22, 0, 19, 0, 0, 0, 0, 0, 17 };  // ShortFloat <--> IEEE 80-bit

#endif  // FLOATS_ARE_64_BIT

/******** ExernalFlag: ********/

Boolean useFullNumericPrecision = true;

/******** Statics: ********/

// static Text fixedDigits[] = "9876543210";

/******** forward declarations ********/

void formatGuts( LongWord radix, Text *buffer, Boolean fullPrecisionP );
Text *fmtInt( LongWord radix, Boolean fullPrecisionP, Boolean exactP ); 
Text *fmtFix( LongWord radix, Boolean fullPrecisionP, Boolean exactP ); 
Text *fmtSci( LongWord radix, Boolean fullPrecisionP, Boolean exactP ); 
Text *xecvt( long double x, LongWord ndig, LongWord *rdxpt, Word *sign, LongWord radix );
Text *xdcvt( LongWord N, LongWord ndig, LongWord *rdxpt, Word *sign, LongWord radix );

/******** Functions: ********/

/**** getDigits -- Copy just digits from one buffer to another.  This
                   auxiliary for "jcvt" assumes the "from" buffer has
                   a number which ends in an "e"-form exponent.          ****/

Text *getDigits( Text *from, Text *putHere )
{
    Text *answer = putHere;

    while( *from && (*from != 'e') ) {
        if( isdigit( *from ) )
            *putHere++ = *from;
        ++from;
        }
    *putHere = 0;
    return answer;
    }

/**** getNDigLeft -- Auxiliary for numeric formatting.

      If given is 0.0, return 1, otherwise return max( 0, ceiling(logToBaseRadix(|given|)) ).

      The result is the number of digits that would be printed to the left of the radix point
      if the number were printed in "fix" fmt with arbitrary precision. ( Zero -> "0. ...".)
                                        
      ****/
                                
LongWord getNDigLeft( long double given, LongWord radix ) // DANGER WILL ROBINSON
{
    long double theLogarithm;

    if( given == (long double)0.0 )
        return 1;

    if( given < 0.0 )
        given = - given;

    if( radix == 10 )
        theLogarithm = log10l( given );
    else if( radix == 2 )
        theLogarithm = log2l( given );
    else if( radix == 8 )
        theLogarithm = log2l( given ) / 3.0;
    else /* radix == 16 */
        theLogarithm = log2l( given ) / 4.0;

    if( theLogarithm < 0.0 )
        return 0;
    else
        return lroundl( ceill( theLogarithm ) );
    }

/**** getRadixPoint -- Auxiliary for numeric formatting.

    Consider the absolute value of the number:

      For zero, return 1.
      For a number greater that one, return the number of digits left of the radix point.
      For a number between zero and one. return minus one times the number of zeros
         between the radix point and the first non-zero character of the number.
                                   
    ****/
                                
LongWord getRadixPoint( long double given, LongWord radix ) // DANGER WILL ROBINSON
{
    long double theLogarithm;

    if( given == (long double)0.0 )
        return 1;

    if( given < 0.0 )
        given = - given;

    if( radix == 10 )
        theLogarithm = log10l( given );
    else if( radix == 2 )
        theLogarithm = log2l( given );
    else if( radix == 8 )
        theLogarithm = log2l( given ) / 3.0;
    else /* radix == 16 */
        theLogarithm = log2l( given ) / 4.0;

    return lroundl( ceill( theLogarithm ) );
    }

/**** getRequiredDigitsShortFloat ****/

LongWord getRequiredDigitsShortFloat( long double given, LongWord radix )
{
    long double x;
    long double logMaxBit;
    long double logMinBit;
    long double minBitDelta;
    long double logMaxRadix;
    long double logMinRadix;

    if( given == 0.0L )  // Special case.
        return 1;

    x = fabsl( given );
    logMaxBit = floorl( log2l( x ) );                      // Bit position of MSB of x.
    logMinBit = logMaxBit - maxShortFloatDigits[ 2 ] + 1;  // Bit position of LSB of x.
    minBitDelta = powl( 2.0L, logMinBit );                 // Delta x associated with changing the LSB of x.
    logMaxRadix = 0.0L;                                    // Radix position of MS digit of x.
    logMinRadix = 0.0L;                                    // To be determined ...

    //  Objective:  Choose logMinRadix such that a change of 1 * (radix ** logMinRadix) does
    //              not exceed minBitDelta.  Then the number of digits required is
    //              logMaxRadix - logMinRadix + 1.
    //
    //                   radix ** logMinRadix <= minBitDelta
    //              ==>  logMinRadix <= log-to-base-radix( minbitDelta )

    switch( radix ) {

        case  2:
            return maxShortFloatDigits[ 2 ];
            break;
        
        case  8:
            logMaxRadix = floorl( log2l( x ) / 3.0L );
            logMinRadix = floorl( log2l( minBitDelta ) / 3.0L );
            return lroundl( logMaxRadix - logMinRadix + 1.0L ) + 1;
            break;

        case 10:
            logMaxRadix = floorl( log10l( x ) );
            logMinRadix = floorl( log10l( minBitDelta ) );
            return lroundl( logMaxRadix - logMinRadix + 1.0L );
            break;

        case 16:
            return maxShortFloatDigits[ 16 ];  //  NOTE -- won't work if four does not evenly divide maxShort`<FloatDigits[ 16 ]
            break;
        }
   resetError( 257 );
   return 0;
   }

/**** parseExpt ****/

Word parseExpt( Text *str )
{
    while( *str++ != 'e')
        ;
    return atoi( str );
    }

/**** fmtInt - Format the number in R as an "int", according to the passed information. ****/

Text *fmtInt( LongWord radix, Boolean fullPrecisionP, Boolean exactP )
{
    static Text myBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    Text *buffer = 0;
    Text *scan = myBuffer;
    long double value = 0.0;
    Word sign = 0;
    LongWord counter = 0;
    LongWord nDigMax;
    LongWord N = 0;
    LongWord rdxpt = 0;

    myBuffer[ 0 ] = 0;  // Defensive.

    INTEGER_P
    if( ! B ) {
        OUTPUT_TO_ERROR
        DISPLAY
        resetError( 221 );
        }

    FIXNUM_P
    if( B ) {
//        debug("Fixnum");
        N = (LongWord)(R.u.c);
        value = (long double)N;
        }
    else {
//        debug("Not fixnum");
//        debug("Long float");
        value = R.u.f;
        }
    FIXNUM_P                
    if( B )
        nDigMax = maxFixDigits[ radix ] + 1;
    else
        nDigMax = getRequiredDigitsShortFloat( value, radix );
    if( value == 0.0 ) {
        if( (radix != 10) && (! exactP) && fullPrecisionP ) {
            *scan++ = SHARP;
            *scan++ = INEXACT;
            }
        *scan++ = getDigitChar( 0 );
        if( (radix == 10) && (! exactP) && fullPrecisionP )
            *scan++ = RADIX_POINT;
        *scan = 0;
        return myBuffer;
        }     
    FIXNUM_P
    if( B ) {
        buffer = xdcvt( N, nDigMax, &rdxpt, &sign, radix );
        }
    else
        buffer = xecvt( value, nDigMax, &rdxpt, &sign, radix );
    if( sign == 1 ) {
        *scan++ = MINUS_SIGN;
        *scan = 0;
        }
    if( (radix != 10) && (! exactP) && fullPrecisionP ) {
        *scan++ = SHARP;
        *scan++ = INEXACT;
        *scan = 0;
        }        

    // Else the radix is ten and we will use a decimal point -- and maybe a "q0" -- to indicate exactness.

    strncpy( scan, buffer, sizeof( myBuffer ) - strlen( myBuffer ) );
    scan = myBuffer + strlen( myBuffer );
    if( rdxpt > nDigMax )  { // Pad at right with zeros.
        for( counter = nDigMax; counter < rdxpt; ++counter )
        *scan++ = getDigitChar( 0 );
        *scan = 0;
        }
    else {
       scan -= (nDigMax - rdxpt);
       *scan = 0;
       }
    if( (scan < myBuffer) || (scan > myBuffer + sizeof( myBuffer )) ) {
       fatalError( 54 );
       }
    if( (radix == 10) && (! exactP) && fullPrecisionP ) {
        *scan++ = RADIX_POINT;
        *scan = 0;
        }      

    return myBuffer;
    }

/**** fmtFix - Format the number in R as a "fix", according to the passed information. ****/

Text *fmtFix( LongWord radix, Boolean fullPrecisionP, Boolean exactP )
{
    static Text myBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    Text *buffer = 0;
    long double value = 0.0;
    Text *scan = myBuffer;
    LongWord nDigMax;
    LongWord rdxpt = 0;
    LongWord counter = 0;
    Word sign = 0;
    LongWord lnth = 0;

//    debug( "In fmtFix");

    myBuffer[ 0 ] = 0;
    nDigMax = 0;

    INTEGER_P
    if( B ) {
//        debug( "fmtFix about to call fmtInt");
        return fmtInt( radix, fullPrecisionP, exactP );
        }
    value = R.u.f;
    if( fullPrecisionP )
        nDigMax = getRequiredDigitsShortFloat( value, radix );
    else
        nDigMax = HEURISTIC_DIGITS;

    buffer = xecvt( value, nDigMax, &rdxpt, &sign, radix );
    lnth = strlen( buffer );
//    tprintf4( stderr, "\nIn fmtFix, xecvt returned %s, rdxpt %d\n", buffer, rdxpt ); // debug

    // Put in a minus sign, if required.

    if( sign == 1 )
        *scan++ = MINUS_SIGN;        
    if( (radix != 10) ) {
        if( (! exactP) && fullPrecisionP ) {
            *scan++ = SHARP;
            *scan++ = INEXACT;
            *scan = 0;
            }        
        }        
    else {    // radix is 10 
        if( (rdxpt < 0) || (rdxpt < lnth) ) {      // Going to need a decimal point.
            if( exactP && fullPrecisionP ) {       // So we need an exact prefix for an exact number.
                *scan++ = SHARP;
                *scan++ = EXACT;
                *scan = 0;
                }        
            }
        else {                                     // Not going to need a decimal point.
            if( (! exactP) && fullPrecisionP ) {   // So we need an	inexact prefix for an inexact number.
                *scan++ = SHARP;
                *scan++ = INEXACT;
                *scan = 0;
                }        
            }
        }        

    if( rdxpt <= 0 ) {          
        if( radix != 10 )
            resetError( 260 );
        *scan++ =  getDigitChar( 0 );                        // Need a zero before the decimal point ...
        *scan++ = RADIX_POINT;                               //     ... and then the decimal point ...
        for( counter = rdxpt; counter < 0; ++counter )       //     ... and maybe some zeros after the decimal point ...
           *scan++ = getDigitChar( 0 );
        *scan = 0;
        strncpy( scan, buffer,                               //     ... and at last the digits.
                 sizeof( myBuffer ) - strlen( myBuffer) );
        scan = myBuffer + strlen( myBuffer );
        }
    else {
        for( counter = 0; counter < rdxpt; ++counter ) {     // Some digits precede the decimal point ...
            if( counter < lnth )
                *scan++ = buffer[ counter ];                 //     ... either copy them in ...
            else
                *scan++ = getDigitChar( 0 );                 //     ... or pad with zeros ...
            }
        *scan = 0;
        if( radix == 10 ) {
            *scan++ = RADIX_POINT;                           //     ... and then put in the decimal point ...
            lnth = strlen( buffer );
            for( ; counter < lnth; ++counter )
                *scan++ = buffer[ counter ];                 //     ... and then any remaining digits.
            *scan = 0;
             }
        else {
            if( counter < (Word)(strlen( buffer )) )
                resetError( 260 );
            }
        }

    if( (scan < myBuffer) || (scan > myBuffer + sizeof( myBuffer )) )
       fatalError( 55 );

    if( radix == 10 ) {
        while( (scan >= myBuffer) && (*( scan - 1 ) == getDigitChar( 0 )) )          // Strip any trailing zeros.
            --scan;
        *scan = 0;
        }

//    tprintf7( stdout, "\nfmtFix gets %s, nDigMax %ld, sign %d, radix %ld, rdxpt %dxs\n",
//              myBuffer, nDigMax,  sign, radix, rdxpt ); // debug

    return myBuffer;
    }

/**** fmtSci - Format the number in R as a "sci", according to the passed information. ****/

Text *fmtSci( LongWord radix, Boolean fullPrecisionP, Boolean exactP )
{
    static Text myBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    long double value = 0.0;
    Text *buffer = 0;
    LongWord rdxpt = 0;
    Word sign = 0;
    Text *scan = myBuffer;
    UnsignedLongWord counter = 0;
    UnsignedLongWord exponentToPrint = 0;
    LongWord nSigDigits = 0;
    LongWord N = 0;

    myBuffer[ 0 ] = 0;
    if( radix != 10 )
        resetError( 261 );
    FIXNUM_P
    if( B ) {
        N = R.u.c;
        value = (long double)N;
        nSigDigits = maxFixDigits[ radix ];
        }
    else {
        value = R.u.f;
        if( fullPrecisionP )
            nSigDigits = getRequiredDigitsShortFloat( value, radix );
        else
            nSigDigits = HEURISTIC_DIGITS;
        }

    buffer = xecvt( value, nSigDigits, &rdxpt, &sign, radix );
    if( sign == 1 ) {
        *scan++ = MINUS_SIGN;
        *scan = 0;
        }

    // In sci format, we are always in base 10 and we always print a decimal point ...

    if( exactP && fullPrecisionP ) {   // .. so we need an exact prefix for an exact number.
        *scan++ = SHARP;
        *scan++ = EXACT;
        *scan = 0;
        }        

    for( counter = 0; counter < 1; ++counter )                             // Lead digit.
        *scan++ =  buffer[ counter ];
    *scan++ = RADIX_POINT;                                                 // Radix point.
    for( ; counter < (UnsignedLongWord)nSigDigits; ++counter )                               // All the rest of the digits.
        *scan++ =  buffer[ counter ];
    if( (scan < myBuffer) || (scan > myBuffer + sizeof( myBuffer )) )
       fatalError( 56 );
    while( (scan >= myBuffer) && (*( scan - 1 ) == getDigitChar( 0 )) )    // Strip trailing zeros.
        --scan;
    *scan = 0;
    if( (exponentToPrint = rdxpt - 1) != 0 ) {                             // And (perhaps) the exponent.
        *scan++ = EXPONENT;
        sprintf( scan, "%jd", (intmax_t)exponentToPrint );
        }
    return myBuffer;
    }

/**** jcvt -- This function is a major auxilary for printing floating-point numbers with
              various amounts of precision.  x is a ShortFloat, ndig is the number of
              digits desired to print (not counting decimal points, signs, exponents
              and so on.  The function returns a pointer to a static buffer containing
              those digits -- and *just* those digits, as determined by using snprintf
              and string-manipulation facilities.  It also returns, via a passed pointer,
              the position in the returned string at which one would have to put a
              decimal point to show the number correctly, assuming that the string
              is padded with zeros as necessary.   For this value, zero means "right before
              the string", a positive number N meaning "right before the Nth character
              (zero-based) of the string, and a negative number -N meaning "right before
              the Nth padded zero added to the left of the string".  The function also
              returns, via a passed pointer, an indication of whether the number is
              positive (value set to zero) or negative (set to one).

              Phew!

              ****/

Text *jcvt( long double x, LongWord ndig, LongWord *decPt, Word *sign )
{
    static Text rawBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text returnBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text formatString[ 16 ];
 
    LongWord expt = 0;

    if( ndig > (LongWord)MAX_OUTPUT_NUMBER_SIZE ) {
        resetError( 222 );
        }

    *decPt = 0;
    *sign = 0;
    *rawBuffer = '\0';
    *returnBuffer = '\0';

#ifdef FLOATS_ARE_64_BIT

    sprintf( formatString, "%%+#.%jdLe", (intmax_t)(ndig - 1) );                  // "%+#.<ndig - 1>Le"

#else

    sprintf( formatString, "%%+#.%jdLe", (intmax_t)(ndig - 1) );                  // "%+#.<ndig - 1>Le"

#endif

    snprintf( rawBuffer, MAX_OUTPUT_NUMBER_SIZE, formatString, x );   // Format +d.ddddde(+/-)dd, ndig d's before the exponent.

    if( *rawBuffer == '+' )
        *sign = 0;
    else
        *sign = 1;

    expt = parseExpt( rawBuffer );
    *decPt = expt + 1;

    getDigits( rawBuffer, returnBuffer );

//    tprintf8( stdout, "\njcvt returning %s (from %s), format %s, ndig %ld, decPt %d, sign %d\n",  returnBuffer, rawBuffer, formatString, ndig, *decPt, *sign ); // debug

    return returnBuffer;
    }
    
/**** octalCharForN ****/

Text octalCharForN( LongWord N ) { return OCTAL_DIGITS[ 7 - N ]; }

/**** threeBitsToOctalChar ****/

Text threeBitsToOctalChar( Text bit2, Text bit1, Text bit0 )
{
    Word fourO = (bit2 == '1' ? 4 : 0);
    Word twoO  = (bit1 == '1' ? 2 : 0);
    Word oneO  = (bit0 == '1' ? 1 : 0);
    Word sum = fourO + twoO + oneO;

    return OCTAL_DIGITS[ 7 - sum ];
    }

/**** hexCharForN ****/

Text hexCharForN( LongWord N ) { return HEX_DIGITS[ 7 - N ]; }

/**** fourBitsToHexChar ****/

Text fourBitsToHexChar( Text bit3, Text bit2, Text bit1, Text bit0 )
{
    Word eightX = (bit3 == '1' ? 8 : 0);
    Word fourX  = (bit2 == '1' ? 4 : 0);
    Word twoX   = (bit1 == '1' ? 2 : 0);
    Word oneX   = (bit0 == '1' ? 1 : 0);
    Word sum = eightX + fourX + twoX + oneX;

    return HEX_DIGITS[ 15 - sum ];
    }

/**** xecvt -- This basic primitive works like "ecvt" except that it accepts an additional
               argument; namely, the radix of the base in which the printing is to be done. ****/
                          
Text *xecvt( long double x, LongWord ndig, LongWord *rdxpt, Word *sign, LongWord radix )
{
    static Text buffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text buffer2[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text buffer3[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    long double shiftedX;
    LongWord basePowerOfTwo = 1;
    Text *back, *scratchPtr;
    Word expTwo;
    long double scratchFloat;
    LongWord rightEndOfUnits = 0;
    LongWord counter = 0;
    Word digitsPrinted = 0;

    *rdxpt = 0;
    *sign = 0;
    *buffer = 0;
    *buffer2 = 0;
    *buffer3 = 0;

    if( ndig > (LongWord)MAX_OUTPUT_NUMBER_SIZE ) {
//        tprintf4( &ErrPort, "In xecvt with %#+.31Le, ndig is %ld\n", x, ndig ); // debug
        resetError( 222 );
        }

    if( x == ((long double)0.0) ) {                   // Zero is a special case
        back = buffer;
        while( ndig-- > 0 )
            *back++ = getDigitChar( 0 );
        *back = '\0';
        return buffer;
        }
    switch( radix ) {
        case 10:  return jcvt( x, ndig, rdxpt, sign ); break;
        case  2:  basePowerOfTwo = 1; break;
        case  8:  basePowerOfTwo = 3; break;
        case 16:  basePowerOfTwo = 4; break;
        default:  resetError( 266 );  break;
        }            
    if( x > 0.0 )
        *sign = 0;
    else
        *sign = 1;

    shiftedX = frexpl( fabsl( x ), &expTwo );

    if( expTwo < 0 )  //  This code won't work -- Scheme only deals with non-decimal numbers which are integers.
        resetError( 260 );

    // shiftedX lies in range [0.5, 1).  Let's write it out in binary.  We'll peel bits off the front of
    // the returned value, doubling it as we go to get the next one out left of the radix point.

    *buffer2 = 0;
    scratchPtr = buffer2;
    scratchFloat = 2.0 * shiftedX;

    while( scratchFloat != 0.0 ) {
        if( scratchFloat >= 1.0 )
            *scratchPtr++ = '1';
        else
            *scratchPtr++ = '0';
        scratchFloat = scratchFloat - truncl( scratchFloat );
        scratchFloat = 2.0 * scratchFloat;
        }
    *scratchPtr = 0;
    if( (scratchPtr < buffer2) || (scratchPtr > buffer2 + sizeof( buffer2 )) )
        resetError( 58 );
    
//    tprintf5( defaultErrorPort, "xecvt 1: scratchFloat %10.6Le, As binary: %s, expTwo %d\n",
//              scratchFloat, buffer2, expTwo ); // debug

    if( radix != 2 ) {

        // Now we need to know how to parse this string into characters appropriate to the radix.
        // The problem is knowing with which bit a given character starts.  What we can count on
        // is that some character ends with the digit just to the left of "expTwo" -- that's the
        // "units" character in the representation of the number.  It takes basePowerOfTwo bits
        // to make up one character, so that the character boundaries are just to the left of
        // positions expTwo + N * basePowerOfTwo, for integer N.  Which of those is the most
        // significant character of the string we want to return?
        //
        // If expTwo is non-negative, that character has its rightmost boundary just to the left
        // of the position: remainder( expTwo, basePowerOfTwo ).  If expTwo is negative, the
        // boundary is at basePowerOfTwo - remainder( |expTwo|, basePowerOfTwo ).

        if( expTwo >= 0 )
            rightEndOfUnits = expTwo - basePowerOfTwo * (expTwo / basePowerOfTwo );
        else                        // Very defensive, we already tested for this.
            resetError( 260 );

//        tprintf3( defaultErrorPort, "xecvt 2: The right end of the units character starts at position %ld.\n", rightEndOfUnits );   // debug

        // If that number is not zero, we must pad the buffer at the left end with (basePowerOfTwo - rightEndOfUnits)
        // zeros, and we must pad on the right with enough zeros so that the whole buffer contains a number of characters
        // that is equal to (basePowerOfTwo * ndig).

        *buffer3 = 0;
        if( rightEndOfUnits > 0 ) {
            for( counter = 0; counter < (basePowerOfTwo - rightEndOfUnits); ++counter )
                buffer3[ counter ] = '0';
            buffer3[ counter ] = 0;
            }
        strncat( buffer3, buffer2, sizeof( buffer3 ) - strlen( buffer3 ) );
        counter = (basePowerOfTwo * ndig) - strlen( buffer3 );
        if( counter > 0 ) {
            scratchPtr = buffer3 + strlen( buffer3 );
            while( counter-- > 0 )
                *scratchPtr++ = '0';
            *scratchPtr = 0;
            }

//        tprintf3( defaultErrorPort, "xecvt 3: Buffer3 is now %s\n", buffer3); // debug

        // Now we can peel off bits, basePowerOfTwo at a time, and write
        // the corresponding characters in the radix.

        *buffer = 0;
        scratchPtr = buffer;
        for(    digitsPrinted = 0, counter = 0;
                (counter < (LongWord)(strlen( buffer3 ))) && (digitsPrinted < ndig);
                 counter += basePowerOfTwo, ++digitsPrinted ) {
            if( radix == 8 )
                *scratchPtr++ = threeBitsToOctalChar( buffer3[ counter ], buffer3[ counter + 1 ], buffer3[ counter + 2 ] );
            else if( radix == 16 )
                *scratchPtr++ = fourBitsToHexChar( buffer3[ counter ], buffer3[ counter + 1 ],
                                                   buffer3[ counter + 2 ], buffer3[ counter + 3 ] );
            }
        while( digitsPrinted++ < ndig )
             *scratchPtr++ = '0';
        *scratchPtr = 0;

        // But what about the exponent?  If expTwo is positive, the desired exponent is
        // (expTwo plus the number of zeros we left-padded with) / basePowerOfTwo.  (The quantity
        // in () is the size of the entire numeric string, after left-padding.)

        if( expTwo >= 0 ) {
            if( rightEndOfUnits > 0 )
                *rdxpt = 1 + (expTwo / basePowerOfTwo);
            else
                *rdxpt = expTwo / basePowerOfTwo;
            }
        else                        // Very defensive, we already tested for this.
            resetError( 260 );

//        tprintf4( defaultErrorPort, "xecvt 4: Buffer is now %s, *rdxpt %d\n", buffer, *rdxpt ); // debug

        }
    else {    // radix is two.
        *buffer = 0;
        scratchPtr = buffer;
        for( digitsPrinted = 0, counter = 0;
             (counter < (LongWord)(strlen( buffer2 ))) && (digitsPrinted < ndig);
             ++counter, ++digitsPrinted )
            *scratchPtr++ = buffer2[ counter ];
        while( digitsPrinted++ < ndig )
             *scratchPtr++ = '0';
        *scratchPtr = 0;
        *rdxpt = expTwo;
        }
    return buffer;
    }

/**** xdcvt -- This basic primitive works like "ecvt" except that it accepts an additional
               argument; namely, the radix of the base in which the printing is to be done,
               and it only accepts LongWord arguments to be formatted.   ****/
                          
Text *xdcvt( LongWord N, LongWord ndig, LongWord *rdxpt, Word *sign, LongWord radix )
{
    static Text buffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text buffer2[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text buffer3[ MAX_OUTPUT_NUMBER_SIZE + 1 ];  // Possible bug-fixer.

    LongWord NShifted = N;
    LongWord thisDigitN = 0;
    Text *back;
    LongWord counter = 0;
    LongWord nDigitsFound = 0;

    *rdxpt = 0;
    *sign = 0;
    buffer[0] = 0;
    buffer2[0] = 0;
    buffer3[0] = 0;

    if( ndig > (LongWord)MAX_OUTPUT_NUMBER_SIZE ) {
        resetError( 222 );
        }

    if( N == 0 ) {                   // Zero is a special case
        back = buffer;
        while( ndig-- > 0 )
            *back++ = getDigitChar( 0 );
        *back = '\0';
        return buffer;
        }

    if( N > 0 )
        *sign = 0;
    else {
#ifndef PIXIE_SCHEME_III
        N = -N;
#else

//        N = - N;    // Broken for iPad/ARM code generation as of 2010.10.11

        LongWord nUpper = N >> 32;
        LongWord nLower = N & 0x0ffffffffLL;

        nUpper = (~nUpper) & 0x0ffffffffLL;
        nLower = (~nLower) & 0x0ffffffffLL;

        N = nLower | (nUpper << 32LL);

        snprintf( buffer3, MAX_OUTPUT_NUMBER_SIZE, "%lld", N );
        N = N + 1;

        snprintf( buffer3, MAX_OUTPUT_NUMBER_SIZE, "%lld", N );
#endif
        NShifted = N;
        *sign = 1;
        }
    
    switch( radix ) {
        case 10:
            if( N == (LongWord)(FENCE_SPLITTER) )
                sprintf( buffer, "%s", "9223372036854775808");
            else
#ifndef PIXIE_SCHEME_III
                sprintf( buffer, "%ld", N );
#else
                sprintf( buffer, "%lld", N );
#endif
            nDigitsFound = strlen( buffer );
            for( counter = nDigitsFound; counter < ndig; ++counter )
                buffer[ counter ] = BINARY_DIGITS[ 1 ];
            buffer[ counter ] = 0;
            *rdxpt = nDigitsFound;
            return buffer;
            break;
        case  2:  
            for( nDigitsFound = 0; NShifted != 0; ++nDigitsFound ) {
                thisDigitN = NShifted & 1;
                buffer2[ nDigitsFound ] = thisDigitN ? BINARY_DIGITS[ 0 ] : BINARY_DIGITS[ 1 ];
                NShifted >>= 1;
                NShifted &= 0x7fffffffffffffffLL;
                }
            break;
        case  8:
            for( nDigitsFound = 0; NShifted != 0; ++nDigitsFound ) {
                thisDigitN = NShifted & 0x7;
                buffer2[ nDigitsFound ] = octalCharForN( thisDigitN );
                NShifted >>= 3;
                NShifted &= 0x1fffffffffffffffLL;
                }
            break;
        case 16:
            for( nDigitsFound = 0; NShifted != 0; ++nDigitsFound ) {
                thisDigitN = NShifted & 0xf;
                buffer2[ nDigitsFound ] = octalCharForN( thisDigitN );
                NShifted >>= 4;
                NShifted &= 0x0fffffffffffffffLL;
                }
            break;
        default:  resetError( 266 );  break;
        }            
    for( counter = 0; counter < nDigitsFound; ++counter )
        buffer[ counter ] = buffer2[ nDigitsFound - counter - 1 ];
    while( counter < ndig )
        buffer[ counter++ ] = BINARY_DIGITS[ 1 ];
    buffer[ counter ] = 0;
    *rdxpt = nDigitsFound;
    return buffer;
    }

/**** formatGuts ****/

void formatGuts( LongWord radix, Text *buffer, Boolean fullPrecisionP )
{
    static Text firstBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];
    static Text secondBuffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    Text *scratch = 0;
    Boolean exactP;

    EXACT_P
    exactP = B;

    NAN_OR_INF_P
    if( B ) {
        NAN_P
        if( B )
            strcpy( buffer, MiscStrings[ 229 ] );
        else
            strcpy( buffer, MiscStrings[ 230 ] );
        return;
        }

    if( radix != 10 ) {
        scratch = fmtFix( radix, fullPrecisionP, exactP );
        strcpy( buffer, scratch );
        }
    else {
        scratch = fmtFix( radix, fullPrecisionP, exactP );
        strcpy( firstBuffer, scratch );
        scratch = fmtSci( radix, fullPrecisionP, exactP );
        strcpy( secondBuffer, scratch );

        if( strlen( firstBuffer ) <= strlen( secondBuffer ) )
            strcpy( buffer, firstBuffer );
#ifndef PIXIE_SCHEME_III
        else
            strcpy( buffer, secondBuffer );
#else
        else {

            //  The second buffer is shorter, but is it right?  The ARM has only a 64-bit long double, so we need to check.

//            INTEGER_P
//            if( B && ((R.u.c > MAX_FIXNUM_AS_64_BIT_FLOAT) || (R.u.c < MIN_FIXNUM_AS_64_BIT_FLOAT)) ) {
//            if( (R.u.c > MAX_FIXNUM_AS_64_BIT_FLOAT) || (R.u.c < MIN_FIXNUM_AS_64_BIT_FLOAT) ) {
            FIXNUM_P            
            if( B
                && (   ((R.u.c > MAX_FIXNUM_AS_64_BIT_FLOAT) && (R.u.c <= MAX_FIXNUM_DECIMAL))
                    || ((R.u.c < MIN_FIXNUM_AS_64_BIT_FLOAT) && (R.u.c >= MIN_FIXNUM_DECIMAL)))
                ) {

                long double intAsLongDouble = (long double)(R.u.c);

                if( (LongWord)intAsLongDouble == R.u.c )
                    strcpy( buffer, secondBuffer );
                else
                    strcpy( buffer, firstBuffer );
                }
            else
                strcpy( buffer, secondBuffer );
            }
#endif
        }
    return;
    }

/**** displayNumberFull ****/

Text *displayNumberFull()
{
    static Text buffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    formatGuts( 10, buffer, TRUE );

    return buffer;
    }

/**** displayNumberHeuristically ****/

Text *displayNumberHeuristically()
{
    static Text buffer[ MAX_OUTPUT_NUMBER_SIZE + 1 ];

    formatGuts( 10, buffer, useFullNumericPrecision );

    return buffer;
    }

/**** formatNumber -- for "number->string" ****/

Text *formatNumber( Text *buffer )
{
    LongWord radix = (STACK( 0 )).u.c;

    formatGuts( radix, buffer, TRUE );

//    tprintf4( stdout, "FormatNumber gets \"%s\", radix is %ld\n", buffer, radix );  // debug

    return buffer;
    }
