/****************************************************************************************

Reader.c -- Copyright (c) by Jay Reynolds Freeman 1987, 1988, 1989, 1990, 1991, 2006

Input reader for Wraith Scheme.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "OpCodes.h"
#import "Reader.h"
#import "SchemeCType.h"

#import "FunctionAndSpecialNames.e"
#import "DataTypes.e"
#import "Globals.e"
#import "Intern.e"
#import "IO.e"
#import "MainAux.e"
#import "ReaderAux.e"
#import "ReaderNumbers.e"
#import "Registers.e"
#import "SchemeAux.e"
#import "SchemeGlobals.e"
#import "StreamHandling.e"
#import "StringLoader.e"

static taggedAval CRPort = sharpF;

Boolean ParserActive = FALSE;
Boolean ReadingString = FALSE;
UnsignedLongWord RParensNeeded = 0;

/******** forward declarations ********/

void readListCloser(),
     readParenthesizedList(),
     readParenthesizedListTailReversed(),
     readSexpInternal();

void skipWhiteSpacePort( taggedAval *aPort );

/******** Functions: ********/

/**** initializeReader ****/

void initializeReader()
{
    UnsignedLongWord theWord = 0;

    ParserActive = FALSE;
    ReadingString = FALSE;
    RParensNeeded = 0;
    CRPort.t = portTag;
    orIntoPortNumberByteNonAtomic( theWord, coninIndex );
    orIntoPortKittenByteNonAtomic( theWord, theKittenNumber );
    orIntoKittenByteNonAtomic( theWord, theKittenNumber );
    CRPort.u.c = (LongWord)theWord;
    reportEvent( PARSER_STATUS_EVENT );
    }

/**** isAtomicToken -- Elementary predicate to tell whether a particular type of token is an atom.  ****/

Boolean isAtomicToken( LexType tokenType )
{
    switch( tokenType ) {
        case LEX_BOOLEAN:
        case LEX_CHARACTER:
        case LEX_CONS_DOTTER:
        case LEX_IDENTIFIER:
        case LEX_NUMBER:
        case LEX_STRING:
        case LEX_LOGIC:
        case LEX_NON_PRINTING:
            return TRUE;
            break;
        default:
            return FALSE;
            break;
        }
    }

/**** getFirstType -- Return type indicated by the given character (which is presumably the first one in a string somewhere). ****/

FirstType getFirstType( LongWord l )
{
    Text c = (Text)l;
        
    if( isEOF( l ) )
        return EOF_TYPE;
    else if( isInitial( c ) )
        return INITIAL_TYPE;
    else if( isPeculiarInitial( c ) && (c != PECULIAR_INITIAL_DOT) )
        return PECULIAR_NON_POINT_TYPE;
    else if( c == SHARP )
        return SHARP_TYPE;
    else if( c == DOUBLE_QUOTE )
        return DOUBLE_QUOTE_TYPE;
    else if( isSchemeDigit( c ) )
        return DIGIT_10_TYPE;
    else if( isDelimiter( c ) )
        return DELIMITER_TYPE;
    else if( c == CONS_DOTTER )
        return POINT_TYPE;
    else if( isWeird( c ) )
        return WEIRD_TYPE;
    else
        return BAD_FIRST_CHARACTER_TYPE;
    }

/**** getDelimiter -- Put the given character into a null-terminated string, all by itself.  ****/

LexType getDelimiter( Text firstChar, Text *answer )
{
    Text *bufferPointer = answer;

    *bufferPointer++ = firstChar;
    *bufferPointer = 0;
    return LEX_DELIMITER;
    }
        
/**** getIdentifier -- Stuff an identifier into the answer. ****/

LexType getIdentifier( Text firstChar, Text *answer, UnsignedLongWord size )
{
    Text *bufferPointer = answer;
    Text *bufferLastFillable = answer + size - 2;
    Text c;
        
    *bufferPointer++ = schemeToLower(firstChar);
    while( (peekPort( &CRPort ), c = (Text)(R.u.c), isSubsequent( c )) ) {
        getNextCharLowerPort( &CRPort );
        if( bufferPointer >= bufferLastFillable ) {
            *bufferPointer = 0;
            OUTPUT_TO_ERROR
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 19 );
            return BAD_LEXTYPE;
            }
        else
            *bufferPointer++ = (R.u.c);
        }
    *bufferPointer = 0;
    return LEX_IDENTIFIER;
    }

/**** getNamedCharacter -- Test for a "named character", e. g., "space" or "newline" ****/

Boolean getNamedCharacter( Text *namedCharacter, Text *bufferPointer, Text firstChar )
{
    Boolean goodCharacterNameP = FALSE;
    LongWord counter, lnth;

    if( (firstChar == schemeToLower( namedCharacter[0] )) || (firstChar == schemeToUpper( namedCharacter[0] )) ) {
        *bufferPointer++ = schemeToLower( namedCharacter[0] );
        lnth = strlen( namedCharacter );
        goodCharacterNameP = TRUE;
        for( counter = 1; counter < lnth; ++counter ) {
            getNextCharLowerPort( &CRPort );
            *bufferPointer++ = (Text)(R.u.c);
            if( (Text)(R.u.c) != namedCharacter[ counter ] ) {
                goodCharacterNameP = FALSE;
                break;
                }
            }
        }
    *bufferPointer = 0;
    return goodCharacterNameP;
    }

/**** getNumber --  Stuff a number into the answer. ****/

LexType getNumber( Text firstChar, Text *answer )
{
    LongWord nDigits = 0;
    Text c;
    Text *bufferPointer = answer;

    *bufferPointer++ = firstChar;
    while( (peekPort( &CRPort ), c = schemeToLower((Text)(R.u.c)), !(isDelimiter( c ))) ) {
        getNextCharPort( &CRPort );
        if( isEOF( R.u.c ) )
            break;
        if( ++nDigits > (LongWord)MAX_NUMBER_SIZE )
            resetError( 20 );
        *bufferPointer++ = c;
        }
    *bufferPointer = 0;
    return LEX_NUMBER;
    }

/**** getPeculiarNonPoint -- Stuff a "peculiar" identifier into the answer. ****/

LexType getPeculiarNonPoint( Text firstChar, Text *answer )
{
    Text *bufferPointer = answer;
    Text c;

    peekPort( &CRPort );
    c = schemeToLower( (Text)(R.u.c) );       
    *bufferPointer++ = firstChar;
    if( isDelimiter( c ) || isEOF( R.u.c ) ) {  // FIrst character is an isolated '+', or '-'.
        *bufferPointer = 0;
        return LEX_IDENTIFIER;
        }
    else {
        getNextCharPort( &CRPort );
        return getNumber( c, answer + 1 );      // Only lexically permitted construct is a number ...
        }
    }

/**** getPoint -- Test for isolated ".", then "...", otherwise assume it's a number. ****/

LexType getPoint( Text firstChar, Text *answer )
{
    Text *bufferPointer = answer;
    Text c;

    getNextCharLowerPort( &CRPort );
    c = (Text)(R.u.c);        
    *bufferPointer++ = firstChar;
    if( isDelimiter( c ) || isEOF( R.u.c ) ) {  // First character is an isolated dot.
        *bufferPointer = 0;
        return LEX_CONS_DOTTER;
        }
    else if( c == PECULIAR_SUBSEQUENT_DOT ) {   // First character is not isolated, and is followed by a '.'.
        *bufferPointer++ = c;                   // Put the second character into the buffer.
        getNextCharLowerPort( &CRPort );         // Get the third character ...
        c = schemeToLower( (Text)(R.u.c) );
        if( isEOF( R.u.c ) )                    //   ... end-of-file is an error.
            resetError( 271 );
        if( c != PECULIAR_SUBSEQUENT_DOT )      //   ... so is anything *other* than another '.'
            resetError( 270 );
        *bufferPointer++ = c;                   // All's well; store the last character ...
        *bufferPointer = 0;                     //   ... terminate the string
        return LEX_IDENTIFIER;                  //   ... report an identifier.
        }
    else                                       // First character is not isolated, and not followed by a '.' --
        return getNumber( c, answer + 1 );     //   the only legal sytax here is a number.
    }

/**** getSharped -- Stuff a "sharped" entity (one preceded by '#') into the answer. ****/

LexType getSharped( Text firstChar, Text *answer )
{
    Text *bufferPointer = answer;
    Text c, e;

    getNextCharLowerPort( &CRPort );
    if( isEOF( R.u.c ) )
        resetError( 193 );
    c = (Text)(R.u.c);        
    *bufferPointer++ = firstChar;
    *bufferPointer++ = c;
    if( (c == TRUTH_T) || (c == FALSE_F) ) {        /* boolean */
        *bufferPointer = 0;
        peekPort( &CRPort );
        e = (Text)(R.u.c);
        if( isDelimiter( e ) || isEOF( R.u.c ) )
            return LEX_BOOLEAN;
        else {
            OUTPUT_TO_ERROR
            *bufferPointer++ = e;
            *bufferPointer = 0;
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 21 );
            return BAD_LEXTYPE;
            }
        }
    else if( (c == FAIL_U) || (c == SUCCEED_S) ) {  /* logic value */
        *bufferPointer = 0;
        peekPort( &CRPort );
        e = (Text)(R.u.c);
        if( isDelimiter( e ) || isEOF( R.u.c ) )
            return LEX_LOGIC;
        else {
            OUTPUT_TO_ERROR
            *bufferPointer++ = e;
            *bufferPointer = 0;
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 291 );
            return BAD_LEXTYPE;
            }
        }
    else if( c == NON_PRINT_N ) {  /* non-printing object */
        *bufferPointer = 0;
        peekPort( &CRPort );
        e = (Text)(R.u.c);
        if( isDelimiter( e ) || isEOF( R.u.c ) )
            return LEX_NON_PRINTING;
        else {
            OUTPUT_TO_ERROR
            *bufferPointer++ = e;
            *bufferPointer = 0;
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 291 );
            return BAD_LEXTYPE;
            }
        }
    else if( c == BACKSLASH ) {                     /* character */
        getNextCharRawPort( &CRPort );
        if( isEOF( R.u.c ) )
            resetError( 194 );
        if( R.u.c == '\n')
            resetError( 280 );
        c = (Text)(R.u.c);        
            peekPort( &CRPort );
        e = (Text)(R.u.c);
        if( isDelimiter( e ) || isEOF( R.u.c ) ) {
            *bufferPointer++ = c;
            *bufferPointer = 0;
            return LEX_CHARACTER;
            }
        else {
            if( schemeToLower( c ) == MiscStrings[ 147 ][ 0 ] ) {
                if( getNamedCharacter( MiscStrings[ 147 ], bufferPointer, c ) )
                    return LEX_CHARACTER;
                }
            else if( schemeToLower( c ) == MiscStrings[ 148 ][ 0 ] ) {
                if( getNamedCharacter( MiscStrings[ 148 ], bufferPointer, c ) )
                    return LEX_CHARACTER;
                }
            else if( schemeToLower( c ) == MiscStrings[ 335 ][ 0 ] ) {
                if( getNamedCharacter( MiscStrings[ 335 ], bufferPointer, c ) )
                    return LEX_CHARACTER;
                }
            else if( schemeToLower( c ) == MiscStrings[ 336 ][ 0 ] ) {
                if( getNamedCharacter( MiscStrings[ 336 ], bufferPointer, c ) )
                    return LEX_CHARACTER;
                }
            else {
                *bufferPointer++ = c;
                *bufferPointer = 0;
                }
            OUTPUT_TO_ERROR
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 22 );
            return BAD_LEXTYPE;
            }
        }
    else if( c == LIST_OPENER )             /* vector */
        return LEX_VECTOR;
    else                                                    /* number */
        return getNumber( schemeToLower( c ), answer + 1 );
    }
        
/**** getString -- Stuff a string into the answer. ****/

LexType getString( Text *answer, UnsignedLongWord size )
{
    Text *bufferPointer = answer;
    Text *bufferLastFillable = answer + size - 2;
    Text c;
        
    while( getNextCharRawPort( &CRPort ), c = (Text)(R.u.c), c != DOUBLE_QUOTE ) {
        if( isEOF( R.u.c ) )
            resetError( 195 );
        if( c == NEWLINE )
            resetError( 98 );
        if( bufferPointer >= bufferLastFillable ) {
            *bufferPointer = 0;
            OUTPUT_TO_ERROR
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 23 );
            break;
            }
        if( c == BACKSLASH ) {
            getNextCharRawPort( &CRPort );
            c = (Text)(R.u.c);
            if( isEOF( R.u.c ) )
                resetError( 196 );
            switch( c ) {
                case '\n': *bufferPointer++ = '\n';                    // Escaped literal newlines are okay.
                    while( (peekPort( &CRPort ), isEOF( R.u.c )) )     // But we must wait till the user types something else.
                        ;
                    break;
                case 'n':  *bufferPointer++ = '\n'; break;
                case 'r':  *bufferPointer++ = '\r'; break;
                case 't':  *bufferPointer++ = '\t'; break;
                default:   *bufferPointer++ = c;    break;
                }
            }
        else
            *bufferPointer++ = c;
        }
    *bufferPointer = 0;
    return LEX_STRING;
    }       

/**** getWeird -- Stuff a weird token into the answer. ****/

LexType getWeird( Text firstChar, Text *answer, UnsignedLongWord size )
{
    Text *bufferPointer = answer;
    Text *bufferLastFillable = answer + size - 2;
    Text c;
    LongWord counter = 1;
        
    *bufferPointer++ = firstChar;
    while( (peekPort( &CRPort ), c = (Text)(R.u.c), c == firstChar ) ) {
        if( ((counter == 1) && (c == OPTION_MINUS)) || (counter == 3) ) {
            *bufferPointer = 0;
            OUTPUT_TO_ERROR
            tprintf3( &CurrentOutputPort, MiscStrings[ 252 ], c );
            resetError( 25 );
            return BAD_LEXTYPE;
            }
        ++counter;
        getNextCharPort( &CRPort );
        if( bufferPointer >= bufferLastFillable ) {
            *bufferPointer = 0;
            OUTPUT_TO_ERROR
            tprintf3( &CurrentOutputPort, MiscStrings[ 201 ], answer );
            resetError( 19 );
            return BAD_LEXTYPE;
            }
        else
            *bufferPointer++ = (Text)(R.u.c);
        }
    if( isSubsequent( c ) ) {
        *bufferPointer = 0;
        OUTPUT_TO_ERROR
        tprintf4( &CurrentOutputPort, MiscStrings[ 253 ], answer, c );
        resetError( 25 );
        return BAD_LEXTYPE;
        }
    else {
        *bufferPointer = 0;
        return LEX_IDENTIFIER;
        }
    }

/**** getToken -- Basic lexical scanner:  Put a token into "answer", as a null-terminated
                                 string.  The length of the string buffer is given by "size".             ****/

LexType getToken( Text *answer, UnsignedLongWord size )
{
    FirstType theFirstType;
    LexType theWholeType = LEX_EOF;
    LongWord firstNonWhiteLong;
    Text firstNonWhite;
        
    skipWhiteSpacePort( &CRPort );
    if(    (theFirstType = getFirstType( (getNextCharPort( &CRPort ), firstNonWhiteLong = (R.u.c)) ) ) 
        ==
           BAD_FIRST_CHARACTER_TYPE )
      {
        OUTPUT_TO_ERROR
        tprintf4( &CurrentOutputPort, MiscStrings[ 314 ], R.u.c, R.u.c );
        resetError( 25 );
        }
    firstNonWhite = (Text)firstNonWhiteLong;
    switch( theFirstType ) {
        case EOF_TYPE:
            theWholeType = LEX_EOF;
            *answer = 0;
            break;
        case INITIAL_TYPE:
            theWholeType = getIdentifier( firstNonWhite, answer, size );
            break;
        case PECULIAR_NON_POINT_TYPE:
            theWholeType = getPeculiarNonPoint( firstNonWhite, answer );
            break;
        case SHARP_TYPE:
            theWholeType = getSharped( firstNonWhite, answer );
            break;
        case DOUBLE_QUOTE_TYPE:
            ReadingString = TRUE;
            reportEvent( PARSER_STATUS_EVENT );
            reportEvent( EXPECT_DOUBLE_QUOTE_EVENT );
            theWholeType = getString( answer, size );
            ReadingString = FALSE;
            reportEvent( PARSER_STATUS_EVENT );
            reportEvent( EXPECT_DOUBLE_QUOTE_EVENT );
            break;
        case DIGIT_10_TYPE:
            theWholeType = getNumber( firstNonWhite, answer );
            break;
        case DELIMITER_TYPE:
            theWholeType = getDelimiter( firstNonWhite, answer );
            break;
        case POINT_TYPE:
            theWholeType = getPoint( firstNonWhite, answer );
            break;
        case WEIRD_TYPE:
            theWholeType = getWeird( firstNonWhite, answer, size );
            break;
        default:
            resetError( 26 );
        }
    return theWholeType;
    }

/**** skipWhiteSpace -- Read and ignore white space. ****/

void skipWhiteSpace()
{ 
    while( (peekPort( &CRPort ), isWhiteSpace( (Text)(R.u.c) )) )
        getNextCharRawPort( &CRPort );
    }

/**** skipWhiteSpacePort -- Read and ignore white space. ****/

void skipWhiteSpacePort( taggedAval *aPort )
{ 
    while( (peekPort( aPort ), isWhiteSpace( (Text)(R.u.c) )) )
        getNextCharRawPort( aPort );
    }

/**** skipWhiteSpacePortNoHangOnEOF -- Read and ignore white space, do not spin on EOF. ****/

void skipWhiteSpacePortNoHangOnEOF( taggedAval *aPort )
{ 
    while( (peekPortNoHangOnEOF( aPort ), isWhiteSpace( (Text)(R.u.c) )) ) {
        getNextCharRawPortNoHangOnEOF( aPort );
        }
    }

/**** readAtom -- Read an atom and return it in R. ****/

PROC( readAtom )

    LexType theType;
    Text buffer[MAX_ATOM_SIZE];
    taggedAval *a;
        
    LOAD_R_TAG( exactShortFixnumBitsMask )
    LOAD_OBJECT( buffer, R )
    PUSH
    theType = getToken( (Text *)(R.u.c), MAX_ATOM_SIZE );
    POP
    switch( theType ) {
        case LEX_EOF:
            resetError( 197 );
            break;
        case LEX_IDENTIFIER:
            LOAD_R_TAG( exactShortFixnumBitsMask )
            CALL( intern )
            break;
        case LEX_BOOLEAN:
            if( buffer[1] == FALSE_F ) {
                LOAD_R_TAG( falseBooleanTag )
                }
            else
                LOAD_R_TAG( trueBooleanTag )
            break;
        case LEX_LOGIC:
            if( buffer[1] == FAIL_U )
                LOAD_R_TAG( failTag )
            else
                LOAD_R_TAG( succeedTag )
            break;
        case LEX_NON_PRINTING:
            LOAD_R_TAG( peculiarLeafTag );
            LOAD_OBJECT( nonPrintingObjectTag, R );
            break;
        case LEX_NUMBER:
            R = constructNumber( (Text *)(R.u.c), 0, TRUE );
            break;
        case LEX_CHARACTER:
            if( strcmp( (Text *)(R.u.c) + 2, MiscStrings[ 147 ] ) == SAME )
                LOAD_CHARACTER_OBJECT( SPACE, R )
            else if( strcmp( (Text *)(R.u.c) + 2, MiscStrings[ 148 ] ) == SAME )
                LOAD_CHARACTER_OBJECT( NEWLINE, R )
            else if( strcmp( (Text *)(R.u.c) + 2, MiscStrings[ 335 ] ) == SAME )
                LOAD_CHARACTER_OBJECT( TAB, R )
            else if( strcmp( (Text *)(R.u.c) + 2, MiscStrings[ 336 ] ) == SAME )
                LOAD_CHARACTER_OBJECT( RETURN, R )
            else
                LOAD_CHARACTER_OBJECT( *((Text *)(R.u.c) + 2), R )
            LOAD_R_TAG( characterTag )
            break;
        case LEX_STRING:
            MAKE_STRING
            a = R.u.a;
            a->t = (a->t) | immutableBitMask;
            break;
        case LEX_DELIMITER:
            if( *((Text *)(R.u.c)) == RETURN )
                resetError( 12 );
            else if( *((Text *)(R.u.c)) == (Text)0x0a )                                       /* FIX?? */
                resetError( 13 );
            else if( *((Text *)(R.u.c)) == TAB )
                resetError( 14 );
            else
                resetError( 15 );
            break;
        case LEX_VECTOR:
            CALL( readParenthesizedList )   /* read contents of vector as if it were a list */
            PUSH                                                    /* S: (contents ) */
            CALL( readListCloser )                  /* read the final ")" */
            MOVE( STACK( 0 ), R )
            LENGTH                                                  /* find the length of the vector-to-be */
            CALL( makeVectorAndStuff )              /* create the vector and fill it from the list */
            break;
        case LEX_CONS_DOTTER:
            LOAD_R_TAG( peculiarLeafTag )                    
            LOAD_OBJECT( 0, R )
            orIntoLeafByteNonAtomic( R.u.c, consDotterPLTag );
            break;
        default:
            LOAD_R_TAG( emptyListTag );
            break;
            }
    CONTINUE

END_PROC
        
/**** readListCloser -- Silently read and ignore a ")", or else complain. ****/

PROC( readListCloser )
 
    skipWhiteSpacePort( &CRPort );
    peekPort( &CRPort );
    if( (Text)(R.u.c) == LIST_CLOSER ) {                      /* what we wanted */
        --RParensNeeded;
        getNextCharPort( &CRPort );
        }
    else {
        getNextCharPort( &CRPort );
        resetError( 16 );
        }
    CONTINUE

END_PROC
        
/**** readParenthesizedList -- Read and return a parenthesized list, but NOT its "()". ****/

PROC( readParenthesizedList )
 
    skipWhiteSpacePort( &CRPort );
    peekPort( &CRPort );
    if( (Text)(R.u.c) == LIST_CLOSER )                        /* bottomed out with ")" */
        LOAD_R_TAG( emptyListTag )
    else {
        CALL( readSexpInternal )
        if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
            resetError( 17 );
            }
        else {                                                
            
             //  Strategy:  Build up the list, reversed, except for the last element (in case
             //             the list we are parsing is improper.  Reverse the non-last-element
             //             part and append something which is either the last element (in case
             //             of an improper list) or a list of just the last element.

            NILCONS                                           // List of just the car of what we want.
            PUSH
            CALL( readParenthesizedListTailReversed )
            CALL( reverseListCode )
            SWAP
            PUSH
            MOVE( two, R )
            CALL( appendCode )
            }
        }
    }
        
/**** readParenthesizedListTailReversed -- Read and return the tail of a parenthesized list, but NOT its final ")". ****/

PROC( readParenthesizedListTailReversed )
 
    while( TRUE ) {

        //  Loop invariant (I hope):  the reversed list, so far, is on top of the stack.

        skipWhiteSpacePort( &CRPort );
        peekPort( &CRPort );
        if( (Text)(R.u.c) == LIST_CLOSER ) {                    // End of proper list.
            LOAD_R_TAG( emptyListTag )                           // So the thing to be appended is '().
            SWAP                           // Put it below the list to be reversed.
            break;
            }
        else {
            CALL( readSexpInternal )
            if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {      // Next S-expression is the thing to be appended.
                CALL( readSexpInternal )                        // So go get it.
                SWAP                       // Put it below the list to be reversed.
                break;
                }
            else {  
                SWAP                       // Swap the new item into the "car" position for consing.
                CONS                                            // Cons the new element onto the existing list.
                PUSH                                            // And push it back onto the stack.
                }
            }
        }

    CONTINUE

END_PROC

/**** complainIfEOForWhiteSpace ****/

void complainIfEOForWhiteSpace( LongWord cInt, LongWord EOFMessage, LongWord WhiteSpaceMessage )
{
    if( isEOF( cInt ) )
        resetError( EOFMessage );
    if( isWhiteSpace( (Text)cInt ) )
        resetError( WhiteSpaceMessage );
    }

/**** readSexpInternal -- "readSexp" for internal use; it uses peculiar leaves
                           to indicate that a "." was seen in a position where it denotes
                           a delimiter between the car and the cdr of a pair. ****/

PROC( readSexpInternal )
 
    skipWhiteSpacePort( &CRPort );
    peekPort( &CRPort );
    if( getPortBigArrayLocation( &CRPort ) == coninIndex ) {        // Reading from the console.
        ParserActive = TRUE;
        }
    if( (Text)(R.u.c) == LIST_CLOSER ) {
        getNextCharPort( &CRPort );
        resetError( 18 );
        }
    else if( (Text)(R.u.c) == LIST_OPENER ) {
        ++RParensNeeded;
        getNextCharPort( &CRPort );
        CALL( readParenthesizedList )
        PUSH
        CALL( readListCloser )
        POP
        }
    else if( (Text)(R.u.c) == QUASIQUOTER ) {
        getNextCharPort( &CRPort );
        peekPort( &CRPort ); 
        complainIfEOForWhiteSpace( R.u.c, 274, 275 );
        LOAD_R_TAG( exactShortFixnumBitsMask )
        LOAD_OBJECT( NameStrings[ 149 ], R )
        CALL( intern )
        PUSH
        CALL( readSexpInternal )
        if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
            resetError( 17 );
            }
        NILCONS
        CONS
        }
    else if( (Text)(R.u.c) == QUOTER ) {
        getNextCharPort( &CRPort );
        peekPort( &CRPort );
        complainIfEOForWhiteSpace( R.u.c, 272, 273 );
        LOAD_R_TAG( exactShortFixnumBitsMask )
        LOAD_OBJECT( NameStrings[ 150 ], R )
        CALL( intern )
        PUSH
        CALL( readSexpInternal )
        if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
            resetError( 17 );
            }
        NILCONS
        CONS
        }
    else if( (Text)(R.u.c) == UNQUOTER ) {
        getNextCharPort( &CRPort );
        peekPort( &CRPort );
        complainIfEOForWhiteSpace( R.u.c, 276, 277 );
        peekPort( &CRPort );
        if( (Text)(R.u.c) == UNQUOTE_SPLICING_SECOND_CHARACTER ) {
            getNextCharPort( &CRPort );
            peekPort( &CRPort );
            complainIfEOForWhiteSpace( R.u.c, 278, 279 );
            LOAD_R_TAG( exactShortFixnumBitsMask )
            LOAD_OBJECT( NameStrings[ 188 ], R )
            }
        else {
            LOAD_R_TAG( exactShortFixnumBitsMask )
            LOAD_OBJECT( NameStrings[ 187 ], R )
            }
        CALL( intern )
        PUSH
        CALL( readSexpInternal )
        if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
            resetError( 17 );
            }
        NILCONS
        CONS
        }
    else {
        CALL( readAtom )
        }
    CONTINUE

END_PROC

/**** readSexp -- This is part of what might be called "read" but for the namespace conflict with the C library. ****/

void readSexp()
{
    CRPort = DefaultInputPort;
    Boolean isConsoleP = TRUE;

    MOVE( CRPort, R )
    PUSH
    isConsoleP = (getPortBigArrayLocation( &CRPort ) == coninIndex);
    skipWhiteSpacePort( &DefaultInputPort );
    peekPort( &DefaultInputPort );

    if( isEOF( R.u.c ) ) {
        if( isConsoleP ) {                                      /* Wait forever, for console input. */
            while( isEOF( R.u.c ) ) {
                skipWhiteSpacePort( &DefaultInputPort );
                peekPort( &DefaultInputPort );
                }
            CALL( readSexpInternal )                            /* At last, something which is not an EOF. */
            if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
                resetError( 17 );
                }
            }
        else {                                                  /* Return EOF-object for EOF on file. */
            LOAD_R_TAG( peculiarLeafTag )
            LOAD_OBJECT( 0, R ) 
            orIntoLeafByteNonAtomic( R.u.c, eofObjectPLTag );
            }
        }
    else {                                                      /* At last, something which is not an EOF. */
        CALL( readSexpInternal )
        if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
            resetError( 17 );
            }
        }

    MOVE( S[0], CRPort )
    DROP
    CONTINUE
    }

/**** readSexpFromPort -- This is part of what would be called "read" but for the namespace conflict with the C library. ****/

void readSexpFromPort() // From the port in STACK( 0 )
{
    MOVE( CRPort, R )
    PUSH
    MOVE( S[1], CRPort )

    skipWhiteSpacePort( &CRPort );
    peekPort( &CRPort );
    if( isEOF( R.u.c ) ) {
        LOAD_R_TAG( peculiarLeafTag )
        LOAD_OBJECT( 0, R ) 
        orIntoLeafByteNonAtomic( R.u.c, eofObjectPLTag );
        }
    else 
        CALL( readSexpInternal )
    if( peculiar_leaf_n_p( R, consDotterPLTag ) ) {
        resetError( 17 );
        }

    MOVE( S[0], CRPort )
    DROP
    CONTINUE
    }

/**** readSexpOrQueueTopLevel ****/

#define WAIT_FOR_SEXP_DELAY 100000        // microseconds
#define WAIT_FOR_CONTROLLER_DELAY 1000    // microseconds

void readSexpOrQueueTopLevel()
{
    while( ! readyToRead )
        eventLoopOnce( WAIT_FOR_CONTROLLER_DELAY );
    while( TRUE ) {
        theHeap->triggerGC();
        RParensNeeded = 0;
        ReadingString = FALSE;
        reportEvent( PARSER_STATUS_EVENT );
        if( ReadingString )
            reportEvent( EXPECT_DOUBLE_QUOTE_EVENT );
        else
            reportEvent( EXPECT_RPARENS_EVENT );
        CALL( kittenDequeueCode )
        if( pair_p( R.t ) ) {
            CAR
            break;       
            }
        PUSH

        skipWhiteSpacePortNoHangOnEOF( &DefaultInputPort );
        peekPortNoHangOnEOF( &DefaultInputPort );

        if( isEOF( R.u.c ) ) {
            if( getPortBigArrayLocation( &DefaultInputPort ) == coninIndex ) {        // Reading from the console.
                POP
                usleep( WAIT_FOR_SEXP_DELAY );           // EXPERIMENT -- in 2010.12.06
//                eventLoopOnce( WAIT_FOR_SEXP_DELAY );  // EXPERIMENT -- out 2010.12.06
                continue;
                }
            }
        POP
        
        readSexp();
        if( empty_list_p( R.t ) )
            resetError( 146 );
        break;
        }

    ParserActive = FALSE;
    if( ReadingString )
        reportEvent( EXPECT_DOUBLE_QUOTE_EVENT );
    else
        reportEvent( EXPECT_RPARENS_EVENT );
    reportEvent( EXPECT_RPARENS_EVENT );
    }
