/****************************************************************************************

ReaderNumbers.c -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006

Input reader numeric parsing functions for Pixie.

        "A computer scientist is a person who refuses to believe that 1.0 is an integer."

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "Numbers.h"
#import "OpCodes.h"
#import "Reader.h"
#import "SECD.h"

#import "DataTypes.e"
#import "Globals.e"
#import "IO.e"
#import "NumberFormats.e"
#import "SECDIO.e"

/******** Statics: ********/

static Text* theNumber;

/******** forward declarations ********/

LongWord getExponent( Text **scan, Text *endString, Boolean *hasExponentString, Text *exponentMarker );
void getDigitsAndSharps(
    Text **scan,
    Text *endString,
    Text *digits,
    Text **firstDigit,
    Text **pastDigits,
    Text **firstSharp,
    Text **pastSharps
    );

void showNumberReset( UnsignedLongWord resetNumber );

/******** Functions: ********/

/**** parseNumber -- Disentangle an entire number into its constituent parts. ****/

void parseNumber(
    Text *numString,
    Text *signChar,
    Text *exactnessChar,
    Text *radixChar,
    LongWord *radixNumber,
    Text *delimiterChar,
    Text **prefixStart,
    Text **pastPrefix,
    Text **firstPreDelimiterDigit,
    Text **pastPreDelimiterDigits,
    Text **firstPreDelimiterSharp,
    Text **pastPreDelimiterSharps,
    Text **firstPostDelimiterDigit,
    Text **pastPostDelimiterDigits,
    Text **firstPostDelimiterSharp,
    Text **pastLastSharp,
    LongWord *exponent,
    Text *exponentMarker
    )
{
    Boolean hasDelimiter    = FALSE;
    Boolean hasExactness;
    Boolean hasRadix;
    Boolean hasExponentString = FALSE;
        
    Text *endString;
    Text *digits;
    Text *scan = numString;

    /* Set defaults for locals and for values to be returned. */
        
    *signChar = PLUS_SIGN;
    if( *exactnessChar != SPACE ) {
        hasExactness = TRUE;
        if( ! strchr( EXACTNESSES, *exactnessChar ) )
            showNumberReset( 27 );
        }
    else
        hasExactness = FALSE;

    if( *radixChar != SPACE ) {
        hasRadix = TRUE;
        if( *radixChar == BINARY_BASE )
            *radixNumber = 2;
        else if( *radixChar == DECIMAL_BASE )
            *radixNumber = 10;
        else if( *radixChar == OCTAL_BASE )
            *radixNumber = 8;
        else if( *radixChar == HEX_BASE )
            *radixNumber = 16;
        else
            showNumberReset( 28 );
        digits = ALL_DIGITS + (16 - *radixNumber) * sizeof( Text );
        }
    else {
        hasRadix = FALSE;
        *radixNumber = 10;
        digits = ALL_DIGITS + 6 * sizeof( Text );
        }

    *delimiterChar              = SPACE;
    *prefixStart                = (Text *)NULL;
    *pastPrefix                 = (Text *)NULL;
    *firstPreDelimiterDigit     = (Text *)NULL;
    *pastPreDelimiterDigits     = (Text *)NULL;
    *firstPreDelimiterSharp     = (Text *)NULL;
    *pastPreDelimiterSharps     = (Text *)NULL;
    *firstPostDelimiterDigit    = (Text *)NULL;
    *pastPostDelimiterDigits    = (Text *)NULL;
    *firstPostDelimiterSharp    = (Text *)NULL;
    *pastLastSharp              = (Text *)NULL;
    *exponent                   = 0;
    *exponentMarker             = SPACE;

    endString = numString + strlen( numString );

    /* Get sign. */

    if( (*scan == MINUS_SIGN) || (*scan == PLUS_SIGN) ) {
        *signChar = *scan;
        if( ++scan >= endString )
            showNumberReset( 29 );
        }
    else
        *signChar = PLUS_SIGN;

    /* Deal with prefixes. */
        
    *prefixStart = scan;
    while( *scan == SHARP_NS ) {
        if( ++scan >= endString )
            showNumberReset( 30 );
        if( strchr( EXACTNESSES, *scan ) )
            if( hasExactness && (*scan != *exactnessChar) )
                showNumberReset( 31 );
            else {
                hasExactness = TRUE;
                *exactnessChar = *scan;
                }
        else if( strchr( RADICES, *scan ) )
            if( hasRadix && (*scan != *radixChar) )
                showNumberReset( 33 );
            else {
                hasRadix = TRUE;
                *radixChar = *scan;
                if( *radixChar == BINARY_BASE )
                    *radixNumber = 2;
                else if( *radixChar == DECIMAL_BASE )
                    *radixNumber = 10;
                else if( *radixChar == OCTAL_BASE )
                    *radixNumber = 8;
                else if( *radixChar == HEX_BASE )
                    *radixNumber = 16;
                else
                    showNumberReset( 34 );
                digits = ALL_DIGITS + (16 - *radixNumber) * sizeof( Text );
                }
        else
            showNumberReset( 35 );
        if( ++scan >= endString )
            showNumberReset( 30 );
        }
    *pastPrefix = scan;
        
    /* At this point, we have established for once and all the sign, exactness, and radix of the number. */
    /* Next legal character should be a radix point (only for radix 10) or a digit -- no slashes allowed. */

    if( strchr( DELIMITERS, *scan ) ) {  // Delimiters are "." and "/"
        if( *scan == SLASH )                                   // Looks like a fraction with no numerator ...
            showNumberReset( 45 );
        if( (*radixNumber != 10) && (*scan == RADIX_POINT) )   // No radix points allowed in non-decimal numbers.
            showNumberReset( 262 );

        // What we have starts out with a radix point.  Now there ought to be some digits, maybe some sharps,
        //  and maybe an exponent.

        *delimiterChar = *scan;
        hasDelimiter = TRUE;
        if( ++scan >= endString )
            showNumberReset( 36 );
        if( ! strchr( digits, *scan ) )
            showNumberReset( 37 );
        getDigitsAndSharps( &scan, endString, digits, firstPostDelimiterDigit, pastPostDelimiterDigits,
                            firstPostDelimiterSharp, pastLastSharp );
        if( *scan && isExponentMarker( *scan ) ) {
            if( *radixNumber == 10 )
                *exponent = getExponent( &scan, endString, &hasExponentString, exponentMarker );
            else
                showNumberReset( 263);
            }
        }
    else {  // We ought to have a string starting with some digits.
        if( ! strchr( digits, *scan ) )
            showNumberReset( 38 );
        getDigitsAndSharps( &scan, endString, digits, firstPreDelimiterDigit, pastPreDelimiterDigits,
                            firstPreDelimiterSharp, pastPreDelimiterSharps );

        /* scan now points past the end of the number, or to a DELIMITER or an exponent marker */
        
        if( scan >= endString )                                     // End of string -- fine.
            ;
        else if( strchr( DELIMITERS, *scan ) ) {                    // Not end of string, but a delimiter ...
            if( (*radixNumber != 10) && (*scan == RADIX_POINT) )    // No radix points allowed in non-decimal numbers.
                showNumberReset( 262 );
            if( hasDelimiter )                                      // Too many delimiters!
                showNumberReset( 40 );
            else  {                                                 // Either a radix point or a slash.
                hasDelimiter = TRUE;
                *delimiterChar = *scan;
                ++scan;

                // Now let's get any digits and sharps that follow the delimiter.  If the delimiter was
                // a slash, these are the denominator of the fraction.  If it was a radix point, these
                // are the digits to the right of the radix point for a number that is not a fraction but
                // that does have a radix point.

                getDigitsAndSharps( &scan, endString, digits, firstPostDelimiterDigit, pastPostDelimiterDigits,
                                    firstPostDelimiterSharp, pastLastSharp );

                if(        (*firstPreDelimiterSharp < *pastPreDelimiterSharps)
                        && (*firstPostDelimiterDigit < *pastPostDelimiterDigits)
                        && (*delimiterChar == RADIX_POINT) )    // We had sharps before the radix point and digits after ... ?!?
                    showNumberReset( 39 );

                // Now we see if there is an exponent.

                if( *scan && isExponentMarker( *scan ) ) {
                    if( *delimiterChar == SLASH )               // There shouldn't be any exponents in fractions.
                        showNumberReset( 265 );
		            if( *radixNumber == 10 )
                        *exponent = getExponent( &scan, endString, &hasExponentString, exponentMarker );
		            else
                        showNumberReset( 263);
                    }
                }
            }
        else if( *scan && isExponentMarker( *scan ) ) {     // Not end of string or a delimiter, but an exponent marker ...
            if( *radixNumber == 10 )
                *exponent = getExponent( &scan, endString, &hasExponentString, exponentMarker );
            else
                showNumberReset( 263);
            }
        else  {                                    // Not end of string, delimiter or exponent marker -- garbage!
            if( *radixNumber == 10 )
                showNumberReset( 41 );
            else
                showNumberReset( 264 );
            }
        }

    // By now, we have the exponent, if any.

    if( (*radixNumber != 10) && hasExponentString )
        showNumberReset( 263 );
    if( scan < endString )
        showNumberReset( 42 );
    }

/**** getDigitsAndSharps -- The pointer *scan either points past endString or to a digit; stuff the various
                            pointers in the obvious manner, and bump scan past the last digit or sharp.        ****/

void getDigitsAndSharps(
        Text **scan,
        Text *endString,
        Text *digits,
        Text **firstDigit,
        Text **pastDigits,
        Text **firstSharp,
        Text **pastSharps
        )
{
    /* Locate the digits. */
        
    *firstDigit = *scan;
    while( (*scan < endString) && strchr( digits, **scan ) )
        ++(*scan);
    *pastDigits = *firstSharp = *scan;
        
    /* Locate the sharps. */

    while( (*scan < endString) && (**scan == SHARP_NS) )
        ++(*scan);
    *pastSharps = *scan;
    }

/**** getExponent -- The pointer *scan either points past endString, or points to a radix point
                     followed by a possibly signed decimal integer, or, in the case of a badly-formatted
                     points past a character -- 'd', 'e', or 'f' -- which the rest of the parser
                     interpreted as a hexadecimal digit but which the user intended to stand for
                     the exponent character in a "hexnum with an exponent" (which Scheme does
                     not allow.

                     This function does some error checking, returns the exponent (if it exists),
                     returns zero if it doesn't, and bumps the pointer "*scan" past the end
                     of the integer.
                                                                                               ****/

LongWord getExponent( Text **scan, Text *endString, Boolean *hasExponentString, Text *exponentMarker )
{
    LongWord answer;
    Text *newScan;

    *hasExponentString = FALSE;

    if( *scan >= endString )
        return (LongWord)0;
    *hasExponentString = TRUE;
    if( **scan && (! isExponentMarker( **scan )) )    
        showNumberReset( 43 );
    *exponentMarker = **scan;
    if( ++(*scan) >= endString )
        showNumberReset( 44 );
    if( (**scan == PLUS_SIGN) || (**scan == MINUS_SIGN) ) {
        if( (*scan + 1) >= endString )
            showNumberReset( 44 );
        if(    (strlen( *scan + 1 ) == ((UnsignedLongWord)(maxFixDigits[ 10 ])))         /* max digits in fixnum */
            && (strcmp( *scan + 1, MAX_DECIMAL_DIGITS )) >= 0 )
            showNumberReset( 204 );
        if( strlen( *scan + 1 ) > ((UnsignedLongWord)(maxFixDigits[ 10 ])) )              /* max digits in fixnum */
            showNumberReset( 205 );
        }
    else {
        if(    (strlen( *scan ) == ((UnsignedLongWord)(maxFixDigits[ 10 ])))             /* max digits in fixnum */
            && (strcmp( *scan, MAX_DECIMAL_DIGITS )) >= 0 )
            showNumberReset( 204 );
        if( strlen( *scan ) > ((UnsignedLongWord)(maxFixDigits[ 10 ])))                  /* max digits in fixnum */
            showNumberReset( 205 );
        }
    answer =  strtol( *scan, &newScan, 10 );
    *scan = newScan;
    return answer;
    }

/**** constructShortFloatWithDigitsPadAndSign ****/

ShortFloat constructShortFloatWithDigitsPadAndSign( Text *start, Text *end, LongWord pad, Text signChar )
{
    Text buffer[ MAX_NUMBER_SIZE ];
    LongWord i = 0; 

    buffer[ i++ ] = signChar;                // The sign.

    strncpy( &(buffer[i]), start, end - start );   // The digits given.
    buffer[ (end - start) + 1 ] = 0;

    i = strlen( buffer );     

    while( pad-- > 0 )
        buffer[ i++ ] = '0';                 // Zeros to pad with.

    buffer[ i++ ] = RADIX_POINT;             // Radix point.

#ifndef FLOATS_ARE_64_BIT
    buffer[ i++ ] = 'L';                     // Marker for "long double".
#endif

    buffer[ i ] = 0;
    
    return strtodShortFloat( buffer, 0 );
    }

static jmp_buf BadNumberBuffer;
static Boolean localFailOnErrorP;

/**** constructNumber -- Return a Scheme number (taggedAval) built from the given string. ****/

taggedAval constructNumber( Text *str, LongWord radix, Boolean failOnErrorP )
{
    Boolean mustBeFloat = false;
    Boolean userInsistsOnFloat = false;
    ShortFloat shortFloatValue = 0.0;
    ShortFloat shortFloatMultiplier, shortRadixNumber, shortFloatNumerator, shortFloatDenominator;
    LongWord i = 0;
    LongWord lnth = 0;
    LongWord lnth2 = 0;
    LongWord longWordMultiplier;
    LongWord longWordValue = 0;
    LongWord nPreDelimiterDigits, nPostDelimiterDigits, nPostDelimiterDigitsIncludingZeros,
             nPreDelimiterSharps, nPostDelimiterSharps;
    Text exactnessChar = ' ';
    Text radixChar;
    Text *p;
    Text digitString[ MAX_NUMBER_SIZE ], *digitStringPointer;
    Text zeroD = getDigitChar( 0 );
    Text exponentMarker = ' ';
	Text scratchNumber[ MAX_NUMBER_SIZE ];

    LongWord radixNumber, exponent;
    Text signChar, delimiterChar;
    Text *prefixStart, *pastPrefix,
         *firstPreDelimiterDigit, *pastPreDelimiterDigits,
         *firstPreDelimiterSharp, *pastPreDelimiterSharps,
         *firstPostDelimiterDigit, *pastPostDelimiterDigits,
         *firstPostDelimiterSharp, *pastLastSharp;

    Boolean exactFractionP = FALSE;
    taggedAval theAnswer;

    localFailOnErrorP = failOnErrorP;
    if( setjmp( BadNumberBuffer ) == 1 ) {
        theAnswer.t = falseBooleanTag;
        theAnswer.u.c =24;
        return theAnswer;
        }
    switch( radix ) {
        case  0: radixChar = ' '; break;   // No radix given -- let the number-parser look for one.
        case  2: radixChar = 'b'; break;
        case  8: radixChar = 'o'; break;
        case 10: radixChar = 'd'; break;
        case 16: radixChar = 'x'; break;
        default: showNumberReset( 259 ); break;
        }

    theNumber = str;

    parseNumber( str, &signChar, &exactnessChar, &radixChar, &radixNumber,
                 &delimiterChar,
                 &prefixStart, &pastPrefix,
                 &firstPreDelimiterDigit, &pastPreDelimiterDigits,
                 &firstPreDelimiterSharp, &pastPreDelimiterSharps,
                 &firstPostDelimiterDigit, &pastPostDelimiterDigits,
                 &firstPostDelimiterSharp, &pastLastSharp,
                 &exponent, &exponentMarker );

    nPostDelimiterDigitsIncludingZeros = pastPostDelimiterDigits - firstPostDelimiterDigit;

    /* Ignore trailing zeros that are to the right of the radix point. */

    if( delimiterChar == RADIX_POINT )
        while( (pastPostDelimiterDigits > firstPostDelimiterDigit) && (*(pastPostDelimiterDigits -1) == zeroD) )
            --pastPostDelimiterDigits;

    nPreDelimiterDigits = pastPreDelimiterDigits - firstPreDelimiterDigit;
    nPreDelimiterSharps = pastPreDelimiterSharps - firstPreDelimiterSharp;
    nPostDelimiterDigits = pastPostDelimiterDigits - firstPostDelimiterDigit;
    nPostDelimiterSharps = pastLastSharp - firstPostDelimiterSharp;

    /* Shall this number be a float or a fixnum? */

    if( exponentMarker != SPACE ) {                // user insists on a specific float representation
        mustBeFloat = TRUE;
        userInsistsOnFloat = TRUE;
        }
    else if( delimiterChar == SLASH )              // fractional format -- store as float
        mustBeFloat = TRUE;
    else if( exponent < nPostDelimiterDigits)      // if not, e.g., 123.45e+02 ==> 12345
        mustBeFloat = TRUE;
    else if( (nPreDelimiterDigits + nPreDelimiterSharps + exponent) < (maxFixDigits[ radixNumber ] - 1) )
        mustBeFloat = FALSE;
    else if( (nPreDelimiterDigits + nPreDelimiterSharps + exponent) > maxFixDigits[ radixNumber ] )
        mustBeFloat = TRUE;
    else {  
        /* (nPreDelimiterDigits + nPreDelimiterSharps + exponent) equals maxFixDigits[ radixNumber ] or is one less */

        // What we do here is actually prepare the (unsigned) digit string and compare it with
        // the largest digit string that is allowed for the given radix.

        digitStringPointer = digitString;
        for( i = 0; i < nPreDelimiterDigits; ++i )
            *digitStringPointer++ = *(firstPreDelimiterDigit + i);
        for( i = 0; i < nPreDelimiterSharps; ++i )
            *digitStringPointer++ = zeroD;
        for( i = 0; i < nPostDelimiterDigits; ++i )
            *digitStringPointer++ = *(firstPostDelimiterDigit + i);
        for( i = 0; i < exponent - nPostDelimiterDigits; ++i )
            *digitStringPointer++ = zeroD;
        *digitStringPointer = '\0';

        lnth = strlen( digitString );

        switch( radixNumber ) {
            case 2:
                if( signChar == PLUS_SIGN )
                    lnth2 = strlen( MAX_BINARY_DIGITS_UNSIGNED );
                else
                    lnth2 = strlen( MAX_BINARY_DIGITS );
                if( lnth < lnth2 )
                    mustBeFloat = FALSE;
                else if( lnth > lnth2 )
                    mustBeFloat = TRUE;
                else
                    if( signChar == PLUS_SIGN )
                        mustBeFloat = ( strcmp( digitString, MAX_BINARY_DIGITS_UNSIGNED ) > 0 );
                    else
                        mustBeFloat = ( strcmp( digitString, MAX_BINARY_DIGITS ) > 0 );
                break;
            case 8:
                if( signChar == PLUS_SIGN )
                    lnth2 = strlen( MAX_OCTAL_DIGITS_UNSIGNED );
                else
                    lnth2 = strlen( MAX_OCTAL_DIGITS );
                if( lnth < lnth2 )
                    mustBeFloat = FALSE;
                else if( lnth > lnth2 )
                    mustBeFloat = TRUE;
                else   
                    if( signChar == PLUS_SIGN )
                        mustBeFloat = ( strcmp( digitString, MAX_OCTAL_DIGITS_UNSIGNED ) > 0 );
                    else
                        mustBeFloat = ( strcmp( digitString, MAX_OCTAL_DIGITS ) > 0 );
                break;
            case 10:
                lnth2 = strlen( MAX_DECIMAL_DIGITS );
                if( lnth < lnth2 )
                    mustBeFloat = FALSE;
                else if( lnth > lnth2 )
                    mustBeFloat = TRUE;
                else
                    mustBeFloat = ( strcmp( digitString, MAX_DECIMAL_DIGITS ) > 0 );
                break;
            case 16:
                if( signChar == PLUS_SIGN )
                    lnth2 = strlen( MAX_HEX_DIGITS_UNSIGNED );
                else
                    lnth2 = strlen( MAX_HEX_DIGITS );
                if( lnth < lnth2 )
                    mustBeFloat = FALSE;
                else if( lnth > lnth2 )
                    mustBeFloat = TRUE;
                else
                    if( signChar == PLUS_SIGN )
                        mustBeFloat = ( strcmp( digitString, MAX_HEX_DIGITS_UNSIGNED ) > 0 );
                    else
                        mustBeFloat = ( strcmp( digitString, MAX_HEX_DIGITS ) > 0 );
                break;
            default:
                fatalError( 18 );
                break;
                }
            }
        
    /* Now calculate the number */

     if( mustBeFloat ) {
        if( radixNumber == 10 ) {
            if( delimiterChar == SLASH ) {     // Input was expressed as a fraction.
                shortFloatNumerator = 0.0;
                shortFloatDenominator = 0.0;
                if( nPreDelimiterDigits == 0 )
                    showNumberReset( 45 );
                if( nPostDelimiterDigits == 0 )
                    showNumberReset( 46 );
                shortFloatDenominator =
                    constructShortFloatWithDigitsPadAndSign( firstPostDelimiterDigit, pastPostDelimiterDigits,
                                                             nPostDelimiterSharps, '+' );
                shortFloatNumerator =
                    constructShortFloatWithDigitsPadAndSign( firstPreDelimiterDigit, pastPreDelimiterDigits,
                                                             nPreDelimiterSharps, signChar );
                if( shortFloatDenominator == 0.0 )
                    showNumberReset( 47 );

                feclearexcept( FE_ALL_EXCEPT );

                shortFloatValue = shortFloatNumerator / shortFloatDenominator;

                if( BigEndianFlag )
                    exactFractionP = FALSE;
                else
                    exactFractionP = (! fetestexcept( FE_INEXACT ));

#ifdef PIXIE_SCHEME_III

#define NO_FP_INEXACT 

#endif

#ifdef NO_FP_INEXACT
                exactFractionP = FALSE;
#endif    

                }
            else {                              // input was not expressed as a fraction.

                // The idea here is to construct something "strtodShortFloat" can read, and use it.

                i = 0;
                scratchNumber[ i++ ] = signChar;     

                //  We just put in the sign; next the actual digits left of the decimal point, and then
                //  zeros in place of any pre-decimal-point sharps.

                strncpy( scratchNumber + i, firstPreDelimiterDigit, (pastPreDelimiterDigits - firstPreDelimiterDigit) );
                i = i + (pastPreDelimiterDigits - firstPreDelimiterDigit);
                lnth = pastPreDelimiterSharps - firstPreDelimiterSharp;
                while( lnth-- > 0 )
                    scratchNumber[ i++ ] = '0';

                // Now the decimal point, and any post-delimiter digits.

                scratchNumber[ i++ ] = RADIX_POINT;
                strncpy( scratchNumber + i, firstPostDelimiterDigit, (pastPostDelimiterDigits - firstPostDelimiterDigit) );
                 i = i + (pastPostDelimiterDigits - firstPostDelimiterDigit);
                
                // Now the exponent, if there is one, and a final 'L'.

                if( exponentMarker != SPACE ) {
                    scratchNumber[ i++ ] = 'e';
                    sprintf( scratchNumber + i, "%jd", (intmax_t)exponent );
                    i = strlen( scratchNumber );
                    }
                scratchNumber[ i++ ] = 'L';
                scratchNumber[ i ] = 0;

                shortFloatValue = strtodShortFloat( scratchNumber, 0 );

                }
            }
        else {                                  // radixNumber is not 10.  Exponent had better be zero!
            shortRadixNumber = (ShortFloat)radixNumber;
            if( delimiterChar == SLASH ) {      // Input was expressed as a fraction.
                shortFloatNumerator = 0.0;
                shortFloatDenominator = 0.0;
                if( nPreDelimiterDigits == 0 )
                    showNumberReset( 45 );
                if( nPostDelimiterDigits == 0 )
                    showNumberReset( 46 );

                // Construct the denominator.

                shortFloatMultiplier = 1.0;
                for( i = 0; i < nPostDelimiterSharps; ++i )
                    shortFloatMultiplier *= shortRadixNumber;
                for( p = pastPostDelimiterDigits; --p >= firstPostDelimiterDigit; ) {
                    shortFloatDenominator += (ShortFloat)translate( *p ) * shortFloatMultiplier;
                    shortFloatMultiplier *= shortRadixNumber;
                    }

                // Construct the numerator.

                shortFloatMultiplier = 1.0;
                for( i = 0; i < nPreDelimiterSharps; ++i )
                    shortFloatMultiplier *= shortRadixNumber;
                for( p = pastPreDelimiterDigits; --p >= firstPreDelimiterDigit; ) {
                    shortFloatNumerator += (ShortFloat)translate( *p ) * shortFloatMultiplier;
                    shortFloatMultiplier *= shortRadixNumber;
                    }
                if( shortFloatDenominator == 0.0 )
                    showNumberReset( 47 );

                // Divide them.

                feclearexcept( FE_ALL_EXCEPT );
                shortFloatValue = shortFloatNumerator / shortFloatDenominator;
                exactFractionP = (! fetestexcept( FE_INEXACT ));
                }
            else {                              // input was not expressed as a fraction
                shortFloatValue = 0.0;
                if( nPostDelimiterDigits > 0 )  // There shouldn't be any fractional part.
                    showNumberReset( 264 );

                // Construct the integer part.

                if( nPreDelimiterDigits > 0 ) {
                    shortFloatMultiplier = 1.0;
                    for( i = 0; i < nPreDelimiterSharps; ++i )
                        shortFloatMultiplier = shortFloatMultiplier * shortRadixNumber;
                    for( p = pastPreDelimiterDigits; --p >= firstPreDelimiterDigit; ) {
                        shortFloatValue = shortFloatValue + ((ShortFloat)translate( *p )) * shortFloatMultiplier;
                        shortFloatMultiplier = shortFloatMultiplier * shortRadixNumber;
                        }
                    }
                }
            if( signChar == MINUS_SIGN )  // Set the sign.
                shortFloatValue = - shortFloatValue;
            }
        }
    else {                  // Fixnum
        longWordValue = 0;
        longWordMultiplier = 1;
        for( i = 0; i < exponent - nPostDelimiterDigits; ++i )
            longWordMultiplier *= radixNumber;
        if( nPostDelimiterDigits > 0 )
            for( p = pastPostDelimiterDigits; --p >= firstPostDelimiterDigit; ) {
                longWordValue += (LongWord)translate( *p ) * longWordMultiplier;
                longWordMultiplier *= radixNumber;
                }
        for( i = 0; i < nPreDelimiterSharps; ++i )
            longWordMultiplier *= radixNumber;
        if( nPreDelimiterDigits > 0 )
            for( p = pastPreDelimiterDigits; --p >= firstPreDelimiterDigit; ) {
                longWordValue += (LongWord)translate( *p ) * longWordMultiplier;
                longWordMultiplier *= radixNumber;
                }
        if( signChar == MINUS_SIGN )
            longWordValue = - longWordValue;
        }

    /* 
       Shall this number be exact or inexact? 
       This code makes it an error to have both an "exact" prefix and sharps in the string.
     */

    if( exactnessChar == INEXACT )                                            //  Fine.
        ;
    else if(    (exactnessChar == EXACT)                                      //  There had better not be any sharps.
             && ((nPreDelimiterSharps > 0) || (nPostDelimiterSharps > 0)))    //  Oops!
            showNumberReset( 267 );
    else if(    (exactnessChar == SPACE)                                      //  No explicit exactness prefix ...
             && (    ((delimiterChar != SPACE) && (delimiterChar != SLASH))   //     ... but the form indicates inexactness.
                  || (exponentMarker != SPACE) 
                  || (nPreDelimiterSharps > 0) || (nPostDelimiterSharps > 0) )  )
            exactnessChar = INEXACT;
    else {

        //  exactnessChar is SPACE or EXACT -- investigate a number of reasons to make it inexact,
        //    and if none of them are valid, make it exact.

        if( mustBeFloat ) {
            if( nanOrInfFloatP( shortFloatValue ) )                            // Nans and infinities are inexact.
                exactnessChar = INEXACT;
            else if( delimiterChar == SLASH )
                exactnessChar = exactFractionP ? EXACT : INEXACT;             // Fraction happens to be exact.
            }
        }
    if( exactnessChar == SPACE )  //  No reason so far to make the number inexact, so make it exact.  (Perhaps is is already.)
        exactnessChar = EXACT;

    /* Now compose the answer */
        
    if( mustBeFloat ) {
        theAnswer.u.f =  shortFloatValue;
        if( exactnessChar == EXACT )
            theAnswer.t = exactShortFloatBitsMask;
         else
            theAnswer.t = inexactShortFloatBitsMask;
        R = theAnswer;
        if( ! userInsistsOnFloat )
            coerceRToShortFixnumIfPossible();
        theAnswer = R;
        }
    else {
        if( exactnessChar == EXACT )
            theAnswer.t = exactShortFixnumBitsMask;
        else
            theAnswer.t = inexactShortFixnumBitsMask;
        theAnswer.u.c = longWordValue;
        }
    return theAnswer;
    }

/**** showNumberReset -- Print out the number, then do the indicated reset. ****/

void showNumberReset( UnsignedLongWord resetNumber )
{
    if( localFailOnErrorP ) {
        tprintf3( &ErrorPort, MiscStrings[ 263 ], theNumber );
        resetError( resetNumber );
        }
    longjmp( BadNumberBuffer, 1 );
    }
