/****************************************************************************************

SECDFloatStuff.c -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006

C source for the functions which the "OpCodes.h" file uses for op codes that have to
do with floating-point operations, for Wraith Scheme.  These include all of the transcendental
functions as well as addition, subtraction, multiplication, division and numeric
type-conversion.

NOTHING in this file should reference any of the macros in "OpCodes.h"

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "WraithHeader.h"
#import "MemoryManager.h"
#import "SECD.h"
#import "Numbers.h"
#import "Registers.h"
#import "Globals.e"
#import "DataTypes.e"
#import "IO.e"
#import "ReaderNumbers.e"
#import "Registers.e"
#import "SECDEval.e"
#import "SECDMachine.e"
#import "SECDRegisters.e"

#ifdef PIXIE_SCHEME_III

#define NO_FP_INEXACT 

#endif

/**** Functions: ****/

/**** Some forward declarations: ****/

void addShortFloats(), subtractShortFloats(), multiplyShortFloats(), divideShortFloats();
void multiplyShortFixnums();
void coerceRToShortFloat(), coerceS0ToShortFloat();
Boolean coerceRToShortFixnumIfPossible();
Boolean exactMinusOneRP(), exactOneRP(), exactZeroRP(), exactZeroS0P();
Boolean subtractShortFixnums();

void exactifyResultForCoercion( Boolean allInputsExactP, Boolean forceExactP );

typedef enum { ALL_SHORT_FIXNUMS, ALL_SHORT_FLOATS } coercionResult;

/****************

  If there is no reliable floating-point inexact exception -- as appears to be the case in the original iPad --
  then we proceed as follows:  We set boolean "isFloatingPointOperation" to FALSE in a preamble function for
  each operation where floating point work is possible, that is, in one of the "coerce" functions.  We then
  set it TRUE if floating point work actually does take place in that operation, and use that value later,
  in an "exactify" function, to make sure to set the Scheme result to inexact.

  All this is controlled by "NO_FP_INEXACT", which is #defined on the basis of "PIXIE_SCHEME_III".

  This approach is also used in "ReaderNumbers.c++", in "reading fractions" -- i.e., in dealing with
  input numbers presented in the form 1/3 -- but is not documented there.

  It is also used in "SECDMachine.c++".

****************/

#ifdef NO_FP_INEXACT
Boolean isFloatingPointOperation = FALSE;
#endif

/**** coerceTwoCanonically -- for floating-point binops -- perform the canonical coercion -- whatever
                              that might mean -- of the operands in R and STACK( 0 ). ****/

coercionResult coerceTwoCanonically( Boolean forceFloat, Boolean *bothExactP )
{
    *bothExactP = (exact_p( R.t ) && exact_p( (S[0]).t ));

    feclearexcept( FE_ALL_EXCEPT );
#ifdef NO_FP_INEXACT
    isFloatingPointOperation = FALSE;
#endif

    //  If either argument is a short float, coerce the other to be one as well.

    if( short_float_p( R.t ) ) {
        if( ! short_float_p( (S[0]).t ) )
            coerceS0ToShortFloat();
        return ALL_SHORT_FLOATS;
       }
    if( short_float_p( (S[0]).t ) ) {
        coerceRToShortFloat();
        return ALL_SHORT_FLOATS;
        }

    //  Now we know that both arguments must be short fixnum.
    //  If we are told to "force floats", coerce both arguments to short float.

    if( forceFloat ) {
        coerceRToShortFloat();
        coerceS0ToShortFloat();
        return ALL_SHORT_FLOATS;
        }

    // Otherwise, short fixnums are okay.

    return ALL_SHORT_FIXNUMS;
    }

/**** coerceOneCanonically -- for floating-point monops -- perform the canonical coercion -- whatever
                              that might mean -- of the operand in R. ****/

coercionResult coerceOneCanonically( Boolean forceFloat, Boolean *oneExactP )
{
    *oneExactP = exact_p( R.t );

    feclearexcept( FE_ALL_EXCEPT );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = FALSE;
#endif

    if( short_float_p( R.t ) )
        return ALL_SHORT_FLOATS;
    if( forceFloat ) {
        coerceRToShortFloat();
        return ALL_SHORT_FLOATS;
        }

    // Otherwise, short fixnum is okay.

    return ALL_SHORT_FIXNUMS;
    }

/**** exactifyResult ****/

void exactifyResult( Boolean allInputsExactP, Boolean forceExactP ) 
{
    ShortFloat value;

    if( forceExactP ) {  //  The caller has some reason to know the result should be exact.
        makeExact( R.t );
        return;
        }

    //  Now do the general case; override later if we have a nan or an inf.
    //  That is to keep the nan-or-inf tests from possibly messing up the result of "fetestexcept".

#ifdef NO_FP_INEXACT

    if( (! isFloatingPointOperation) && allInputsExactP )
        makeExact( R.t );
    else
        makeInexact( R.t );

//    if( isFloatingPointOperation || (! allInputsExact) )
//        makeInexact( R.t );
//    else
//        makeExact( R.t );

#else

    if( allInputsExactP &&(! fetestexcept( FE_INEXACT )) )
        makeExact( R.t );
    else
        makeInexact( R.t );

#endif

    //  Here come the overrides.

    if( short_float_p( R.t ) ) {
        value = R.u.f;
        if( fabsShortFloat( value ) > MAX_SHORT_FLOAT_EXACT_INTEGER ) {
            makeInexact( R.t );
            return;
            }
        }
    if( short_float_p( R.t ) && nanOrInfShortFloatP( R.u.f ) ) {
        makeInexact( R.t );
        return;
        }
    }

/**** exactifyResultForCoercion ****/

void exactifyResultForCoercion( Boolean allInputsExactP, Boolean forceExactP ) 
{
    ShortFloat value;

    if( forceExactP ) {  //  The caller has some reason to know the result should be exact.
        makeExact( R.t );
        return;
        }

    //  Now do the general case; override later if we have a nan or an inf.

    if( ! allInputsExactP ) { 
        makeInexact( R.t );
        return;
        }
    else
        makeExact( R.t );    //  And do not return -- must check overrides.

    if( short_float_p( R.t ) ) {
        value = R.u.f;
        if( fabsShortFloat( value ) >= MAX_SHORT_FLOAT_EXACT_INTEGER ) {
            makeInexact( R.t );
            return;
            }
        }
    if( short_float_p( R.t ) && nanOrInfShortFloatP( R.u.f ) ) {
        makeInexact( R.t );
        return;
        }
    }

/**** acosMC -- Replace R with acos(R). ****/

void acosMC()
{
    Boolean exactP;
    Boolean exactMinusOneInputP = exactMinusOneRP();

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    if( (R.u.f <= 1.0) && (R.u.f >= -1.0) )
        R.u.f = acosShortFloat( R.u.f );
    else {
        useErrorOutput();
        displayObject();
        resetError( 159 );
        }
    exactifyResult( (exactP && (! exactMinusOneInputP)), FALSE );
    }

/**** add --            R <-- STACK(0) + R, and pop the stack.  ****/

void add()
{
    Boolean bothExactP;
    Boolean sameSignP = false;
    LongWord scratch = 0;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            addShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            sameSignP = ((R.u.c >= 0) && ((S[0]).u.c >= 0)) || ((R.u.c <= 0) && ((S[0]).u.c <= 0));
            if( sameSignP ) {
                scratch = R.u.c + (S[0]).u.c;
                if( ((R.u.c >= 0) && ((S[0]).u.c >= 0) && (scratch >= 0)) || ((R.u.c <= 0) && ((S[0]).u.c <= 0) && (scratch <= 0)) ) {
                    R.u.c = scratch;
                    SDrop();
                    }
                else {     // Must coerce to floating-point.
                    coerceRToShortFloat();
                    coerceS0ToShortFloat();
                    addShortFloats();
                    coerceRToShortFixnumIfPossible();
                    }
                }
            else {
                R.u.c += (S[0]).u.c;
                SDrop();
                }
            break;
        default:
            fatalError( 60 );
            break;
        }
    exactifyResult( bothExactP, FALSE );
    }

/**** addShortFloats    R <-- STACK(0) + R, and pop the stack.  ****/

void addShortFloats()
{

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f += (S[0]).u.f;
    SDrop();
    }

/**** asinMC -- Replace R with asin(R). ****/

void asinMC()
{
    Boolean exactP;
    Boolean exactZeroInputP = exactZeroRP();
    Boolean exactOneInputP = exactOneRP();
    Boolean exactMinusOneInputP = exactMinusOneRP();

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    if( (R.u.f <= 1.0) && (R.u.f >= -1.0) )
        R.u.f = asinShortFloat( R.u.f );
    else {
        useErrorOutput();
        displayObject();
        resetError( 160 );
        }
    exactifyResult( (exactP && (! exactOneInputP) && (! exactMinusOneInputP)), exactZeroInputP );
    }

/**** atanMC -- Replace R with atan(R). ****/

void atanMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = atanShortFloat( R.u.f ); 
    exactifyResult( exactP, FALSE );
    }

/**** atan2MC -- Replace R with atan2( STACK( 0 ), R ). ****/

void atan2MC()
{
    Boolean exactP;

    coerceTwoCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = atan2ShortFloat( (S[0]).u.f, R.u.f );
    exactifyResult( exactP, FALSE );
    }

/**** ceilingMC -- Replace R with ceiling(R).  R is assumed to be some kind of float.  ****/

void ceilingMC()
{
    LongWord theLong;
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = ceilShortFloat( R.u.f );
    if( R.u.f > MAX_FIXNUM_AS_FLOAT || R.u.f < MIN_FIXNUM_AS_FLOAT )
        ;  // Nothing to do.
    else{
        theLong = R.u.f;
        R.u.c = theLong;
        R.t = shortFixnumBitsMask;
        }
    exactifyResultForCoercion( exactP, FALSE );
    }

/**** coerceToShortFloat ****/

ShortFloat coerceToShortFloat( taggedAval *x )
{
    LongWord theInt;

    if( ! float_p( x->t ) ) {
        theInt = ((x->u).c);
        return (ShortFloat)theInt;
        }
    else
        return (ShortFloat)((x->u).f);
    }

/**** coerceRToShortFloat ****/

void coerceRToShortFloat()
{
    Tag savedExactness = R.t & exactBitMask;

    if( ! float_p( R.t ) )
        R.u.f = (ShortFloat)(R.u.c);

    // else nothing

    R.t = savedExactness | shortFloatBitsMask;
    }

/**** coerceS0ToShortFloat ****/

void coerceS0ToShortFloat()
{
    Tag savedExactness = (S[0]).t & exactBitMask;

    if( ! float_p( (S[0]).t ) )
        (S[0]).u.f = (ShortFloat)((S[0]).u.c);

    // else nothing

    (S[0]).t = savedExactness | shortFloatBitsMask;
    }

/**** coerceRToShortFixnum ****/

void coerceRToShortFixnum()
{
    ShortFloat integerPart;
    ShortFloat theShortFloat;
    LongWord theLongWord;

    if( ! fixnum_p( R.t ) ) {
        if( long_p( R.t ) ) {
            theShortFloat = R.u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT ) {
                useErrorOutput();
                displayObject();
                resetError( 161 );
                }
            (void)modfShortFloat(theShortFloat, &integerPart );
            theLongWord = (LongWord)integerPart;
            R.u.c = theLongWord;
            R.t = (R.t & exactBitMask) | shortFixnumBitsMask;
            }
        else {
            theShortFloat = R.u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT ) {
                useErrorOutput();
                displayObject();
                resetError( 161 );
                }
            (void)modfShortFloat( (ShortFloat)theShortFloat, &integerPart );
            theLongWord = (LongWord)integerPart;
            R.u.c = theLongWord;
            R.t = (R.t & exactBitMask) | shortFixnumBitsMask;
            }
        }
    }

/**** coerceRToShortFixnumIfPossible ****/

//  That this function does what I intend depends on "modfl" not setting any of the
//  floating-point exception flags.  These are generally tested herein, after calling
//  this function, and any changes to them by this function will confuse the results
//  of those tests.

Boolean coerceRToShortFixnumIfPossible()
{
    ShortFloat integerPart;
    ShortFloat theShortFloat;
    LongWord theLongWord;
    Tag savedExactness = R.t & exactBitMask;

    if( ! fixnum_p( R.t ) ) {
        if( long_p( R.t ) ) {
            theShortFloat = R.u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT )   // Can't.  Leave R unchanged.
                return FALSE;
            if( modfShortFloat(theShortFloat, &integerPart ) == 0.0 ) {
                theLongWord = (LongWord)integerPart;
                R.u.c = theLongWord;
                R.t = shortFixnumBitsMask | savedExactness;
                return TRUE;
                }
            else
               return FALSE;
            }
        else {
            theShortFloat = R.u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT )  // Can't.  Leave R unchanged.
                return FALSE;
            if( modfShortFloat( (ShortFloat)theShortFloat, &integerPart ) == 0.0 ) {
                theLongWord = (LongWord)integerPart;
                R.u.c = theLongWord;
                R.t = shortFixnumBitsMask | savedExactness;
                return TRUE;
                }
            else
                return FALSE;
            }
        }
    return TRUE;
    }

/**** coerceS0ToShortFixnum ****/

void coerceS0ToShortFixnum()
{
    ShortFloat integerPart;
    ShortFloat theShortFloat;
    LongWord theLongWord;

    if( ! fixnum_p( (S[0]).t ) ) {
        if( long_p( (S[0]).t ) ) {
            theShortFloat = (S[0]).u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT ) {
                SPopR();
                useErrorOutput();
                displayObject();
                resetError( 161 );
                }
            (void)modfShortFloat(theShortFloat, &integerPart );
            theLongWord = (LongWord)integerPart;
            (S[0]).u.c = theLongWord;
            (S[0]).t = ((S[0]).t & exactBitMask) | shortFixnumBitsMask;
            }
        else {
            theShortFloat = (S[0]).u.f;
            if( theShortFloat > MAX_FIXNUM_AS_FLOAT || theShortFloat < MIN_FIXNUM_AS_FLOAT ) {
                SPopR();
                useErrorOutput();
                displayObject();
                resetError( 161 );
                }
            (void)modfShortFloat( (ShortFloat)theShortFloat, &integerPart );
            theLongWord = (LongWord)integerPart;
            (S[0]).u.c = theLongWord;
            (S[0]).t = ((S[0]).t & exactBitMask) | shortFixnumBitsMask;
            }
        }
    }

/**** cosMC -- Replace R with cos(R). ****/

void cosMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = cosShortFloat( R.u.f );
    exactifyResult( exactP, FALSE );
    }

/**** divide --                 R <-- STACK(0) / R, and pop the stack.  ****/

void divide()
{
    Boolean exactZeroInputP = exactZeroS0P();
    Boolean bothExactP;    
    LongWord scratch = 0;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            divideShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            if( R.u.c == 0 ) {
                coerceRToShortFloat();
                coerceS0ToShortFloat();
                divideShortFloats();
                }
            else {
                scratch = (S[0]).u.c / R.u.c;
                if( scratch * R.u.c == (S[0]).u.c ) {
                    R.u.c = scratch;
                    SDrop();
                    }
                else {
                    coerceRToShortFloat();
                    coerceS0ToShortFloat();
                    divideShortFloats();
                    coerceRToShortFixnumIfPossible();
                    }
                }
            if( short_float_p( R.t ) && nanOrInfShortFloatP( R.u.f ) )
                makeInexact( R.t );
            break;
        default:
            fatalError( 75 );
            break;
        }
    exactifyResult( bothExactP, exactZeroInputP );
    }

/**** divideShortFloats R <-- STACK(0) / R, and pop the stack.  ****/

void divideShortFloats()
{

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = (S[0]).u.f / R.u.f;
    SDrop();
    }

/**** expMC -- Replace R with exp(R). ****/

void expMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = expShortFloat( R.u.f );
    exactifyResult( exactP, FALSE );
    }

/**** exptMC -- Replace R with STACK(0) raised to the power R. ****/

void exptMC()
{
    Boolean exactP;

    coerceTwoCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = powShortFloat( (S[0]).u.f, R.u.f );
    exactifyResult( exactP, FALSE );
    }

/**** exactMinusOneRP ****/

Boolean exactMinusOneRP()
{
    if( exact_p( R.t ) ) {
        if( short_float_p( R.t ) && (R.u.f == -1.0) )
            return TRUE;
        else if( ( R.u.c ) == -1 ) 
            return TRUE;
        else
            return FALSE;
        }
    else
        return FALSE;
    }

/**** exactOneRP ****/

Boolean exactOneRP()
{
    if( exact_p( R.t ) ) {
        if( short_float_p( R.t ) && (R.u.f == 1.0) )
            return TRUE;
        else if( ( R.u.c ) == 1 ) 
            return TRUE;
        else
            return FALSE;
        }
    else
        return FALSE;
    }

/**** exactZeroRP ****/

Boolean exactZeroRP()
{
    if( exact_p( R.t ) ) {
        if( short_float_p( R.t ) && (R.u.f == 0.0) )
            return TRUE;
        else if( ( R.u.c ) == 0 ) 
            return TRUE;
        else
            return FALSE;
        }
    else
        return FALSE;
    }

/**** exactZeroS0P ****/

Boolean exactZeroS0P()
{
    if( exact_p( (S[0]).t ) ) {
        if( short_float_p( (S[0]).t ) && ((S[0]).u.f == 0.0) )
            return TRUE;
        else if( ( (S[0]).u.c ) == 0 ) 
            return TRUE;
        else
            return FALSE;
        }
    else
        return FALSE;
    }

/**** floorMC -- Replace R with floor(R). ****/

void floorMC()
{
    LongWord theLong;
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = floorShortFloat( R.u.f );
    if( R.u.f > MAX_FIXNUM_AS_FLOAT || R.u.f < MIN_FIXNUM_AS_FLOAT )
        ;  // Nothing to do.
    else{
        theLong = R.u.f;
        R.u.c = theLong;
        R.t = shortFixnumBitsMask;
        }
    exactifyResultForCoercion( exactP, FALSE );
    }

/**** gcd1 -- Calculate the GCD of the passed nonnegative LongWords. ****/

LongWord gcd1( LongWord M, LongWord N )
{
    LongWord temp;

    while( N != 0 ) {
        temp = N;
        N = M % N;
        M = temp;
        }
    return M;
    }

/**** gcdShortFloats -- Calculate the GCD of the passed nonnegative ShortFloats. ****/

ShortFloat gcdShortFloats( ShortFloat M, ShortFloat N )
{
    ShortFloat temp;

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    while( N >= (ShortFloat)(1.0) ) {
        temp = N;
        N = fmodShortFloat( M, N );
        M = temp;
        }
    return M;
    }

/**** gcdMC -- Return the GCD of the objects in R and in STACK( 0 ).  Pop the stack.  ****/

void gcdMC()
{
    LongWord M, N;
    ShortFloat fM, fN;
    Boolean bothExactP;
 
    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            fM = R.u.f;
            fN = (S[0]).u.f;
            R.u.f = gcdShortFloats( fabsShortFloat( fM ), fabsShortFloat( fN ) );
            break;
        case ALL_SHORT_FIXNUMS:
            M = (R.u.c);
            N = ((S[0]).u.c);
            R.u.c = gcd1( (intmax_t)llabs( M ), (intmax_t)llabs( N ) );
            R.t = shortFixnumBitsMask;
            break;
        default:
            resetError( 76 );
            break;
        }
    exactifyResult( bothExactP, FALSE );
    SDrop();
    }

/**** infPMC -- Is the object in R an INF? ****/

void infPMC()
{
    if( short_float_p( R.t ) )
        B = infShortFloatP( R.u.f );
    else
        B = FALSE;
    }

/**** lcm1 -- Calculate the LCM of the passed nonnegative LongWords. ****/

LongWord lcm1( LongWord M, LongWord N ) 
{
    ShortFloat fAnswer;
    LongWord g;

    if( (M == 0) || (N == 0) )
        return 0;

    g = gcd1( M, N );
    fAnswer = (((ShortFloat)M) / (ShortFloat)g) * ((ShortFloat)N);
    if( fAnswer > MAX_FIXNUM_AS_FLOAT || fAnswer < MIN_FIXNUM_AS_FLOAT )
        resetError( 224 );
    return (M / g) * N;
    }       

/**** lcmShortFloats -- Calculate the LCM of the passed nonnegative ShortFloats. ****/

ShortFloat lcmShortFloats( ShortFloat M, ShortFloat N )
{
    ShortFloat answer, g, quotient;

    if( (M == 0.0) || (N == 0.0) )
        return (ShortFloat)0;

    g = gcdShortFloats( M, N );
    quotient = M / g;
    answer = quotient * N;
    if( fabsShortFloat( answer ) >= MAX_SHORT_FLOAT_EXACT_INTEGER )
        resetError( 224 );
    return answer;
    }       

/**** lcmMC -- Return the LCM of the objects in R and in STACK( 0 ).  Pop the stack.  ****/

void lcmMC()
{
    LongWord M, N;
    ShortFloat fM, fN;
    Boolean bothExactP;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            fM = R.u.f;
            fN = (S[0]).u.f;
            R.u.f = lcmShortFloats( fabs( fM ), fabs( fN ) );
            break;
        case ALL_SHORT_FIXNUMS:
            M = (R.u.c);
            N = ((S[0]).u.c);
            R.u.c = lcm1( fabsl( M ), fabsl( N ) );
            R.t = shortFixnumBitsMask;
            break;
        default:
            resetError( 77 ); 
            break;
        }
    exactifyResult( bothExactP, FALSE );
    SDrop();
    }

/**** logMC -- Replace R with log(R). ****/

void logMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
        isFloatingPointOperation = TRUE;
#endif

    if( R.u.f > 0.0 )
        R.u.f = logShortFloat( R.u.f );
    else {
        useErrorOutput();
        displayObject();
        resetError( 163 );
        }
    exactifyResult( exactP, FALSE );
    }

/**** multiply --               R <-- STACK(0) * R, and pop the stack.  ****/

void multiply()
{
    Boolean bothExactP;
    Boolean oneWasExactZeroP = exactZeroRP() || exactZeroS0P();
    ShortFloat scratchR;
    ShortFloat scratchS0;
    ShortFloat scratchResult;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            multiplyShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            scratchR  = (ShortFloat)(R.u.c);            
            scratchS0 = (ShortFloat)((S[0]).u.c);            
            scratchResult = scratchR * scratchS0;
            if( (scratchResult > MAX_FIXNUM_AS_FLOAT) || (scratchResult < MIN_FIXNUM_AS_FLOAT) ) {

#ifdef NO_FP_INEXACT
                isFloatingPointOperation = TRUE;
#endif

                R.u.f = scratchResult;
                R.t = shortFloatBitsMask;
                SDrop();
                }
            else {
                feclearexcept( FE_ALL_EXCEPT );

#ifdef NO_FP_INEXACT
                isFloatingPointOperation = FALSE;
#endif

                R.u.c *= (S[0]).u.c;
                SDrop();
                }
            break;
        default:
            fatalError( 78 );
            break;
        }
    // Multiplication by an exact zero yields an exact zero, UNLESS
    //   the result was a nan or an inf, which "exactifyResult" checks for.

    exactifyResult( (bothExactP || oneWasExactZeroP), FALSE );
    }

/**** multiplyShortFloats       R <-- STACK(0) * R, and pop the stack.  ****/

void multiplyShortFloats()
{

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = R.u.f * (S[0]).u.f;
    SDrop();
    }

/**** multiplyShortFixnums      R <-- STACK(0) * R, and pop the stack.  ****/

void multiplyShortFixnums()
{
    LongWord answer, first, second;
    taggedAval firstArg;

    feclearexcept( FE_ALL_EXCEPT );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = FALSE;
#endif

    firstArg = S[0];
    first = (firstArg.u.c);
    second = (R.u.c);
    answer = first * second;
    if( FALSE ) {                                           /* need overflow test */
        R.t &= ~exactBitMask;
        coerceRToShortFloat();
        coerceS0ToShortFloat();
        multiplyShortFloats();
        }
    else {
        R.u.c = answer;
        R.t = shortFixnumBitsMask;
        if( ! exact_p( firstArg.t ) )
            R.t &= ~exactBitMask;
        SDrop();
        }
    }

/**** nanPMC -- Is the object in R a NAN? ****/

void nanPMC()
{
    if( short_float_p( R.t ) )
        B = nanShortFloatP( R.u.f );
    else
        B = FALSE;
    }

/**** nanOrInfPMC -- Is the object in R a NAN or an INF? ****/

void nanOrInfPMC()
{
    if( short_float_p( R.t ) ) 
        B = nanOrInfShortFloatP( R.u.f );
    else
        B = FALSE;
    }

/**** quotientShortFloats --             R <-- STACK(0) div R    ****/

void quotientShortFloats()
{
    ShortFloat scratch;

    if( R.u.f == 0.0 ) {
        R = S[0];
        useErrorOutput();
        displayObject();
        resetError( 162 );
        }

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    scratch = (S[0]).u.f / R.u.f;
    if( scratch >= 0.0 )
        R.u.f = floorShortFloat( scratch );
    else
        R.u.f = ceilShortFloat( scratch );
    }

/**** quotient2 --              R <-- STACK(0) div R.  Pop the stack.   ****/

void quotient2()
{
    Boolean bothExactP;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            quotientShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            quotient1();
            break;
        default:       
            fatalError( 87 );
            break;
        }
    exactifyResult( bothExactP, FALSE );
    SDrop();
    }

/**** rangeCheckAsUnsignedFixnum -- Expects a number! ****/

Boolean rangeCheckAsUnsignedFixnum( taggedAval num )
{
    if( fixnum_p( num.t ) )
        return TRUE;
    else {          /* short float */
        if( floorShortFloat( num.u.f ) != num.u.f )
            return FALSE;
        else
            return( (num.u.f >= 0.0) && (fabsShortFloat( num.u.f ) <= MAX_SHORT_FLOAT_EXACT_INTEGER) );
        }
    }

/**** rangeCheckAsSignedFixnumMC ****/

void rangeCheckAsSignedFixnumMC()
{
    if( number_p( R.t ) ) {
        if( fixnum_p( R.t ) )
            B = (R.u.c <= MAX_FIXNUM_DECIMAL) && (R.u.c >= MIN_FIXNUM_DECIMAL);
        else {          /* short float */
            if( floorShortFloat( R.u.f ) != R.u.f )
                B = FALSE;
            else
                B = fabsShortFloat( R.u.f ) <= MAX_SHORT_FLOAT_EXACT_INTEGER;
            }
        }
    else
        B = FALSE;
    }

/**** rangeCheckAsUnsignedFixnumMC ****/

void rangeCheckAsUnsignedFixnumMC()
{
    if( number_p( R.t ) ) {
        if( fixnum_p( R.t ) )
            B = TRUE;
        else {          /* short float */
            if( floorShortFloat( R.u.f ) != R.u.f )
                B = FALSE;
            else
                B = (R.u.f >= 0.0) && (fabsShortFloat( R.u.f ) <= MAX_SHORT_FLOAT_EXACT_INTEGER);
            }
        }
    else
        B = FALSE;
    }

/**** remainderShortFloats --    R <-- remainder( STACK(0) / R )         ****/

void remainderShortFloats()
{
    if( R.u.f == 0.0 ) {
        R = S[0];
        useErrorOutput();
        displayObject();
        resetError( 94 );
        }

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = fmod( (S[0]).u.f, R.u.f );
    }

/**** remainder2 --             R <-- remainder( STACK(0) / R ).  Pop the stack.        ****/

void remainder2()
{
    Boolean bothExactP;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            remainderShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            remainder1();
            break;
        default:
            fatalError( 79 );
            break;
        }
    exactifyResult( bothExactP, FALSE );
    SDrop();
    }

/**** roundMC -- Replace R with round(R). ****/

void roundMC()
{
    ShortFloat integerPart, fraction;
    LongWord theLong;
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    fraction = modfShortFloat( R.u.f, &integerPart );
    if( (integerPart >= 0.0) && (fraction > 0.5) )
        R.u.f = integerPart + 1.0;    
    else if( (integerPart <= 0.0) && (fraction < -0.5) )
        R.u.f = integerPart - 1.0;
    else if( fabsShortFloat(fraction) == 0.5 ) {
        if( fmodShortFloat( integerPart, 2.0 ) == 0.0 )
            R.u.f = integerPart;
        else if( integerPart >= 0.0 )
            R.u.f = integerPart + 1.0;
        else
            R.u.f = integerPart - 1.0;
        }
    else
        R.u.f = integerPart;

    if( R.u.f > MAX_FIXNUM_AS_FLOAT || R.u.f < MIN_FIXNUM_AS_FLOAT )
        ;    // All done.
    else{
        theLong = R.u.f;
        R.u.c = theLong;
        R.t = shortFixnumBitsMask;
        }
    exactifyResultForCoercion( exactP, FALSE );
    }

/**** sinMC -- Replace R with sin(R). ****/

void sinMC()
{
    Boolean exactP;
    Boolean exactZeroInput = exactZeroRP();

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = sinShortFloat( R.u.f );
    exactifyResult( exactP, exactZeroInput );
    }

/**** sqrtMC -- Replace R with sqrt(R). ****/

void sqrtMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    if( R.u.f >= 0.0 )
        R.u.f = sqrtShortFloat( R.u.f );
    else {
        useErrorOutput();
        displayObject();
        resetError( 164 );
        }
    exactifyResult( exactP, FALSE );
    }

/**** subtract --               R <-- STACK(0) - R, and pop the stack.  ****/

void subtract()
{
    Boolean bothExactP;
    Boolean sameSignP = false;
    LongWord scratch = 0;

    switch( coerceTwoCanonically( FALSE, &bothExactP ) ) {
        case ALL_SHORT_FLOATS:
            subtractShortFloats();
            break;
        case ALL_SHORT_FIXNUMS:
            sameSignP = ((R.u.c <= 0) && ((S[0]).u.c >= 0)) || ((R.u.c >= 0) && ((S[0]).u.c <= 0));
            if( sameSignP ) {
                scratch = (S[0]).u.c - R.u.c;
                if( ((R.u.c <= 0) && ((S[0]).u.c >= 0) && (scratch >= 0)) || ((R.u.c >= 0) && ((S[0]).u.c <= 0) && (scratch <= 0)) ) {
                    R.u.c = scratch;
                    SDrop();
                    }
                else {     // Must coerce to floating-point.
                    coerceRToShortFloat();
                    coerceS0ToShortFloat();
                    subtractShortFloats();
                    coerceRToShortFixnumIfPossible();
                    }
                }
            else {
                R.u.c = (S[0]).u.c - R.u.c;
                SDrop();
                }
            break;
        default:
            fatalError( 80 );
            break;
        }
    exactifyResult( bothExactP, FALSE );
    }

/**** subtractShortFloats       R <-- STACK(0) - R, and pop the stack.  ****/

void subtractShortFloats()
{

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = (S[0]).u.f - R.u.f;
    SDrop();
    }

/**** subtractShortFixnums      R <-- STACK(0) - R, and pop the stack.  ****/

Boolean subtractShortFixnums()
{
    LongWord answer, first, second;
    taggedAval firstArg;
    Boolean overflowed = FALSE;

    firstArg = S[0];
    first = (firstArg.u.c);
    second = (R.u.c);
    answer = first - second;
    if(    ((first > 0) && (second < 0) && (answer < 0))
        || ((first < 0) && (second > 0) && (answer > 0))
        || ((unsigned)answer == 0x80000000) ) 
      {                               /* overflow */
        overflowed = TRUE;
        coerceRToShortFloat();
        coerceS0ToShortFloat(); 
        R.u.f = (S[0]).u.f - R.u.f;
        }
    else {
        R.t = shortFixnumBitsMask;
        R.u.c = answer;
        }
    SDrop();
    return overflowed;
    }

/**** tanMC -- Replace R with tan(R). ****/

void tanMC()
{
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    R.u.f = tanShortFloat( R.u.f );
    exactifyResult( exactP, FALSE );
    }

/**** truncateMC -- Replace R with truncate(R). ****/

void truncateMC()
{
    ShortFloat integerPart;
    LongWord theLong;
    Boolean exactP;

    coerceOneCanonically( TRUE, &exactP );

#ifdef NO_FP_INEXACT
    isFloatingPointOperation = TRUE;
#endif

    (void)modfShortFloat( R.u.f, &integerPart );
    R.u.f = integerPart;
    if( R.u.f > MAX_FIXNUM_AS_FLOAT || R.u.f < MIN_FIXNUM_AS_FLOAT )
        ;  //  Nothing too d.
    else{
        theLong = R.u.f;
        R.u.c = theLong;
        R.t = inexactShortFixnumBitsMask;
        }
    exactifyResultForCoercion( exactP, FALSE );
    }          

/**** Dummy functions used in NumberFormats.c++, for dealing with bug in ARM code-generation  ****/

Text stupidBuffer[ 100 ];

void stupidDummyFunction( taggedAval *port, const Text *dummy, LongWord n )
{
#ifdef PIXIE_SCHEME_III
    sprintf( stupidBuffer, "%lld %s", n, dummy );
#endif
    }

