/****************************************************************************************

SECDRegisters.c -- Copyright (c) by Jay Reynolds Freeman 1988, 1989, 1990, 1991, 2006, 2008

C source for the functions which the "OpCodes.h" file uses to reference registers, for Wraith
Scheme.  NOTHING in this file should reference any of the macros in "OpCodes.h"

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact information:

        EMail:       Jay_Reynolds_Freeman@mac.com
        Web Site:    http://web.mac.com/Jay_Reynolds_Freeman

        Paper Mail:  Jay Reynolds Freeman
                     Post Office Box 60628
                     Palo Alto, CA, 94306-0628

****************************************************************************************/

#import "LocalTypes.h"
#import "WraithHeader.h"
#import "MemoryManager.h"

#import "SECD.h"
#import "Globals.e"
#import "DataTypes.e"
#import "IO.e"
#import "SECDEval.e"
#import "SECDFloatStuff.e"
#import "SECDMachine.e"
#import "Registers.h"

#define SNIFF_POSITION 300

/**** Statics: ****/

/* R for result, A for scratch, F is frame pointer, L is for the current lambda (if any). */
                
taggedAval *S = (taggedAval *)0xFFFFFFFF;

/* Non-simple registers, defined elsewhere, included here commented out as a reminder ... */

// taggedAval IVR;             // The interrupt vector register.
// taggedAval TopLevelE;       // The top-level environment.
// taggedAval ObList;          // The oblist
// taggedAval TLC;             // For the code that runs the top-level loop.

/* Simple registers */

taggedAval E;               // The entire environment -- the "E" register.
taggedAval R;               // The results register, for returned values, and often the logical top of the stack.
taggedAval A;               // The auxiliary register, used in address allocation, function calling, and elsewhere.
taggedAval SavedE;          // For saving E.  Used in macro expansion.
taggedAval *F;              // Points into the current frame, at the location where F is stored.

taggedAval L;               //  Stores the current lambda expression, if any -- saved in frame for backtrace.

taggedAval CurrentInputPort;
taggedAval CurrentOutputPort;
taggedAval ErrorPort;

taggedAval DefaultInputPort;
taggedAval MasterDefaultInputPort;
taggedAval TranscriptPort;

taggedAval *CS, *CF;        // C stack pointer and frame pointer.
taggedAval NArgs;           // For the evaluation engine.
taggedAval ArgCounter;      // For the evaluation engine.
Text *FrameName = (Text *)NULL;

Boolean B;                  // Place to put Booleans returned by microcode primitives 

UnsignedLongWord ESR;       // Evaluation status register.

//  Here come some "registers" that are not Scheme-machine registers, but more like
//  I/O status registers.  They are for storing current values of what is in the
//  various sensory devices.

UnsignedLongWord senseSwitchPositions = 0;

UnsignedLongWord sliderValues[ N_SLIDERS ] 
    = { SLIDER_INITIAL_VALUE, SLIDER_INITIAL_VALUE, SLIDER_INITIAL_VALUE, SLIDER_INITIAL_VALUE };

UnsignedLongWord pushbuttonInterruptNumbers[ N_PUSHBUTTONS ] = { 0, 0, 0, 0 };
Boolean pushbuttonEnabled[ N_PUSHBUTTONS ] = { FALSE, FALSE, FALSE, FALSE };

UnsignedLongWord bigRedButtonInterruptNumber = 0;
Boolean bigRedButtonEnabled = FALSE;

/******** End of statics ********/

/**** Functions: ****/

/**** Some forward declarations: ****/

void compareContinuations( Text *msg, Boolean forceP );
void describeCurrentCFrame( Text *msg );
void compareFirstContinuationObjects( Text *msg );

void displayFrame( Text *msg )
{
    tprintf3( &CurrentOutputPort, MiscStrings[ 504 ], msg );
    tprintf2( &CurrentOutputPort, "S:          ");
    printAnyTaggedAval( &CurrentOutputPort, S );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "F:          ");
    printAnyTaggedAval( &CurrentOutputPort, F );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "\n");

    tprintf2( &CurrentOutputPort, "E:          ");
    printAnyTaggedAval( &CurrentOutputPort, F + FRAME_E_OFFSET );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "FP:         ");
    printAnyTaggedAval( &CurrentOutputPort, F + FRAME_FP_OFFSET );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "CF:         ");
    printAnyTaggedAval( &CurrentOutputPort, F + FRAME_CF_OFFSET );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "magic:      ");
    printAnyTaggedAval( &CurrentOutputPort, F + FRAME_NEW_MN_OFFSET );
    tprintf2( &CurrentOutputPort, "\n");
    tprintf2( &CurrentOutputPort, "L:          ");
    printAnyTaggedAval( &CurrentOutputPort, F + FRAME_NEW_L_OFFSET );
    tprintf2( &CurrentOutputPort, "\n");
    }

/**** initializeMachine ****/

void initializeMachine()
{
    UnsignedLongWord theWord = 0;

    // Non-simple registers are initialized during clowder initialization.

//    IVR.t = emptyListTag;
//    TopLevelE.t = emptyListTag;
//    ObList.t = emptyListTag;
//    TLC.t = emptyListTag;

    E = TopLevelE;
    SavedE.t = emptyListTag;
    F  =  S = (taggedAval *)(theHeap->schemeStackEnd());
    CF = CS = (taggedAval *)(theHeap->schemeContinuationEnd());
    NArgs = ArgCounter = zero;
    R.t = emptyListTag;
    A.t = emptyListTag;
    L.t = emptyListTag;

    CurrentInputPort.t = portTag;
    theWord = 0;
    orIntoPortNumberByteNonAtomic( theWord, coninIndex );
    orIntoPortKittenByteNonAtomic( theWord, theKittenNumber );
    orIntoKittenByteNonAtomic( theWord, theKittenNumber );
    CurrentInputPort.u.c = theWord;

    CurrentOutputPort.t = portTag;
    theWord = 0;
    orIntoPortNumberByteNonAtomic( theWord, conoutIndex );
    orIntoPortKittenByteNonAtomic( theWord, theKittenNumber );
    orIntoKittenByteNonAtomic( theWord, theKittenNumber );
    CurrentOutputPort.u.c = theWord;

    ErrorPort.t = portTag;
    theWord = 0;
    orIntoPortNumberByteNonAtomic( theWord, conerrIndex );
    orIntoPortKittenByteNonAtomic( theWord, theKittenNumber );
    orIntoKittenByteNonAtomic( theWord, theKittenNumber );
    ErrorPort.u.c = theWord;

    DefaultInputPort = CurrentInputPort;
    MasterDefaultInputPort = CurrentInputPort;

    TranscriptPort.t = portTag;
    theWord = 0;
    orIntoPortNumberByteNonAtomic( theWord, transcriptIndex );
    orIntoPortKittenByteNonAtomic( theWord, theKittenNumber );
    orIntoKittenByteNonAtomic( theWord, theKittenNumber );
    TranscriptPort.u.c = theWord;

    KIQ.t = emptyListTag;
    B = FALSE;
    ESR = 0;
    blankChar.u.c = MiscChars[ 1 ];
    newlineChar.u.c = MiscChars[ 7 ];
    stackLowLimit        = theHeap->schemeStackSpace()        + SNIFF_POSITION;
    continuationLowLimit = theHeap->schemeContinuationSpace() + SNIFF_POSITION;

    }

/**** reinitializeMachine ****/

void reinitializeMachine()
{
    E = TopLevelE;
    SavedE.t = emptyListTag;
    F  =  S = (taggedAval *)(theHeap->schemeStackEnd());
    CF = CS = (taggedAval *)(theHeap->schemeContinuationEnd());
    NArgs = ArgCounter = zero;
    R.t = emptyListTag;
    A.t = emptyListTag;
    L.t = emptyListTag;

    DefaultInputPort = CurrentInputPort = MasterDefaultInputPort;

    KIQ.t = emptyListTag;
    B = FALSE;
    ESR = 0;
    stackLowLimit = theHeap->schemeStackSpace() + SNIFF_POSITION;
    continuationLowLimit = theHeap->schemeContinuationSpace() + SNIFF_POSITION;
    }

/**** restoreFrame -- R contains a number, which is to become F ****/

void restoreFrame() { F = R.u.a; }

/**** restoreContinuationFrame -- R contains a number, which is to become CF ****/

void restoreContinuationFrame() { CF = R.u.a; }

/**** restoreStack -- R is assumed to point to a CDR CODED list.  Make that list be
                      the Scheme stack.  Fix up the cdr codes so they are all notCar.     ****/
                                  
void restoreStack()
{
    Byte *startOfStackCopy;
    taggedAval *scratch;
    UnsignedLongWord stackLengthInBytes;
    UnsignedLongWord numberOfStackElements;
        
    SPushR();
    numberOfStackElements = lengthFunction();
    stackLengthInBytes = numberOfStackElements * sizeof( taggedAval );
    SPopR();

    if( stackLengthInBytes > theHeap->schemeStackSize() )
        resetError( 229 );

    R = spliceOutNotCarForwardingPointers( R );
    startOfStackCopy = (Byte *)(R.u.a);
    S = scratch = (taggedAval *)(theHeap->schemeStackEnd() - stackLengthInBytes);
    memcpy( (Byte *)S, startOfStackCopy, (size_t)stackLengthInBytes );

    while( numberOfStackElements-- > 0 ) {
        scratch->t = (scratch->t) & ~(cdrCodingBitsMask);
        ++scratch;
        }
    }

/**** restoreContinuation -- R is assumed to point to a CDR CODED list.  Make that list be
                             the Scheme continuation.  Fix up the cdr codes so they are all notCar.     ****/
                                  
void restoreContinuation()
{
    Byte *startOfContinuationCopy;
    taggedAval *scratch;
    UnsignedLongWord continuationLengthInBytes;
    UnsignedLongWord numberOfContinuationElements;
        
    SPushR();
    numberOfContinuationElements = lengthFunction();
    continuationLengthInBytes = numberOfContinuationElements * sizeof( taggedAval );
    SPopR();

    if( continuationLengthInBytes > theHeap->schemeContinuationSize() )
        resetError( 228 );

    R = spliceOutNotCarForwardingPointers( R );
    startOfContinuationCopy = (Byte *)(R.u.a);
    CS = scratch = (taggedAval *)(theHeap->schemeContinuationEnd() - continuationLengthInBytes);
    memcpy( (Byte *)CS, startOfContinuationCopy, (size_t)continuationLengthInBytes );

    while( numberOfContinuationElements-- > 0 ) {
        scratch->t = (scratch->t) & ~(cdrCodingBitsMask);
        ++scratch;
        }
    }

/**** restoreStackAndFrame -- The stack is assumed to have the appropriate frame pointers on top of it.    ****/
                                                         
void restoreStackAndFrame()
{
    restoreStack();
    SPopR();
    restoreFrame();
    }
        
/**** restoreContinuationAndFrame -- The stack is assumed to have the appropriate frame pointers on top of it.    ****/
                                                         
void restoreContinuationAndFrame()
{
    restoreContinuation();
    CSPopR();
    restoreContinuationFrame();
    }
  
/**** restoreForCWCC -- Restore the stack and continuation as provided by call-with-current-continuation.    ****/

void restoreForCWCC()
{
    A = R;
    car();
    restoreContinuationAndFrame();
    R = A;
    cdr();
    restoreStackAndFrame();
    }

/**** saveStack -- Allocate space for that portion of the Scheme stack that is now in use,
                   in the current memory space.  Block-copy the stack thereto.  Fix up the
                   tags so it looks like a cdr-coded list.  Return that list (a pointer
                   thereto, of course) in A.                                                ****/
                                  
void saveStack()
{
    Byte *startOfCopiedStack;
    taggedAval *scratch;
    Tag newTag = 0;
    memoryBlockSize stackLengthInBytes = (Byte *)(theHeap->schemeStackEnd()) - (Byte *)S;
    LongWord numberOfStackElements = stackLengthInBytes / sizeof( taggedAval );
    LongWord counter = 0;


    A.t = emptyListTag;
    if( numberOfStackElements == 0 )
        return;

    theHeap->allocate( stackLengthInBytes );
    scratch = (taggedAval *)(startOfCopiedStack = (Byte *)(A.u.a));

    for( counter = numberOfStackElements - 1; counter >= 0; --counter ) {
        newTag =((S + counter)->t);
        newTag &= ~cdrCodingBitsMask;
        if( counter == (numberOfStackElements - 1) )
            newTag |= carCdrNil;
        else
            newTag |= carCdrNext;
        ((scratch + counter)->u).c = ((S + counter)->u).c;
        (scratch + counter)->t = newTag;
        }
    A.t = pairTag;
    }
        
/**** saveContinuation -- Allocate space for that portion of the Scheme continuation that is now in use,
                          in the current memory space.  Block-copy the continuation thereto.  Fix up the
                          tags so it looks like a cdr-coded list.  Return that list (a pointer
                          thereto, of course) in A.                                                ****/
                                  
void saveContinuation()
{
    Byte *startOfCopiedContinuation;
    taggedAval *scratch;
    Tag newTag = 0;
    memoryBlockSize continuationLengthInBytes = (Byte *)(theHeap->schemeContinuationEnd()) - (Byte *)CS;
    LongWord numberOfContinuationElements = continuationLengthInBytes / sizeof( taggedAval );
    LongWord counter = 0;


    A.t = emptyListTag;
    if( numberOfContinuationElements == 0 )
        return;

    theHeap->allocate( continuationLengthInBytes );
    scratch = (taggedAval *)(startOfCopiedContinuation = (Byte *)(A.u.a));

    for( counter = numberOfContinuationElements - 1; counter >= 0; --counter ) {
        newTag =((CS + counter)->t);
        newTag &= ~cdrCodingBitsMask;
        if( counter == (numberOfContinuationElements - 1) )
            newTag |= carCdrNil;
        else
            newTag |= carCdrNext;
        ((scratch + counter)->u).c = ((CS + counter)->u).c;
        (scratch + counter)->t = newTag;
        }
    A.t = pairTag;
    }
        
/**** saveForCWCC ****/

void saveForCWCC()
{
    saveStack();
    saveContinuation();
    }

/**** stackDepth ****/

void stackDepth()
{
    R.t = exactShortFixnumBitsMask;
    R.u.c = (LongWord)(((Byte *)(theHeap->schemeStackEnd()) - (Byte *)S) / sizeof( taggedAval ));
    }

/**** stackUnZap -- ****/

void stackUnZap() { reinitializeMachine(); }

/**** stackZappedP -- Is the stack zapped? ****/

Boolean stackZappedP() {  return( S < ((taggedAval *)theHeap->schemeStackSpace() + SNIFF_POSITION)); }

/**** SDropFunction ****/

void SDropFunction() { SDrop(); }

/**** SPopEFunction ****/

void SPopEFunction() { SPopE(); }

/**** SPopRFunction ****/

void SPopRFunction() { SPopR(); }

/**** SPushFunction ****/

void SPushFunction() { SPushR(); }

/**** nilConsPush ****/

void nilConsPush()
{
    nilCons();
    SPushR();
    }

/**** SwapCons ****/

void SwapCons()
{
    SPushR();
    S[0] = S[1];
    S[1] = R;
    SPopR();
    cons();
    }

/**** CDropFunction ****/

void CDropFunction() { CSDrop(); }

/**** CPopRFunction ****/

void CPopRFunction() { CSPopR(); }

/**** CPushFunction ****/

void CPushFunction() { CSPush( R ); }

/**** ClearLFunction ****/

void ClearLFunction() { L.t = emptyListTag; }

/**** MakeRTrueFunction ****/

void MakeRTrueFunction() { R.t = trueBooleanTag; }

/**** popToEFunction -- E gets S(0), and pop stack. ****/

void popToEFunction()
{
    E = S[0];
    SDrop();
    }

/**** lengthOfC ****/

LongWord lengthOfC()
{
    LongWord answer = 0;

    SPushR();
    answer = (LongWord)lengthFunction();
    SPopR();
    return answer;
    }

/**** lengthOfCFrame****/

LongWord lengthOfCFrame() { return( CF - CS ); }

/**** compareContinuations ****/

void compareContinuations( Text *msg, Boolean forceP )
{
    LongWord lC    = lengthOfC();
    LongWord lCF   = lengthOfCFrame();
    LongWord lAllC = ((taggedAval *)theHeap->schemeContinuationEnd()) - CS;

    if( lC != lCF )
        tprintf8( &CurrentOutputPort,
                  "\n%s: Continuation lengths DIFFER -- C %jd, current *CF %jd, all *CFs %jd (CS 0x%016jx, CF 0x%016jx) \n",
                  msg, (intmax_t)lC, (intmax_t)lCF, (intmax_t)lAllC, (intmax_t)CS, (intmax_t)CF );
    else if( forceP )
        tprintf7( &CurrentOutputPort,
                  "\n%s: Continuation lengths are both %jd, all *CFs %jd (CS 0x%016jx, CF 0x%016jx) \n",
                  msg, (intmax_t)lC, (intmax_t)lAllC, (intmax_t)CS, (intmax_t)CF );
    if( (lC > 0) && (lAllC > 0) ) {
        SPushR();
        car();
        if( (type_code( R.t) != type_code( CS->t )) || (R.u.a != (CS->u).a) ) {
            tprintf8( &CurrentOutputPort,
                      "\n%s: Continuation first objects DIFFER -- C %jd, current *CF %jd, all *CFs %jd (CS 0x%016jx, CF 0x%016jx) \n",
                      msg, (intmax_t)lC, (intmax_t)lCF, (intmax_t)lAllC, (intmax_t)CS, (intmax_t)CF );
            tprintf2( &CurrentOutputPort, "The first object in the continuation (C) is: ");
            printAnyTaggedAval( &CurrentOutputPort, &R );
            tprintf2( &CurrentOutputPort, "\n");
            displayObject();
            tprintf2( &CurrentOutputPort, "\n");
            CSTopR();
            tprintf2( &CurrentOutputPort, "The first object in the continuation stack (*CS) is: ");
            printAnyTaggedAval( &CurrentOutputPort, CS );
            tprintf2( &CurrentOutputPort, "\n");
            displayObject();
            tprintf2( &CurrentOutputPort, "\n");
            }
        SPopR();
        }
    }

/**** compareFirstContinuationObjects ****/

void compareFirstContinuationObjects( Text *msg )
{
    LongWord lC    = lengthOfC();
    LongWord lAllC = ((taggedAval *)theHeap->schemeContinuationEnd()) - CS;

    tprintf3( &CurrentOutputPort, "%s\n", msg );

    if( lC > 0 ) {
        SPushR();
        car();
        tprintf2( &CurrentOutputPort, "The first object in the continuation (C) is: ");
        displayObject();
        tprintf2( &CurrentOutputPort, "\n");
        SPopR();
        }
    if( lAllC > 0 ) {
        SPushR();
        CSTopR();
        tprintf2( &CurrentOutputPort, "The first object in the continuation stack (*CS) is: ");
        displayObject();
        tprintf2( &CurrentOutputPort, "\n");
        SPopR();
        }
    if( lAllC > 1 ) {
        SPushR();
        R = *(CS + 1);
        tprintf2( &CurrentOutputPort, "The second object in the continuation stack (*CS) is: ");
        displayObject();
        tprintf2( &CurrentOutputPort, "\n");
        SPopR();
        }
    }

/**** CSPushProperList ****/

void CSPushProperList()
{
    LongWord length = 0;

    while( ! empty_list_p( R.t ) ) {
        carCdrMixForCSPushProperList();
        ++length;
        }
    while( length-- > 0 )
        *(--CS) = *S++;
    }

/**** describeCurrentCFrame ****/

void describeCurrentCFrame( Text *msg )
{
    taggedAval *tp = CS;
    taggedAval *theEnd = (taggedAval *)(theHeap->schemeContinuationEnd());

    tprintf5( &CurrentOutputPort, "Current continuation frame: %s --  CS 0x%016jx, CF 0x%016jx\n",
              msg, (intmax_t)CS, (intmax_t)CF );
    SPushR();
    while( tp < CF ) {
        R = *tp;
        tprintf3( &CurrentOutputPort, "At 0x%016jx: ", (intmax_t)tp );
        displayObject();
        tprintf2( &CurrentOutputPort, "\n");
        ++tp;
        }
    tprintf2( &CurrentOutputPort, "The rest of the continuation stack is:\n");
    while( tp < theEnd ) {
        R = *tp;
        tprintf3( &CurrentOutputPort, "At 0x%016jx: ", (intmax_t)tp );
        displayObject();
        tprintf2( &CurrentOutputPort, "\n");
        ++tp;
        }
    SPopR();
    }

/**** applicativeModeP -- return the applicative mode bit from the ESR ****/

Boolean applicativeModeP()
{
    return( (ESR & ESR_applicativeBitMask) != 0 );
    }

/**** applicativeModeNDown -- return the applicative mode bit from the (0-based) Nth stack frame below the current one.  ****/

Boolean applicativeModeNDown( UnsignedLongWord N )
{
    taggedAval *whereItIs = F + (N * FrameSize) + FRAME_ESR_OFFSET;

    return( (((UnsignedLongWord)(whereItIs->u).c) & ESR_applicativeBitMask) != 0 );
    }

/**** applicativeModeOneDown -- return the applicative mode bit from the stack frame immediately below the current one.  ****/

Boolean applicativeModeOneDown() {

    taggedAval *whereItIs = F + FrameSize + FRAME_ESR_OFFSET;

    return( (((UnsignedLongWord)(whereItIs->u).c) & ESR_applicativeBitMask) != 0 );
    }

/**** setApplicativeModeNDown -- set the applicative mode bit in the (0-based) Nth stack frame below the current one.  ****/

void setApplicativeModeNDown( UnsignedLongWord N )
{
    taggedAval *whereItIs = F + (N * FrameSize) + FRAME_ESR_OFFSET;

    (whereItIs->u).c = ((whereItIs->u).c) | ESR_applicativeBitMask;
    }

/**** clearApplicativeModeNDown -- set the applicative mode bit in the (0-based) Nth stack frame below the current one.  ****/

void clearApplicativeModeNDown( UnsignedLongWord N )
{
    taggedAval *whereItIs = F + (N * FrameSize) + FRAME_ESR_OFFSET;

    (whereItIs->u).c = ((whereItIs->u).c) & (~(ESR_applicativeBitMask));
    }

